#!/bin/bash
# ===========================================================================
#
#                            PUBLIC DOMAIN NOTICE
#               National Center for Biotechnology Information
#
#  This software/database is a "United States Government Work" under the
#  terms of the United States Copyright Act.  It was written as part of
#  the author's official duties as a United States Government employee and
#  thus cannot be copyrighted.  This software/database is freely available
#  to the public for use. The National Library of Medicine and the U.S.
#  Government have not placed any restriction on its use or reproduction.
#
#  Although all reasonable efforts have been taken to ensure the accuracy
#  and reliability of the software and data, the NLM and the U.S.
#  Government do not and cannot warrant the performance or results that
#  may be obtained by using this software or data. The NLM and the U.S.
#  Government disclaim all warranties, express or implied, including
#  warranties of performance, merchantability or fitness for any particular
#  purpose.
#
#  Please cite the author in any work or product based on this material.
#
# ===========================================================================

# prepare script name
SELF_NAME="$(basename $0)"
SCRIPT_BASE="${0%.sh}"

# os
OS="$1"
shift

# architecture
ARCH="$1"
shift

# binary loader tool
LD="$1"
shift

# configuration
unset SHLX
unset DYLX
unset LIBX
unset OBJX
unset LOBX

# parameters
TYPE=exe
STATIC=0
DYLD=0
CHECKSUM=0
unset BUILD
unset LDIRS
unset XDIRS
unset SRCDIR
unset BINDIR
unset VERSFILE
unset VERSDIR
unset TARG
unset EXT
unset OBJS
unset LIBS
unset DEPFILE

while [ $# -ne 0 ]
do

    case "$1" in
    --build)
        BUILD="$2"
        shift
        ;;

    --ldflags)
        LDFLAGS="$2"
        shift
        ;;

    --checksum)
        CHECKSUM=1
        ;;

    --shlx)
        SHLX="$2"
        shift
        ;;

    --dylx)
        SHLX="$2"
        shift
        ;;

    --libx)
        LIBX="$2"
        shift
        ;;

    --objx)
        OBJX="$2"
        shift
        ;;

    --srcdir)
        SRCDIR="$2"
        shift
        ;;

    --bindir)
        BINDIR="$2"
        shift
        ;;

    -MD)
        DEPFILE=1
        ;;

    -L*)
        ARG="${1#-L}"
        if [ "$ARG" = "" ]
        then
            ARG="$2"
            shift
        fi
        LDIRS="$LDIRS:$ARG"
        ;;
        
    -X*)
        ARG="${1#-X}"
        if [ "$ARG" = "" ]
        then
            ARG="$2"
            shift
        fi
        XDIRS="$XDIRS:$ARG"
        ;;

    --dlib)
        TYPE=dlib
        ;;

    --slib)
        TYPE=slib
        ;;

    --stub)
        TYPE=stub
        ;;

    --exe)
        TYPE=exe
        ;;

    --static)
        STATIC=1
        ;;

    --vers)
        if [ -f "$2" ]
        then
            VERSFILE="$2"
        elif [ -d "$2" ]
        then
            VERSDIR="$2"
        else
            echo "$SELF_NAME: expected version file or source directory"
            exit 3
        fi
        shift
        ;;

    -o*)
        ARG="${1#-o}"
        if [ "$ARG" = "" ]
        then
            ARG="$2"
            shift
        fi
        TARG="$ARG"
        ;;

    -l*|-s*)
        LIBS="$LIBS $1"
        ;;

    -d*)
        LIBS="$LIBS $1"
        DYLD=1
        ;;

    *.$OBJX)
        OBJS="$OBJS $1"
        ;;
        
    esac

    shift
done

# correct for prefixes
LDIRS="${LDIRS#:}"
XDIRS="${XDIRS#:}"
LIBS="${LIBS# }"
OBJS="${OBJS# }"

# split target
OUTDIR=$(dirname "$TARG")
NAME=$(basename "$TARG")

# dependency file
[ "$DEPFILE" != "" ] && DEPFILE="$NAME.d"

# parse target
if [ "$TYPE" = "dlib" ] && [ "$DYLX" != "" ]
then
    EXT="$DYLX"
    NAME="${NAME%.$DYLX}"
fi
unset VERS
if [ "$NAME" != "${NAME%.[0-9][0-9]*}" ]
then
    ARG="${NAME%.[0-9][0-9]*}"
    VERS="${NAME#$ARG}"
    NAME="${ARG#.}"

    if [ "$NAME" != "${NAME%.[0-9][0-9]*}" ]
    then
        ARG="${NAME%.[0-9][0-9]*}"
        VERS="${NAME#$ARG}.$VERS"
        NAME="${ARG#.}"

        if [ "$NAME" != "${NAME%.[0-9][0-9]*}" ]
        then
            ARG="${NAME%.[0-9][0-9]*}"
            VERS="${NAME#$ARG}.$VERS"
            NAME="${ARG#.}"
        fi
    fi
fi
case "$TYPE" in
dlib)
    if [ "$SHLX" != "" ]
    then
        EXT="$SHLX"
        NAME="${NAME%.$SHLX}"
    fi
    ;;
slib)
    EXT="$LIBX"
    NAME="${NAME%.$LIBX}"
esac

unset DBGAP
if [ "$NAME" != "${NAME%-dbgap}" ]
then
	DBGAP=-dbgap
	NAME="${NAME%-dbgap}"
fi

# locate version file and version
[ "$VERSDIR" != "" ] && VERSFILE="$VERSDIR/$NAME.vers"
if [ "$VERSFILE" != "" ]
then
    if [ ! -f "$VERSFILE" ]
    then
        echo "$SELF_NAME: warning - creating version file '$VERSFILE'"
        echo 1.0.0 > $VERSFILE
    fi

    if [ ! -r "$VERSFILE" ]
    then
        echo "$SELF_NAME: version file '$VERSFILE' is unreadable"
        exit 5
    fi

    ARG=$(cat $VERSFILE)
    if [ "$VERS" != "" ] && [ "$VERS" != "$ARG" ]
    then
        echo "$SELF_NAME: version from file '$VERSFILE' does not match '$VERS'"
        exit 5
    fi
    VERS="$ARG"
fi

#echo "# $SELF_NAME"
#echo "#    BUILD     : $BUILD"
#echo "#    OS        : $OS"
#echo "#    ARCH      : $ARCH"
#echo "#    tool      : $LD"
#echo "#    dep file  : $DEPFILE"
#echo "#    LDFLAGS   : $LDFLAGS"
#echo "#    CHECKSUM  : $CHECKSUM"
#echo "#    static    : $STATIC"
#echo "#    type      : $TYPE"
#echo "#    srcdir    : $SRCDIR"
#echo "#    bindir    : $BINDIR"
#echo "#    LDIRS     : $LDIRS"
#echo "#    XDIRS     : $XDIRS"
#echo "#    vers file : $VERSFILE"
#echo "#    vers dir  : $VERSDIR"
#echo "#    target    : $TARG"
#echo "#    outdir    : $OUTDIR"
#echo "#    name      : $NAME"
#echo "#    dbgap     : $DBGAP"
#echo "#    extension : $EXT"
#echo "#    version   : $VERS"
#echo "#    objects   : $OBJS"
#echo "#    libraries : $LIBS"
#echo "#  script-base : $SCRIPT_BASE"

# overwrite dependencies
[ -f "$DEPFILE" ] && rm -f "$DEPFILE"

# generate mode
MODE=$(expr $DYLD + $DYLD + $STATIC)

# perform link
"$SCRIPT_BASE.$OS.$TYPE.sh" "$LD" "$ARCH" "$BUILD" "$SRCDIR" "$BINDIR" "$OUTDIR" "$TARG" "$NAME" "$DBGAP" "$VERS" "$VERSFILE" "$DEPFILE" "$MODE" "$LDFLAGS" "$LDIRS" "$XDIRS" "$OBJS" "$LIBS" || exit $?

# *** START SCM-code ***************************************
if [ $CHECKSUM -eq 1 ]
    then
    # prepare to call the module-scm-script...
    BUILD_DIR=${SCRIPT_BASE%/*}
    # calling the scm-script
    "$BUILD_DIR/scm.sh" "$NAME" "$OBJS"

    # concatenate all md5-files for the source-dir into one file...
    # this works because this script is invoked from within the obj-directory
    #CMD="cat *.md5"
    #echo "$CMD > $TARG.temp.md5"
    #if $CMD > "$TARG.temp.md5"
        # then
        # build the scm-dir (parallel to build) out of script base
        # and create it if it does not exist ( just to make sure... )
        #mkdir -p "$SCM_DIR"

        # build one single checksum-file for the module in the scm-dir
        #CMD="md5sum -b $TARG.temp.md5"
        #echo "$CMD | head -c32 > $TARG.md5"
        #if $CMD | head -c32 > "$SCM_DIR/$NAME.md5"
        #    then
        #    echo "# $NAME ---> checksum build OK"
            # test if the release-md5-file exists
        #    TMP_FILE="$SCM_DIR/$NAME.md5"
        #    REL_FILE="$SCM_DIR/$NAME.md5.release"
        #    CAN_FILE="$SCM_DIR/$NAME.md5.candidate"
        #    STA_FILE="$SCM_DIR/$NAME.status"
        #    if [ -f "$REL_FILE" ]
        #        then
        #        echo "release-file for $NAME exists"
        #        cmp -s $TMP_FILE $REL_FILE
        #        if [ $? -ne 0 ]
        #            then
        #            echo "current-md5 and last-released-md5 differ!"
        #            
        #            else
        #            echo "current-md5 and last-released-md5 are the same!"
        #            echo "---> nothing to be done for module $NAME"
        #        fi
        #        else
        #        echo "$NAME never released before!"
        #        if [ -f "$CAN_FILE" ]
        #            then
        #            echo "but we have a candidate-file"
        #
        #            else
        #            echo "and no candidate-file found!"
        #            echo "---> increment version"
        #            echo "---> save current md5 as candidate"
        #            echo "---> trigger rebuild"
        #        fi
        #
        #    fi
        #        
        #fi
        # remove the temp. file
        #rm "$TARG.temp.md5"
        #fi
fi
# *** END SCM-code ***************************************

# establish links
if [ "$VERS" != "" ] && [ "$OS" != "win" ]
then
    $SCRIPT_BASE.$OS.ln.sh "$TYPE" "$OUTDIR" "$TARG" "$NAME" "$DBGAP" "$EXT" "$VERS"
fi
