/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_kdb_meta_
#define _hpp_kdb_meta_

#ifndef _h_kdb_meta_
#include <kdb/meta.h>
#endif


/*--------------------------------------------------------------------------
 * KMetadata
 *  a versioned, hierarchical structure
 */
struct KMetadata
{
    /* AddRef
     * Release
     *  all objects are reference counted
     *  NULL references are ignored
     */
    inline rc_t AddRef () const
    { return KMetadataAddRef ( this ); }

    inline rc_t Release () const
    { return KMetadataRelease ( this ); }


    /* Version
     *  returns the metadata format version
     */
    inline rc_t Version ( uint32_t *version ) const
    { return KMetadataVersion ( this, version ); }


    /* ByteOrder
     *  indicates whether original byte order is reversed
     *  under current architecture.
     *
     *  the byte order of the column is established by
     *  the host architecture when created.
     *
     *  "reversed" [ OUT ] - if true, the original byte
     *  order is reversed with regard to host native byte order.
     */
    inline rc_t ByteOrder ( bool *reversed ) const
    { return KMetadataByteOrder ( this, reversed ); }


    /* Revision
     *  returns current revision number
     *  where 0 ( zero ) means tip
     */
    inline rc_t Revision ( uint32_t *revision ) const
    { return KMetadataRevision ( this, revision ); }


    /* MaxRevision
     *  returns the maximum revision available
     */
    inline rc_t MaxRevision ( uint32_t *revision ) const
    { return KMetadataMaxRevision ( this, revision ); }


    /* Commit
     *  ensure any changes are committed to disk
     */
    inline rc_t Commit ()
    { return KMetadataCommit ( this ); }


    /* Freeze
     *  freezes current metadata revision
     *  further modification will begin on a copy
     */
    inline rc_t Freeze ()
    { return KMetadataFreeze ( this ); }


    /* OpenRevision
     *  opens a read-only indexed revision of metadata
     */
    inline rc_t OpenRevision ( const KMetadata **meta, uint32_t revision ) const
    { retyrb KMetadataOpenRevision ( this, meta, revision ); }


    /* GetSequence
     * SetSequence
     * NextSequence
     *  access a named sequence
     *
     *  "seq" [ IN ] - NUL terminated sequence name
     *
     *  "val" [ OUT ] - return parameter for sequence value
     *  "val" [ IN ] - new sequence value
     */
    inline rc_t GetSequence ( const char *seq, int64_t *val ) const
    { return KMetadataGetSequence ( this, seq, val ); }

    inline rc_t SetSequence ( const char *seq, int64_t val )
    { return KMetadataSetSequence ( this, seq, val ); }

    inline rc_t NextSequence ( const char *seq, int64_t *val )
    { return KMetadataNextSequence ( this, seq, val ); }
};


/*--------------------------------------------------------------------------
 * KMDataNode
 *  a node with an optional value,
 *  optional attributes, and optional children
 *
 *  nodes are identified by path, relative to a starting node,
 *  where "/" serves as a path separator.
 */
struct KMDataNode
{
    /* AddRef
     * Release
     *  ignores NULL references
     */
    inline rc_t AddRef () const
    { return KMDataNodeAddRef ( this ); }

    inline rc_t Release () const
    { return KMDataNodeRelease ( this ); }


    inline rc_t OpenNodeUpdate ( KMDataNode **node,
        const char *path, ... )
    {
        va_list args;

        va_start ( args, path );
        rc_t rc = KMDataNodeOpenNodeUpdate (this, node, path, args);
        va_end ( args );
        return rc;
    }

    inline rc_t Read ( uint32_t *value ) const
    { return KMDataNodeReadAsU32 ( this, value ); }

    inline rc_t Read ( uint64_t *value ) const
    { return KMDataNodeReadAsU64 ( this, value ); }

    inline rc_t Read ( char *buffer, size_t bsize, size_t *size ) const
    { return KMDataNodeReadCString ( this, buffer, bsize, size ); }

    inline rc_t Read ( int32_t *i ) const
    { return KMDataNodeReadAsI32 ( this, i ); }

    inline rc_t WriteAttr ( const char *name, const char *value )
    { return KMDataNodeWriteAttr ( this, name, value ); }

private:
    KMDataNode ();
    ~ KMDataNode ();
    KMDataNode ( const KMDataNode& );
    KMDataNode &operator = ( const KMDataNode& );
};


#endif // _hpp_kdb_meta_
