/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_kfs_directory_
#define _hpp_kfs_directory_

#include <kfs/directory.h>

/*--------------------------------------------------------------------------
 * KDirectory
 *  a container node within file system
 */
struct KDirectory {
    /* AddRef
     * Release
     *  ignores NULL references
     */
    rc_t AddRef ( void ) const
    { return KDirectoryAddRef ( this ); }
    rc_t Release ( void ) const
    { return KDirectoryRelease ( this ); }

    /* List
     *  create a directory listing
     *
     *  "list" [ OUT ] - return parameter for list object
     *
     *  "f" [ IN, NULL OKAY ] and "data" [ IN, OPAQUE ] - optional
     *  filter function to execute on each path. receives a base directory
     *  and relative path for each entry. if "f" returns true, the name will
     *  be added to the list.
     *
     *  "path" [ IN, NULL OKAY ] - optional parameter for target
     *  directory. if NULL, interpreted to mean "."
     */
    rc_t List ( struct KNamelist **list,
        bool ( CC * f ) ( const KDirectory *dir, const char *name, void *data ),
        void *data,
        const char *path, ... ) const
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = KDirectoryList ( this, list, f, data, path, args );
        va_end ( args );
        return rc;
    }

    /* PathType
     *  returns a KPathType
     *
     *  "path" [ IN ] - NUL terminated string in directory-native character set
     */
    uint32_t PathType ( const char *path, ... ) const
    {
        va_list args;
        va_start ( args, path );
        uint32_t r = KDirectoryPathType ( this, path, args );
        va_end ( args );
        return r;
    }

    /* OpenFileRead
     *  opens an existing file with read-only access
     *
     *  "f" [ OUT ] - return parameter for newly opened file
     *
     *  "path" [ IN ] - NUL terminated string in directory-native
     *  character set denoting target file
     */
    rc_t OpenFileRead ( struct KFile const **f, const char *path, ... ) const
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = KDirectoryVOpenFileRead ( this, f, path, args );
        va_end ( args );
        return rc;
    }
    rc_t OpenFileRead ( struct KFile const **f, const char *path, va_list args ) const
    { return KDirectoryVOpenFileRead ( this, f, path, args ); }

    /* NativeDir
     *  returns a native file-system directory node reference
     *  the directory root will be "/" and set to the native
     *  idea of current working directory
     *
     *  NB - the returned reference will be non-const, allowing
     *  modification operations to be attempted. these operations
     *  may still fail if the underlying FS disallows them.
     *
     *  "dir" [ OUT ] - return parameter for native directory
     */
    static rc_t NativeDir ( KDirectory **dir )
    { return KDirectoryNativeDir ( dir ); }
};

#endif /* _hpp_kfs_directory_ */
