/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_kfs_toc_
#define _h_kfs_toc_

#ifndef _h_kfs_extern_
#include <kfs/extern.h>
#endif

#ifndef _h_klib_defs_
#include <klib/defs.h>
#endif

#ifndef _h_kfs_directory_
#include <kfs/directory.h>
#endif

#ifndef _h_klib_pbstree_
#include <klib/pbstree.h>
#endif

#ifndef _h_kfs_sra_
#include <kfs/sra.h>
#endif


#ifdef __cplusplus
extern "C" {
#endif

/*==========================================================================
 * The TOC (Table of Contents) is the basis of non-O/S level file systems
 * directory structures.
 *
 * The TOC is based on a set of entries.  An entry can represent a file
 * (collection of bytes), a directory (a list of entries, or a link (an
 * entry that refers to another entry).
 *
 * Files are represented either as a contiguous array of bytes or a
 * sequence of smaller chunks of contiguous bytes.  Other than the initial
 * creation of the chunked version, most operations on the chunked file
 * will be able to treat it as a contiguous sequence with any gaps between
 * contiguous chunks being treated as if they were all zero bytes.  So
 * for the most part the two types of file are the same to the outside
 * world.
 *
 * A directory is viewed as a sorted list of named references to files.
 * In the first versions the form is of a binary search tree per directory
 * but for the most part that detail is irrelevant.  In the tree a directory
 * is an entry that contains its own tree.  A path is a sequence of named
 * entries in order - the representation of a path is the Unix/Posix style
 * path.
 *
 * A link is either a hard or soft link 9using Unix/Posix terminology. A
 * soft link is a named item that contains a path that names another item.
 * A hard link will as much as possible refer directly to the same file,
 * directory or another link as another item.
 *
 * Each entry is represented by a TocEntry.
 *
 * A path as mentioned above is a list of facets (names of directories) up to
 * a final leaf item that can be any of the item types.
 *
 * Position is the logical offset within a file or a particualr byte,  The
 * first byte is at position 0.
 */


/*--------------------------------------------------------------------------
 * forwards
 */
struct BSTree;
struct Vector;
struct String;
struct KFile;


/*--------------------------------------------------------------------------
 * KTocChunk
 *  Describes a chunk of valid data within a file.  The concept is that a file
 *  can be made up of discrete chunks that might not be consecutive on the 
 *  storage system.  There might also be gaps between the chunks - gaps that 
 *  read as all zeroes.  These could be tarred up versions of sparse files
 *  where the only things stored on disk are small regions of non-zero data. 
 *  Or it could be some archival format where data is broken into some form of
 *  packets.
 *
 *  64 bit values are used for file sizes up to 18,446,744,073,709,551,617 bytes
 *  up to exabytes.  The human genome is in the gigabyte range so 32 bit values
 *  are too close to the limits so the larger range was used.
 *
 *  This struct is very clear and public so chunks can be built up or used  freely
 *  by non-file-system specific code.
 */
typedef struct KTocChunk KTocChunk;
struct KTocChunk
{
    /* position of chunk within logical file */
    uint64_t logical_position;

    /* position of chunk within source container file */
    uint64_t source_position;

    /* chunk size */
    uint64_t size;
};


/*--------------------------------------------------------------------------
 * KTocEntry
 */
typedef struct KTocEntry KTocEntry;
typedef enum KTocEntryType
{
    ktocentrytype_unknown = -1,
    ktocentrytype_notfound,
    ktocentrytype_dir,
    ktocentrytype_file,
    ktocentrytype_chunked,
    ktocentrytype_softlink,
    ktocentrytype_hardlink,
    ktocentrytype_emptyfile
} KTocEntryType;

/* four fields are common to all entries.*/


/* GetName
 *
 *  [RET] rc_t                  0 for success; anything else for a failure
 *                       see itf/klib/rc.h for general details
 *  [IN]     const KTocEntry *  self        object oriented object reference
 *  [OUT] const char **         name        where to put a reference to the name
 */
KFS_EXTERN rc_t CC KTocEntryGetName(const KTocEntry *self,
                     const char ** name);

/* GetTime
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] KTime_t *      mtime       where to put unix/posix style permission flags
 */
KFS_EXTERN rc_t CC KTocEntryGetTime(const KTocEntry *self,
                     KTime_t *mtime);

/* GetAccess
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                             see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] uint32_t       access      where to put unix/posix style permission flags
 */
KFS_EXTERN rc_t CC KTocEntryGetAccess(const KTocEntry * self, uint32_t * access);


/* GetType
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] KTocEntryType *    type        where to put the type specific node type (differs
 *                      from KDirectory type)
 */
KFS_EXTERN rc_t CC KTocEntryGetType(const KTocEntry *self,
                     KTocEntryType * type);
/* the type retrieved by GetType will permit several more fields specialized by type */
/* for a directory the only specialzied field is the "BSTree" */

#if 0
KFS_EXTERN const char * CC KTocEntryGetTypeString(KTocEntryType type);
#endif

/* GetBSTree
 *
 * Get a reference to the BSTree of a TocEntry that represents a directory.
 * This will fail for all ites taht are not directories.
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] const BSTree **    ptree       where to put a reference to the kcont/BSTree of a
 *                      directory node
 */
KFS_EXTERN rc_t CC KTocEntryGetBSTree( const KTocEntry * self,
                     struct BSTree const ** ptree);

/* for a file there are two fields: a container offset and a file size */


/* GetFileSize (works for contiguous or chunked files)
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] uint64 *       size        where to put the size of a referenced node
 */
KFS_EXTERN rc_t CC KTocEntryGetFileSize( const KTocEntry * self,
                     uint64_t * size);


/* GetFileOffset
 *
 * For normal contiguous file nodes 'offset' gets the offset to the start of the file.
 * For all other nodes, 'offset' gets 0 with an appropriate rc_t return.
 *
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] uint64 *       offset      where to put the offset of a referenced node
 */
KFS_EXTERN rc_t CC KTocEntryGetFileOffset( const KTocEntry * self,
                     uint64_t * offset);


/* GetHardTarget
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] const KTocEntry ** target      where to put a reference from a hard link node
 */
KFS_EXTERN rc_t CC KTocEntryGetHardTarget( const KTocEntry *self,
                     const KTocEntry ** target);

/* there are two specialzied fields for chunked files - both retrieved with the saem method */


/* GetChunks
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] const KTocChunk *  chunks      where to put the size of a referenced node
 */
KFS_EXTERN rc_t CC KTocEntryGetChunks (const KTocEntry * self,
                     uint32_t * num_chunks,
                     const KTocChunk ** chunks);

/* GetSoftTarget
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KTocEntry *  self        object oriented object reference
 * [OUT] const char **      target      where to put a reference from a hard link node
 */
KFS_EXTERN rc_t CC KTocEntryGetSoftTarget (const KTocEntry *self,
                     const char ** target);


/*--------------------------------------------------------------------------
 * KToc
 *  archive table of contents
 *
 *  the "CreateXXX" series of methodss take a KCreateMode ( see kfs/directory.h ).
 *  when mode == kcmOpen, create if needed, overwrite if exists.
 *  when mode == kcmInit, create if needed, clear if exists.
 *  when mode == kcmCreate, fail if exists, create otherwise.
 *  if mode has kcmParents set, fill in missing parents on create.
 *
 * These is the only open API into a TOC with in the archive package of classes.
 * Functions to parse archives (Tar was the first format but is not expected to 
 * be the only format) use these methods to add entries to the TOC.  New parsers
 * should fit the signature of 
 *
 *  rc_t PARSER (KToc * <TOC>, const KFile * <ARCHIVE>)
 *
 * "PARSER" [IN] is the parsing function
 *
 * "TOC" [OUT] is a opaque class that is never exposed other than to these create
 * entry methods
 *
 * "FILE" [IN] is an archive opaquely opened as a KFile and then given to a
 * KDirectory and should not be accessed from outside this parsing function.
 *
 * An Archive is opened via a call to friend function.  It has by design been
 * made to mimic the signature of KDirectoryOpenDirRead with only a change in
 * parameters after path.  No "..." or va_list is supported.  To use that
 * functionality one would have to call KDirectoryResolvePath and use the
 * new path generated by that call as the path parameter for this function.
 *
 * rc_t KDirectoryOpenArcDirRead (const KDirectory * dir, 
 *                    KDirectory ** pdir,
 *                    bool chroot,
 *                    const char * path,
 *                    rc_t (CC* parse )(KToc *,const KFile*));
 *
 * "dir" [IN] is an existing KDirectory in which to find the archive file.
 *
 * "pdir" [OUT] is a pointer which will after the function is called point to the new
 * KDirectory that is the archive opened as a directory with the same name as
 * its containing file.
 *
 * "chroot" [IN] is false(1)/true(1 [or any non zero]) as to whether the archive will
 * treat its contents as a virtual root directory ('/') or not.
 *
 * "path" [IN] is the address with in directory of the archive.
 *
 * "parse" [IN] is the function that will create TOC entries for that archive.
 *
 * No member element or function of an object of class TOC will be exposed
 * other than these create entry methods.
 */
typedef struct KToc KToc;

/* Init
 *  construct a TOC tied to a KFile with a given path
 * At this point, this method has been made private and will only be called by 
 * KDirectoryOpenArcDirRead().
 */
typedef enum KTocFSType
{
    tocUnknown,
    tocKFile,
    tocKDirectory,
    tocKVirtual
} KTocFSType;

KFS_EXTERN rc_t CC KTocInit( KToc**self, struct String const * path, KTocFSType arctype,
        const void * archive, KSRAFileAlignment alignment);

/* AddRef
 * Release
 *  ignores NULL references
 */
KFS_EXTERN rc_t CC KTocAddRef( const KToc *self );
KFS_EXTERN rc_t CC KTocRelease( const KToc *self );


/* AddRef
 * Release
 *  ignores NULL references
 * At this point, these methods have been made private and will only be called by 
 * with in the archive package.  References to the TOC will be handled within 
 * the constructors and destructors for KDirectorys and KFiles created by
 * KDirectory OpenArcDirRead() and then any KDirectory OpenFileRead() and 
 * KDirectory OpenDirRead() with in that initial KDirectory().  At this point
 * archives are read only so OpenFileWrite() OpenDirWrite() CreateDir() CreateFile()
 * and other such methods are not supported for archives.
 */
KFS_EXTERN rc_t CC KTocAddRef ( const KToc *self );
KFS_EXTERN rc_t CC KTocRelease ( const KToc *self );


/* ResolvePathNode
 *
 *
 * [RET] rc_t                   0 for success; anything else for a failure
 *                      see itf/klib/rc.h for general details
 * [IN]  const KToc *   self        TOC self reference: object oriented in C
 * [IN]  const char *       path        the path to be parsed - it will start at the root 
 *                      of the TOC's address
 * [OUT] const KTocEntry **     pnode       the address of the final node traversed is put here
 * [OUT] uint32_t *         type        the type of node for the final node
 * [OUT] char **        unusedpath  This will point to any part of the path not used up
 *                      in parsing against the TOC.
 *                      NULL if the leaf node was found
 *                      the start of a node that could not be parsed
 *                      the start of the node after a softlink (or NULL)
 */
KFS_EXTERN rc_t CC KTocResolvePathTocEntry( const KToc *self,
                         const KTocEntry ** pnode,
                         const char *path,
                         size_t path_len,
                         KTocEntryType * ptype,
                         const char ** unusedpath);

/*--------------------------------------------------------------------------
 * Constructors/factories
 */

/* CreateDir
 *  create a sub-directory
 *
 *  "mode" [ IN ] - a creation mode ( see explanation above ).
 *
 *  "path" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target directory
 */
KFS_EXTERN rc_t CC KTocCreateDir ( KToc *self, KTime_t mtime, uint32_t access, 
        KCreateMode mode, const char *path, ... );
KFS_EXTERN rc_t CC KTocVCreateDir ( KToc *self, KTime_t mtime, uint32_t access,
        KCreateMode mode, const char *path, va_list args );

/* CreateFile
 *  create an entry for a contiguous file
 *
 *  "source_position" [ IN ] and "size" [ IN ] - description of the contiguous
 *  region of the archive file occupied by the file
 *
 *  "mode" [ IN ] - a creation mode ( see explanation above ).
 *
 *  "path" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target file
 *
 * This is used also to create an empty file.  When size is zero the underlying 
 * library might create a special zero length entry type.
 */
KFS_EXTERN rc_t CC KTocCreateFile ( KToc *self, uint64_t source_position, uint64_t size,
              KTime_t mtime, uint32_t access, KCreateMode mode,
              const char *path, ... );
KFS_EXTERN rc_t CC KTocVCreateFile ( KToc *self, uint64_t source_position, uint64_t size,
               KTime_t mtime, uint32_t access, KCreateMode mode,
               const char *path, va_list args );

/* CreateChunkedFile
 *  create an entry for a chunked file
 *
 *  a chunked file consists of 0 or more possibly discontiguous chunks
 *  positioned within the logical file
 *
 *  "size" [ IN ] - the size of the logical file
 *
 *  "num_chunks" [ IN ] - the number of chunks in "chunks"
 *
 *  "chunks" [ IN ] - an array of chunk descriptors, where each chunk
 *  contains valid data that are mapped to form file
 *
 *  "mode" [ IN ] - a creation mode ( see explanation above ).
 *
 *  "path" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target file
 */
KFS_EXTERN rc_t CC KTocCreateChunkedFile ( KToc *self, uint64_t size, KTime_t mtime,
                 uint32_t access, uint32_t num_chunks,
                 const KTocChunk chunks [], KCreateMode mode,
                 const char *path, ... );
KFS_EXTERN rc_t CC KTocVCreateChunkedFile ( KToc *self, uint64_t size, KTime_t mtime,
                  uint32_t access, uint32_t num_chunks,
                  const KTocChunk chunks [], KCreateMode mode,
                  const char *path, va_list args );

/* CreateHardLink
 *  creates a hard link to an existing file within TOC
 *
 *  "mode" [ IN ] - a creation mode ( see explanation above ).
 *
 *  "targ" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target object
 *
 *  "link" [ IN ] - NUL terminated string in directory-native
 *  character set denoting alias
 */
KFS_EXTERN rc_t CC KTocCreateHardLink ( KToc *self, KTime_t mtime, uint32_t access,
              KCreateMode mode, const char *targ,
              const char *link, ... );
KFS_EXTERN rc_t CC KTocVCreateHardLink ( KToc *self, KTime_t mtime, uint32_t access,
               KCreateMode mode, const char *targ,
               const char *link, va_list args );

/* CreateSoftLink
 *  creates a soft link to any file or directory
 *  that is interpreted upon acces.
 *
 *  "mode" [ IN ] - a creation mode ( see explanation above ).
 *
 *  "targ" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target object
 *
 *  "alias" [ IN ] - NUL terminated string in directory-native
 *  character set denoting target alias
 */
KFS_EXTERN rc_t CC KTocCreateSoftLink ( KToc * self, KTime_t mtime, uint32_t access,
              KCreateMode mode, const char * targ,
              const char * alias, ... );
KFS_EXTERN rc_t CC KTocVCreateSoftLink ( KToc * self, KTime_t mtime, uint32_t access,
               KCreateMode mode, const char * targ,
               const char * alias, va_list args );

/* OpenArcDirRead
 *  Open an archive file as a KDirectory derived type: made to match 
 *  KDirectoryOpenDirRead() where parse could be the first element of arg
 *
 * [IN]  dir    A KDirectory (of any derived type) to reach the archive file
 * [OUT] pdir   The KDirectory (of type KArcDir) that will be created
 * [IN]  chroot if non-zero, the new directory becomes chroot'd and interprets paths
 *      stating with '/'
 * [IN]  path   The path to the archive - this will become the directory path
 * [IN]  parse  A pointer to the function needed to build a TOC by parsing the file
 *
 * parameters to parse
 * REQUIRED:    KToc *
 * REQUIRED:    void *  KFile* | KDirectory* | ???
 * OPTIONAL:    bool (*)(const char*)           Filter: include or skip this path
 * OPTIONAL:    rc_t (*)(???)               Sort: reorder name list
 */

KFS_EXTERN rc_t CC KDirectoryOpenArcDirRead (const KDirectory * dir, 
            const KDirectory ** pdir,
            bool chroot,
            const char * path,
            KTocFSType baseType,
            rc_t (CC*parse)(KToc*,const void *,
                bool(CC*)(const KDirectory*,const char *, void*),
                void *),
                bool (CC* filter )(const KDirectory*, const char *, void *),
            void * filterparam );

/* ParseTAR
 *  A parse function for tar files suitable for KDirectoryOpenArcDirRead
 *
 * [INOUT] KToc*    self    The TOC (table of contents) to fill in with the parse
 * [IN]    const KFile* kfile   The opened (as KFile) archive to parse.
 */
KFS_EXTERN rc_t CC KTocParseTAR( KToc * self, 
             const void * kfile,
             bool(CC*f)(const KDirectory*,const char *, void*),
             void *filterparam);
KFS_EXTERN rc_t CC KTocParseKDir( KToc * self,
             const void * kdir,
             bool(CC*f)(const KDirectory*,const char *, void*),
             void *filterparam);
KFS_EXTERN rc_t CC KTocParseSRA( KToc * self,
             const void * kvoid,
             bool (CC*ignored) (const KDirectory*,  const char*, void*),
             void *ignored_data);

#if 0
KFS_EXTERN rc_t CC KTocResolvePathTocEntry( const KToc *self,
                 const KTocEntry ** pnode,
                 const char *path,  /* must be entirely in the TOC */
                 size_t path_len,
                 KTocEntryType * ptype,
                 const char ** unusedpath);
#endif



KFS_EXTERN const KTocEntry* CC KTocGetRoot( const KToc * self );

#if 0
KFS_EXTERN const char * CC KTocGetRootPath( const KToc * self, size_t * len );
#endif

KFS_EXTERN rc_t CC KTocGetPath( const KToc * self, struct String const ** path );
KFS_EXTERN rc_t CC KTocResolvePathFromOffset( const KToc *self,
                const char ** path,
                uint64_t * ppos,
                uint64_t offset );

KFS_EXTERN KSRAFileAlignment CC KTocAlignmentGet( KToc * self );
KFS_EXTERN rc_t CC KTocAlignmentSet( KToc * self,KSRAFileAlignment align );

KFS_EXTERN const void * CC KTocGetArchive( const KToc * self );

KFS_EXTERN rc_t CC KTocPersist( const KToc * self,
          void ** buffer, size_t * buffer_size,
          uint64_t * virtual_file_size,
          struct Vector const * files );

KFS_EXTERN rc_t CC KTocEntryPersist( void *param, const void * node,
               size_t * num_writ, 
               PTWriteFunc write, void * write_param );

KFS_EXTERN rc_t CC KDirectoryOpenTocFileRead( const KDirectory * self,
                struct KFile const ** pfile,
                KSRAFileAlignment align,
                bool (CC * filter) (const KDirectory*,const char*,void*),
                void * filter_param,
                rc_t (CC * usort)(const KDirectory*, struct Vector*) );


#ifdef __cplusplus
}
#endif

#endif /* _h_kfs_toc_ */
