/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_vdb_manager_
#define _hpp_vdb_manager_

#ifndef _h_vdb_manager_
#include <vdb/manager.h>
#endif


/*--------------------------------------------------------------------------
 * VDBManager
 *  opaque handle to library
 */
struct VDBManager
{
    /* AddRef
     * Release
     *  all objects are reference counted
     *  NULL references are ignored
     */
    inline rc_t AddRef ()
    const { return VDBManagerAddRef ( this ); }

    inline rc_t Release () const
    { return VDBManagerRelease ( this ); }


    /* MakeRead
     * MakeUpdate
     *  create library handle for specific use
     *  NB - only one of the functions will be implemented
     *
     *  "wd" [ IN, NULL OKAY ] - optional working directory for
     *  accessing the file system. mgr will attach its own reference.
     */
    static inline rc_t MakeRead ( const VDBManager **mgr, const KDirectory *wd = 0 )
    { return VDBManagerMakeRead ( mgr, wd ); }

    static inline rc_t MakeUpdate ( VDBManager **mgr, KDirectory *wd = 0 )
    { return VDBManagerMakeUpdate ( mgr, wd ); }


    /* Version
     *  returns the library version
     */
    inline rc_t Version ( uint32_t *version ) const
    { return VDBManagerVersion ( this, version ); }


    /* Writable
     *  returns 0 if object is writable
     *  or a reason why if not
     *
     *  "path" [ IN ] - NUL terminated path
     */
    inline rc_t Writable ( const char *path, ... ) const
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = VDBManagerVWritable ( this, path, args );
        va_end ( args );
        return rc;
    }

    /* Lock
     *  apply lock
     *
     *  if object is already locked, the operation is idempotent
     *  and returns an rc state of rcLocked
     *
     *  "path" [ IN ] - NUL terminated path
     */
    inline rc_t Lock ( const char *path, ... )
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = VDBManagerVLock ( this, path, args );
        va_end ( args );
        return rc;
    }

    /* Unlock
     *  remove lock
     *
     *  if object is already unlocked, the operation is idempotent
     *  and returns an rc state of rcUnlocked
     *
     *  "path" [ IN ] - NUL terminated path
     */
    inline rc_t Unlock ( const char *path, ... )
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = VDBManagerVUnlock ( this, path, args );
        va_end ( args );
        return rc;
    }


    /* AddSchemaIncludePath
     *  add a path to schema for locating input files
     */
    inline rc_t AddSchemaIncludePath ( const char *path, ... )
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = VDBManagerVAddSchemaIncludePath ( this, path, args );
        va_end ( args );
        return rc;
    }


    /* AddLoadLibraryPath
     *  add a path to loader for locating dynamic libraries
     */
    inline rc_t AddLoadLibraryPath ( const char *path, ... )
    {
        va_list args;
        va_start ( args, path );
        rc_t rc = VDBManagerVAddLoadLibraryPath ( this, path, args );
        va_end ( args );
        return rc;
    }


    /* GetUserData
     * SetUserData
     *  store/retrieve an opaque pointer to user data
     *
     *  "data" [ OUT ] - return parameter for getting data
     *  "data" [ IN ] - parameter for setting data
     *
     *  "destroy" [ IN, NULL OKAY ] - optional destructor param
     *  invoked from destructor of "self"
     */
    inline rc_t GetUserData ( void **data ) const
    { return VDBManagerGetUserData ( this, data ); }

    inline rc_t SetUserData ( void *data, void ( CC * destroy ) ( void *data ) = 0 ) const
    { return VDBManagerSetUserData ( this, data, destroy ); }
};

#endif // _hpp_vdb_manager_
