/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_vdb_table_
#define _hpp_vdb_table_

#ifndef _h_vdb_table_
#include <vdb/table.h>
#endif


/*--------------------------------------------------------------------------
 * VTable
 *  a collection of columns indexed by row id, metadata, indices
 */
struct VTable
{
    /* AddRef
     * Release
     *  all objects are reference counted
     *  NULL references are ignored
     */
    rc_t VTableAddRef ( const VTable *self );
    rc_t VTableRelease ( const VTable *self );


    /* Locked
     *  returns true if locked
     */
    bool VTableLocked ( const VTable *self );


    /* Writable
     *  returns 0 if object is writable
     *  or a reason why if not
     *
     *  "type" [ IN ] - a KDBPathType
     *  valid values are kptIndex and kptColumn
     *
     *  "path" [ IN ] - NUL terminated path
     */
    rc_t VTableWritable ( const VTable *self, uint32_t type,
                          const char * name, ... );
    rc_t VTableVWritable ( const VTable *self, uint32_t type,
                           const char * name, va_list args );

    /* Lock
     *  apply lock
     *
     *  if object is already locked, the operation is idempotent
     *  and returns an rc state of rcLocked
     *
     *  "type" [ IN ] - a KDBPathType
     *  valid values are kptIndex and kptColumn
     *
     *  "path" [ IN ] - NUL terminated path
     */
    rc_t VTableLock ( VTable *self, uint32_t type,
                      const char * name, ... );
    rc_t VTableVLock ( VTable *self, uint32_t type,
                       const char * name, va_list args );

    /* Unlock
     *  remove lock
     *
     *  if object is already unlocked, the operation is idempotent
     *  and returns an rc state of rcUnlocked
     *
     *  "type" [ IN ] - a KDBPathType
     *  valid values are kptIndex and kptColumn
     *
     *  "path" [ IN ] - NUL terminated path
     */
    rc_t VTableUnlock ( VTable *self, uint32_t type,
                        const char * name, ... );
    rc_t VTableVUnlock ( VTable *self, uint32_t type,
                         const char * name, va_list args );


    /* OpenMetadataRead
     * OpenMetadataUpdate
     *  opens metadata file
     *
     *  "meta" [ OUT ] - return parameter for metadata
     */
    rc_t VTableOpenMetadataRead ( const VTable *self, struct KMetadata const **meta );
    rc_t VTableOpenMetadataUpdate ( VTable *self, struct KMetadata **meta );
    

    /* ColumnCreateParams
     *  sets the creation parameters for physical columns
     *
     *  "cmode" [ IN ] - creation mode
     *
     *  "checksum" [ IN ] - the type of checksum information to
     *  apply when writing blobs
     *
     *  "pgsize" [ IN, DEFAULT ZERO ] - size of internal column "pages"
     *  the default value is indicated by 0 ( zero ).
     *  NB - CURRENTLY THE ONLY SUPPORTED PAGE SIZE IS 1 ( ONE ) BYTE.
     */
    rc_t VTableColumnCreateParams ( VTable *self,
                                    KCreateMode cmode, KChecksum checksum, size_t pgsize );


    /* CreateIndex
     * VCreateIndex
     *  create a new or open an existing index
     *
     *  "idx" [ OUT ] - return parameter for newly opened index
     *
     *  "type" [ IN ] - type of index to create
     *
     *  "cmode" [ IN ] - creation mode
     *
     *  "name" [ IN ] - NUL terminated string in UTF-8 giving simple name of idx
     */
    rc_t VTableCreateIndex ( VTable *self, KIndex **idx,
                             KIdxType type, KCreateMode cmode, const char *name, ... );
    rc_t VTableVCreateIndex ( VTable *self, KIndex **idx,
                              KIdxType type, KCreateMode cmode, const char *name, va_list args );


    /* OpenIndexRead
     * VOpenIndexRead
     *  open an index for read
     *
     *  "idx" [ OUT ] - return parameter for newly opened index
     *
     *  "name" [ IN ] - NUL terminated string in UTF-8 giving simple name of idx
     */
    rc_t VTableOpenIndexRead ( const VTable *self,
                               const KIndex **idx, const char *name, ... );
    rc_t VTableVOpenIndexRead ( const VTable *self,
                                const KIndex **idx, const char *name, va_list args );


    /* OpenIndexUpdate
     * VOpenIndexUpdate
     *  open an index for read/write
     *
     *  "idx" [ OUT ] - return parameter for newly opened index
     *
     *  "name" [ IN ] - NUL terminated string in UTF-8 giving simple name of idx
     */
    rc_t VTableOpenIndexUpdate ( VTable *self,
                                 KIndex **idx, const char *name, ... );
    rc_t VTableVOpenIndexUpdate ( VTable *self,
                                  KIndex **idx, const char *name, va_list args );


    /* List
     *  create table listings
     */
    rc_t VTableListCol ( const VTable *self, struct KNamelist **names );
    rc_t VTableListIdx ( const VTable *self, struct KNamelist **names );
   

    /* ColumnDatatypes
     *  returns list of typedecls for named column
     *
     *  "col" [ IN ] - column name
     *
     *  "dflt_idx" [ OUT, NULL OKAY ] - returns the zero-based index
     *  into "typedecls" of the default datatype for the named column
     *
     *  "typedecls" [ OUT ] - list of datatypes available for named column
     */
    rc_t VTableColumnDatatypes ( const VTable *self, const char *col,
                                 uint32_t *dflt_idx, struct KNamelist **typedecls );


    /* Reindex
     *  optimize column indices
     */
    rc_t VTableReindex ( VTable *self );


    /* OpenManager
     *  duplicate reference to manager
     *  NB - returned reference must be released
     */
    rc_t VTableOpenManagerRead ( const VTable *self, struct VDBManager const **mgr );
    rc_t VTableOpenManagerUpdate ( VTable *self, struct VDBManager **mgr );


    /* OpenParent
     *  duplicate reference to parent database
     *  NB - returned reference must be released
     */
    rc_t VTableOpenParentRead ( const VTable *self, struct VDatabase const **db );
    rc_t VTableOpenParentUpdate ( VTable *self, struct VDatabase **db );


    /* OpenSchema
     *  duplicate reference to table schema
     *  NB - returned reference must be released
     */
    rc_t VTableOpenSchema ( const VTable *self, struct VSchema const **schema );


    /* GetUserData
     * SetUserData
     *  store/retrieve an opaque pointer to user data
     *
     *  "data" [ OUT ] - return parameter for getting data
     *  "data" [ IN ] - parameter for setting data
     *
     *  "destroy" [ IN, NULL OKAY ] - optional destructor param
     *  invoked from destructor of "self"
     */
    rc_t VTableGetUserData ( const VTable *self, void **data );
    rc_t VTableSetUserData ( const VTable *self,
        void *data, void ( CC * destroy ) ( void *data ) );
};

#endif //  _hpp_vdb_table_
