/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 */

#include <kfg/extern.h>
#include <kfg/config.h>
#include <klib/token.h>
#include <klib/container.h>
#include <klib/refcount.h>
#include <klib/text.h>
#include <klib/rc.h>
#include <klib/log.h>
#include <kfs/directory.h>
#include <kfs/file.h>
#include <kfs/dyload.h>
#include <kfs/mmap.h>
#include <sysalloc.h>

#include <string.h>
#include <stdlib.h>
#include <assert.h>


/*--------------------------------------------------------------------------
 * KConfig
 */
static
rc_t KConfigSever ( const KConfig *self );


/*--------------------------------------------------------------------------
 * KConfigNode
 *  node within configuration tree
 */
struct KConfigNode
{
    BSTNode n;

    /* needs to hold a dependency reference to mgr */
    KConfig *mgr;

    /* uncounted reference to parent node */
    KConfigNode *dad;

    /* named children - always unique */
    BSTree children;

    /* named attributes */
    BSTree attr;

    /* node value */
    char * val_buffer;
    String value;

    /* node name */
    String name;

    KRefcount refcount;

    bool read_only;
};

/* replace this once we introduce attributes */
#define KConfigAttrWhack NULL


/* Whack
 */
static
void CC KConfigNodeWhack ( BSTNode *n, void * data )
{
    KConfigNode *self = ( KConfigNode* ) n;
    KConfig *mgr = data;

    if ( mgr == NULL )
    {
        /* just releasing reference */
        KConfigSever ( self -> mgr );
        self -> mgr = NULL;
        self -> read_only = false;
    }
    else
    {
        /* tearing down structure */
        BSTreeWhack ( & self -> children, KConfigNodeWhack, mgr );
        BSTreeWhack ( & self -> attr, KConfigAttrWhack, mgr );
        free ( self -> val_buffer );
        free ( self );
    }
}

/* Init
 */
static
void KConfigNodeInit ( KConfigNode *self, const String *name )
{
    /* initialize name early for the sake of KRefcountInit */
    string_copy ( ( char* ) ( self + 1 ), name -> size + 1, name -> addr, name -> size );
    StringInit ( & self -> name, ( char* ) ( self + 1 ), name -> size, name -> len );

    KRefcountInit ( & self -> refcount, 0, "KConfigNode", "init", self -> name . addr );

    self -> mgr = NULL;
    self -> dad = NULL;
    BSTreeInit ( & self -> children );
    BSTreeInit ( & self -> attr );
    self -> val_buffer = NULL;
    StringInit ( & self -> value, "", 0, 0 );
    self -> read_only = false;
}

/* Make
 */
static
rc_t KConfigNodeMake ( KConfigNode **n, const String *name )
{
    KConfigNode *cn = malloc ( sizeof * cn + name -> size + 1 );
    if ( cn == NULL )
    {
        rc_t rc = RC ( rcKFG, rcNode, rcCreating, rcMemory, rcExhausted );
        PLOGERR (klogErr, (klogErr, rc, "Unable to create a config item for $(i)", "i=%S", name));
        return rc;
    }
    KConfigNodeInit ( cn, name );
    * n = cn;

    return 0;
}


/* Cmp
 * Sort
 */
static
int CC KConfigNodeCmp ( const void *item, const BSTNode *n )
{
    const String *a = ( const String* ) item;
    const KConfigNode *b = ( const KConfigNode* ) n;
    return StringCompare ( a, & b -> name );
}

static
int CC KConfigNodeSort ( const BSTNode *item, const BSTNode *n )
{
    const KConfigNode *a = ( const KConfigNode* ) item;
    const KConfigNode *b = ( const KConfigNode* ) n;
    return StringCompare ( & a -> name, & b -> name );
}


/* AddRef
 * Release
 *  all objects are reference counted
 *  NULL references are ignored
 */
LIB_EXPORT rc_t CC KConfigNodeAddRef ( const KConfigNode *self )
{
    if ( self != NULL )
    {
        switch ( KRefcountAdd ( & self -> refcount, "KConfigNode" ) )
        {
        case krefLimit:
            return RC ( rcKFG, rcNode, rcAttaching, rcRange, rcExcessive );
        }
    }
    return 0;
}

LIB_EXPORT rc_t CC KConfigNodeRelease ( const KConfigNode *self )
{
    if ( self != NULL )
    {
        switch ( KRefcountDrop ( & self -> refcount, "KConfigNode" ) )
        {
        case krefWhack:
            KConfigNodeWhack ( & ( ( KConfigNode* ) self ) -> n, NULL );
	    break;
        case krefLimit:
            return RC ( rcKFG, rcNode, rcReleasing, rcRange, rcExcessive );
        }
    }
    return 0;
}


/*--------------------------------------------------------------------------
 * KConfig
 *  configuration paramter manager
 */
struct KConfig
{
    BSTree tree;
    KDualRef refcount;
};


static KConfig *G_kfg = NULL;


/* Whack
 */
static
rc_t KConfigWhack ( KConfig *self )
{
    if ( self == G_kfg )
#if KFG_COMMON_CREATION
        return 0;
#else
        G_kfg = NULL;
#endif

    BSTreeWhack ( & self -> tree, KConfigNodeWhack, self );
    free ( self );
    return 0;
}

/* Init
 */
static
void KConfigInit ( KConfig *self, KConfigNode * root )
{
    KDualRefInit ( & self -> refcount, 1, 0, "KConfig", "init", "kfg" );
    BSTreeInit ( & self -> tree );
    BSTreeInsert ( & self -> tree, & root -> n, KConfigNodeSort );
}


/* AddRef
 * Release
 */
LIB_EXPORT rc_t CC KConfigAddRef ( const KConfig *self )
{
    if ( self != NULL )
    {
        switch ( KDualRefAdd ( & self -> refcount, "KConfig" ) )
        {
        case krefLimit:
            return RC ( rcKFG, rcMgr, rcAttaching, rcRange, rcExcessive );
        }
    }
    return 0;
}

LIB_EXPORT rc_t CC KConfigRelease ( const KConfig *self )
{
    if ( self != NULL )
    {
        switch ( KDualRefDrop ( & self -> refcount, "KConfig" ) )
        {
        case krefWhack:
            return KConfigWhack ( ( KConfig* ) self );
        case krefLimit:
            return RC ( rcKFG, rcMgr, rcReleasing, rcRange, rcExcessive );
        }
    }
    return 0;
}

static
KConfig *KConfigAttach ( const KConfig *self )
{
    if ( self != NULL )
    {
        switch ( KDualRefAddDep ( & self -> refcount, "KConfig" ) )
        {
        case krefLimit:
            return NULL;
        }
    }
    return ( KConfig* ) self;
}

static
rc_t KConfigSever ( const KConfig *self )
{
    if ( self != NULL )
    {
        switch ( KDualRefDropDep ( & self -> refcount, "KConfig" ) )
        {
        case krefWhack:
            return KConfigWhack ( ( KConfig* ) self );
        case krefLimit:
            return RC ( rcKFG, rcMgr, rcReleasing, rcRange, rcExcessive );
        }
    }
    return 0;
}


/* init_token_source
 */
static
rc_t init_token_source ( KTokenText *tt, KTokenSource *src,
                         char *full, size_t fsize, const char *srcpath, const char *path, va_list args )
{
    size_t num_writ;
    rc_t rc = 0;

    if (args == NULL)
        num_writ = string_copy ( full, fsize, path, string_size ( path ));
    else
        rc = string_vprintf ( full, fsize, & num_writ, path, args );
    if ( rc == 0 )
    {
        String text, fpath;
        StringInit ( & text, full, num_writ, string_len ( full, num_writ ) );
        StringInitCString ( & fpath, srcpath );
        KTokenTextInit ( tt, & text, & fpath );
        KTokenSourceInit ( src, tt );
    }
    return rc;
}

/* Find
 */
static
KToken *KConfigNodeFind ( const KConfigNode *self, const KConfigNode **n, KTokenSource *src, KToken *t )
{
    * n = NULL;

    while ( t -> id != eEndOfInput )
    {
        switch ( t -> id )
        {
        case ePeriod:
            break;
        case eDblPeriod:
            if ( self -> dad == NULL )
                return NULL;
            self = self -> dad;
            break;
        case eDecimal:
        case eHex:
        case eOctal:
        case eIdent:
        case eName:
            self = ( const KConfigNode* ) BSTreeFind
                ( & self -> children, &t -> str, KConfigNodeCmp );
            if ( self == NULL )
                return t;
            break;
        default:
            * n = self;
            return t;
        }

        if ( KTokenizerNext ( kPOSIXPathTokenizer, src, t ) -> id != eFwdSlash )
            break;

        KTokenizerNext ( kPOSIXPathTokenizer, src, t );
    }

    * n = self;
    return t;
}

/* Create
 */
static
KToken *KConfigNodeCreate ( KConfigNode *self, KConfigNode **n, KTokenSource *src, KToken *t )
{
    KConfigNode * nself;
    * n = NULL;

    while ( t -> id != eEndOfInput )
    {
        switch ( t -> id )
        {
        case ePeriod:
            break;
        case eDblPeriod:
            if ( self -> dad == NULL )
                return NULL;
            self = self -> dad;
            break;
        case eDecimal:
        case eHex:
        case eOctal:
        case eName:
        case eIdent:
            nself = ( KConfigNode* ) BSTreeFind
                ( & self -> children, & t -> str, KConfigNodeCmp );
            if ( nself == NULL )
            {
                KConfigNode *child;
                rc_t rc = KConfigNodeMake ( & child, & t -> str );
                if ( rc != 0 )
                    return t;
                BSTreeInsert ( & self -> children, & child -> n, KConfigNodeSort );  
                child -> dad = self;
                self = child;
            }
            else
            {
                self = nself;
            }
            break;
        default:
            * n = self;
            return t;
        }

        if ( KTokenizerNext ( kPOSIXPathTokenizer, src, t ) -> id != eFwdSlash )
            break;

        KTokenizerNext ( kPOSIXPathTokenizer, src, t );
    }

    * n = self;
    return t;
}


/* OpenNodeRead
 * VOpenNodeRead
 *  opens a configuration node
 *
 *  "node" [ OUT ] - return parameter for indicated configuration node
 *
 *  "path" [ IN, NULL OKAY ] - optional path for specifying named
 *  node within configuration hierarchy. paths will be interpreted as
 *  if they were file system paths, using '/' as separator. the
 *  special values NULL and "" are interpreted as "."
 */
static
rc_t KConfigNodeVOpenNodeReadInt ( const KConfigNode *self, const KConfig *mgr,
                                   const KConfigNode **node, const char *path, va_list args )
{
    rc_t rc;

    if ( node == NULL )
    {
        rc = RC ( rcKFG, rcNode, rcOpening, rcParam, rcNull );
        PLOGERR (klogErr, (klogErr, rc, "faile to provide node to open $(n)", "n=%s", path));
    }
    else
    {
        if ( self == NULL )
        {
            rc = RC ( rcKFG, rcNode, rcOpening, rcSelf, rcNull );
            PLOGERR (klogErr, (klogErr, rc, "failed to provide node reference for opening $(n)", "n=%s", path));
        }
        else
        {
            if ( path == NULL || path [ 0 ] == 0 )
            {
                * node = self;
                rc = 0;
            }
            else
            {
                KTokenText tt;
                KTokenSource src;
                char full [ 4096 ];
        
                rc = init_token_source ( & tt, & src, full, sizeof full, "", path, args );
                if ( rc == 0 )
                {
                    /* look ahead */
                    KToken t;

                    /* skip over fwd slashes */
                    do
                        KTokenizerNext ( kPOSIXPathTokenizer, & src, & t );
                    while ( t.id == eFwdSlash );

                    /* follow path */
                    if ( KConfigNodeFind ( self, node, & src, & t ) == NULL )
                    {
                        rc = RC ( rcKFG, rcNode, rcOpening, rcPath, rcInvalid );
                        PLOGERR (klogErr, (klogErr, rc, "bad path $(p)", "p=%s", path));
                    }
                    if ( ( self = * node ) == NULL )
                    {
                        rc = RC ( rcKFG, rcNode, rcOpening, rcPath, rcNotFound );
                        /* don't complain about this
                           PLOGERR (klogErr, (klogErr, rc, "can't find symbol $(p)", "p=%s", path));
                        */
                    }
                    else if ( t . id != eEndOfInput )
                    {
                        rc = RC ( rcKFG, rcNode, rcOpening, rcPath, rcInvalid );
                        PLOGERR (klogErr, (klogErr, rc, "bad path $(p)", "p=%s", path));
                    }
                }
            }

            if ( rc == 0 )
            {
                /* open node for read */
                if ( self -> read_only )
                {
                    assert ( self -> mgr == mgr );
                    return KConfigNodeAddRef ( self );
                }

                /* check to see if already open */
                if ( atomic32_read ( & self -> refcount ) == 0 )
                {
                    ( ( KConfigNode* ) self ) -> mgr = KConfigAttach ( mgr );
                    ( ( KConfigNode* ) self ) -> read_only = true;
                    return KConfigNodeAddRef ( self );
                }

                rc = RC ( rcKFG, rcNode, rcOpening, rcNode, rcBusy );
            }
        }

        * node = NULL;
    }

    return rc;
}

LIB_EXPORT rc_t CC KConfigNodeVOpenNodeRead ( const KConfigNode *self,
                                              const KConfigNode **node, const char *path, va_list args )
{
    if ( self != NULL )
        return KConfigNodeVOpenNodeReadInt ( self, self -> mgr, node, path, args );

    if ( node == NULL )
        return RC ( rcKFG, rcNode, rcOpening, rcParam, rcNull );

    * node = NULL;
    return RC ( rcKFG, rcNode, rcOpening, rcSelf, rcNull );
}

LIB_EXPORT rc_t CC KConfigNodeOpenNodeRead ( const KConfigNode *self,
                                             const KConfigNode **node, const char *path, ... )
{
    rc_t rc;
    va_list args;

    va_start ( args, path );
    rc = KConfigNodeVOpenNodeRead ( self, node, path, args );
    va_end ( args );

    return rc;
}

LIB_EXPORT rc_t CC KConfigVOpenNodeRead ( const KConfig *self,
                                          const KConfigNode **node, const char *path, va_list args )
{
    rc_t rc;

    if ( node == NULL )
        rc = RC ( rcKFG, rcMgr, rcOpening, rcParam, rcNull );
    else
    {
        if ( self == NULL )
            rc = RC ( rcKFG, rcMgr, rcOpening, rcSelf, rcNull );
        else if (self->tree.root == NULL)
            rc = RC ( rcKFG, rcMgr, rcOpening, rcPath, rcNotFound );
        else
        {
            return KConfigNodeVOpenNodeReadInt
                ( (const KConfigNode *) self -> tree . root, self, node, path, args );
        }

        * node = NULL;
    }

    return rc;
}

LIB_EXPORT rc_t CC KConfigOpenNodeRead ( const KConfig *self,
                                         const KConfigNode **node, const char *path, ... )
{
    rc_t rc;
    va_list args;

    va_start ( args, path );
    rc = KConfigVOpenNodeRead ( self, node, path, args );
    va_end ( args );

    return rc;
}


/* OpenNodeUpdate
 * VOpenNodeUpdate
 *  opens a configuration node
 *
 *  "node" [ OUT ] - return parameter for indicated configuration node
 *
 *  "path" [ IN, NULL OKAY ] - optional path for specifying named
 *  node within configuration hierarchy. paths will be interpreted as
 *  if they were file system paths, using '/' as separator. the
 *  special values NULL and "" are interpreted as "."
 */
static
rc_t KConfigNodeVOpenNodeUpdateInt ( KConfigNode *self, KConfig *mgr,
                                     KConfigNode **node, const char *path, va_list args )
{
    rc_t rc;

    if ( node == NULL )
        rc = RC ( rcKFG, rcNode, rcOpening, rcParam, rcNull );
    else
    {
        if ( self == NULL )
            rc = RC ( rcKFG, rcNode, rcOpening, rcSelf, rcNull );
        else
        {
            if ( path == NULL || path [ 0 ] == 0 )
            {
                * node = self;
                rc = 0;
            }
            else
            {
                KTokenText tt;
                KTokenSource src;
                char full [ 4096 ];
        
                rc = init_token_source ( & tt, & src, full, sizeof full, "", path, args );
                if ( rc == 0 )
                {
                    /* look ahead */
                    KToken t;

                    do
                        KTokenizerNext ( kPOSIXPathTokenizer, & src, & t );
                    while ( t.id == eFwdSlash);

                    /* follow path */
                    if ( KConfigNodeCreate ( self, node, & src, & t ) == NULL )
                        return RC ( rcKFG, rcNode, rcOpening, rcPath, rcInvalid );
                    if ( ( self = * node ) == NULL )
                        rc = RC ( rcKFG, rcNode, rcOpening, rcMemory, rcExhausted );
                    else if ( t . id != eEndOfInput )
                        rc = RC ( rcKFG, rcNode, rcOpening, rcPath, rcInvalid );
                }
            }

            if ( rc == 0 )
            {
                /* check to see if open */
                if ( atomic32_read ( & self -> refcount ) == 0 )
                {
                    self -> mgr = KConfigAttach ( mgr );
                    assert ( ! self -> read_only );
                    return KConfigNodeAddRef ( self );
                }

                rc = RC ( rcKFG, rcNode, rcOpening, rcNode, rcBusy );
            }
        }

        * node = NULL;
    }

    return rc;
}

LIB_EXPORT rc_t CC KConfigNodeVOpenNodeUpdate ( KConfigNode *self,
                                                KConfigNode **node, const char *path, va_list args )
{
    if ( self != NULL )
        return KConfigNodeVOpenNodeUpdateInt ( self, self -> mgr, node, path, args );

    if ( node == NULL )
        return RC ( rcKFG, rcNode, rcOpening, rcParam, rcNull );

    * node = NULL;
    return RC ( rcKFG, rcNode, rcOpening, rcSelf, rcNull );
}

LIB_EXPORT rc_t CC KConfigNodeOpenNodeUpdate ( KConfigNode *self,
                                               KConfigNode **node, const char *path, ... )
{
    rc_t rc;
    va_list args;

    va_start ( args, path );
    rc = KConfigNodeVOpenNodeUpdate ( self, node, path, args );
    va_end ( args );

    return rc;
}

LIB_EXPORT rc_t CC KConfigVOpenNodeUpdate ( KConfig *self,
                                            KConfigNode **node, const char *path, va_list args )
{
    rc_t rc;

    if ( node == NULL )
        rc = RC ( rcKFG, rcMgr, rcOpening, rcParam, rcNull );
    else
    {
        if ( self == NULL )
            rc = RC ( rcKFG, rcMgr, rcOpening, rcSelf, rcNull );
        else if (self->tree.root == NULL)
            rc = RC ( rcKFG, rcMgr, rcOpening, rcSelf, rcCorrupt );
        else
        {
            return KConfigNodeVOpenNodeUpdateInt
                ( ( KConfigNode* ) self -> tree . root, self, node, path, args );
        }

        * node = NULL;
    }

    return rc;
}

LIB_EXPORT rc_t CC KConfigOpenNodeUpdate ( KConfig *self,
                                           KConfigNode **node, const char *path, ... )
{
    rc_t rc;
    va_list args;

    va_start ( args, path );
    rc = KConfigVOpenNodeUpdate ( self, node, path, args );
    va_end ( args );

    return rc;
}


/* Read
 *  read a node value
 *
 *  "offset" [ IN ] - initial offset into configuration
 *
 *  "buffer" [ OUT ] and "bsize" [ IN ] - return buffer for read
 *
 *  "num_read" [ OUT ] - number of bytes actually read
 *
 *  "remaining" [ OUT, NULL OKAY ] - optional return parameter for
 *  the number of bytes remaining to be read.
 *  specifically, "offset" + "num_read" + "remaining" == sizeof node data
 */
LIB_EXPORT rc_t CC KConfigNodeRead ( const KConfigNode *self,
                                     size_t offset, char *buffer, size_t bsize,
                                     size_t *num_read, size_t *remaining )
{
    rc_t rc;
    size_t dummy;

    if ( remaining == NULL )
        remaining = & dummy;

    if ( num_read == NULL )
        rc = RC ( rcKFG, rcNode, rcReading, rcParam, rcNull );
    else
    {
        if ( self == NULL )
            rc = RC ( rcKFG, rcNode, rcReading, rcSelf, rcNull );
        else if ( buffer == NULL && bsize != 0 )
            rc = RC ( rcKFG, rcNode, rcReading, rcBuffer, rcNull );
        else if ( offset >= self -> value . size )
            rc = 0;
        else
        {
            size_t avail = * remaining = self -> value . size - offset;
            if ( avail > bsize )
                avail = bsize;
            if ( avail > 0 )
                memcpy ( buffer, & self -> value . addr [ offset ], avail );
            * num_read = avail;
            * remaining -= avail;
            return 0;
        }

        * num_read = 0;
    }

    * remaining = 0;

    return rc;
}


/* Write
 *  write a node value or attribute
 *  replaces anything already there
 *
 *  "buffer" [ IN ] and "size" [ IN ] - new value data
 */
LIB_EXPORT rc_t CC KConfigNodeWrite ( KConfigNode *self, const char *buffer, size_t size )
{
    rc_t rc;

    if ( self == NULL )
        rc = RC ( rcKFG, rcNode, rcWriting, rcSelf, rcNull );
    else if ( self -> read_only )
        rc = RC ( rcKFG, rcNode, rcWriting, rcSelf, rcReadonly );
    else if ( size == 0 )
    {
        free ( self -> val_buffer ), self -> val_buffer = NULL;
        StringInit ( & self -> value, "", 0, 0 );
        rc = 0;
    }
    else if ( buffer == NULL )
        rc = RC ( rcKFG, rcNode, rcWriting, rcBuffer, rcNull );
    else
    {
        if ( size != self -> value . size )
        {
            void *new_buffer = realloc ( self -> val_buffer, size + 1 );
            if ( new_buffer == NULL )
                return RC ( rcKFG, rcNode, rcWriting, rcMemory, rcExhausted );
            self -> val_buffer = new_buffer;
            self -> value . size = size;
            self -> value . addr = new_buffer;
        }

        assert ( self -> val_buffer != NULL );
        string_copy ( self -> val_buffer, self -> value . size + 1, buffer, size );
        self -> value . len = string_len ( self -> val_buffer, size );
        rc = 0;
    }

    return rc;
}


/* Append
 *  append data to value
 *
 *  "buffer" [ IN ] and "size" [ IN ] - value data to be appended
 */
LIB_EXPORT rc_t CC KConfigNodeAppend ( KConfigNode *self, const char *buffer, size_t size )
{
    rc_t rc;

    if ( self == NULL )
        rc = RC ( rcKFG, rcNode, rcWriting, rcSelf, rcNull );
    else if ( self -> read_only )
        rc = RC ( rcKFG, rcNode, rcWriting, rcSelf, rcReadonly );
    else if ( size == 0 )
        rc = 0;
    else if ( buffer == NULL )
        rc = RC ( rcKFG, rcNode, rcWriting, rcBuffer, rcNull );
    else
    {
        void *new_buffer = realloc ( self -> val_buffer, self -> value . size + size + 1 );
        if ( new_buffer == NULL )
            return RC ( rcKFG, rcNode, rcWriting, rcMemory, rcExhausted );
        self -> val_buffer = new_buffer;
        string_copy ( & self -> val_buffer [ self -> value . size ], self -> value . size + size + 1, buffer, size );
        self -> value . size += size;
        self -> value . len = string_len ( self -> val_buffer, self -> value . size );
        rc = 0;
    }

    return rc;
}


/* ReadAttr
 *  reads as NUL-terminated string
 *
 *  "name" [ IN ] - NUL terminated attribute name
 *
 *  "buffer" [ OUT ] and "bsize" - return parameter for attribute value
 *
 *  "size" [ OUT ] - return parameter giving size of string
 *  not including NUL byte. the size is set both upon success
 *  and insufficient buffer space error.
 */
LIB_EXPORT rc_t CC KConfigNodeReadAttr ( const KConfigNode *self, const char *name,
                                         char *buffer, size_t bsize, size_t *size )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}


/* WriteAttr
 *  writes NUL-terminated string
 *
 *  "name" [ IN ] - NUL terminated attribute name
 *
 *  "value" [ IN ] - NUL terminated attribute value
 */
LIB_EXPORT rc_t CC KConfigNodeWriteAttr ( KConfigNode *self,
                                          const char *name, const char *value )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}


/* Drop
 * VDrop
 *  drop some or all node content
 */
LIB_EXPORT rc_t CC KConfigNodeDropAll ( KConfigNode *self )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}

LIB_EXPORT rc_t CC KConfigNodeDropAttr ( KConfigNode *self, const char *attr )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}

LIB_EXPORT rc_t CC KConfigNodeVDropChild ( KConfigNode *self, const char *path, va_list args )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}

LIB_EXPORT rc_t CC KConfigNodeDropChild ( KConfigNode *self, const char *path, ... )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}


/* Rename
 *  renames a contained object
 *
 *  "from" [ IN ] - NUL terminated string in UTF-8
 *  giving simple name of existing attr
 *
 *  "to" [ IN ] - NUL terminated string in UTF-8
 *  giving new simple attr name
 */
LIB_EXPORT rc_t CC KConfigNodeRenameAttr ( KConfigNode *self, const char *from, const char *to )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}

LIB_EXPORT rc_t CC KConfigNodeRenameChild ( KConfigNode *self, const char *from, const char *to )
{
    PLOGMSG (klogFatal, (klogFatal, "$(F) unimplmented", "F=%s", __func__));
    return -1;
}


/*--------------------------------------------------------------------------
 * KConfig
 */


/*
  Grammar:
  EOL = "\n";
  lower = [a-z]
  alpha = [A-Za-z]
  file = <line>*;
  line = comment | path-value;
  comment = script-comment | cpp-comment;
  script-comment = "#", not_pound , EOL;
  cpp-comment = "//", any, EOL;

*/

typedef struct parse_block
{
    KConfig * self;
    const KTokenizer * tz;
    KTokenSource * src;
    KToken t;
    const String * symbol;
    const String * value;
} parse_block;

static
rc_t write_nvp (parse_block * pb)
{
    rc_t rc;
    KConfigNode * node;

    rc = KConfigVOpenNodeUpdate ( pb->self, &node, pb->symbol->addr, NULL);

    if (rc == 0)
    {
        /* +1 and -2 get rid of containing quote marks */
        rc = KConfigNodeWrite (node, pb->value->addr + 1, StringSize (pb->value) - 2);

        KConfigNodeRelease ( node );
    }

    return rc;
}

static
rc_t parse_value (parse_block * pb)
{
    rc_t rc;
    size_t value_size;
    char value_buffer [4096];


    switch ((KTokenizerNext (pb->tz, pb->src, &pb->t))->id)
    {
    default:
        rc = RC (rcKFG, rcMgr, rcParsing, rcString, rcUnexpected);
        break;
    case eString:
    case eEscapedString:
        rc = KTokenToString (&pb->t, value_buffer, sizeof value_buffer, &value_size);
        if (rc == 0)
        {
            StringCopy (&pb->value, &pb->t.str);
            rc = write_nvp (pb);
            free ((void*)pb->value);
        }
        break;
    }
    return rc;
}

static
rc_t validate_symbol (parse_block * pb)
{
    /* yeah this should do something eventually */
    return 0;
}
static
rc_t parse_name (parse_block * pb)
{
    const String * concat;
    rc_t rc;

    switch ( KTokenizerNext ( pb -> tz, pb -> src, & pb -> t ) -> id )
    {
    default:
        rc = RC (rcKFG, rcMgr, rcParsing, rcName, rcUnexpected);
        break;
        /* any of these can be part of a path -- but not yet */
        /*     case eDecimal: */
        /*     case eHex: */
        /*     case eOctal: */
        /*     case eFloat: */
        /*     case eExpFloat: */
        /*     case eMajMinRel: */
    case eIdent:
    case eName:
        /*     case ePeriod: */
        /*     case eTilde: */
        /*     case eMinus: */
    case eFwdSlash:     
        /*     case eDblPeriod: */
        /*     case eEllipses: */
        rc = StringConcat (&concat, pb->symbol, &pb->t.str);
        if (rc == 0)
        {
            pb->symbol = concat;
            rc = parse_name (pb);
            free ((void*)concat);
        }
        break;
    case eAssign:
        rc = validate_symbol (pb);
        if (rc == 0)
        {
            rc = parse_value (pb);            
        }
        break;
    }
    return rc;
}

static
rc_t parse_nvp (parse_block * pb)
{
    const String * symbol;
    rc_t rc;

redo:
    switch ((KTokenizerNext (pb->tz, pb->src, &pb->t))->id)
    {
    case eEndOfInput:
        rc = 0;
        break;

    case eEndOfLine:
        LOGMSG (klogInfo, "End Of Line");
        goto redo;

    default:
        rc = RC (rcKFG, rcMgr, rcParsing, rcToken, rcUnexpected);
        break;

        /* a symbol must look like a legal posix path
         * thetokenizer eats things awkwardly for us so we'll muddle through
         *
         * We aren't getting whitespace and we aren't getting all legal path
         * characters as a path at this point so we'll just face it.
         * Gather up all legal symbols and concatenate them.
         *
         * this currently allows the bad syntax of
         *
         *   path/name = "value"; 
         *
         * to be the same as
         *
         *   path 
         *   /
         *      # woo hoo
         *          name
         *   =
         *     "value"
         *
         *       ;
         */
        /* any of these can be the first part of a path -- but not yet */
        /*     case eDecimal: */
        /*     case eHex: */
        /*     case eOctal: */
        /*     case eFloat: */
        /*     case eExpFloat: */
        /*     case eMajMinRel: */
    case eIdent:
    case eName:
        /*     case ePeriod: */
        /*     case eTilde: */
        /*     case eMinus: */
    case eFwdSlash:
        /*     case eDblPeriod: */
        /*     case eEllipses: */

        rc = StringCopy (&symbol, &pb->t.str);
        if (rc == 0)
        {
            pb->symbol = symbol;
            rc = parse_name (pb);
            free ((void*)symbol);
        }
        break;
    }
    return rc;
}


/*
 * Set up the parameter block and start prasing lines
 */
static
rc_t parse_file (KConfig * self, KTokenSource * src)
{
    parse_block pb;
    rc_t rc;

    pb.self = self;
    pb.tz = kPOSIXPathTokenizer;
    pb.src = src;
    memset (&pb.t, 0, sizeof pb.t);
    pb.symbol = NULL;
    pb.value = NULL;
    do 
    {
        rc = parse_nvp (&pb);
        if (rc)
            break;
    }
    while (pb.t.id != eEndOfInput);
    return rc;
}

/* LoadFile
 * loads a configuration file
 */
LIB_EXPORT rc_t CC KConfigLoadFile (KConfig * self, const char * path, const KFile * file)
{
    const KMMap * mm;
    rc_t rc;

    rc = KMMapMakeRead (&mm, file);
    if (rc == 0)
    {
        const void * ptr;
        size_t size;

        rc = KMMapAddrRead (mm, &ptr);
        if (rc == 0)
        {
            rc = KMMapSize (mm, &size);
            if (rc == 0)
            {
                KTokenText tt;
                KTokenSource src;
                String str;
                String spath;

                StringInitCString (&spath, path);
                StringInit (&str, ptr, size, string_len (ptr, size));
                KTokenTextInit (&tt, &str, &spath);
                KTokenSourceInit (&src, &tt);

                parse_file (self, &src);
            }
        }
        KMMapRelease (mm);
    }
    return rc;
}


static
const KFile * get_config_file (const KDirectory * dir, const char ** ppath)
{
    static const char * file_list[] = { "ncbi/config.kfg", "config.kfg", NULL };
    uint32_t idx;

    for (idx = 0; file_list[idx] != NULL; ++idx)
    {
        const KFile * file;
        rc_t rc;

        rc = KDirectoryVOpenFileRead (dir, &file, file_list[idx], NULL);

        if (rc == 0)
        {
            *ppath = file_list[idx];
            return file;
        }
    }
    *ppath = NULL;
    return NULL;
}


static
rc_t load_config_files (KConfig * self, const KDirectory * indir)
{
    static const char * env_list[] = { "VDB_CONFIG", "VDBCONFIG", NULL };
    static char config_path [4096];
    const char * filepath;
    const KDirectory * basedir;
    const KDirectory * subdir;
    const KFile * cfgfile;
    const char * env_var;
    const char * env_val;

    KPathType type;
    uint32_t idx;
    rc_t rc;


    cfgfile = NULL;
    rc = 0;

    if (indir)
        basedir = indir;
    else
        rc = KDirectoryNativeDir ((KDirectory**)&basedir);

    if (rc == 0)
    {
        /* first try to get a file at the passed in parameter */
        if (basedir == indir)
        {
            cfgfile = get_config_file (basedir, &filepath);
            if (cfgfile)
            {
                strncpy (config_path, filepath, sizeof config_path);
            }
        }
        else
        {
            /* second try the environment variables */

            /* loop through those configured by being in the NULL terminated list */
            for (env_var = env_list [idx = 0]; env_var != NULL; env_var = env_list [++idx])
            {
                env_val = getenv (env_var);
                if (env_val != NULL)
                {
                    /* check that value of the environment variable against
                     * the base directory as a path
                     */
                    type = KDirectoryVPathType (basedir, env_val, NULL);
                    switch (type & ~kptAlias)
                    {
                    case kptFile:
                        rc = KDirectoryVOpenFileRead (basedir, &cfgfile, env_val, NULL);
                        if (rc == 0)
                            strncpy (config_path, env_val, sizeof config_path);
                        else
                            cfgfile = NULL; /* assure that cfg is NULL if we failed to open */
                        break;

                    case kptDir:
                        rc = KDirectoryVOpenDirRead (basedir, &subdir, false, env_val, NULL);
                        if (rc == 0)
                        {
                            cfgfile = get_config_file (subdir, &filepath);
                            if (cfgfile)
                            {
                                strncpy (config_path, env_val, sizeof config_path);
                                strncat (config_path, "/", sizeof config_path);
                                strncat (config_path, filepath, sizeof config_path);
                            }
                            KDirectoryRelease (subdir);
                        }
                        break;
                    }
                    /* break out of the loop if we have opened a potential cfgfile */
                    if (cfgfile != NULL)
                        break;
                }

            }

            /* if we still don't have a cfgfile yet we third try to trace back to
             * the directory containing the binary executable or library
             */
            if (cfgfile == NULL)
            {
                KDyld * dyld;

                rc = KDyldMake (&dyld);
                if (rc == 0)
                {
                    rc = KDyldHomeDirectory (dyld, &subdir, (fptr_t)KConfigMake);
                    if (rc == 0)
                    {
                        cfgfile = get_config_file (subdir, &filepath);
                        if (cfgfile)
                            strncpy (config_path, filepath, sizeof config_path);

                        KDirectoryRelease (subdir);
                    }
                    KDyldRelease (dyld);
                }
            }
            /* we could put a loop on default well-known locations here
             * we can make it system dependant by putting an array of locations
             * is a system specific file and reach into that array here like
             * we use the static lists
             */
        }
    }

    /* if we opened basedir release it */
    if (basedir != indir)
        KDirectoryRelease (basedir);

    /* if we have an open KFile try to load it as a config file */
    if (cfgfile != NULL)
    {
        rc = KConfigLoadFile (self, config_path, cfgfile);
        KFileRelease (cfgfile);
    }
    else
        rc = RC ( rcKFG, rcMgr, rcCreating, rcFile, rcNotFound );

    return rc;
}


/* Make
 *  create a process-global configuration manager
 *
 *  "cfg" [ OUT ] - return parameter for mgr
 */
LIB_EXPORT rc_t CC KConfigMake ( KConfig **cfg, const KDirectory * cfgdir )
{
    rc_t rc;
    if ( cfg == NULL )
        rc = RC ( rcKFG, rcMgr, rcCreating, rcParam, rcNull );
    else
    {
        KConfig *mgr = G_kfg;
        if ( mgr != NULL )      /* if already made, just attach */
        {
            KConfigAddRef ( mgr );
            * cfg = mgr;
            return 0;
        }

        mgr = malloc ( sizeof * mgr );
        if ( mgr == NULL )
            rc = RC ( rcKFG, rcMgr, rcCreating, rcMemory, rcExhausted );
        else
        {
            KConfigNode * root;
            String slash;

            CONST_STRING (&slash, "/");

            rc = KConfigNodeMake (&root, &slash);
            if (rc == 0)
            {
                KConfigInit ( mgr, root );

                rc = load_config_files (mgr, cfgdir);

                if ( rc == 0 )
                {
                    G_kfg = mgr;
                    * cfg = mgr;
                    return 0;
                }

                /* the node '/' is already in mgr so doesn't need a separate release */
            }

            KConfigWhack ( mgr );
        }

        * cfg = NULL;
    }

    return rc;
}
