/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 */

#include <kfs/extern.h>
#include <kfs/buffile.h>

/* #include <klib/container.h> */
/* #include <klib/vector.h> */
/* #include <klib/pbstree.h> */
/* #include <klib/text.h> */
#include <klib/log.h>
#include <klib/rc.h>
#include <kfs/file.h>
#include <sysalloc.h>

#include <assert.h>
/* #include <limits.h> */
/* #include <stdio.h> */
#include <stdlib.h>
/* #include <string.h> */

/* #include <klib/container.h> */
/* #include <klib/vector.h> */
/* #include <klib/pbstree.h> */
/* #include <klib/text.h> */
/* #include <kfs/file.h> */
/* #include <assert.h> */
/* #include <limits.h> */
/* #include <stdio.h> */
/* #include <stdlib.h> */
#include <string.h>

/* ======================================================================
 * KBufFile
 *  a file inside an archive
 */
typedef struct KBufFile KBufFile;

/* -----
 * define the specific types to be used in the templatish/inheritancish
 * definition of vtables and their elements
 */
#define KFILE_IMPL struct KBufFile
#include <kfs/impl.h>

static rc_t CC KBufFileDestroy (KBufFile *self);
static struct KSysFile *CC KBufFileGetSysFile (const KBufFile *self,
					    uint64_t *offset);
static rc_t CC KBufFileRandomAccess (const KBufFile *self);
static uint32_t CC KBufFileType (const KBufFile *self);
static rc_t CC KBufFileSize (const KBufFile *self, uint64_t *size);
static rc_t CC KBufFileSetSize (KBufFile *self, uint64_t size);
static rc_t CC KBufFileRead (const KBufFile *self, uint64_t pos,
			  void *buffer, size_t bsize, size_t *num_read);
static rc_t CC KBufFileWrite (KBufFile *self, uint64_t pos, const void *buffer,
			   size_t size, size_t *num_writ);


static const KFile_vt_v1 vtKBufFile_v1 =
{
    /* version */
    1, 1,

    /* 1.0 */
    KBufFileDestroy,
    KBufFileGetSysFile,
    KBufFileRandomAccess,
    KBufFileSize,
    KBufFileSetSize,
    KBufFileRead,
    KBufFileWrite,

    /* 1.1 */
    KBufFileType
};


/*-----------------------------------------------------------------------
 * KBufFile
 *  an archive file including tar and sra
 */
typedef struct Buff
{
    uint64_t	pos;
    uint64_t	top;
    uint64_t	limit;
    uint8_t *	data;
} Buff;

static
rc_t BuffInit ( Buff * self, uint64_t p, void * b, size_t z )
{
    self->pos = p;
    self->top = p;
    self->limit = p + z;
    self->data = b;
    return 0;
}
struct KBufFile
{
    KFile		dad;
    const KFile *	original;
    Buff		buffs[2];
    size_t		bufsize;
    uint8_t		low_page;
    uint8_t		_buf	[1];
};



/* ----------------------------------------------------------------------
 * KBufFileMake
 *  create a new file object
 */
LIB_EXPORT rc_t CC KFileMakeBuf ( const KFile ** self, const KFile * original, size_t gobackmax )
{
    rc_t	rc;
    size_t	use_max;
    union K
    {
    KBufFile *		rw;
    const KBufFile *	ro;
    } pF;

    /* -----
     * we can not accept any of the pointer parameters as NULL
     */
    assert (self != NULL);
    assert (original != NULL);

    /* -----
     * Minimal go back size we accept is 64Kb
     */
    for (use_max = 4096; use_max < gobackmax; use_max += use_max)
	;

    /* -----
     * get space for the object
     */
    pF.rw = malloc (sizeof (KBufFile) - 1 + use_max + use_max);
    if (pF.rw == NULL)	/* allocation failed */
    {
	/* fail */
	rc = RC (rcFS, rcFile, rcConstructing, rcMemory, rcExhausted);
    }
    else
    {
	rc = KFileInit (&pF.rw->dad,		/* initialize base class */
			(const KFile_vt*)&vtKBufFile_v1,/* VTable for KBufFile */
			true,				/* read allowed */
			false);				/* write disallowed */
	if (rc == 0)
	{
	    /* succeed */
	    pF.rw->bufsize = use_max;
	    pF.rw->original = original;
	    pF.rw->low_page = 0;
	    rc = BuffInit (pF.rw->buffs, 0, pF.rw->_buf, use_max);
	    if (rc == 0)
	    {
		rc = BuffInit (pF.rw->buffs+1, use_max, pF.rw->_buf+ use_max,
			       use_max);
		if (rc == 0)
		{
		    *self = &pF.ro->dad;
		    return 0;
		}
	    }
	}
	free (pF.rw);
    }
    return rc;
}

/* ----------------------------------------------------------------------
 * Destroy
 *
 */
static
rc_t CC KBufFileDestroy (KBufFile *self)
{
    rc_t rc = 0;
    if (self != NULL)
    {
	rc = KFileRelease (self->original);
	if (rc == 0)
	    free (self);
    }
    return rc;
}

/* ----------------------------------------------------------------------
 * GetSysFile
 *  returns an underlying system file object
 *  and starting offset to contiguous region
 *  suitable for memory mapping, or NULL if
 *  no such file is available.
 *
 * It is unlikely that this would have been created over a KFile
 * that would allow this but we'll let it decide.
 */

static
struct KSysFile *CC KBufFileGetSysFile (const KBufFile *self, uint64_t *offset)
{
    return KFileGetSysFile (self->original, offset);
}

/* ----------------------------------------------------------------------
 * RandomAccess
 *
 *  returns 0 if random access, error code otherwise
 *
 * Update needs to be able to seek both original and copy while read
 * only needs to be able to seek the original.
 */
static
rc_t CC KBufFileRandomAccess (const KBufFile *self)
{
    return 0; /* its limited but it is random access */
}

/* ----------------------------------------------------------------------
 * Type
 *  returns a KFileDesc
 *  not intended to be a content type,
 *  but rather an implementation class
 */
static
uint32_t CC KBufFileType (const KBufFile *self)
{
    return KFileType (self->original);
}


/* ----------------------------------------------------------------------
 * Size
 *  returns size in bytes of file
 *
 *  "size" [ OUT ] - return parameter for file size
 */
static
rc_t CC KBufFileSize (const KBufFile *self, uint64_t *size)
{
    rc_t	rc;
    uint64_t	fsize;

    assert (self != NULL);
    assert (size != NULL);

    rc = KFileSize (self->original, &fsize);

    if (rc == 0)
    {
	/* success */
	*size = fsize;
    }
    /* pass along RC value */
    return rc;
}

/* ----------------------------------------------------------------------
 * SetSize
 *  sets size in bytes of file
 *
 *  "size" [ IN ] - new file size
 */
static
rc_t CC KBufFileSetSize (KBufFile *self, uint64_t size)
{
    return RC (rcFS, rcFile, rcUpdating, rcSelf, rcUnsupported);
}

/* ----------------------------------------------------------------------
 * Read
 *  read file from known position
 *
 *  "pos" [ IN ] - starting position within file
 *
 *  "buffer" [ OUT ] and "bsize" [ IN ] - return buffer for read
 *
 *  "num_read" [ OUT, NULL OKAY ] - optional return parameter
 *  giving number of bytes actually read
 */
static
rc_t CC KBufFileRead	(const KBufFile *cself,
			 uint64_t pos,
			 void *_buffer,
			 size_t bsize,
			 size_t *_num_read)
{
    KBufFile * 	self;
    Buff *	low;
    Buff *	high;
    uint8_t *	buffer;
    uint64_t	new_top;
    uint64_t	top;
    size_t	this_read;
    size_t 	to_read;
    size_t 	to_copy;
    size_t	num_read;
    rc_t	rc = 0;


    /* -----
     */
    assert (cself != NULL);
    assert (_buffer != NULL);

    self = (KBufFile *)cself;

    rc = 0;
    low = self->buffs + self->low_page;
    high = self->buffs + (1 ^ self->low_page);
    this_read = 0xFFFF; /* just setting to a non-zero value as a flag not EOF */

    if (_num_read == NULL)
	_num_read = &to_read; /* just pick a safe spot */

   /* see if our cache is already past this requested position */
    if (low->pos > pos)
    {
	/* we are so fail */
	return RC (rcFS, rcFile, rcReading, rcOffset, rcOutofrange);
    }
    /* asked to do nothing, so do nothing */
    if (bsize == 0)
    {
	*_num_read = 0;
	return 0;
    }
    buffer = _buffer; 	/* need a pointer to byte not pointer to void */
    top = pos + bsize;	/* top is theposition one past where we need to read */
    *_num_read = num_read = 0;

    /* do we need to fill in the low buffer? */
    if ((low->top < top) && (low->top < low->limit))
    {
	for ( to_read = (size_t)( ( ( top < low->limit ) ? top : low->limit ) - low->top );
	      to_read > 0;
	      to_read -= this_read )
	{
	    rc = KFileRead (self->original, low->top, low->data + low->top - low->pos,
			    to_read, &this_read);
	    if (this_read == 0)
		goto end_or_fail;
	    low->top += this_read;
	    if (rc != 0)
		goto end_or_fail;
	}
    }
    /* do we need to fill in the high buffer? */
    if ((high->top < top) && (high->top < high->limit))
    {
	for ( to_read = (size_t)( ( ( top < high->limit ) ? top : high->limit ) - high->top );
	      to_read > 0;
	      to_read -= this_read)
	{
	    rc = KFileRead (self->original, high->top, high->data + high->top - high->pos,
			    to_read, &this_read);
	    if (this_read == 0)
		goto end_or_fail;
	    high->top += this_read;
	    if (rc != 0)
		goto end_or_fail;
	}
    }
    /* read new buffers up to pos */
    if (pos > high->limit)
    {
	/* more than a partial page */
	while (pos > high->limit + self->bufsize)
	{
	    /* switch low and high */
	    high = self->buffs + self->low_page;
	    self->low_page ^= 1;
	    low = self->buffs + self->low_page;
	    high->top = high->pos = low->limit;
	    high->limit = high->pos + self->bufsize;
	    for (to_read = self->bufsize; to_read > 0; to_read -= this_read)
	    {
		rc = KFileRead (self->original, high->top, high->data,
			    self->bufsize, &this_read);
		if (this_read != 0)
		    high->top += this_read;
		else
		    goto end_or_fail;
		if (rc != 0)
		    goto end_or_fail;
	    }
	    
	} /* while (pos > high->limit + self->bufsize) */
	/* a partial page left before pos */
	self->low_page ^= 1;
	low = self->buffs + self->low_page;
	high = self->buffs + (1 ^ self->low_page);
	high->top = high->pos = low->limit;
	high->limit = high->pos + self->bufsize;
	for ( to_read = (size_t)( ( top > high->limit ) ? self->bufsize : top - high->pos );
	      to_read > 0;
	      to_read -= this_read )
	{
	    rc = KFileRead (self->original, high->top, high->data,
			    self->bufsize, &this_read);
	    if (this_read != 0)
	    {
		high->top += this_read;
	    }
	    else
		goto end_or_fail;
	    if (rc != 0)
		goto end_or_fail;
	}
    } /* (pos > high->limit) */
end_or_fail:
    if (top <= low->top) /* reading all from lower cache page */
    {
	memcpy (buffer, low->data + pos - low->pos, bsize);
	*_num_read = bsize;
	return rc;
    }
    if ((pos < low->top) && (low->top > low->pos)) /* reading some from lower cache page */
    {
	to_copy = (size_t)( low->top - pos );
	memcpy ( buffer, low->data + pos - low->pos, to_copy );
	num_read = to_copy;
    }
    if ((top <= high->top) && (high->top > high->pos)) /* reading up to end from upper cache page */
    {
	to_copy = (size_t)( top - ( ( pos >= high->pos ) ? pos : high->pos ) );
	memcpy ( buffer + num_read, high->data + pos - high->pos, to_copy );
	*_num_read = num_read + to_copy;
	return rc;
    }
    if ((pos < high->top) && (high->top > high->pos)) /* reading some from upper cache page but not to end */
    {
	to_copy = (size_t)( high->top - ( ( pos >= high->pos ) ? pos : high->pos ) );
	memcpy (buffer + num_read, high->data + pos - high->pos, to_copy);
	num_read += to_copy;
    }

    if ((high->top == high->limit) && (top > high->top))
    {
	/* after all that if we are here, we have read up to POS with more to read. */
	new_top = high->top;
        if ((rc == 0)&&(this_read != 0)) /* don't read any if we've hit failure or EOF */
	{
	    for ( to_read = (size_t)( top - new_top );
		      to_read > 0;
		      to_read -= this_read )
	    {
		rc = KFileRead (self->original, new_top, buffer + new_top - pos,
				bsize - num_read, &this_read);
		if (this_read != 0)
		{
		    num_read += this_read;
		    new_top += this_read;
		}
		else
		    break;
		if (rc != 0)
		    break;
	    }
	}
	if (new_top != (high->top > high->pos) ? high->top : low->top) /* fix up buffers */
	{
	    
	    if (new_top > high->top + self->bufsize) /* just set new low */
	    {
		low->pos = new_top - self->bufsize;
		high->top = low->top = low->limit = high->pos = new_top;
		high->limit = new_top + self->bufsize;
		memcpy (low->data, buffer + low->pos - pos, self->bufsize);
	    }
	    else
	    {
		/* switch low and high */
		high = self->buffs + self->low_page;
		self->low_page ^= 1;
		low = self->buffs + self->low_page;
		high->pos = low->limit;
		high->limit = high->pos + self->bufsize;
		high->top = new_top;;
		memcpy ( high->data, buffer + high->pos - pos, (size_t)( new_top - high->pos ) );
	    }
	}
    }


    *_num_read = num_read;
    return rc;
}

/* ----------------------------------------------------------------------
 * Write
 *  write file at known position
 *
 *  "pos" [ IN ] - starting position within file
 *
 *  "buffer" [ IN ] and "size" [ IN ] - data to be written
 *
 *  "num_writ" [ OUT, NULL OKAY ] - optional return parameter
 *  giving number of bytes actually written
 *
 * Unsupported
 */
static
rc_t CC KBufFileWrite (KBufFile *self, uint64_t pos,
		       const void *buffer, size_t bsize,
		       size_t *num_writ)
{
    assert (self != NULL);
    assert (buffer != NULL);
    assert (num_writ != NULL);
    assert (bsize != 0);

    *num_writ = 0;
    return RC (rcFS, rcFile, rcWriting, rcSelf, rcUnsupported);
}

/* end of file arc.c */

