/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_kfs_toc_priv_h_
#define _h_kfs_toc_priv_h_

#include <klib/container.h>
#include <klib/text.h>
#include <kfs/toc.h>
#include <atomic.h>

#ifdef _DEBUGGING
#define FUNC_ENTRY() DBGMSG (DBG_KFS, DBG_FLAG(DBG_KFS_TOCENTRY), ("Enter: %s\n", __func__))
#define TOC_DEBUG(msg) DBGMSG (DBG_KFS, DBG_FLAG(DBG_KFS_TOC), msg)
#else
#define FUNC_ENTRY()
#define TOC_DEBUG(msg)
#endif



/* arbitrary number that was originially set much much higher than expected needs. */
#define	KARC_LINK_RESOLVE_LOOPMAX		(16)

struct KTocEntry
{
    /* -----
     * In C putting this first allows a weak form of polymorphism for
     * a entry class.  A pointer to a BSTNode can be an alias to a pointer to 
     * this form of node.
     */
    BSTNode		node;
    /* -----
     * The name of the directory/file/link that this Entry represents.  It is not
     * the full path which must be assembled going up the tree.
     *
     * Be careful what String functions you try as some want to reallocate memory.
     */
    String		name;

    /* -----
     * Unix style file modification time (same as a 64-bit time_t)
     */
    KTime_t		mtime;

    /* -----
     * Unix style file access
     */
    uint32_t		access;

    /* -----
     * what type of entry is this?
     */
    KTocEntryType	type;
    union KTocEntryUnion
    {
	struct KTocEntryDir
	{
	    /* -----
	     * Tree of directories, files and links below this directory
	     */
	    BSTree	tree;	    
	} dir;
	struct KTocEntryFile
	{
	    /*
	     * how long is this file?
	     */
	    uint64_t	file_size;
	    /* 
	     * location within the archive where this file's data starts
	     */
	    uint64_t	archive_offset;
	} contiguous_file;
	struct KTocEntryChunkFile
	{
/* KEEP the first two fields here match the first two fields above
 * file_size and archive_offset so the chunked is sort of a derived type */
	    /*
	     * Size of the whole file including all chunks and any zero byte holes
	     * between chunks
	     */
	    uint64_t	file_size;
	    /* 
	     * location within the archive where this file's data starts if viewws
	     * as a virtual "*.sra"
	     */
	    uint64_t	archive_offset;

	    KTocChunk *	chunks;			/* this will piggy back in the same allocation */
	    uint32_t	num_chunks;	    
	} chunked_file;
	struct KTocEntrySoftLink
	{
	    String	link_path;
	} symbolic_link;
	struct KTocEntryHardLink
	{
	    struct KTocEntry * ref;
	    /* -----
	     * This approach can lead to failures if the target of the link is deleted and this
	     * reference is left dangling.
	     */
	} hard_link;
    } u;
};

typedef struct KTocEntryIndex KTocEntryIndex;
struct KTocEntryIndex
{
    BSTNode	node;
    KTocEntry *	entry;
    String	fullpath;
};


/* ----------------------------------------------------------------------
 * KTocEntryParam
 *
 * An internal structure used only with in this compilation unit / module
 */
/* ----------------------------------------------------------------------
 * NOTE: This oddness of specific to general to specific can probably be 
 * refactored into specific functions (the published API) that uses common
 * functions instead of a common target function.
 */
typedef struct KTocEntryParam	KTocEntryParam;
struct KTocEntryParam
{
    KTocEntryType	type;
    KTime_t		mtime;
    uint32_t		access;
    union
    {
	struct
	{
	    uint64_t		size;
	    uint64_t		source_position;
	} file;
	struct
	{
	    uint64_t		size;
	    uint32_t		num_chunks;
	    const KTocChunk *	chunks;
	} chunked;
	struct
	{
	    const char *	targ;
	} hardlink;
	struct
	{
	    const char *	targ;
	} softlink;
    } u;
};

/* ----------------------------------------------------------------------
 * KArcTOCNodeWhack
 * [IN] BSTNode * 		n	
 * [IN] void *    		data	
 *
 * This function matches the signature for a function pointer in a 
 * call to BSTreeWhack().  It ignores the data pointer second parameter
 * which can safely be called with NULL. 
 */
void	CC KTocEntryWhack	(BSTNode * n, void * data);
int	CC KTocEntryCmpVoid (const void * item, const BSTNode * node);
rc_t	KTocEntryDelete	(KTocEntry * self);
int	CC KTocEntryCmp2 (const BSTNode * n, const BSTNode * p);
rc_t	KTocEntryNewDirectory	(KTocEntry ** new_entry,
					 const char * name,
					 size_t name_size,
					 KTime_t mtime,
					 uint32_t access);
rc_t		KTocEntryNewFile	(KTocEntry ** new_entry,
					 const char * name,
					 size_t name_size,
					 KTime_t mtime,
					 uint32_t access,
					 uint64_t offset,
					 uint64_t size);
rc_t		KTocEntryNewChunked	(KTocEntry ** new_entry,
					 const char * name,
					 size_t name_size,
					 KTime_t mtime,
					 uint32_t access,
					 uint64_t size,
					 const KTocChunk * chunks,
					 uint32_t num_chunks);
rc_t		KTocEntryNewSoft	(KTocEntry ** new_entry,
					 const char * name,
					 size_t name_size,
					 KTime_t mtime,
					 uint32_t access,
					 const char * link,
					 size_t link_size);
rc_t		KTocEntryNewHard	(KTocEntry ** new_entry,
					 const char * name,
					 size_t name_size,
					 KTime_t mtime,
					 uint32_t access);


typedef struct	KTocEntryStack 	KTocEntryStack;
rc_t KTocEntryStackNew(KTocEntryStack ** self);
rc_t KTocEntryStackDel(KTocEntryStack * self);
rc_t KTocEntryStackDepth(KTocEntryStack * self, int * depth);
rc_t KTocEntryStackPush(KTocEntryStack * self, KTocEntry * pathaddr);
rc_t KTocEntryStackPop(KTocEntryStack * self, KTocEntry ** pathaddr);

#if _LOGGING
const char * KTocEntryTypeGetString(KTocEntryType t);
const char * get_mode_string (KCreateMode mode);
#endif

typedef struct KTocEntryPersistWriteFuncData
{
    uint8_t *	buffptr;
    uint8_t *	limit;
} KTocEntryPersistWriteFuncData;
rc_t CC KTocEntryPersistWriteFunc (void * param, const void * buffer, size_t size,
				size_t * num_writ);

rc_t KTocEntryPersistNodeDir (void *param, const KTocEntry * n,
			      size_t * num_writ, 
			      PTWriteFunc write, void * write_param);

/* ======================================================================
 * KToc struct
 */
struct KArcDir;
struct KToc
{
    KTocFSType	arctype;
    KSRAFileAlignment alignment;
    /* -----
     * File containing the archive: details of its type hidden behind KFile.
     * We should be able to do an archive in an archive by having this KFile
     * point to a KArcFile with no problems.
     */
    union
    {
	const void *		v;
	struct KFile const *f;
	const KDirectory * 	d;
    } archive;

    /* -----
     * KArcDir that implements the KDirectory interface for the contents of 
     * the file opened as archive.  This link is needed to get to the 
     * KDirectory interface from a KFile symbolic link.
     *
     * This reference is not counted in refcount as it is a link back to the
     * tracked reference from the KArcDir to the KARcTOC.
     */
/*     const struct KArcDir *	karchive; */

    /* -----
     * count of references to this opened archive
     */
    atomic32_t		refcount;

    /* -----
     * A entry outside of a tree: instantiated as a special KTocEntry that is part of
     * no BSTree with type set to ktocentrytype_dir (directory).
     * Does not use the base dir BSTNode but makes processing easier to have a 
     *"entry" type...
     */
    KTocEntry		entry;

    /* -----
     * These next two are used only in the KFile interface on top of a KToc
     * on a KDirectory.  This needs to be rewritten as an interface so these
     * don't exist in all KToc implmentations.
     *
     * Index  The main tree in 'entry' is sorted by names with in directories.
     * This index is a directory less tree sorted by offset with in the archive.
     *
     * This is not currently used in tar file versions of the TOC archive.
     *
     * Each entry is of type KTocEntryIndex;
     */
    BSTree		offset_index;
    KSraHeader *	header;


    /* -----
     * This is the full path of the archive file as used to open it as a KFile.
     */
    String		path;

};


rc_t KTocInflatePBSTree (KToc * self, const
void * treestart, uint32_t maxsize,
			 uint64_t offset, bool rev, const char * path);

#endif /* #ifndef _h_kfs_toc_priv_h_ */
/* end of file */
