/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <klib/extern.h>
#include "log-priv.h"
#include <klib/writer.h>
#include <klib/log.h>
#include <klib/text.h>
#include <klib/sort.h>
#include <klib/rc.h>
#include <sysalloc.h>
#include <os-native.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


/*--------------------------------------------------------------------------
 * LogLevel
 * defaults to the error level which is the lowest error level filtering
 * warning and informational messages
 */
KLogLevel G_log_level = klogErr;


/* LogLevelAdjust
 *  adjust process-global log level by an offset from current
 */
LIB_EXPORT rc_t CC KLogLevelAdjust( int32_t adjust )
{
    return KLogLevelSet((KLogLevel)((int32_t)KLogLevelGet() + adjust));
}


/* Get
 *  retrieve current process-global log level
 */
#ifdef LogLevelGet
#undef LogLevelGet
#endif
LIB_EXPORT KLogLevel CC KLogLevelGet ( void )
{
    return G_log_level;
}


/* Set
 *  set process-global log level
 */
#ifdef LogLevelSet
#undef LogLevelSet
#endif
LIB_EXPORT rc_t CC KLogLevelSet ( KLogLevel lvl )
{
    /* -----
     * Don't allow an out of range set
     */
    if (/*(lvl >= klogLevelMin)&&*/(lvl <= klogLevelMax))
    {
        G_log_level = lvl;
        return 0;
    }
    else
        return RC (rcApp, rcLog, rcUpdating, rcRange, rcInvalid);
}


/*
 * These need to be kept in sync with type LogLevel
 */
static const char * logLevelParamStrings [] = 
{
    "fatal",
    "sys",
    "int",
    "err",
    "warn",
    "info"
};

LIB_EXPORT const char ** CC KLogGetParamStrings ( void )
{
    return logLevelParamStrings;
}


/* LogLevelExplain
 *  inserts level description
 */
LIB_EXPORT rc_t CC KLogLevelExplain ( KLogLevel lvl,
                        char *buffer, size_t bsize, size_t *num_writ )
{
    static const char undefined[] = "undefined";
    size_t len;
    const char *t;

    /* catch unaccounted for or bad changes in log.h
     * if the asserts fail then the table above and the rest
     * of the function will fail */
    assert (klogLevelMin == 0);
    assert (klogLevelMin == klogFatal);
    assert ((klogFatal + 1) == klogSys);
    assert ((klogSys + 1) == klogInt);
    assert ((klogInt + 1) == klogErr);
    assert ((klogErr + 1) == klogWarn);
    assert ((klogWarn + 1) == klogInfo);
    assert ((klogInfo) == klogLevelMax);

    if (/*(lvl < klogLevelMin) ||*/ (lvl > klogLevelMax))
        t = undefined;
    else
        t = logLevelParamStrings[lvl];

    len = strlen (t);
    if (len > bsize)
    {
        if (num_writ)
            *num_writ = 0;
        return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );
    }
    if (num_writ)
        *num_writ = strlen (t);
    strncpy (buffer, t, bsize);
    return 0;
    
}


/*--------------------------------------------------------------------------
 * Log
 */


static rc_t s_LastErrorRc;


/* Init
 *  initialize the logging module with executable identity and version,
 *  plus logging level. actual behavior is determined by the logging
 *  implementation.
 *
 *  "ident" [ IN ] - identity of executable
 *
 *  "vers" [ IN ] - 4-part version code: 0xMMmmrrrr, where
 *      MM = major release
 *      mm = minor release
 *    rrrr = bug-fix release
 *
 *  "lvl" [ IN ] - a LogLevel specifying cutoff point for logs, such
 *  that any log generated with a level above "lvl" will be ignored.
 */
LIB_EXPORT rc_t CC KLogInit ( void )
{
    rc_t rc;

    G_log_level = klogWarn;

    rc = KLogHandlerSetStdErr();

    if (rc == 0)
        rc = KLogLibHandlerSet (NULL, NULL);

    return rc;
}


/* LastErrorCode
 * LastErrorCodeReset
 *  return or clear the last reported rc_t with log level LogErr or above
 */
LIB_EXPORT rc_t CC KLogLastErrorCode ( void )
{
    return s_LastErrorRc;
}

LIB_EXPORT void CC KLogLastErrorCodeReset ( void )
{
    s_LastErrorRc = 0;
}


/* LastErrorCodeSet
 *  set the last reported rc_t with log level LogErr or above
 */
LIB_EXPORT void CC KLogLastErrorCodeSet ( KLogLevel lvl, rc_t rc )
{
    if ( lvl <= klogErr )
        s_LastErrorRc = rc;
}


/* KLogFlushLine
 *  trims excess trailing newlines and spaces
 *  ensures single trailing newline
 *  writes line to log
 */
rc_t KLogFlushLine ( KWrtHandler * handler, char *buffer, size_t bsize, size_t total )
{
    size_t num_written;
    size_t remaining;
    rc_t rc;

    assert (handler);
    assert (buffer);

    /* strip trailing newlines */
    while ( total > 0 && isblank ( buffer [ total - 1 ] ) )
        -- total;

    /* add single, trailing newline */
    if ( total == bsize )
        return RC ( rcApp, rcLog, rcWriting, rcBuffer, rcInsufficient );
    buffer [ total ] = '\n';

    /* add trailing NUL as courtesy */
    if ( ++ total < bsize )
        buffer [ total ] = 0;

    rc = 0;

    /* log the string */
    for (remaining = total; remaining > 0; remaining -= num_written)
    {
        rc = handler->writer (handler->data, buffer, remaining, &num_written);
        if (rc)
            break;
    }
    
    return rc;
}


/* InsertSpace
 *  inserts a division after current text
 *
 *  "spacer" [ IN, NULL OKAY ] - optional characters to insert
 */
rc_t LogInsertSpace ( const char *spacer,
    char *buffer, size_t bsize, size_t *num_writ )
{
    int len;

    if ( spacer == NULL )
    {
        if ( bsize < 2 )
            return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );
        buffer [ 0 ] = ' ';
        buffer [ 1 ] = 0;
        * num_writ = 1;
        return 0;
    }

    len = snprintf ( buffer, bsize, spacer );

    * num_writ = len;

    if ( len < 0 || (size_t)len >= bsize )
    {
        if ( len < 0 )
            * num_writ = 0;
        return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );
    }

    return 0;
}


/* AppendMsg
 *  appends a message
 */
rc_t LogAppendMsg ( const char *msg,
    char *buffer, size_t bsize, size_t *num_writ )
{
    int len = snprintf ( buffer, bsize, msg );
    * num_writ = len;
    if ( len < 0 || ( size_t ) len >= bsize )
    {
        if ( len < 0 )
            * num_writ = 0;
        return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );
    }
    return 0;
}

/*--------------------------------------------------------------------------
 * nvp - name/value pair
 */

static
int CC nvp_cmp ( const void *a, const void *b, void * ignored )
{
    int i;
    const char *key = a;
    const char *name = ( ( const nvp* ) b ) -> name;

    for ( i = 0; key [ i ] == name [ i ]; ++ i )
    {
        /* key should always terminate in right-paren */
        if ( key [ i ] == 0 )
            return -1;
    }

    /* treat right-paren as key terminator */
    if ( key [ i ] != ')' )
        return key [ i ] - name [ i ];

    return 0 - name [ i ];
}

static
int CC nvp_sort ( const void *a, const void *b, void * ignored )
{
    const nvp *left = a;
    const nvp *right = b;
    return strcmp ( left -> name, right -> name );
}




/* logsubstituteparams
 *  perform param substitution by name
 */
static
rc_t CC logsubstituteparams ( const char *msg, uint32_t argc, const nvp argv [],
    char *buffer, size_t bsize, size_t *num_writ )
{
    rc_t rc;
    size_t i, sz;
    for ( rc = 0, i = 0, sz = 0; msg [ i ] != 0; ++ sz, ++ i )
    {
        /* copy msg character */
        if ( sz < bsize )
            buffer [ sz ] = msg [ i ];

        /* detect substitution param */
        if ( msg [ i ] == '$' && msg [ i + 1 ] == '(' )
        {
            const char *value;

            /* find param by name */
            const nvp *arg = ( const nvp* )
                kbsearch ( & msg [ i + 2 ], argv, argc, sizeof argv [ 0 ], nvp_cmp, NULL );
            if ( arg == NULL )
            {
                rc = RC ( rcApp, rcLog, rcInserting, rcParam, rcNotFound );
                break;
            }

            /* substitute param value */
            for ( value = arg -> value; value [ 0 ] != 0; ++ value, ++ sz )
            {
                if ( sz < bsize )
                    buffer [ sz ] = value [ 0 ];
            }

            /* compensate for outer loop's increment */
            -- sz;

            /* advance past param token */
            i += strlen ( arg -> name ) + 2;
            assert ( msg [ i ] == ')' );
        }
    }

    * num_writ = sz;

    if ( rc != 0 )
        return rc;

    if ( sz >= bsize )
        return RC ( rcApp, rcLog, rcInserting, rcBuffer, rcInsufficient );

    buffer [ sz ] = 0;
    return 0;
}


/* logappendpmsg
 *  appends a message and parameters
 */
static
int CC logmatchname ( const char *fmt, char *pdata )
{
    int i;
    for ( i = 0; fmt [ i ] == pdata [ i ]; ++ i )
    {
        switch ( fmt [ i ] )
        {
        case '=':
            if ( i > 0 )
            {
                pdata [ i ] = 0;
                return i;
            }
        case 0:
        case ',':
            return -1;
        }
    }
    return 0;
}

static
rc_t CC logtokenizeparams ( const char *fmt, char *pdata,
    uint32_t *argcp, nvp argv [], uint32_t arg_max )
{
    /* now split into parameters */
    uint32_t argc;
    int len = logmatchname ( fmt, pdata );
    if ( len ++ <= 0 )
        return RC ( rcApp, rcLog, rcTokenizing, rcParam, rcInvalid );

    argv [ 0 ] . name = pdata;
    argv [ 0 ] . value = & pdata [ len ];
    for ( argc = 1, fmt += len; ; ++ argc )
    {
        /* find end of formal param fmt */
        char *end = strchr ( fmt, ',' );
        if ( end == NULL )
            break;
        fmt = end + 1;

        /* find end of param value by
           finding beginning of next name */
        end = strchr ( argv [ argc - 1 ] . value, ',' );
        while ( end != NULL )
        {
            len = logmatchname ( fmt, end + 1 );
            if ( len > 0 )
                break;
            if ( len < 0 )
                return RC ( rcApp, rcLog, rcTokenizing, rcParam, rcInvalid );
            end = strchr ( end + 1, ',' );
        }

        /* detect badly formed parameter value or va_list */
        if ( end ++ == NULL )
            return RC ( rcApp, rcLog, rcTokenizing, rcParam, rcInvalid );

        /* detect too many parameters */
        if ( argc == arg_max )
            return RC ( rcApp, rcLog, rcTokenizing, rcRange, rcExcessive );

        /* account for NUL */
        ++ len;

        /* NUL terminate value */
        end [ -1 ] = 0;

        /* record param name and value addresses */
        argv [ argc ] . name = end;
        argv [ argc ] . value = end + len;

        /* advance over formal param name */
        fmt += len;
    }

    /* if multiple parameters, order by name */
    if ( argc > 1 )
        ksort ( argv, argc, sizeof argv [ 0 ], nvp_sort, NULL );

    * argcp = argc;
    return 0;
}

/* ProduceMsg
 *  given a formatted msg string and parameters,
 *  produce a normalized message string and
 *  parse parameters into name,value pairs
 *
 *  "msg" [ IN ] - formatted message string
 *
 *  "fmt" [ IN ] and "args" [ IN ] - printf-style parameters
 *
 *  "argc" [ OUT ] - return parameter for number of
 *  name,value pairs
 *
 *  "argv" [ OUT ] and "asize" [ IN ] - output array
 *  of nvp where "asize" is size in bytes of "argv"
 *
 *  "result" [ OUT ] - pointer to resultant message
 *  with parameters substituted. normally, should be
 *  the value of "buffer" passed in. if not, then it
 *  was allocated with malloc and should be freed.
 *
 * NB - only free "result" after finished with argv
 *
 *  "offset" [ IN ] - starting offset into buffer
 *  allows existing data in buffer to be passed to
 *  allocation if needed
 *
 *  "buffer" [ OUT ] and "bsize" [ IN, OUT ] -
 *  caller-supplied buffer for assembling text
 *  on output "bsize" is size of "result"
 *
 *  "num_writ" [ OUT ] - number of bytes written to "result"
 */
rc_t LogProduceMsg ( const char *msg, const char *fmt, va_list args,
    uint32_t *argc, nvp argv [], size_t asize, char *pbuffer, size_t pbsize,
    size_t offset, char *buffer, size_t bsize, size_t *num_writ )
{
    rc_t rc;

    /* first, create the parameter list */ 
#if USE_OLD_FORMATTING
    int len = vsnprintf ( pbuffer, pbsize, fmt, args );
#else
    int len = vknprintf ( pbuffer, pbsize, fmt, args );
#endif

    if ( len < 0 || ( size_t ) len >= pbsize )
        return RC ( rcApp, rcLog, rcTokenizing, rcBuffer, rcInsufficient );

    /* account for trailing NUL and 2 bytes after message */
    len += 3;

    /* tokenize the parameters into name/value pairs */
    rc = logtokenizeparams ( fmt, pbuffer, argc, argv, (uint32_t)( asize / sizeof argv [ 0 ] ) );
    if ( rc == 0 )
    {
        /* substitute parameters into message */
        rc = logsubstituteparams ( msg, * argc, argv,
            & buffer [ offset ], bsize - offset - len, num_writ );
    }

    return rc;
}


/* pLogMsg
 *  creates an info node with parameters
 *  <info time="" message="" ... />
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "msg" [ IN ] - NUL terminated string with named parameters
 *
 *  "fmt" [ IN ] - NUL terminated format string, behaves according
 *  to standard printf-style formatting
 */
LIB_EXPORT rc_t CC pLogMsg ( KLogLevel lvl, const char *msg, const char *fmt, ... )
{
    rc_t rc;
    va_list args;

    if ( lvl > G_log_level )
        return 0;

    va_start ( args, fmt );
    rc = vLogMsg ( lvl, msg, fmt, args );
    va_end ( args );

    return rc;
}

LIB_EXPORT rc_t CC pLogLibMsg ( KLogLevel lvl, const char *msg, const char *fmt, ... )
{
    rc_t rc;
    va_list args;

    if ( lvl > G_log_level )
        return 0;

    va_start ( args, fmt );
    rc = vLogLibMsg ( lvl, msg, fmt, args );
    va_end ( args );

    return rc;
}


/* pLogErr
 *  reports an error to log file
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "status" [ IN ] - return code to be decoded
 *
 *  "msg" [ IN ] - NUL terminated message string with named parameters
 *
 *  "fmt" [ IN ] - NUL terminated format string, behaves according
 *  to standard printf-style formatting
 */
LIB_EXPORT rc_t CC pLogErr ( KLogLevel lvl, rc_t status, const char *msg, const char *fmt, ... )
{
    rc_t rc;
    va_list args;

    if ( lvl > G_log_level )
        return 0;

    va_start ( args, fmt );
    rc = vLogErr ( lvl, status, msg, fmt, args );
    va_end ( args );

    return rc;
}

LIB_EXPORT rc_t CC pLogLibErr ( KLogLevel lvl, rc_t status, const char *msg, const char *fmt, ... )
{
    rc_t rc;
    va_list args;

    va_start ( args, fmt );
    rc = vLogLibErr ( lvl, status, msg, fmt, args );
    va_end ( args );

    return rc;
}

LIB_EXPORT KWrtHandler G_log_handler;
LIB_EXPORT KWrtHandler G_log_lib_handler;

#ifdef KLogHandlerSetStdOut
#undef KLogHandlerSetStdOut
#endif
LIB_EXPORT rc_t CC KLogHandlerSetStdOut ( void )
{
    return KLogHandlerSet(DefaultWriter,DefaultWriterDataStdOut);
}
#ifdef KLogLibHandlerSetStdOut
#undef KLogLibHandlerSetStdOut
#endif
LIB_EXPORT rc_t CC KLogLibHandlerSetStdOut ( void )
{
    return KLogLibHandlerSet(DefaultWriter,DefaultWriterDataStdOut);
}

#ifdef KLogHandlerSetStdErr
#undef KLogHandlerSetStdErr
#endif
LIB_EXPORT rc_t CC KLogHandlerSetStdErr ( void )
{
    return KLogHandlerSet(DefaultWriter,DefaultWriterDataStdErr);
}

#ifdef KLogLibHandlerSetStdErr
#undef KLogLibHandlerSetStdErr
#endif
LIB_EXPORT rc_t CC KLogLibHandlerSetStdErr ( void )
{
    return KLogLibHandlerSet(DefaultWriter,DefaultWriterDataStdErr);
}


#ifdef KLogHandlerSet
#undef KLogHandlerSet
#endif
LIB_EXPORT rc_t CC KLogHandlerSet ( KWrtWriter writer, void * writer_data )
{
    G_log_handler.writer = writer;
    G_log_handler.data = writer_data;
    return 0;
}

#ifdef KLogLibHandlerSet
#undef KLogLibHandlerSet
#endif
LIB_EXPORT rc_t CC KLogLibHandlerSet ( KWrtWriter writer, void * writer_data )
{
    G_log_lib_handler.writer = writer;
    G_log_lib_handler.data = writer_data;
    return 0;
}

#ifdef KLogDataGet
#undef KLogDataGet
#endif
LIB_EXPORT void * CC KLogDataGet ( void )
{
    return ( KLogHandlerGet()->data );
}

#ifdef KLogLibDataGet
#undef KLogLibDataGet
#endif
LIB_EXPORT void * CC KLogLibDataGet ( void )
{
    return ( KLogLibHandlerGet()->data );
}

#ifdef KLogWriterGet
#undef KLogWriterGet
#endif
LIB_EXPORT KWrtWriter CC KLogWriterGet ( void )
{
    return ( KLogHandlerGet()->writer );
}

#ifdef KLogLibWriterGet
#undef KLogLibWriterGet
#endif
LIB_EXPORT KWrtWriter CC KLogLibWriterGet ( void )
{
    return ( KLogLibHandlerGet()->writer );
}

#ifdef KLogHandlerGet
#undef KLogHandlerGet
#endif
LIB_EXPORT KWrtHandler * CC KLogHandlerGet ( void )
{
    return ( &G_log_handler );
}

#ifdef KLogLibHandlerGet
#undef KLogLibHandlerGet
#endif
LIB_EXPORT KWrtHandler * CC KLogLibHandlerGet ( void )
{
    return ( &G_log_lib_handler );
}


/* EOF */
