/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*/

#include <klib/extern.h>
#include "status-priv.h"
#include <klib/status.h>
#include <klib/writer.h>
#include <klib/rc.h>
#include <sysalloc.h>

LIB_EXPORT KStsLevel G_sts_level = ksts_0;

LIB_EXPORT KWrtHandler G_sts_handler;
LIB_EXPORT KWrtHandler G_sts_lib_handler;

LIB_EXPORT rc_t CC KStsInit ( void )
{
    rc_t rc;

    rc = KStsHandlerSetStdOut();

    if (rc == 0)
        rc = KStsLibHandlerSet ( NULL, NULL );

    return rc;
}


LIB_EXPORT rc_t CC KStsMsg ( const char * _fmt, ... )
{
    va_list ap;
    rc_t rc;
    size_t used;
    size_t remaining;
    char * pbuf;
    char fmt [16 * 1024];
    char buffer [16 * 1024];

    if (_fmt == NULL)
        return RC (rcRuntime, rcLog, rcLogging, rcParam, rcNull);

    used = knprintf (fmt, sizeof fmt, "%s\n", _fmt);

    if (used < (sizeof fmt - 1))
        fmt[used] = '\0';
    else
        fmt[sizeof fmt - 1] = '\0';

    /* we are piggy-backing on the identity and time stamp from the status */
    rc = logstampident (buffer, sizeof (buffer), &used);
    if (rc)
        return rc;

    pbuf = buffer + used;
    remaining = sizeof (buffer) - used;

    used = knprintf (pbuf, remaining, " status-%u: ", KStsLevelGet());
    if (used > remaining)
        return RC (rcRuntime, rcLog, rcFormatting, rcBuffer, rcInsufficient);

    pbuf += used;
    remaining -= used;

    va_start (ap, _fmt);

    used = vknprintf (pbuf, remaining, fmt, ap);

    if (used == 0) /* drop blank messages (blank other than the time stamp) */
        rc = 0;
    else if (used > sizeof (buffer))
        rc = RC (rcRuntime, rcLog, rcFormatting, rcBuffer, rcInsufficient);
    else
    {
        size_t to_write;
        size_t written;

        to_write = (pbuf + used) - buffer;
        pbuf = buffer;

        do
        {
            void * pv;
            pv = KStsDataGet();
            rc = KStsWriterGet()(pv, pbuf, to_write, &written);
            if (rc)
                break;
            to_write -= written;
        }        
        while (to_write > 0);
    }

    va_end (ap);

    return rc;
}


LIB_EXPORT rc_t CC KStsLibMsg ( const char * _fmt, ... )
{
    va_list ap;
    rc_t rc;
    size_t used;
    size_t remaining;
    char * pbuf;
    char fmt [16 * 1024];
    char buffer [16 * 1024];

    if (_fmt == NULL)
        return RC (rcRuntime, rcLog, rcLogging, rcParam, rcNull);

    used = knprintf (fmt, sizeof fmt, "%s\n", _fmt);

    if (used < (sizeof fmt - 1))
        fmt[used] = '\0';
    else
        fmt[sizeof fmt - 1] = '\0';

    /* we are piggy-backing on the identity and time stamp from the status */
    rc = logstampident (buffer, sizeof (buffer), &used);
    if (rc)
        return rc;

    pbuf = buffer + used;
    remaining = sizeof (buffer) - used;

    used = knprintf (pbuf, remaining, " status-%u: ", KStsLevelGet());
    if (used > remaining)
        return RC (rcRuntime, rcLog, rcFormatting, rcBuffer, rcInsufficient);

    pbuf += used;
    remaining -= used;

    va_start (ap, _fmt);

    used = vknprintf (pbuf, remaining, fmt, ap);

    if (used == 0) /* drop blank messages (blank other than the time stamp) */
        rc = 0;
    else if (used > sizeof (buffer))
        rc = RC (rcRuntime, rcLog, rcFormatting, rcBuffer, rcInsufficient);
    else
    {
        size_t to_write;
        size_t written;

        to_write = (pbuf + used) - buffer;
        pbuf = buffer;

        do
        {
            void * pv;
            pv = KStsLibDataGet();
            rc = KStsLibWriterGet()(pv, pbuf, to_write, &written);
            if (rc)
                break;
            to_write -= written;
        }        
        while (to_write > 0);
    }

    va_end (ap);

    return rc;
}


LIB_EXPORT rc_t CC KStsLevelAdjust ( int32_t adjust )
{
    KStsLevel local_level = KStsLevelGet() + adjust;
    return KStsLevelSet (local_level);
}

#ifdef KStsLevelGet
#undef KStsLevelGet
#endif
LIB_EXPORT KStsLevel CC KStsLevelGet ( void )
{
    return G_sts_level;
}

#ifdef KStsLevelSet
#undef KStsLevelSet
#endif
LIB_EXPORT rc_t CC KStsLevelSet ( KStsLevel lvl )
{
    if ( (lvl >= ksts_min) || (lvl <= ksts_max) )
    {
        G_sts_level = lvl;
        return 0;
    }
    return RC (rcApp, rcLog, rcUpdating, rcParam, rcOutofrange);
}

#ifdef KStsHandlerSetStdOut
#undef KStsHandlerSetStdOut
#endif
LIB_EXPORT rc_t CC KStsHandlerSetStdOut ( void )
{
    return KStsHandlerSet( DefaultWriter, DefaultWriterDataStdOut );
}
#ifdef KStsLibHandlerSetStdOut
#undef KStsLibHandlerSetStdOut
#endif
LIB_EXPORT rc_t CC KStsLibHandlerSetStdOut ( void )
{
    return KStsLibHandlerSet( DefaultWriter, DefaultWriterDataStdOut );
}

#ifdef KStsHandlerSetStdErr
#undef KStsHandlerSetStdErr
#endif
LIB_EXPORT rc_t CC KStsHandlerSetStdErr ( void )
{
    return KStsHandlerSet( DefaultWriter, DefaultWriterDataStdErr );
}

#ifdef KStsLibHandlerSetStdErr
#undef KStsLibHandlerSetStdErr
#endif
LIB_EXPORT rc_t CC KStsLibHandlerSetStdErr ( void )
{
    return KStsLibHandlerSet( DefaultWriter, DefaultWriterDataStdErr );
}


#ifdef KStsHandlerSet
#undef KStsHandlerSet
#endif
LIB_EXPORT rc_t CC KStsHandlerSet ( KWrtWriter writer, void * writer_data )
{
    G_sts_handler.writer = writer;
    G_sts_handler.data = writer_data;
    return 0;
}

#ifdef KStsLibHandlerSet
#undef KStsLibHandlerSet
#endif
LIB_EXPORT rc_t CC KStsLibHandlerSet ( KWrtWriter writer, void * writer_data )
{
    G_sts_lib_handler.writer = writer;
    G_sts_lib_handler.data = writer_data;
    return 0;
}

#ifdef KStsDataGet
#undef KStsDataGet
#endif
LIB_EXPORT void * CC KStsDataGet ( void )
{
    return ( KStsHandlerGet()->data );
}

#ifdef KStsLibDataGet
#undef KStsLibDataGet
#endif
LIB_EXPORT void * CC KStsLibDataGet ( void )
{
    return ( KStsLibHandlerGet()->data );
}

#ifdef KStsWriterGet
#undef KStsWriterGet
#endif
LIB_EXPORT KWrtWriter CC KStsWriterGet ( void )
{
    return ( KStsHandlerGet()->writer );
}

#ifdef KStsLibWriterGet
#undef KStsLibWriterGet
#endif
LIB_EXPORT KWrtWriter CC KStsLibWriterGet ( void )
{
    return ( KStsLibHandlerGet()->writer );
}

#ifdef KStsHandlerGet
#undef KStsHandlerGet
#endif
LIB_EXPORT KWrtHandler * CC KStsHandlerGet ( void )
{
    return ( &G_sts_handler );
}

#ifdef KStsLibHandlerGet
#undef KStsLibHandlerGet
#endif
LIB_EXPORT KWrtHandler * CC KStsLibHandlerGet ( void )
{
    return ( &G_sts_lib_handler );
}

/* EOF */
