/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <klib/extern.h>
#include "log-priv.h"
#include <klib/log.h>
#include <klib/text.h>
#include <klib/rc.h>
#include <sysalloc.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


static
rc_t CC msg_int ( KWrtHandler * handler, KLogLevel lvl, bool use_rc, rc_t status, 
    const char * msg, const char * fmt, va_list args )
{
    rc_t rc;

    /* check for errors we can't get past */
    if ( msg == NULL )
        return RC ( rcApp, rcLog, rcLogging, rcString, rcNull );
    if ( msg [ 0 ] == 0 )
        return RC ( rcApp, rcLog, rcLogging, rcString, rcEmpty );

    do
    {
        char * pbuffer;
        char * nbuffer;
        size_t num_writ;
        size_t remaining;
        size_t bsize;
        size_t used;
        char   buffer [8192];

#define FIX_UP()       if (rc) break; remaining -= num_writ; pbuffer += num_writ

        nbuffer = pbuffer = buffer;
        bsize = remaining = sizeof (buffer);

        /* create timestamp and identity */
        rc = logstampident ( pbuffer, remaining, & num_writ );

        FIX_UP();

        /* insert a single space */
        rc = LogInsertSpace (NULL, pbuffer, remaining, &num_writ);

        FIX_UP();

        /* describe log level */
        rc = KLogLevelExplain ( lvl, pbuffer, remaining, &num_writ);

        FIX_UP();

        /* insert space */
        rc = LogInsertSpace ( ": ", pbuffer, remaining, &num_writ);

        FIX_UP();

        if (use_rc)
        {
            do
            {
                rc = RCExplain ( status, pbuffer, remaining, &num_writ);

                FIX_UP();

                /* insert space */
                rc = LogInsertSpace ( " - ", pbuffer, remaining, &num_writ);

                FIX_UP();

            } while (0);

            if (rc)
                break;
        }

        used = sizeof (buffer) - remaining;
        
        if (fmt == NULL)
        {
            /* print message */
            rc = LogAppendMsg ( msg, pbuffer, remaining, &num_writ);

            if (GetRCState (rc) == rcInsufficient)
            {
                if (num_writ == 0)
                    bsize = 2 * sizeof (buffer);
                else
                    bsize = used + num_writ + 2;

                nbuffer = malloc (bsize);
                if (nbuffer == NULL)
                {
                    rc =  RC ( rcApp, rcLog, rcLogging, rcMemory, rcExhausted );
                    break;
                }
                
                memcpy (nbuffer, buffer, used );

                remaining = bsize - used;
                    
                rc = LogAppendMsg (msg, nbuffer, remaining, &num_writ);
                if (rc)
                {
                    free (nbuffer);
                    break;
                }
            }
            else if (rc)
                break;

            used = bsize - (remaining - num_writ);
        }
        else
        {
            uint32_t argc;
            nvp argv [ 16 ];
            char lbuffer [ 4096 ];

            /* build parameterized message */
            rc = LogProduceMsg ( msg, fmt, args,
                                 &argc, argv, sizeof argv, lbuffer, sizeof lbuffer,
                                 used, buffer, sizeof buffer, &num_writ );

            if (rc)
                break;

            used = bsize - (remaining - num_writ);
        }

        /* write the string */
        rc = KLogFlushLine ( handler, nbuffer, bsize, used );

        if ((rc == 0 ) && use_rc)
            KLogLastErrorCodeSet ( lvl, status );

        if (nbuffer != buffer)
            free (nbuffer);

    } while (0);

    return rc;
}

/* LogMsg
 *  makes an entry to the log file
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "msg" [ IN ] - NUL terminated string
 */

LIB_EXPORT rc_t CC LogMsg ( KLogLevel lvl, const char *msg )
{
    /* early ignore - violates param checking protocol, but... */
    if ( lvl > KLogLevelGet() )
        return 0;

    return msg_int ( KLogHandlerGet(), (unsigned)lvl, false, 0, msg, NULL, NULL );
}

LIB_EXPORT rc_t CC LogLibMsg ( KLogLevel lvl, const char *msg )
{
    /* early ignore - violates param checking protocol, but... */
    if ( lvl > KLogLevelGet() )
        return 0;

    return msg_int ( KLogLibHandlerGet(), (unsigned)lvl, false, 0, msg, NULL, NULL );
}


/* LogErr
 *  reports an error to log file
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "status" [ IN ] - return code to be decoded
 *
 *  "msg" [ IN, NULL OKAY ] - optional NUL terminated message string
 */
LIB_EXPORT rc_t CC LogErr ( KLogLevel lvl, rc_t status, const char *msg )
{
    /* early ignore - violates param checking protocol, but... */
    if ( lvl > KLogLevelGet() )
        return 0;

    return msg_int ( KLogHandlerGet(), (unsigned)lvl, true, status, msg, NULL, NULL );
}

LIB_EXPORT rc_t CC LogLibErr ( KLogLevel lvl, rc_t status, const char *msg )
{
    /* early ignore - violates param checking protocol, but... */
    if ( lvl > KLogLevelGet() )
        return 0;

    return msg_int ( KLogLibHandlerGet(), (unsigned)lvl, true, status, msg, NULL, NULL );
}


/* vLogMsg
 *  makes a parameterized entry to the log file
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "msg" [ IN ] - NUL terminated string with named parameters
 *
 *  "fmt" [ IN ] - NUL terminated format string, behaves according
 *  to standard printf-style formatting
 */
LIB_EXPORT rc_t CC vLogMsg ( KLogLevel lvl, const char *msg, const char *fmt, va_list args )
{
    return msg_int ( KLogHandlerGet(), (unsigned)lvl, false, 0, msg, fmt, args );
}

LIB_EXPORT rc_t CC vLogLibMsg ( KLogLevel lvl, const char *msg, const char *fmt, va_list args )
{
    return msg_int ( KLogLibHandlerGet(), (unsigned)lvl, false, 0, msg, fmt, args );
}


/* vLogErr
 *  reports an error to log file
 *
 *  "lvl" [ IN ] - severity of message
 *
 *  "status" [ IN ] - return code to be decoded
 *
 *  "msg" [ IN ] - NUL terminated message string with named parameters
 *
 *  "fmt" [ IN ] - NUL terminated format string, behaves according
 *  to standard printf-style formatting
 */
LIB_EXPORT rc_t CC vLogErr ( KLogLevel lvl, rc_t status, const char *msg, const char *fmt, va_list args )
{
    return msg_int ( KLogHandlerGet(), (unsigned)lvl, true, status, msg, fmt, args );
}

LIB_EXPORT rc_t CC vLogLibErr ( KLogLevel lvl, rc_t status, const char *msg, const char *fmt, va_list args )
{
    return msg_int ( KLogLibHandlerGet(), (unsigned)lvl, true, status, msg, fmt, args );
}

/* EOF */
