/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <klib/extern.h>
#include "writer-priv.h"
#include <klib/out.h>
#include <klib/status.h>
#include <klib/log.h>
#include <klib/debug.h>
#include <klib/text.h>
#include <klib/rc.h>
#include <sysalloc.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include <os-native.h>

size_t wrt_ident_size;

char wrt_ident [ 256 ];

uint32_t wrt_version;

/*
 *  "ident" [ IN ] - identity of executable
 *
 *  "vers" [ IN ] - 4-part version code: 0xMMmmrrrr, where
 *      MM = major release
 *      mm = minor release
 *    rrrr = bug-fix release
 */
LIB_EXPORT rc_t CC KWrtInit ( const char * appname, uint32_t vers )
{
    rc_t rc;

    if ( appname == NULL )
        return RC ( rcApp, rcLog, rcConstructing, rcString, rcNull );
    if ( appname [ 0 ] == 0 )
        return RC ( rcApp, rcLog, rcConstructing, rcString, rcEmpty );

    do
    {
        rc = KWrtSysInit(appname, vers);
        if (rc) break;

        wrt_version = vers;




        rc = KOutInit();
        if (rc) break;

        rc = KLogInit();
        if (rc) break;

        rc = KStsInit();
#if _DEBUGGING
        if (rc) break;

        rc = KDbgInit();
#endif
    } while (0);

    return rc;
}


static
rc_t RCLiteral ( rc_t rc, char *buffer, size_t bsize, size_t *num_writ )
{
    int len = snprintf ( buffer, bsize
                         , "rc = %u.%u.%u.%u.%u"
                         , ( unsigned int ) GetRCModule ( rc )
                         , ( unsigned int ) GetRCTarget ( rc )
                         , ( unsigned int ) GetRCContext ( rc )
                         , ( unsigned int ) GetRCObject ( rc )
                         , ( unsigned int ) GetRCState ( rc )
        );

    if ( len < 0 )
    {
        * num_writ = 0;
        return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );
    }

    * num_writ = len;

    if ( ( size_t ) len >= bsize )
        return RC ( rcApp, rcLog, rcLogging, rcBuffer, rcInsufficient );

    return 0;
}

LIB_EXPORT rc_t CC RCExplain ( rc_t rc, char *buffer, size_t bsize, size_t *num_writ )
{
    return RCExplain2
        ( rc, buffer, bsize, num_writ, eRCExOpt_CompleteMsg );
}

LIB_EXPORT rc_t CC RCExplain2 ( rc_t rc, char *buffer, size_t bsize, size_t *num_writ,
    int rCExplain2Options /* ERCExplain2Options */ )
{
    bool noMessageIfNoError =
        (rCExplain2Options == eRCExOpt_NoMessageIfNoError || 
         rCExplain2Options == eRCExOpt_ObjAndStateOnlyIfError);
    int len;
    size_t total = 0;

    const char *mod = GetRCModuleText ( GetRCModule ( rc ) );
    const char *targ = GetRCTargetText ( GetRCTarget ( rc ) );
    const char *ctx = GetRCContextText ( GetRCContext ( rc ) );
    const char *obj = GetRCObjectText ( GetRCObject ( rc ) );
    const char *state = GetRCStateText ( GetRCState ( rc ) );

    assert ( buffer && num_writ );

    *num_writ = 0;

    if ( rc == 0 && noMessageIfNoError )
    {
        buffer [ 0 ] = '\0';
        return 0;
    }

    /* English'ish formatting */
    if ( obj != NULL )
    {
        len = snprintf ( buffer + total, bsize - total, "%s ", obj );
        if ( len < 0 || ( total + len ) >= bsize )
            return RCLiteral ( rc, buffer, bsize, num_writ );
        total += len;
    }

    if ( state != NULL )
    {
        len = snprintf ( buffer + total, bsize - total, "%s ", state );
        if ( len < 0 || ( total + len ) >= bsize )
            return RCLiteral ( rc, buffer, bsize, num_writ );
        total += len;
    }

    if ( rc != 0 && rCExplain2Options == eRCExOpt_CompleteMsg )
    {
        if ( ctx != NULL )
        {
            len = snprintf ( buffer + total, bsize - total, "while %s ", ctx );
            if ( len < 0 || ( total + len ) >= bsize )
                return RCLiteral ( rc, buffer, bsize, num_writ );
            total += len;

            if ( targ != NULL )
            {
                len = snprintf ( buffer + total, bsize - total, "%s ", targ );
                if ( len < 0 || ( total + len ) >= bsize )
                    return RCLiteral ( rc, buffer, bsize, num_writ );
                total += len;
            }
        }
        else if ( targ != NULL )
        {
            len = snprintf ( buffer + total,
                bsize - total, "while acting upon %s ", targ );
            if ( len < 0 || ( total + len ) >= bsize )
                return RCLiteral ( rc, buffer, bsize, num_writ );
            total += len;
        }
    }

    if ( mod != NULL && rCExplain2Options == eRCExOpt_CompleteMsg )
    {
        len = snprintf
            ( buffer + total, bsize - total, "within %s module ", mod );
        if ( len < 0 || ( total + len ) >= bsize )
            return RCLiteral ( rc, buffer, bsize, num_writ );
        total += len;
    }

    * num_writ = total;
    return 0;
}

/*--------------------------------------------------------------------------
 * RC
 */


/* GetRCModuleText
 */
const char *GetRCModuleText ( enum RCModule mod )
{
    if ( ( int ) mod < 0 || ( int ) mod >= ( int ) rcLastModule_v1_1 )
        return "<INVALID-MODULE>";
    return gRCModule_str [ ( int ) mod ];
}

/* GetRCModuleIdxText
 */
const char *GetRCModuleIdxText ( enum RCModule mod )
{
    if ( ( int ) mod < 0 || ( int ) mod >= ( int ) rcLastModule_v1_1 )
        return "<INVALID-MODULE>";
    return gRCModuleIdx_str [ ( int ) mod ];
}

/* GetRCTargetText
 */
const char *GetRCTargetText ( enum RCTarget targ )
{
    if ( ( int ) targ < 0 || ( int ) targ >= ( int ) rcLastTarget_v1_1 )
        return "<INVALID-TARGET>";
    return gRCTarget_str [ ( int ) targ ];
}

/* GetRCTargetIdxText
 */
const char *GetRCTargetIdxText ( enum RCTarget targ )
{
    if ( ( int ) targ < 0 || ( int ) targ >= ( int ) rcLastTarget_v1_1 )
        return "<INVALID-TARGET>";
    return gRCTargetIdx_str [ ( int ) targ ];
}

/* GetRCContextText
 */
const char *GetRCContextText ( enum RCContext ctx )
{
    if ( ( int ) ctx < 0 || ( int ) ctx >= ( int ) rcLastContext_v1_1 )
        return "<INVALID-CONTEXT>";
    return gRCContext_str [ ( int ) ctx ];
}

/* GetRCContextIdxText
 */
const char *GetRCContextIdxText ( enum RCContext ctx )
{
    if ( ( int ) ctx < 0 || ( int ) ctx >= ( int ) rcLastContext_v1_1 )
        return "<INVALID-CONTEXT>";
    return gRCContextIdx_str [ ( int ) ctx ];
}

/* GetRCObjectText
 */
const char *GetRCObjectText ( int obj )
{
    if ( ( int ) obj < 0 || ( int ) obj >= ( int ) rcLastObject_v1_1 )
        return "<INVALID-OBJECT>";
    if ( ( int ) obj < ( int ) rcLastTarget_v1_1 )
        return gRCTarget_str [ ( int ) obj ];
    return gRCObject_str [ ( int ) obj - ( int ) ( rcLastTarget_v1_1 - 1 ) ];
}

/* GetRCObjectIdxText
 */
const char *GetRCObjectIdxText ( int obj )
{
    if ( ( int ) obj < 0 || ( int ) obj >= ( int ) rcLastObject_v1_1 )
        return "<INVALID-OBJECT>";
    if ( ( int ) obj < ( int ) rcLastTarget_v1_1 )
        return gRCTargetIdx_str [ ( int ) obj ];
    return gRCObjectIdx_str [ ( int ) obj - ( int ) ( rcLastTarget_v1_1 - 1 ) ];
}

/* GetRCStateText
 */
const char *GetRCStateText ( enum RCState state )
{
    if ( ( int ) state < 0 || ( int ) state >= ( int ) rcLastState_v1_1 )
        return "<INVALID-STATE>";
    return gRCState_str [ ( int ) state ];
}

/* GetRCStateIdxText
 */
const char *GetRCStateIdxText ( enum RCState state )
{
    if ( ( int ) state < 0 || ( int ) state >= ( int ) rcLastState_v1_1 )
        return "<INVALID-STATE>";
    return gRCStateIdx_str [ ( int ) state ];
}

LIB_EXPORT rc_t CC logstampident ( char *buffer, size_t bsize, size_t *num_writ )
{
    /* apply timestamp */
    rc_t rc = LogTimestamp ( buffer, bsize, num_writ );
    if ( rc == 0 )
    {
        size_t total = * num_writ;

        /* insert space */
        rc = LogInsertSpace ( NULL, buffer + total, bsize - total, num_writ );
        if ( rc == 0 )
        {
            ++ total;

            /* apply identity */
            if ( wrt_ident_size == 0 || total + wrt_ident_size >= bsize )
                rc = LogPID ( buffer + total, bsize - total, num_writ );
            else
            {
                * num_writ = string_copy ( buffer + total,
                    bsize - total, wrt_ident, wrt_ident_size );
                assert ( * num_writ == wrt_ident_size );
                assert ( * num_writ + total < bsize );
            }
            if ( rc == 0 )
                total += * num_writ;
        }

        * num_writ = total;
    }
    return rc;
}




/* EOF */
