/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <search/extern.h>
#include <compiler.h>
#include <os-native.h>
#include <sysalloc.h>
#include "search-priv.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define BITTYPE int64_t
#define UBITTYPE uint64_t

struct MyersSearch {
    int32_t m;
    BITTYPE PEq[256];
    BITTYPE PEq_R[256];
};

static const char NA2KEY [] = "ACGT";
static const char NA4KEY [] = " ACMGRSVTWYHKDBN";

static
int32_t any_non_4na_chars(const char *pattern)
{
    int32_t len = strlen(pattern);
    int32_t i;
    char *p;

    for (i=0; i<len; i++) {
        p = index(NA4KEY, pattern[i]);
        if (p == NULL)
            return 1;
    }
    return 0;
}

static
int32_t na4key_matches(AgrepFlags mode, char na4, char acgt)
{
    char *p;
    int32_t pos4, pos2;
    p = index(NA4KEY, na4);
    if (p == NULL) {
        p = index(NA4KEY, 'N');
    }
    pos4 = p - NA4KEY;
    pos2 = index(NA2KEY, acgt) - NA2KEY;
    if (pos4 & (1<<pos2))
        return 1;
    return 0;
}

void printbits( BITTYPE in )
{
    char buf[9];
    int32_t j, k;
    unsigned char byte;
    buf[8] = '\0';
    for (j=0; j<sizeof(BITTYPE); j++) {
        byte = in >> (8*(sizeof(BITTYPE) - j - 1));
        for (k=0; k<8; k++) {
            buf[k] = "01"[1&(byte>>(7-k))];
        }
        printf("%s ", buf);
    }
}

void AgrepMyersFree( MyersSearch *self )
{
    free(self);
}
  
/* Returns -1 if error. */

rc_t AgrepMyersMake( MyersSearch **self, AgrepFlags mode, const char *pattern )
{
    const unsigned char *upattern = (const unsigned char *)pattern;
    int32_t rc;
    int32_t len = strlen(pattern);
    int32_t plen = len;
    uint32_t max_pattern_length;
    int32_t i, j;
    int32_t m;
    BITTYPE bits;

    max_pattern_length = sizeof(BITTYPE) * 8 - 1;

    if (mode & AGREP_ANCHOR_LEFT) 
        /* Unsupported */
        return RC( rcText, rcString, rcSearching, rcParam, rcUnsupported );

    if( len > max_pattern_length ) {
        /* Should return return code of pattern too long */
        return RC( rcText, rcString, rcSearching, rcParam, rcExcessive);
    }

    if (!(mode & AGREP_ANYTHING_ELSE_IS_N) &&
        any_non_4na_chars(pattern))
        /* This should be a return code. */
        return RC( rcText, rcString, rcSearching, rcParam, rcUnrecognized);

    *self = malloc(sizeof(MyersSearch));
    m = (*self)->m = len;
  
    memset((*self)->PEq, 0, sizeof((*self)->PEq));
    for(j = 0; j < m; j++) {
        (*self)->PEq[upattern[j]] |= (BITTYPE)1 << j;
        if( pattern[j] == 'a' ) {
            (*self)->PEq['t'] |= (BITTYPE)1 << j; /* t == a */
        }
    }
  
    memset((*self)->PEq_R, 0, sizeof((*self)->PEq_R));
    for(j = 0; j < m; j++) {
        (*self)->PEq_R[upattern[m-j-1]] |= (BITTYPE)1 << j;
        if( pattern[m-j-1] == 'a' ) {
            (*self)->PEq_R['t'] |= (BITTYPE)1 << j; /* t == a */
        }
    }


    for (i=0; i<4; i++) {
        unsigned char acgt = NA2KEY[i];
        bits = 0;
        for (j=0; j<plen; j++) {
            if (na4key_matches(mode, pattern[j], acgt)) {
                bits |= (UBITTYPE)1<<j;
            }
        }
       (*self)->PEq[acgt] |= bits;
       /* Set bits for both the 0-5 characters, as well as ascii */
       if (mode & AGREP_TEXT_EXPANDED_2NA) {
           acgt = i;
           (*self)->PEq[acgt] |= bits;
       }
    }
    for (i=0; i<4; i++) {
        unsigned char acgt = NA2KEY[i];
        bits = 0;
        for (j=0; j<plen; j++) {
            if (na4key_matches(mode, pattern[plen-j-1], acgt)) {
                bits |= (UBITTYPE)1<<j;
            }
        }
        (*self)->PEq_R[acgt] |= bits;
       /* Set bits for both the 0-5 characters, as well as ascii */
        if (mode & AGREP_TEXT_EXPANDED_2NA) {
            acgt = i;
            (*self)->PEq_R[acgt] |= bits;
        }
    }
    return 0;

    free(*self);
    *self = NULL;
    return rc;
}

/* 
   This finds the first match forward in the string less than or equal
   the threshold.  If nonzero, this will be a prefix of any exact match.
*/
uint32_t MyersFindFirst( MyersSearch *self, int32_t threshold, 
                   const char* text, size_t n,
                   AgrepMatch *match )
{
    const unsigned char *utext = (const unsigned char *)text;
    BITTYPE Pv;
    BITTYPE Mv;

    int32_t m = self->m;
    int32_t Score;
    int32_t BestScore = m;
    int32_t from = 0;
    int32_t to = -1;

    int32_t j;
    BITTYPE Eq, Xv, Xh, Ph, Mh;

    Score = m;
    Pv = (BITTYPE)-1;
    Mv = (BITTYPE)0;
    
    for(j = 0; j < n; j++) {
        Eq = self->PEq[utext[j]];
        Xv = Eq | Mv;
        Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
        Ph = Mv | ~ (Xh | Pv);
        Mh = Pv & Xh;
        if(Ph & ((BITTYPE)1 << (m - 1))) {
            Score++;
        } else if(Mh & ((BITTYPE)1 << (m - 1))) {
            Score--;
        }
        Ph <<= 1;
        Mh <<= 1;
        Pv = Mh | ~(Xv | Ph);
        Mv = Ph & Xv;
        if (Score <= threshold) {
            BestScore = Score;
            to = j;
            break;
        }
    }

    if (BestScore <= threshold) {

        /* Continue while score decreases under the threshold */
        for(j++; j < n; j++) {
            Eq = self->PEq[utext[j]];
            Xv = Eq | Mv;
            Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
            Ph = Mv | ~ (Xh | Pv);
            Mh = Pv & Xh;
            if(Ph & ((BITTYPE)1 << (m - 1))) {
                Score++;
            } else if(Mh & ((BITTYPE)1 << (m - 1))) {
                Score--;
            }
            Ph <<= 1;
            Mh <<= 1;
            Pv = Mh | ~(Xv | Ph);
            Mv = Ph & Xv;
            if (Score < BestScore) {
                BestScore = Score;
                to = j;
            } else {
                break;
            }
        }
    }

    if (BestScore <= threshold) {
        /* Re-initialize for next scan! */
        Score = m;
        Pv = (BITTYPE)-1;
        Mv = (BITTYPE)0;
        
        for(j = to; j >= 0; j--) {
            Eq = self->PEq_R[utext[j]];
            Xv = Eq | Mv;
            Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
            Ph = Mv | ~ (Xh | Pv);
            Mh = Pv & Xh;
            if(Ph & ((BITTYPE)1 << (m - 1))) {
                Score++;
            } else if(Mh & ((BITTYPE)1 << (m - 1))) {
                Score--;
            }
            Ph <<= 1;
            Mh <<= 1;
            Pv = Mh | ~(Xv | Ph);
            Mv = Ph & Xv;
            if(Score <= BestScore) {
                from = j;
                break;
            }
        }
    }
        
    if (BestScore <= threshold) {
        match->position = from;
        match->length = to-from+1;
        match->score = BestScore;
        return 1;
    } else {
        return 0;
    }
}

  

/* 
   Returns non-negative if something found.
   Return value is the number of mismatches.
*/
LIB_EXPORT int32_t CC MyersFindBest ( MyersSearch *self, const char* text, 
    size_t n, int32_t *pos, int32_t *len )
{
    const unsigned char *utext = (const unsigned char *)text;
    BITTYPE Pv;
    BITTYPE Mv;

    int32_t m = self->m;
    int32_t Score;
    int32_t BestScore = m;
    int32_t from = 0;
    int32_t to = -1;

    int32_t j;
    BITTYPE Eq, Xv, Xh, Ph, Mh;

    Score = m;
    Pv = (BITTYPE)-1;
    Mv = (BITTYPE)0;
    
    for(j = 0; j < n; j++) {
        Eq = self->PEq[utext[j]];
        Xv = Eq | Mv;
        Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
        Ph = Mv | ~ (Xh | Pv);
        Mh = Pv & Xh;
        if(Ph & ((BITTYPE)1 << (m - 1))) {
            Score++;
        } else if(Mh & ((BITTYPE)1 << (m - 1))) {
            Score--;
        }
        Ph <<= 1;
        Mh <<= 1;
        Pv = Mh | ~(Xv | Ph);
        Mv = Ph & Xv;
        /* printf("%3d. score %d\n", j, Score); */
        if( Score < BestScore ) {
            BestScore = Score;
            to = j;
        }
    }

    /* Re-initialize for next scan! */
    Score = m;
    Pv = (BITTYPE)-1;
    Mv = (BITTYPE)0;
    
    for(j = to; j >= 0; j--) {
        Eq = self->PEq_R[utext[j]];
        Xv = Eq | Mv;
        Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
        Ph = Mv | ~ (Xh | Pv);
        Mh = Pv & Xh;
        if(Ph & ((BITTYPE)1 << (m - 1))) {
            Score++;
        } else if(Mh & ((BITTYPE)1 << (m - 1))) {
            Score--;
        }
        Ph <<= 1;
        Mh <<= 1;
        Pv = Mh | ~(Xv | Ph);
        Mv = Ph & Xv;
        /* printf("%3d. score %d\n", j, Score); */
        if(Score <= BestScore) {
            /* printf("rev match at position %d\n", j); */
            from = j;
            break;
        }
    }
    
    *pos = from;
    *len = to-from+1;

    return BestScore;

    /* printf("Found [%d,%d]\n", from, to);
       printf("Found '%.*s'\n", to - from + 1, &text[from]); */
}

/* 
   This finds the first match forward in the string less than or equal
   the threshold.  If nonzero, this will be a prefix of any exact match.
*/
void MyersFindAll ( const AgrepCallArgs *args )
{
    AgrepFlags mode = args->self->mode;
    MyersSearch *self = args->self->myers;
    int32_t threshold = args->threshold;
    const unsigned char *utext = (const unsigned char *)args->buf;
    int32_t n = args->buflen;
    AgrepMatchCallback cb = dp_end_callback;
    const void *cbinfo = args;

    AgrepMatch match;
    AgrepContinueFlag cont;

    BITTYPE Pv;
    BITTYPE Mv;

    int32_t m = self->m;
    int32_t Score;
    int32_t BestScore;

    int32_t j;
    BITTYPE Eq, Xv, Xh, Ph, Mh;

    int32_t curscore = 0;
    int32_t curlast = 0;
    int32_t continuing = 0;
    
    BestScore = m;
    Score = m;
    Pv = (BITTYPE)-1;
    Mv = (BITTYPE)0;
    
    for(j = 0; j < n; j++) {
        Eq = self->PEq[utext[j]];
        Xv = Eq | Mv;
        Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
        Ph = Mv | ~ (Xh | Pv);
        Mh = Pv & Xh;

        /* Debug code */
        /*
        printf("%d: Ph ", j);
        printbits(Ph);
        printf(" Mh ");
        printbits(Mh);
        printf("\n");
        */

        if(Ph & ((BITTYPE)1 << (m - 1))) {
            Score++;
        } else if(Mh & ((BITTYPE)1 << (m - 1))) {
            Score--;
        }
        Ph <<= 1;
        Mh <<= 1;
        Pv = Mh | ~(Xv | Ph);
        Mv = Ph & Xv;
        if (Score <= threshold) {
            /* At this point we let the DP algorithm find the starting point. */
            /* This just passes the end position. */

            if (continuing) {
                if (Score < curscore &&
                    ((mode & AGREP_EXTEND_BETTER) ||
                     (mode & AGREP_EXTEND_SAME))) {
                    curscore = Score;
                    curlast = j;
                } else if (Score == curscore &&
                           ((mode & AGREP_EXTEND_BETTER) ||
                            (mode & AGREP_EXTEND_SAME))) {
                    if (mode & AGREP_EXTEND_SAME) {
                        curlast = j;
                    }
                } else {
                    continuing = 0;
                    match.score = curscore;
                    match.position = curlast;
                    match.length = -1;
                    cont = AGREP_CONTINUE;
                    (*cb)(cbinfo, &match, &cont);
                    if (cont != AGREP_CONTINUE)
                        return;
                }
            } else if ((mode & AGREP_EXTEND_SAME) ||
                       (mode & AGREP_EXTEND_BETTER)) {
                curscore = Score;
                curlast = j;
                continuing = 1;
            } else {
                match.score = Score;
                match.position = j;
                match.length = -1;
                cont = AGREP_CONTINUE;
                (*cb)(cbinfo, &match, &cont);
                if (cont != AGREP_CONTINUE)
                    return;
            }
            /* If we're no longer under the threshold, we might
               have been moving forward looking for a better match 
            */
        } else if (continuing) {
            continuing = 0;
            match.score = curscore;
            match.position = curlast;
            match.length = -1;
            cont = AGREP_CONTINUE;
            (*cb)(cbinfo, &match, &cont);
            if (cont != AGREP_CONTINUE)
                return;
        }
        /* print_col_as_row(nxt, plen); */
    }
    if (continuing) {
        continuing = 0;
        match.score = curscore;
        match.position = curlast;
        match.length = -1;
        (*cb)(cbinfo, &match, &cont);
    }
}
