/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <search/extern.h>
#include <sysalloc.h>
#include "search-priv.h"

#include <stdlib.h>
#include <string.h>

LIB_EXPORT void CC FgrepFree( FgrepParams *self )
{
    if (self->dumb != NULL)
        FgrepDumbSearchFree(self->dumb);
    if (self->boyer != NULL)
        FgrepBoyerSearchFree(self->boyer);
    if (self->aho != NULL)
        FgrepAhoFree(self->aho);
    free(self);
}


LIB_EXPORT rc_t CC FgrepMake( FgrepParams **self, FgrepFlags mode, const char *strings[], uint32_t numstrings )
{
    *self = malloc(sizeof(FgrepParams));
    memset(*self, 0, sizeof(FgrepParams));
    (*self)->mode = mode;
    if (mode & FGREP_ALG_DUMB) {
        FgrepDumbSearchMake(&(*self)->dumb, strings, numstrings);
    }
    if (mode & FGREP_ALG_BOYERMOORE) {
        FgrepBoyerSearchMake(&(*self)->boyer, strings, numstrings);
    }
    if (mode & FGREP_ALG_AHOCORASICK) {
        FgrepAhoMake(&(*self)->aho, strings, numstrings);
    }
    return 0;
}

LIB_EXPORT uint32_t CC FgrepFindFirst( const FgrepParams *self, const char *buf, size_t len, FgrepMatch *match )
{
    if (self->mode & FGREP_ALG_DUMB) {
        return FgrepDumbFindFirst(self->dumb, buf, len, match);
    }
    if (self->mode & FGREP_ALG_BOYERMOORE) {
        return FgrepBoyerFindFirst(self->boyer, buf, len, match);
    }
    if (self->mode & FGREP_ALG_AHOCORASICK) {
        return FgrepAhoFindFirst(self->aho, buf, len, match);
    }
    /* Should maybe return error code, not 1/0 */
    return 0;
}

LIB_EXPORT void CC AgrepWhack( AgrepParams *self )
{
    if (self->wu != NULL)
        AgrepWuFree(self->wu);
    if (self->dp != NULL)
        AgrepDPFree(self->dp);
    if (self->myers != NULL)
        AgrepMyersFree(self->myers);
    if (self->myersunltd != NULL)
        MyersUnlimitedFree(self->myersunltd);
    free(self);
}

LIB_EXPORT rc_t CC AgrepMake( AgrepParams **self, AgrepFlags mode, const char *pattern )
{
    rc_t rc = 0;
    *self = malloc(sizeof(AgrepParams));
    memset(*self, 0, sizeof(AgrepParams));

    (*self)->mode = mode;
    if (mode & AGREP_ALG_WUMANBER) {
        rc = AgrepWuMake(&(*self)->wu, mode, pattern);
        if (rc)
            return rc;
        rc = AgrepDPMake(&(*self)->dp, mode, pattern);
        if (rc) {
            AgrepWuFree((*self)->wu);
            return rc;
        }
        return 0;
    }
    if (mode & AGREP_ALG_MYERS) {
        rc = AgrepMyersMake(&(*self)->myers, mode, pattern);
        if (rc)
            return rc;
        rc = AgrepDPMake(&(*self)->dp, mode, pattern);
        if (rc) {
            AgrepMyersFree((*self)->myers);
            return rc;
        }
        return 0;
    }
    if (mode & AGREP_ALG_MYERS_UNLTD) {
        rc = MyersUnlimitedMake(&(*self)->myersunltd, mode, pattern);
        if (rc)
            return rc;
        rc = AgrepDPMake(&(*self)->dp, mode, pattern);
        if (rc) {
            MyersUnlimitedFree((*self)->myersunltd);
            return rc;
        }
        return 0;
    }
    if (mode & AGREP_ALG_DP) {
        return AgrepDPMake(&(*self)->dp, mode, pattern);
        return 0;
    }
    /* Should return algorithm unspecified */
    (*self)->mode |= AGREP_ALG_DP;
    return AgrepDPMake(&(*self)->dp, mode, pattern);
}

LIB_EXPORT void CC AgrepFindAll( const AgrepCallArgs *args )
{
    const AgrepParams *self = args->self;

    if (self->mode & AGREP_ALG_WUMANBER) {
        AgrepWuFindAll(args);
        return;
    }
    if (self->mode & AGREP_ALG_MYERS) {
        MyersFindAll(args);
        return;
    }
    if (self->mode & AGREP_ALG_MYERS_UNLTD) {
        MyersUnlimitedFindAll(args);
        return;
    }
    if (self->mode & AGREP_ALG_DP) {
        AgrepDPFindAll(args);
        return;
    }
}

LIB_EXPORT uint32_t CC AgrepFindFirst( const AgrepParams *self, int32_t threshold, const char *buf, size_t len, AgrepMatch *match )
{
    if (self->mode & AGREP_ALG_WUMANBER) {
        return AgrepWuFindFirst(self->wu, threshold, buf, (int32_t)len, match);
    }
    if (self->mode & AGREP_ALG_MYERS) {
        return MyersFindFirst(self->myers, threshold, buf, len, match);
    }
    if (self->mode & AGREP_ALG_MYERS_UNLTD) {
        return MyersUnlimitedFindFirst(self->myersunltd, threshold, buf, len, match);
    }
    if (self->mode & AGREP_ALG_DP) {
        return AgrepDPFindFirst(self->dp, threshold, self->mode, buf, (int32_t)len, match);
    }
    /* Not sure this is the right thing to return. */
    return 0;
}

static 
rc_t CC AgrepFindBestCallback(const void *cbinfo, const AgrepMatch *matchinfo, AgrepContinueFlag *flag)
{
    AgrepMatch *best = (AgrepMatch *)cbinfo;
    if (best->score == -1 ||
        best->score > matchinfo->score) {
        *best = *matchinfo;
    }
    return 0;
}

LIB_EXPORT uint32_t CC AgrepFindBest( const AgrepParams *self, int32_t threshold, const char *buf, int32_t len, AgrepMatch *match )
{
    AgrepCallArgs args;

    args.self = self;
    args.threshold = threshold;
    args.buf = buf;
    args.buflen = len;
    args.cb = AgrepFindBestCallback;
    args.cbinfo = match;

    match->score = -1;
    
    AgrepFindAll(&args);
    if (match->score != -1)
        return 1;
    return 0;
}
