/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <vdb/extern.h>
#include <vdb/xform.h>
#include <arch-impl.h>
#include <klib/rc.h>
#include <sysalloc.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#if 1
static struct {
	uint8_t num_bits;
	uint8_t val[4];
} map[16] = {
/*0000*/{4,{1,2,4,8}}, 
/*0001*/{1,{1}},
/*0010*/{1,{2}},
/*0011*/{2,{1,2}},
/*0100*/{1,{4}},
/*0101*/{2,{1,4}},
/*0110*/{2,{2,4}},
/*0111*/{3,{1,2,4}},
/*1000*/{1,{8}},
/*1001*/{2,{1,8}},
/*1010*/{2,{2,8}},
/*1011*/{3,{1,2,8}},
/*1100*/{2,{4,8}},
/*1101*/{3,{1,4,8}},
/*1110*/{3,{2,4,8}},
/*1111*/{4,{1,2,4,8}}
};

static uint8_t randomize_ambiguity(uint8_t base, unsigned *seed, bool *reseed) {
    assert(base < 16);
    
    if (map[base].num_bits == 1) {
        *seed = (*seed << 4) | base;
	*reseed = true;
	return base;
    }
    else {
        unsigned rval;
	uint8_t rand_base;

	if (*reseed) {
	    *reseed = false;
	    srand(*seed);
	}
        rval = rand() * map[base].num_bits / RAND_MAX;
	assert(rval < map[base].num_bits);
	rand_base = map[base].val[rval];
	assert(map[rand_base].num_bits == 1);
	return rand_base;
    }
}
#else
typedef uint8_t dna4na_t;

static const bool isExact[16]
  = { 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 };

static const struct { dna4na_t val[12]; } possibleBase[16] = {
    {{ 1, 2, 4, 8, 1, 2, 4, 8, 1, 2, 4, 8 }}, /* 0000 1 2 4 8 */
    {{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }}, /* 0001 1 */
    {{ 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2 }}, /* 0010 2 */
    {{ 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2 }}, /* 0011 1 2 */
    {{ 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 }}, /* 0100 4 */
    {{ 1, 4, 1, 4, 1, 4, 1, 4, 1, 4, 1, 4 }}, /* 0101 1 4 */
    {{ 2, 4, 2, 4, 2, 4, 2, 4, 2, 4, 2, 4 }}, /* 0110 2 4 */
    {{ 1, 2, 4, 1, 2, 4, 1, 2, 4, 1, 2, 4 }}, /* 0111 1 2 4 */
    {{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8 }}, /* 1000 8 */
    {{ 1, 8, 1, 8, 1, 8, 1, 8, 1, 8, 1, 8 }}, /* 1001 1 8 */
    {{ 2, 8, 2, 8, 2, 8, 2, 8, 2, 8, 2, 8 }}, /* 1010 2 8 */
    {{ 1, 2, 8, 1, 2, 8, 1, 2, 8, 1, 2, 8 }}, /* 1011 1 2 8 */
    {{ 4, 8, 4, 8, 4, 8, 4, 8, 4, 8, 4, 8 }}, /* 1100 4 8 */
    {{ 1, 4, 8, 1, 4, 8, 1, 4, 8, 1, 4, 8 }}, /* 1101 1 4 8 */
    {{ 2, 4, 8, 2, 4, 8, 2, 4, 8, 2, 4, 8 }}, /* 1110 2 4 8 */
    {{ 1, 2, 4, 8, 1, 2, 4, 8, 1, 2, 4, 8 }}  /* 1111 1 2 4 8 */
};

static uint8_t randomize_ambiguity(uint8_t base, unsigned *seed, bool *reseed) {
    assert(base < 16);
    
    if (isExact[base]) {
        typedef uint8_t dna2na_t;

        static const dna2na_t map4na2na[16]
	  = { 0, 0, 1, 0, 2, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0 };
    
        *seed = (*seed << 2) | map4na2na[base];
	*reseed = true;
	return base;
    }
    else {
        unsigned rval;

	if (*reseed) {
	    *reseed = false;
	    srand(*seed);
	}
        rval = ((uint8_t)(rand() / (RAND_MAX >> 8))) % 12;
	assert(rval < 12);
	return possibleBase[base].val[rval];
    }
}
#endif

static
rc_t CC rand_4na_2na_impl  ( void *data,
    const VXformInfo *info, int64_t row_id, const VFixedRowResult *rslt,
    uint32_t argc, const VRowData argv [] )
{
    unsigned i;
    unsigned seed = 1;
    bool reseed = true;
    uint8_t *dst = rslt -> base;
    const uint8_t *src = argv [ 0 ] . u . data . base;
    
    dst += rslt ->first_elem;
    src += argv[0].u.data.first_elem;
    
    assert(rslt->elem_count == argv[0].u.data.elem_count);
    assert(rslt->elem_count == (unsigned)rslt->elem_count);
    
    for ( i = 0; i != rslt -> elem_count; ++ i )
        dst[i] = randomize_ambiguity(src[i], &seed, &reseed);
    return 0;
}


VTRANSFACT_IMPL ( INSDC_SEQ_rand_4na_2na, 1, 0, 0 ) ( const void *self, const VXfactInfo *info,
    VFuncDesc *rslt, const VFactoryParams *cp, const VFunctionParams *dp )
{
    rslt -> self = NULL;
    rslt -> whack = NULL;
    rslt -> u . pf = rand_4na_2na_impl;
    rslt -> variant = vftFixedRow;

    return 0;
}
