/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 */

#ifndef _h_page_map_
#define _h_page_map_

#ifndef _h_vdb_extern_
#include <vdb/extern.h>
#endif

#ifndef _h_klib_defs_
#include <klib/defs.h>
#endif

struct KDataBuffer;
typedef struct PageMap PageMap;

/* a pessimistic estimate - actual size will *always* be less, except when it's not */
size_t PageMapGetMaxBufferSize( const PageMap *self );

rc_t PageMapSerialize( const PageMap *self, struct KDataBuffer *buffer, uint64_t byte_offset, uint64_t *actual_bytes );

rc_t PageMapDeserialize( PageMap **lhs, const void *src, uint64_t src_bytes, uint64_t row_count );

rc_t PageMapRelease( const PageMap *self );

rc_t PageMapAddRef( const PageMap *self );

/*  PageMapGetIdxRowInfo
 *
 *  Get row length and starting element number for a row.
 *  This is a potentially expensive operation; the first time
 *  it is called for any page map, the memory used by the page
 *  map nearly doubles and an index is built on the row length
 *  runs and the data runs.  Subsequent calls are O(N log N).
 *  However, asking for the information about the first row
 *  (i.e. idx = 0) never causes allocation or indexing and is
 *  always O(1).
 *
 *  Params:
 *      self: [IN] the page map
 *      idx:  the row number starting from 0
 *      starting_element: [OUT, OPTIONAL]
 *
 *  Returns:
 *      the length of the row
 *      or 0 if not found
 */
uint32_t PageMapGetIdxRowInfo( const PageMap *self, uint32_t idx, uint32_t *starting_element );

rc_t PageMapNew( PageMap **lhs, uint32_t reserve );

rc_t PageMapNewSingle( PageMap **lhs, uint64_t row_count, uint64_t row_length );

rc_t PageMapNewFixedRowLength( PageMap **lhs, uint64_t row_count, uint64_t row_len );

uint32_t PageMapFastRowCount( const PageMap *self );

uint32_t PageMapFixedRowLength( const PageMap *self );

uint32_t PageMapHasSimpleStructure( const PageMap *self );

rc_t PageMapAppendRows(PageMap *self, uint64_t row_length, uint64_t run_length, bool same_data);

#define PageMapAppendRow(SELF, ROW_LENGTH, SAME_DATA) (PageMapAppendRows((SELF), (ROW_LENGTH), 1, SAME_DATA))

/* append some rows of the same data */
#define PageMapAppendSomeRows(SELF, ROW_LENGTH, RUN_LENGTH) (PageMapAppendRows((SELF), (ROW_LENGTH), (RUN_LENGTH), false))

/* concatenate two page maps */
rc_t PageMapAppend( PageMap *self, const PageMap *other );


/*
 -1: error
 0: not same
 1: compatible (i.e. all rows same length)
 else: identical
 */
int PageMapCompare( const PageMap *a, const PageMap *b );
/* same but static columns */
int PageMapCompareStatic( const PageMap *a, const PageMap *b );

typedef struct PageMapIterator PageMapIterator;
struct PageMapIterator {
    const uint32_t *length;
    const uint32_t *offset;
    uint64_t end_row;
    uint64_t cur_row;
    int variant;
};

rc_t PageMapNewIterator( const PageMap *self, PageMapIterator *lhs, uint64_t first_row, uint64_t num_rows );

#define PageMapIteratorNext(SELF) \
    (((SELF).cur_row + 1 >= (SELF).end_row) ? false : ( \
            (++(SELF).cur_row), \
            true \
        ) \
    )

#define PageMapIteratorRowLength(SELF) ((SELF).variant == 0 ? (SELF).length[0] : ( (SELF).variant == 1 ? (SELF).length[0] : (SELF).length[(SELF).cur_row]))
#define PageMapIteratorOffset(SELF) ((SELF).variant == 0 ? 0 : ( (SELF).variant == 1 ? (SELF).length[0] * (SELF).cur_row : (SELF).offset[(SELF).cur_row]))

/* Page map compressed iterator
 * for iterating a page map in its compressed form
 */
typedef struct PageMapCIterator PageMapCIterator;
struct PageMapCIterator {
    const uint32_t *length;
    const uint32_t *leng_reps;
    const uint32_t *data_reps;
    uint32_t lcnt;  /* length record count */
    uint32_t dcnt;  /* data record count */
    uint32_t cleng; /* current length record */
    uint32_t cdata; /* current data record */
    uint32_t lrc;   /* length repeat counter */
    uint32_t drc;   /* data repeat counter */
    uint32_t doff;  /* current data offset */
    uint32_t variant;
};

rc_t PageMapNewCIterator( const PageMap *self, PageMapCIterator *lhs );

static __inline__ bool PageMapCIteratorAdvance(PageMapCIterator *self, uint32_t rows) {
    if (self->cleng == self->lcnt || self->cdata == self->dcnt)
        return false;
    
    if (rows == 0)
        return true;
    
    switch (self->variant) {
    case 3:
    case 1:
        while (rows >= self->drc) {
            rows -= self->drc;
            self->doff += self->length[self->cleng];
            
            self->lrc -= self->drc;
            if (self->lrc == 0) {
                if (++self->cleng == self->lcnt)
                    rows = 0;
                else
                    self->lrc = self->leng_reps[self->cleng];
            }
            if (++self->cdata == self->dcnt)
                rows = 0;
            else
                self->drc = self->data_reps[self->cdata];
        }
        self->drc -= rows;
        self->lrc -= rows;
        return true;
    case 0:
        if (self->drc < rows) {
            self->cleng = self->lcnt;
            self->cdata = self->dcnt;
            return false;
        }
        self->drc -= rows;
        return true;
    case 2:
        if (self->lrc < rows) {
            self->cleng = self->lcnt;
            self->cdata = self->dcnt;
            return false;
        }
        self->lrc -= rows;
        self->doff += rows * self->length[0];
        return true;
    default:
        break;
    }
    
}

#define PageMapCIteratorRepeatCount(SELF) (((const PageMapCIterator *)(&(SELF)))->drc)
#define PageMapCIteratorDataOffset(SELF) (((const PageMapCIterator *)(&(SELF)))->doff)
#define PageMapCIteratorDataLength(SELF) (((const PageMapCIterator *)(&(SELF)))->length[((const PageMapCIterator *)(&(SELF)))->cleng])

#endif /* _h_page_map_ */
