/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <kapp/main.h>
#include <klib/log.h>
#include <klib/rc.h>
#include <kfs/directory.h>

#include <sra/types.h>
#include <sra/sradb.h>

#include <os-native.h>
#include <sysalloc.h>

#include <stdio.h>
#include <stdlib.h>
#if 0
#include <unistd.h>
#endif
#include <assert.h>
#include <string.h>
#include <ctype.h>
#if 0
#include <sys/stat.h>
#include <sys/errno.h>
#endif

#include "abi-dump.vers.h"
#include "debug.h"

#define MIN_READ_LEN 25      /* Max read length that will be output */
#define MAX_READ_NUM 4       /* Max number of reads expected within a spot */
#define MAX_QUALITY 93       /* Max expected quality value from abi */
#define MAX_ABI_NAME_SIZE 100 /* Max expected name size for abi file output file */

#if USE_NCBI_SRADB
static char const defaultRepServer[] = "/panfs/traces01.be-md.ncbi.nlm.nih.gov";
static char const defaultVolume[] = "sra0";
#endif

typedef struct
{
    FILE *fastaF3;
    FILE *qualF3;
    FILE *sigftcF3;
    FILE *sigcy3F3;
    FILE *sigtxrF3;
    FILE *sigcy5F3;
    char *fastaF3Name;
    char *qualF3Name;
    char *sigftcF3Name;
    char *sigcy3F3Name;
    char *sigtxrF3Name;
    char *sigcy5F3Name;

    FILE *fastaR3;
    FILE *qualR3;
    FILE *sigftcR3;
    FILE *sigcy3R3;
    FILE *sigtxrR3;
    FILE *sigcy5R3;
    char *fastaR3Name;
    char *qualR3Name;
    char *sigftcR3Name;
    char *sigcy3R3Name;
    char *sigtxrR3Name;
    char *sigcy5R3Name;
} abiFilePtrs;

typedef struct
{
    const SRAColumn *nameCol;
    const SRAColumn *readCol;
    const SRAColumn *csCol;
    const SRAColumn *sdescCol;
    const SRAColumn *rdescCol;
    const SRAColumn *qualCol;
    const SRAColumn *platCol;
    const SRAColumn *sigCol;
} colPtrs;

/* Collection of constraints on dumping to abi files */

typedef struct
{
    spotid_t minSpotId;
    spotid_t maxSpotId;
    int minReadLen;
    int applyClip;
    int dumpOrigFmt;
    int noDotReads;
    int noSignal;
    const uint8_t *platform;
} dumpConstraints;

/*******************************************************************************
 * KAppVersion is needed to utilize the common main defined in the kapp library
 *******************************************************************************/
ver_t CC KAppVersion ( void )
{
    return ABI_DUMP_VERS ;
}

static KDirectory *pwd;
static KDirectory *wd;

static
rc_t init_wd ()
{
    rc_t rc = KDirectoryNativeDir ( & pwd );
    if ( rc == 0 ) 
    {
        rc = KDirectoryAddRef ( pwd );
        if ( rc != 0 )
            KDirectoryRelease ( pwd ), pwd = NULL;
        else
            wd = pwd;
    }
    return rc;
}

static
void whack_wd ()
{
    KDirectoryRelease ( wd );
    KDirectoryRelease ( pwd );
}


static
int abi_chdir ( const char *new_path )
{
    int res = 1;
    rc_t rc;
    
    uint32_t path_type = KDirectoryPathType ( pwd, new_path );
    if ( path_type != kptDir )
    {
        rc = KDirectoryCreateDir ( pwd, 0775, kcmCreate, new_path );
        res = ( rc == 0 );
    }

    if ( res )
    {
        KDirectory *temp_dir;
        
        res = 0;
        rc = KDirectoryOpenDirUpdate( wd, &temp_dir, false, new_path );
        if ( rc == 0 )
        {
            KDirectoryRelease ( wd );
            wd = temp_dir;
            res = 1;
        }
    }
    
    return ( res );
}

#ifdef WINDOWS
static
void translate_path_to_native( char *path )
{
    if ( isalpha ( path [ 1 ] ) && path [ 2 ] == '/' )
    {
        /* rewrite drive letter */
        path [ 0 ] = toupper ( path [ 1 ] );
        path [ 1 ] = ':';
    }
}
#endif

static
FILE *abi_fopen ( const char *path, const char *mode )
{
/* use "wd" to generate a full path */
/* return fopen on real path */
    char abs_path[ 1024 ];
    FILE *res = NULL;
    
    rc_t rc = KDirectoryResolvePath ( wd, true, abs_path, sizeof abs_path, path );
    if ( rc == 0 )
    {
#ifdef WINDOWS    
        translate_path_to_native( abs_path );
#endif
        res = fopen ( abs_path, mode );
/*        PLOGMSG ( klogInfo , ( klogInfo , "abi_fopen: $(f) success" , "f=%s" , abs_path ) ); */
    }
    else
    {
        PLOGERR ( klogErr, ( klogErr, rc , "abi_fopen: $(f)", "d=%s", path ) ); 
     }
    return res;
}

/*******************************************************************************
 * abiDumpFileCloseRemoveEmpty - removes a file if it is empty
 *******************************************************************************/
static
void abiDumpFileCloseRemoveEmpty ( FILE *file, char **filename )
{
#if 0
    struct stat fileStat;
    
    fclose (file);
    
    if ( stat( *filename , &fileStat ) == 0 )
        {
            if(fileStat.st_size == 0)
                {
                    if(remove(*filename) != 0)
                        PLOGMSG ( klogErr , ( klogErr , "abiDumpFileCloseRemoveEmpty: Unable to remove empty file $(f)" ,
                                              "f=%s" , *filename ));
                }
        }
    else
        PLOGMSG ( klogErr , ( klogErr , "abiDumpFileCloseRemoveEmpty: Unable to stat file $(f): $(s)" ,
                              "f=%s,s=%s", *filename, strerror(errno) ));
#else
    uint64_t fsize;
    rc_t rc;
    
    fclose (file);
    
    rc = KDirectoryFileSize(wd, &fsize, *filename);
    if (rc == 0) {
        if (fsize == 0) {
            rc = KDirectoryRemove(wd, true, *filename);
            if (rc)
                PLOGERR ( klogErr , ( klogErr, rc , "abiDumpFileCloseRemoveEmpty: Unable to delete empty file $(f)" ,
                                     "f=%s", *filename));
        }
    }
    else
        PLOGERR ( klogErr , ( klogErr, rc , "abiDumpFileCloseRemoveEmpty: Unable to stat file $(f)" ,
                             "f=%s", *filename));
#endif
    free (*filename);
}

/*******************************************************************************
 * abiDumpFileName
 *******************************************************************************/
static
char *abiDumpFileName ( char const *accession, char const *label, char const *suffix, char **name )
{
    **name = 0;
    strcpy(*name,accession);
    if(label != NULL)
        {
            strcat(*name,"_");
            strcat(*name,label);
        }
    strcat(*name,suffix);
    return *name;
}

/*******************************************************************************
 * abiDumpFilesOpen - opens abi files for output (potentially 1 fragment file
 * and two pair files)
 *******************************************************************************/
static
rc_t abiDumpFilesOpen ( char const *accession, abiFilePtrs *abiFiles,
                        dumpConstraints *constraints )
{
    rc_t rc;
    char label1[3];
    char label2[3];

    abiFiles->fastaF3 = NULL;
    abiFiles->qualF3 = NULL;
    abiFiles->sigftcF3 = NULL;
    abiFiles->sigcy3F3 = NULL;
    abiFiles->sigtxrF3 = NULL;
    abiFiles->sigcy5F3 = NULL;
    abiFiles->fastaF3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->qualF3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigftcF3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigcy3F3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigtxrF3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigcy5F3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );

    abiFiles->fastaR3 = NULL;
    abiFiles->qualR3 = NULL;
    abiFiles->sigftcR3 = NULL;
    abiFiles->sigcy3R3 = NULL;
    abiFiles->sigtxrR3 = NULL;
    abiFiles->sigcy5R3 = NULL;
    abiFiles->fastaR3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->qualR3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigftcR3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigcy3R3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigtxrR3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );
    abiFiles->sigcy5R3Name = (char *) malloc ( MAX_ABI_NAME_SIZE );

    if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID )
        {
            strcpy ( label1, "F3" );
            strcpy ( label2, "R3" );
        }
    else
        {
            strcpy ( label1, "1" );
            strcpy ( label2, "2" );
        }

    rc = RC ( rcExe, rcFile, rcOpening, rcNoObj, rcNull);
    
    if ( ( abiFiles->fastaF3 = abi_fopen ( abiDumpFileName ( accession, label1, ".csfasta",
                                                         &abiFiles->fastaF3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open fastaF3 file");

    else if ( ( abiFiles->qualF3 = abi_fopen ( abiDumpFileName ( accession, label1, "_QV.qual",
                                                             &abiFiles->qualF3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open qualF3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigftcF3 = abi_fopen ( abiDumpFileName ( accession, label1, "_intensity.ScaledFTC.fasta",
                                                               &abiFiles->sigftcF3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigftcF3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigcy3F3 = abi_fopen ( abiDumpFileName ( accession, label1, "_intensity.ScaledCY3.fasta",
                                                               &abiFiles->sigcy3F3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigcy3F3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigtxrF3 = abi_fopen ( abiDumpFileName ( accession, label1, "_intensity.ScaledTXR.fasta",
                                                               &abiFiles->sigtxrF3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigtxrF3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigcy5F3 = abi_fopen ( abiDumpFileName ( accession, label1, "_intensity.ScaledCY5.fasta",
                                                               &abiFiles->sigcy5F3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigcy5F3 file");

    else if ( ( abiFiles->fastaR3 = abi_fopen ( abiDumpFileName ( accession, label2, ".csfasta",
                                                              &abiFiles->fastaR3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open fastaR3 file");

    else if ( ( abiFiles->qualR3 = abi_fopen ( abiDumpFileName ( accession, label2,"_QV.qual",
                                                             &abiFiles->qualR3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open qualR3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigftcR3 = abi_fopen ( abiDumpFileName ( accession, label2, "_intensity.ScaledFTC.fasta",
                                                               &abiFiles->sigftcR3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigftcR3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigcy3R3 = abi_fopen ( abiDumpFileName ( accession, label2, "_intensity.ScaledCY3.fasta",
                                                               &abiFiles->sigcy3R3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigcy3R3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigtxrR3 = abi_fopen ( abiDumpFileName ( accession, label2, "_intensity.ScaledTXR.fasta",

                                                               &abiFiles->sigtxrR3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigtxrR3 file");

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              ( abiFiles->sigcy5R3 = abi_fopen ( abiDumpFileName ( accession, label2, "_intensity.ScaledCY5.fasta",
                                                               &abiFiles->sigcy5R3Name ), "wb" ) ) == NULL )
        LOGERR ( klogFatal, rc, "abiDumpFilesOpen: Failed to open sigcy5R3 file");

    else
        rc = rcNoErr;

    return rc;
}

/*******************************************************************************
 * abiDumpFilesClose - close generated abi files, removing empty
 * files if they exist.
 *******************************************************************************/
static
rc_t abiDumpFilesClose ( char const *accession, abiFilePtrs *abiFiles )
{
    if ( abiFiles->fastaF3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->fastaF3, &abiFiles->fastaF3Name );

    if ( abiFiles->qualF3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->qualF3, &abiFiles->qualF3Name );

    if ( abiFiles->sigftcF3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigftcF3, &abiFiles->sigftcF3Name );

    if ( abiFiles->sigcy3F3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigcy3F3, &abiFiles->sigcy3F3Name );

    if ( abiFiles->sigtxrF3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigtxrF3, &abiFiles->sigtxrF3Name );

    if ( abiFiles->sigcy5F3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigcy5F3, &abiFiles->sigcy5F3Name );

    if ( abiFiles->fastaR3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->fastaR3, &abiFiles->fastaR3Name );

    if ( abiFiles->qualR3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->qualR3, &abiFiles->qualR3Name );

    if ( abiFiles->sigftcR3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigftcR3, &abiFiles->sigftcR3Name );

    if ( abiFiles->sigcy3R3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigcy3R3, &abiFiles->sigcy3R3Name );

    if ( abiFiles->sigtxrR3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigtxrR3, &abiFiles->sigtxrR3Name );

    if ( abiFiles->sigcy5R3 != NULL )
        abiDumpFileCloseRemoveEmpty ( abiFiles->sigcy5R3, &abiFiles->sigcy5R3Name );

    return rcNoErr;
}

/*******************************************************************************
 * abiDumpColumnsOpen - open data columns for reading spot data
 *******************************************************************************/
static
rc_t abiDumpColumnsOpen (const SRATable *sraTable, colPtrs *columns)
{
    rc_t rc;

    columns->nameCol = columns->readCol = columns->qualCol = NULL;
    columns->csCol = columns->rdescCol = columns->sdescCol = columns->sigCol = NULL;

    if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->nameCol, "NAME", vdb_ascii_t ) )
         != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open NAME column" );

    else if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->csCol, "CSREAD", insdc_csfasta_t ) )
              != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open CSREAD column" );

    else if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->qualCol, "QUALITY", ncbi_qual1_t ) )
              != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open QUALITY column" );

    else if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->sdescCol, "SPOT_DESC", sra_spot_desc_t ) )
              != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open SPOT_DESC column" );

    else if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->rdescCol, "READ_DESC", sra_read_desc_t ) )
              != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open READ_DESC column" );

    else if ( ( rc = SRATableOpenColumnRead ( sraTable, &columns->platCol, "PLATFORM", sra_platform_id_t ) )
              != rcNoErr )
        LOGERR ( klogFatal , rc , "abiDumpColumnsOpen: Unable to open PLATFORM column" );

    else /* Signal may not be present */
        {
            SRATableOpenColumnRead ( sraTable, &columns->sigCol, "SIGNAL", ncbi_fsamp4_t) ;
            rc = rcNoErr;
        }

    return rc;
}

/*******************************************************************************
 * abiDumpColumnsRelease - release data columns after reading spot data from them
 *******************************************************************************/
static
rc_t abiDumpColumnsRelease (colPtrs *columns)
{
    rc_t rc = 0;

    if ( columns->nameCol != NULL &&
         ( rc = SRAColumnRelease ( columns->nameCol ) ) != rcNoErr )
        LOGERR( klogErr , rc, "abiDumpColumnsRelease: SRAColumnRelease error on NAME col" );

    if ( columns->csCol != NULL &&
         ( rc = SRAColumnRelease ( columns->csCol ) ) != rcNoErr )
        LOGERR( klogErr , rc, "abiDumpColumnsRelease: SRAColumnRelease error on CSREAD col" );

    if ( columns->qualCol != NULL &&
         ( rc = SRAColumnRelease ( columns->qualCol ) ) != rcNoErr )
        LOGERR( klogErr , rc , "abiDumpColumnsRelease: SRAColumnRelease error on QUALITY col" );
    
    if ( columns->sdescCol != NULL &&
         ( rc = SRAColumnRelease ( columns->sdescCol ) ) != rcNoErr )
        LOGERR( klogErr , rc , "abiDumpColumnsRelease: SRAColumnRelease error on SPOT_DESC col" );
    
    if ( columns->rdescCol != NULL &&
         ( rc = SRAColumnRelease ( columns->rdescCol ) ) != rcNoErr )
        LOGERR( klogErr , rc , "abiDumpColumnsRelease: SRAColumnRelease error on READ_DESC col" );
    
    if ( columns->sigCol != NULL &&
         ( rc = SRAColumnRelease ( columns->sigCol ) ) != rcNoErr )
        LOGERR( klogErr , rc , "abiDumpColumnsRelease: SRAColumnRelease error on SIGNAL col" );

    if ( columns->platCol != NULL &&
         ( rc = SRAColumnRelease ( columns->platCol ) ) != rcNoErr )
        LOGERR( klogErr , rc , "abiDumpColumnsRelease: SRAColumnRelease error on PLATFORM col" );

    return rc;
}

/*******************************************************************************
 * abiDumpReadHdr - Outputs header line to abi file
 *******************************************************************************/
static
rc_t abiDumpReadHdr ( FILE *abiFile, const char *accession, spotid_t spot,
                      const char *spotName, int spotNameSize, const SRAReadDesc *rdesc,
                      const int readIdx, dumpConstraints *constraints )
{
    if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID )
        {
            if ( constraints->dumpOrigFmt == true )
                fprintf ( abiFile, ">%.*s_%s\n", spotNameSize, spotName, rdesc[readIdx].label );
            else
                fprintf ( abiFile, ">%s.%u %.*s_%s\n",
                          accession, (unsigned int) spot, spotNameSize, spotName, rdesc[readIdx].label );
        }
    else
        {
            if ( constraints->dumpOrigFmt == true )
                fprintf ( abiFile, ">%.*s\n", spotNameSize, spotName );
            else
                fprintf ( abiFile, ">%s.%u %.*s\n",
                          accession, (unsigned int) spot, spotNameSize, spotName);
        }
    return rcNoErr;
}

/*******************************************************************************
 * abiDumpReadCsFasta - Outputs csfasta line to csfasta file
 *******************************************************************************/
static
rc_t abiDumpReadCsFasta ( FILE *abiFile, const SRAReadDesc *rdesc,
                          const int readIdx, const char *cs, const uint16_t *readLen )
{
    fprintf ( abiFile, "%c", (char) rdesc[readIdx].cs_key);
    fprintf ( abiFile, "%.*s\n", readLen[readIdx], cs + rdesc[readIdx].seg.start );
    return rcNoErr;
}

/*******************************************************************************
 * abiDumpReadQual - Outputs quality line to quality file
 *******************************************************************************/
static
rc_t abiDumpReadQual ( FILE *abiFile, const SRAReadDesc *rdesc,
                       const int readIdx, const uint8_t *quality, const uint16_t *readLen )
{
    int i;
    for ( i = 0 ; i < readLen[readIdx] ; i++ )
        fprintf ( abiFile, "%u ", *(quality + rdesc[readIdx].seg.start + i) );
    fprintf ( abiFile, "\n");
    return rcNoErr;
}

/*******************************************************************************
 * abiDumpReadSig - Outputs signal line to signal file
 *******************************************************************************/
static
rc_t abiDumpReadSig ( FILE *abiFile, const SRAReadDesc *rdesc, const int readIdx,
                      const float *sig, int offset, const uint16_t *readLen )
{
    int i;
    for ( i = 0 ; i < readLen[readIdx] ; i++ )
        fprintf ( abiFile, "%.6g ", *(sig + (i+rdesc[readIdx].seg.start)*4 + offset ) );
    fprintf ( abiFile, "\n");
    return rcNoErr;
}

/*******************************************************************************
 * abiDumpRead - Outputs cs or base abi for a single read to a abi file
 *******************************************************************************/
static
rc_t abiDumpRead ( abiFilePtrs *abiFiles, const char *accession, spotid_t spot,
                   const char *spotName, int spotNameSize, const char *cs,
                   const uint8_t *quality, const float *sig,
                   const SRAReadDesc *rdesc, const int readIdx,
                   dumpConstraints *constraints, const uint16_t *readLen )
{
    rc_t rc;
    rc = rcNoErr;

    /* Print to F3 files */

    if ( ! ( strcmp ( rdesc[readIdx].label , "F3" ) ) ||
         ( *(constraints->platform) == SRA_PLATFORM_ILLUMINA &&
           readIdx == 0 ) ||
         ( *(constraints->platform) == SRA_PLATFORM_454 &&
           readIdx == 1 ) )
        {
            /* Output csfasta */

            abiDumpReadHdr ( abiFiles->fastaF3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
            abiDumpReadCsFasta ( abiFiles->fastaF3, rdesc, readIdx, cs, readLen );

            /* Output qual */

            abiDumpReadHdr ( abiFiles->qualF3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
            abiDumpReadQual ( abiFiles->qualF3, rdesc, readIdx, quality, readLen );

            if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
                 sig != NULL &&
                 constraints->noSignal == false )
                {
                    /* Output ftc */

                    abiDumpReadHdr ( abiFiles->sigftcF3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigftcF3, rdesc, readIdx, sig, 0, readLen );

                    /* Output cy3 */

                    abiDumpReadHdr ( abiFiles->sigcy3F3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigcy3F3, rdesc, readIdx, sig, 1, readLen );

                    /* Output txr */

                    abiDumpReadHdr ( abiFiles->sigtxrF3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigtxrF3, rdesc, readIdx, sig, 2, readLen );

                    /* Output cy5 */

                    abiDumpReadHdr ( abiFiles->sigcy5F3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigcy5F3, rdesc, readIdx, sig, 3, readLen );
                }
        }

    /* Print to R3 files */

    else if ( ! ( strcmp ( rdesc[readIdx].label , "R3" ) )  ||
              ( *(constraints->platform) == SRA_PLATFORM_ILLUMINA &&
                readIdx == 1 ) ||
              ( *(constraints->platform) == SRA_PLATFORM_454 &&
                readIdx == 3 ) )
        {
            /* Output csfasta */

            abiDumpReadHdr ( abiFiles->fastaR3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
            abiDumpReadCsFasta ( abiFiles->fastaR3, rdesc, readIdx, cs, readLen );

            /* Output qual */

            abiDumpReadHdr ( abiFiles->qualR3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
            abiDumpReadQual ( abiFiles->qualR3, rdesc, readIdx, quality, readLen );

            if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
                 sig != NULL  &&
                 constraints->noSignal == false )
                {
                    /* Output ftc */

                    abiDumpReadHdr ( abiFiles->sigftcR3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigftcR3, rdesc, readIdx, sig, 0, readLen );

                    /* Output cy3 */

                    abiDumpReadHdr ( abiFiles->sigcy3R3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigcy3R3, rdesc, readIdx, sig, 1, readLen );

                    /* Output txr */

                    abiDumpReadHdr ( abiFiles->sigtxrR3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigtxrR3, rdesc, readIdx, sig, 2, readLen );

                    /* Output cy5 */

                    abiDumpReadHdr ( abiFiles->sigcy5R3, accession, spot, spotName, spotNameSize, rdesc, readIdx, constraints );
                    abiDumpReadSig ( abiFiles->sigcy5R3, rdesc, readIdx, sig, 3, readLen );
                }
        }

    /* Unrecognized label */

    else
        {
            PLOGMSG ( klogFatal, ( klogFatal, "abiDumpRead: Unexpected label $(s) encountered\n","s=%s",rdesc[readIdx].label));
            rc = RC ( rcExe, rcFile, rcReading, rcData, rcUnexpected );
        }

    return rc;
}

/*******************************************************************************
 * abiDumpSpotNameOffset - gets offset to beyond common prefix
 *******************************************************************************/
static
int abiDumpSpotNameOffset ( const char *spotName, size_t nm_sz, const uint8_t platform )
{
    int spotNameOffset;

    if ( platform == SRA_PLATFORM_454 )
        {
            spotNameOffset = 0;
        }
    else
        {
            spotNameOffset = nm_sz - 1;
            while ( isdigit ( *(spotName + spotNameOffset) ) ) spotNameOffset--;
            if(*(spotName + spotNameOffset) == 'F' ||
               *(spotName + spotNameOffset) == 'R')
                {
                    spotNameOffset--; /* Skip F or R */
                    spotNameOffset--; /* Skip separator */
                    while ( isdigit ( *(spotName + spotNameOffset) ) ) spotNameOffset--;
                }
            spotNameOffset--; /* Skip separator */
            while ( isdigit ( *(spotName + spotNameOffset) ) ) spotNameOffset--;
            spotNameOffset--; /* Skip separator */
            while ( isdigit ( *(spotName + spotNameOffset) ) ) spotNameOffset--;

            if ( platform == SRA_PLATFORM_ILLUMINA)
                {
                    spotNameOffset--; /* Skip separator */
                    while ( isdigit ( *(spotName + spotNameOffset) ) ) spotNameOffset--;
                }

            spotNameOffset++; /* Back to first digit */
        }

    return spotNameOffset;
}

/*******************************************************************************
 * abiDumpSpotPrefix - one time output of ABI Spot Prefix to every file
 *******************************************************************************/
static
rc_t abiDumpSpotPrefix ( abiFilePtrs *abiFiles, char *spotNamePrefix,
                         const SRASpotDesc *sdesc, const SRAReadDesc *rdesc,
                         const float *sig, dumpConstraints *constraints )
{
    int i;

    for ( i = 0 ; i < sdesc->num_reads ; i++ ) /* Loop once or twice */
        {
            if ( rdesc[i].type != SRA_READ_TYPE_BIOLOGICAL)
                continue;

            if ( ! ( strcmp ( rdesc[i].label , "F3" ) ) ||
                 ( *(constraints->platform) == SRA_PLATFORM_ILLUMINA &&
                   i == 0 ) ||
                 ( *(constraints->platform) == SRA_PLATFORM_454 &&
                   i == 1 ) )
                {
                    fprintf ( abiFiles->fastaF3, "# Title: %s\n",spotNamePrefix);
                    fprintf ( abiFiles->qualF3, "# Title: %s\n",spotNamePrefix);
                    if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
                         sig != NULL )
                        {
                            fprintf ( abiFiles->sigftcF3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigcy3F3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigtxrF3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigcy5F3, "# Title: %s\n",spotNamePrefix);
                        }
                }
            else if ( ! ( strcmp ( rdesc[i].label , "R3" ) )  ||
                      ( *(constraints->platform) == SRA_PLATFORM_ILLUMINA &&
                        i == 1 ) ||
                      ( *(constraints->platform) == SRA_PLATFORM_454 &&
                        i == 3 ) )
                {
                    fprintf ( abiFiles->fastaR3, "# Title: %s\n",spotNamePrefix);
                    fprintf ( abiFiles->qualR3, "# Title: %s\n",spotNamePrefix);
                    if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
                         sig != NULL )
                        {
                            fprintf ( abiFiles->sigftcR3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigcy3R3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigtxrR3, "# Title: %s\n",spotNamePrefix);
                            fprintf ( abiFiles->sigcy5R3, "# Title: %s\n",spotNamePrefix);
                        }
                }
        }

    return rcNoErr;
}

/*******************************************************************************
 * abiDumpSpot - outputs abi for a single spot to appropriate abi files
 *******************************************************************************/
static
rc_t abiDumpSpot ( spotid_t spot, char const *accession, colPtrs *columns,
                   abiFilePtrs *abiFiles, dumpConstraints *constraints )
{
    const char *spotName;
    bitsz_t nameOffset, nameSize;
    const char *cs;
    bitsz_t csOffset, csSize;
    const uint8_t *qual;
    bitsz_t qualOffset, qualSize;
    const SRASpotDesc *sdesc;
    bitsz_t sdescOffset, sdescSize;
    const SRAReadDesc *rdesc;
    bitsz_t rdescOffset, rdescSize;
    const float *sig;
    bitsz_t sigOffset, sigSize;
    int i;
    rc_t rc;
    uint16_t readLen[MAX_READ_NUM+1];
    static int spotNameOffset = -1;
    static char spotNamePrefix[500];
    for ( i = 0; i <= MAX_READ_NUM ; i++ )
        readLen[i] = 0;

    spotName = NULL;
    cs = NULL;
    qual = NULL;
    sig = NULL;
    sdesc = NULL;
    rdesc = NULL;
    nameOffset = nameSize = 0;
    csOffset = csSize = 0;
    qualOffset = qualSize = 0;
    sigOffset = sigSize = 0;
    sdescOffset = sdescSize = 0;
    rdescOffset = rdescSize = 0;

    /* Just for testing: make use of the static variable wd ( to see if it is still valid ) */

    /* Get spot name */

    rc = SRAColumnRead ( columns->nameCol, spot, (const void **) &spotName,
                         &nameOffset, &nameSize );

    if ( rc != rcNoErr )
    {
        PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpSpot: Unable to read spot name for spot $(s)" ,
                               "s=%d" , (int) spot ));
    }

    if ( (nameOffset % 8 != 0) || ( (nameSize % 8) != 0) )
    {
       PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpSpot: unexpected Read(spotName) results for spot $(s)",
                               "s=%d" , (int) spot) );
	   rc = -1;
    }

    /* Get cs sequence */

    else if ( ( rc = SRAColumnRead ( columns->csCol, spot, (const void **) &cs,
                                     &csOffset,  &csSize ) ) != rcNoErr )
        PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpSpot: Unable to read cs for spot $(s)" ,
                               "s=%d" , (int) spot ));

    /* Get quality */

    else if ( ( rc = SRAColumnRead ( columns->qualCol, spot, (const void **) &qual,
                                     &qualOffset,  &qualSize ) ) != rcNoErr )
        PLOGERR ( klogFatal ,  (klogFatal , rc , "abiDumpSpot: Unable to read qualily for spot $(s)" ,
                                "s=%d" , (int) spot ));

    /* Get signal */

    else if ( *(constraints->platform) == SRA_PLATFORM_ABSOLID &&
              columns->sigCol != NULL &&
              constraints->noSignal == false &&
              ( rc = SRAColumnRead ( columns->sigCol, spot, (const void **) &sig,
                                     &sigOffset,  &sigSize ) ) != rcNoErr )
        PLOGERR ( klogFatal ,  (klogFatal , rc , "abiDumpSpot: Unable to read signal for spot $(s)" ,
                                "s=%d" , (int) spot ));

    /* Get spot descriptor */

    else if ( ( rc = SRAColumnRead ( columns->sdescCol, spot, (const void **) &sdesc,
                                     &sdescOffset,  &sdescSize ) ) != rcNoErr )
        PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpSpot: Unable to read sdesc for spot $(s)" ,
                               "s=%d" , (int) spot ));

    /* Get read descriptor */

    else if ( ( rc = SRAColumnRead ( columns->rdescCol, spot, (const void **) &rdesc,
                                     &rdescOffset,  &rdescSize ) ) != rcNoErr )
        PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpSpot: Unable to read rdesc for spot $(s)" ,
                               "s=%d" , (int) spot ));

    if ( rc == rcNoErr )
        {
            assert ( csSize == qualSize );
			assert ( sdesc->num_reads < 5 );

			/* Execute one time initializations */

	/* these values are in bits, make them in bytes... */
	spotName += nameOffset / 8;

            if ( spotNameOffset == -1 )
                {
                    if ( ( spotNameOffset = abiDumpSpotNameOffset ( spotName, nameSize / 8, *(constraints->platform) ) ) != 0 )
					{
                        strncpy ( spotNamePrefix, spotName, spotNameOffset);
					}

                    rc = abiDumpSpotPrefix ( abiFiles, spotNamePrefix, sdesc, rdesc, sig, constraints );
                }

            
            /* Apply right clip or just copy lengths */

            for ( i = 0 ; i < sdesc->num_reads ; i++ )
                {
                    if ( constraints->applyClip &&
                         sdesc->clip_qual_right < ( rdesc[i].seg.start + rdesc[i].seg.len ) )
                        {
                            if( rdesc[i].seg.start > sdesc->clip_qual_right )
                                readLen[i] = 0;
                            else
                                readLen[i] = (uint16_t) ( sdesc->clip_qual_right - rdesc[i].seg.start ) ;
                        }
                    else
                        readLen[i] = rdesc[i].seg.len;
                }

            /* Output F3 or R3 reads */

            for ( i = 0 ; i < sdesc->num_reads && rc == rcNoErr ; i++ ) /* Loop once or twice */
                {
                    if ( rdesc[i].type != SRA_READ_TYPE_BIOLOGICAL ||
                         readLen[i] < constraints->minReadLen ||
                         ( constraints->noDotReads == true &&
                           ( ! ( strncmp ( cs + rdesc[i].seg.start, "..........", 10) ) ) &&
                           ( ! ( strncmp ( cs + rdesc[i].seg.start + readLen[i] - 10 , "..........", 10) ) ) ) )
                        continue;
                    else
                        rc = abiDumpRead ( abiFiles, accession, spot, spotName+spotNameOffset, nameSize / 8,
                                           cs, qual, sig, rdesc, i, constraints, readLen);
                }

        }

    return rc;
}

/*******************************************************************************
 * abiDumpColumns - dump data contained in run columns to abi files
 *******************************************************************************/
static
rc_t abiDumpColumns ( const SRATable        *sraTable,
                      char const            *accession,
                      dumpConstraints       *constraints)
{
    colPtrs columns;
    abiFilePtrs abiFiles;
    spotid_t spot;
    rc_t rc, temp_rc;
    bitsz_t platOffset, platSize;

    platOffset = 0;
    platSize = 0;
    rc = rcNoErr;
    temp_rc = rcNoErr;

    /* Open Columns */

    if ( rc == rcNoErr &&
         ( rc = abiDumpColumnsOpen ( sraTable, &columns ) ) != rcNoErr )
        LOGERR ( klogFatal  , rc , "abiDumpColumns: Error occurred while opening columns" );

    /* Get platform for name dimension and orientation output */

    else if ( ( rc = SRAColumnRead ( columns.platCol, constraints->minSpotId, (const void **) &constraints->platform,
                                     &platOffset,  &platSize ) ) != rcNoErr )
        PLOGERR ( klogFatal,  (klogFatal, rc , "abiDumpColumns: Unable to do initial platform read for spot $(s)" ,
                               "s=%d" , (int)constraints->minSpotId ));

    /* Open abi files for output */

    else if ( rc == rcNoErr &&
              ( rc = abiDumpFilesOpen( accession, &abiFiles, constraints ) ) != rcNoErr )
        LOGERR ( klogFatal  , rc , "abiDumpColumns: Error occurred while opening abi files for output" );

    else if ( rc == rcNoErr )
        {
            /* Loop through spots, outputting to abi files */

            for ( spot = constraints->minSpotId ; rc == rcNoErr && spot <= constraints->maxSpotId ; spot++ )

                rc = abiDumpSpot ( spot, accession, &columns, &abiFiles, constraints ) ;

            /* Output error if return code is non-zero */

            if ( rc != rcNoErr )
                LOGERR ( klogErr  , rc , "abiDumpColumns: Error occurred while processing columns" );
        }

    /* Release columns and close files, removing empty files */

    if ( ( temp_rc = abiDumpColumnsRelease ( &columns ) ) != rcNoErr )
        LOGERR ( klogErr  , temp_rc , "abiDumpColumns: Error occurred while releasing columns" );

    /* Close abi files, removing empty files */

    abiDumpFilesClose ( accession, &abiFiles );

    return rc;
}

/*******************************************************************************
 * abiDumpRun - Opens SRAMgr and SRATable for SRR; call abiDumpColumns to output abi files
 *******************************************************************************/
static
rc_t abiDumpRun (
#if USE_NCBI_SRADB
                 char const* const rep, char const* const vol,
#else
                 char const* const path,
#endif
                 char const* const accession,
                 dumpConstraints *constraints)
{
    const SRAMgr        *sraMGR;
    const SRATable      *sraTable;
    spotid_t            maxRunSpotId;
    rc_t                rc, temp_rc; /* return codes */

    KLogLevelSet ( klogLevelMax );

    /* Instantiate the SRAMgr */

    rc = SRAMgrMakeRead ( &sraMGR );
    if (rc) return rc;

#if USE_NCBI_SRADB

    /* Tell SRAMgr about file system */
    rc = SRAMgrAddRepPath ( sraMGR, rep );
    if ( rc == rcNoErr )
        rc = SRAMgrAddVolPath ( sraMGR, vol );

    /* Open a run on the Archive */

    if ( rc == rcNoErr )
        rc = SRAMgrOpenRunRead ( sraMGR, &sraTable, accession );
#else
    rc = SRAMgrOpenTableRead ( sraMGR, & sraTable, "%s/%s", path, accession );
#endif

    if (rc) /* handle error */
        {
            temp_rc = SRAMgrRelease ( sraMGR );
                
            if (temp_rc)
                LOGERR( klogErr , temp_rc , "abiDumpRun: SRAMgrRelease error" );
            return rc;
        }

    /* Determine the max spot id for this archive */

    rc = SRATableMaxSpotId ( sraTable, &maxRunSpotId );

    if (rc) /* handle error */
        {
            temp_rc = SRATableRelease ( sraTable );
                
            if (temp_rc)
                LOGERR( klogErr , temp_rc , "abiDumpRun: SRATableRelease error" );

            temp_rc = SRAMgrRelease ( sraMGR );
                
            if (temp_rc)
                LOGERR( klogErr , temp_rc , "abiDumpRun: SRAMgrRelease error" );

            return rc;
        }
    else /* Check min/max spot ids against run constraints */
        {
            if(constraints->maxSpotId > maxRunSpotId ||
               constraints->maxSpotId == 0)
                constraints->maxSpotId = maxRunSpotId;

            if(constraints->minSpotId > maxRunSpotId)
                constraints->minSpotId = maxRunSpotId;
            else if(constraints->minSpotId <= 0)
                constraints->minSpotId = 1;
        }

    /* Dump data to abi raw files from appropriate columns */

    rc = abiDumpColumns ( sraTable, accession, constraints );

    /* release resources */

    temp_rc = SRATableRelease ( sraTable );

    if (temp_rc)
        LOGERR ( klogErr , temp_rc , "abiDumpRun: SRATableRelease error" );

    temp_rc = SRAMgrRelease ( sraMGR );
                
    if (temp_rc)
        LOGERR ( klogErr , temp_rc , "abiDumpRun: SRAMgrRelease error" );

    return rc;
}

/*******************************************************************************
 * Usage 
 *******************************************************************************/
rc_t CC Usage( const Args * args )
{
    /* dummy for newer args system will be filled when system is complete!!! */
    return 0;
}

static
void Local_Usage ( const char *progName )
{
    fprintf (stderr,
                     "\nUsage:\n"
                     "     %s\n"
#if USE_NCBI_SRADB
                     "     [-repserver <repserver>] [-volume <volume>]\n"
#else
                     "     -path <path-to-run>\n"
#endif
                     "     [-outdir <outdir>] [-minSpotId <minSpotId>] [-maxSpotId <maxSpotId>] \n"
                     "     [-minReadLen <minReadLen>] [-noclip] [-origfmt] [-noDotReads]\n"
                     "     <accession>\n"
                     "\nWhere:\n"
             , progName);

#if USE_NCBI_SRADB
    fprintf (stderr,
                     "     <accession>    Starts with ERR/SRR for production and with XXX for test (e.g. SRR000001)\n"
                     "     -repserver     Replication server path. Default: '%s'\n"
                     "     -volume        Volume for locating runs. Default: '%s'\n"
             , defaultRepServer, defaultVolume
             );
#else
     fprintf (stderr,
                    "     -accession     Run accession - top level directory for SRA format either downloaded or locally generated\n"
                    "     -path          Path to (but not including) accession\n"
              );
#endif

    fprintf (stderr,
                     "     -outdir        Output directory (defaults to current directory)\n"
                     "     -minSpotId     Minimum spot id to output\n"
                     "     -maxSpotId     Maximum spot id to output\n"
                     "     -minReadLen    Minimum read length to output (defaults is 25)\n"
                     "     -noclip        Indicates don't apply clip quality right\n"
                     "     -origfmt       Excludes SRR accession & length on defline\n"
                     "     -noDotReads    No reads consisting only of dots are output\n"
                     "     -noSignal      Do not write sginal files\n"
                     "\n"
             );
}

static
bool tmp_GetArg(char const* const abbr, char const* const full, int* i, int argc, char *argv[], const char** value)
{
    rc_t rc = 0;
    const char* arg = argv[*i];
    while(*arg == '-' && *arg != '\0') {
        arg++;
    }
    if( abbr != NULL && strcmp(arg, abbr) == 0 ) {
        SRA_DUMP_DBG(9, ("GetArg key: '%s'\n", arg));
        arg = arg + strlen(abbr);
        if( value != NULL && arg[0] == '\0' && (*i + 1) < argc ) {
            if( argv[*i + 1][0] != '-' ) {
                /* advance only if next is not an option with '-' */
                *i = *i + 1;
                arg = argv[*i];
            }
        } else {
            arg = NULL;
        }
    } else if( full != NULL && strcmp(arg, full) == 0 ) {
        SRA_DUMP_DBG(9, ("GetArg key: '%s'\n", arg));
        if( value != NULL && (*i + 1) < argc ) {
            if( argv[*i + 1][0] != '-' ) {
                /* advance only if next is not an option with '-' */
                *i = *i + 1;
                arg = argv[*i];
            }
        } else {
            arg = NULL;
        }
    } else {
        return false;
    }

    SRA_DUMP_DBG(9, ("GetArg val: '%s'\n", arg));
    if( value == NULL && arg != NULL ) {
        rc = RC(rcApp, rcArgv, rcAccessing, rcParam, rcUnexpected);
    } else if( value != NULL ) {
        if( arg == NULL && *value == '\0' ) {
            rc = RC(rcApp, rcArgv, rcAccessing, rcParam, rcNotFound);
        } else if( arg != NULL && arg[0] != '\0' ) {
            *value = arg;
        }
    }
    if( rc != 0 ) {
        PLOGERR(klogErr, (klogErr, rc, "$(a), $(f): $(v)",
            PLOG_3(PLOG_S(a),PLOG_S(f),PLOG_S(v)), abbr, full, arg));
        Local_Usage(argv[0]);
    }
    return rc == 0 ? true : false;
}

/*******************************************************************************
 * KMain - defined for use with kapp library
 *******************************************************************************/
rc_t CC KMain ( int argc, char *argv [] )
{
    int i;
    rc_t rc;
#if USE_NCBI_SRADB
    char const  *rep = defaultRepServer;
    char const  *vol = defaultVolume;
#else
    char const *path = NULL;
#endif
    char const  *accession = NULL;
    char const  *outdir = NULL;
    dumpConstraints constraints;

    constraints.minSpotId = 1;
    constraints.maxSpotId = UINT32_MAX;
    constraints.minReadLen = MIN_READ_LEN;
    constraints.applyClip = true;
    constraints.dumpOrigFmt = false;
    constraints.noDotReads = false;
    constraints.noSignal = false;

    /* Parse command line arguments */
    for ( i = 1 ; i < argc ; ++ i ) {
        const char *arg = argv [ i ];
        if ( arg [ 0 ] != '-' ) {
            accession = arg;
        } else {
            if( tmp_GetArg("L", "log-level", &i, argc, argv, &arg) ) {
                if( (rc = LogLevelSet(arg)) != 0 ) {
                    PLOGERR(klogErr, (klogErr, rc, "log level $(lvl)", PLOG_S(lvl), arg));
                    return 3;
                }
#if _DEBUGGING
            } else if( tmp_GetArg("+", "debug", &i, argc, argv, &arg) ) {
                if( (rc = KDbgSetString(arg)) != 0 ) {
                    PLOGERR(klogErr, (klogErr, rc, "debug level $(lvl)", PLOG_S(lvl), arg));
                    return 3;
                }
#endif
            } else if(tmp_GetArg("H", "help", &i, argc, argv, NULL) ||
                      tmp_GetArg("?", "h", &i, argc, argv, NULL) ) {
                Local_Usage(argv[0]);
            } else if( tmp_GetArg(NULL, "accession", &i, argc, argv, &accession) ) {
#if USE_NCBI_SRADB
            } else if( tmp_GetArg(NULL, "repserver", &i, argc, argv, &rep) ) {
            } else if( tmp_GetArg(NULL, "volume", &i, argc, argv, &vol) ) {
#else
            } else if( tmp_GetArg(NULL, "path", &i, argc, argv, &path) ) {
#endif
            } else if( tmp_GetArg(NULL, "outdir", &i, argc, argv, &outdir) ) {
            } else if( tmp_GetArg(NULL, "origfmt", &i, argc, argv, NULL) ) {
                constraints.dumpOrigFmt = true;
            } else if( tmp_GetArg(NULL, "minSpotId", &i, argc, argv, &arg) ) {
                constraints.minSpotId = AsciiToU32 ( arg, NULL, NULL );
            } else if( tmp_GetArg(NULL, "maxSpotId", &i, argc, argv, &arg) ) {
                constraints.maxSpotId = AsciiToU32 ( arg, NULL, NULL );
            } else if( tmp_GetArg(NULL, "minReadLen", &i, argc, argv, &arg) ) {
                constraints.minReadLen = AsciiToU32 ( arg, NULL, NULL );
            } else if( tmp_GetArg(NULL, "noDotReads", &i, argc, argv, &accession) ) {
                constraints.noDotReads = true;
            } else if( tmp_GetArg(NULL, "noclip", &i, argc, argv, &accession) ) {
                constraints.applyClip = false;
            } else if( tmp_GetArg(NULL, "noSignal", &i, argc, argv, &accession) ) {
                constraints.noSignal = true;
            } else {
                fprintf(stderr,"abidump: Unknown switch: \'%s\'\n",arg);
                return 3;
            }
        }
    }


    /* If no accession, then output usage */

    if (!accession)
        {
        fprintf ( stderr, "abidump: Expected accession\n" );
        Local_Usage ( argv [ 0 ] );
            return 3;
        }
#if ! USE_NCBI_SRADB
    else if (!path)
        {
        fprintf ( stderr, "abidump: Expected path\n" );
        Local_Usage ( argv [ 0 ] );
            return 3;
        }
#endif
    else
        {

        rc = init_wd ();
        if ( rc != 0 )
        {
            PLOGERR ( klogErr, ( klogErr, rc , "Failed to init working directory", "d=%s", accession ) );
            return 3;
        }

            /*
             * Change to output directory, if specified, creating the directory if it doesn't exist
             */

        if ( outdir )
            {
            if ( !abi_chdir ( outdir ) )
                {
                PLOGMSG ( klogFatal , ( klogFatal , "abidump: Unable to change directory to $(d)\n",
                                                "d=%s" , outdir ) );
                }
            }

            /*
             * Dump SRR as ABI raw files
             */

            rc = abiDumpRun (
#if USE_NCBI_SRADB
                             rep, vol,
#else
                             path,
#endif
                             accession, &constraints );

            whack_wd();
            
            if (rc != rcNoErr)
                {
                    PLOGERR ( klogErr, ( klogErr, rc , "Failed for accession $(a)", "a=%s", accession ));
                    return rc;
                }
            else
            {
                PLOGMSG ( klogInfo, ( klogInfo, "Completed for accession $(a)", "a=%s", accession ));
                return 0;
            }
        }
}
