/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <stdlib.h>
#include <string.h>
#include <os-native.h>
#include <sysalloc.h>

#include <kapp/log.h>
#include <kapp/main.h>
#include <kfs/directory.h>
#include <kcont/container.h>

#include "files.h"
#include "debug.h"

#define SRADUMPER_MAKE_NAME(obj, key) "%s%s%s%s", obj->prefix ? obj->prefix : "", \
                                                  ((key && strlen(key) > 0) ? "_" : ""), \
                                                  key ? key : "", obj->suffix ? obj->suffix : ""

typedef struct DumperFile_struct {
    SLNode dad;
    KDirectory* dir;
    char* key;
    KFile* file;
    uint64_t pos;
} DumperFile;

struct SRADumperFiles {
    KDirectory* dir;
    char* prefix;
    char* suffix;
    bool key_as_dir;
    SLList list;
    const SRADumperFiles_hooks* hooks;
};

rc_t SRADumperFilesMake(const SRADumperFiles** self, const char* prefix, const char* suffix, bool key_as_dir,
                        const SRADumperFiles_hooks* hooks, const char* path, ...)
{
    rc_t rc = 0;
    SRADumperFiles* obj = NULL;

    if( self == NULL || prefix == NULL ) {
        return RC(rcExe, rcFile, rcConstructing, rcParam, rcNull);
    }
    *self = NULL;
    if( (obj = calloc(1, sizeof(*obj))) == NULL ) {
        return RC(rcExe, rcFile, rcConstructing, rcMemory, rcExhausted);
    }
    SLListInit(&obj->list);
    if( prefix ) {
        obj->prefix = strdup(prefix);
    }
    if( suffix ) {
        obj->suffix = strdup(suffix);
    }
    obj->key_as_dir = key_as_dir;
    if( (prefix && !obj->prefix) || (suffix && !obj->suffix) ) {
        rc = RC(rcExe, rcFile, rcConstructing, rcMemory, rcExhausted);
    }
    obj->hooks = hooks;
    if( rc == 0 && (rc = KDirectoryNativeDir(&obj->dir)) == 0 ) {
        if( path != NULL ) {
            va_list args;
            va_start(args, path);
            if( (rc = KDirectoryVCreateDir(obj->dir, 0775, kcmCreate | kcmParents, path, args)) == 0 ) {
                KDirectory* sub = NULL;
                va_end(args);
                va_start(args, path);
                if( (rc = KDirectoryVOpenDirUpdate(obj->dir, &sub, true, path, args)) == 0 ) {
                    KDirectoryRelease(obj->dir);
                    obj->dir = sub;
                }
            }
            va_end(args);
        }
    }
    if( rc == 0 ) {
        *self = obj;
    } else {
        SRADumperFilesRelease(obj, true);
    }
    return rc;
}

struct SRADumperFiles_WhackFilesData {
    SRADumperFiles* obj;
    bool deleteEmpty;
};

static
void SRADumperFiles_WhackFiles( SLNode *node, void *data )
{
    struct SRADumperFiles_WhackFilesData* d = (struct SRADumperFiles_WhackFilesData*)data;
    DumperFile* n = (DumperFile*)node;
    uint64_t sz = ~0;

    SRA_DUMP_DBG(5, ("Close file: '%s_%s%s'\n", d->obj->prefix, n->key, d->obj->suffix));
    if( d->obj->hooks && d->obj->hooks->pre_close ) {
        (*(d->obj->hooks->pre_close))(n->key, n->file);
    }
    if( d->deleteEmpty ) {
        KFileSize(n->file, &sz);
    }
    KFileRelease(n->file);
    if( d->deleteEmpty && sz <= 0 ) {
        if( n->dir ) {
            KDirectoryRemove(n->dir, false, SRADUMPER_MAKE_NAME(d->obj, ""));
        } else {
            KDirectoryRemove(d->obj->dir, false, SRADUMPER_MAKE_NAME(d->obj, n->key));
        }
    }
    KDirectoryRelease(n->dir);
    free(n->key);
    free(node);
}

rc_t SRADumperFilesRelease(const SRADumperFiles* cself, bool deleteEmpty)
{
    if( cself != NULL ) {
        SRADumperFiles* self = (SRADumperFiles*)cself;
        struct SRADumperFiles_WhackFilesData data;
        data.obj = self;
        data.deleteEmpty = deleteEmpty;

        SLListWhack(&self->list, SRADumperFiles_WhackFiles, &data);
        KDirectoryRelease(self->dir);
        free(self->prefix);
        free(self->suffix);
        free(self);
    }
    return 0;
}

struct SRADumperFiles_FindByKeyData {
    const char* key;
    DumperFile* file;
};

static
bool SRADumperFiles_FindByKey( SLNode *node, void *data )
{
    struct SRADumperFiles_FindByKeyData* d = (struct SRADumperFiles_FindByKeyData*)data;
    DumperFile* n = (DumperFile*)node;

    if( d->key == n->key || strcmp(n->key, d->key) == 0 ) {
        d->file = n;
        return true;
    }
    return false;
}

rc_t SRADumperFilesWrite(const SRADumperFiles* self, const char* key, const void* buffer, size_t size)
{
    rc_t rc = 0;
    struct SRADumperFiles_FindByKeyData data;

    if( self == NULL || buffer == NULL ) {
        return RC(rcExe, rcFile, rcOpening, rcParam, rcNull);
    }
    data.key = key;
    data.file = NULL;

    if( !SLListDoUntil(&self->list, SRADumperFiles_FindByKey, &data) ) {
        char* k = key ? strdup(key) : NULL;
        data.file = calloc(1, sizeof(*data.file));
        if( data.file == NULL || (key && !k) ) {
            rc = RC(rcExe, rcFile, rcOpening, rcMemory, rcExhausted);
        } else if( self->key_as_dir && k ) {
            /* open new file in key sub dir */
            SRA_DUMP_DBG(5, ("New file: '%s/%s%s'\n", k, self->prefix, self->suffix));
            if( (rc = KDirectoryCreateDir(self->dir, 0775, kcmCreate | kcmInit, k)) == 0 ) {
                if( (rc = KDirectoryOpenDirUpdate(self->dir, &data.file->dir, true, k)) == 0 ) {
                    rc = KDirectoryCreateFile(data.file->dir, &data.file->file, false, 0664,
                                              kcmCreate | kcmInit, SRADUMPER_MAKE_NAME(self, ""));
                }
            }
        } else {
            /* open new file */
            SRA_DUMP_DBG(5, ("New file: '%s_%s%s'\n", self->prefix, key, self->suffix));
            rc = KDirectoryCreateFile(self->dir, &data.file->file, false, 0664,
                                           kcmCreate | kcmInit, SRADUMPER_MAKE_NAME(self, key));
        }
        if( rc == 0 ) {
            data.file->key = k;
            SLListPushTail((SLList*)&self->list, &data.file->dad);
            if( self->hooks && self->hooks->post_open ) {
                rc = (*(self->hooks->post_open))(data.file->key, self);
            }
        } else {
            if( data.file ) {
                KFileRelease(data.file->file);
                KDirectoryRelease(data.file->dir);
            }
            free(data.file);
            free(k);
        }
    }
    if( rc == 0 && size > 0 ) {
        size_t writ = 0;
        rc = KFileWrite(data.file->file, data.file->pos, buffer, size, &writ);
        if( rc == 0 && writ != size ) {
            rc = RC(rcExe, rcFile, rcWriting, rcTransfer, rcIncomplete);
        }
        data.file->pos += writ;
    }
    return rc;
}
