/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <klib/rc.h>
#include <klib/log.h>
#include <klib/debug.h>
#include <klib/container.h>
#include <klib/trie.h>
#include <sra/types.h>
#include <os-native.h>

#include "experiment-xml.h"
#include "debug.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if _DEBUGGING
#define _TRACING 0
#else
#endif

#define MAX_POOL_MEMBER_BASECALLS 64
#define MAX_SPOT_DESCRIPTOR_READS 256

typedef struct PoolReadSpec_struct {
    int16_t start;
    int16_t length;
    /* chosen basecalls for EXPECTED_BASECALL_TABLE for defined member_name, NULL-terminated array */
    const ReadSpecXML_read_BASECALL* item[MAX_POOL_MEMBER_BASECALLS];
} PoolReadSpec;

typedef struct PoolMember_struct {
    BSTNode node;
    char* name;
    PoolReadSpec spec[MAX_SPOT_DESCRIPTOR_READS];
    uint32_t added;
    uint32_t removed;
} PoolMember;

struct ExperimentXML {
    const PlatformXML* platform;
    ProcessingXML processing;
    const ReadSpecXML* reads;
    BSTree* member_pool;
    PoolMember* member_null;
    SLList member_stats;
};

static
void PoolMember_Whack(BSTNode* node, void* data)
{
    if( node != NULL ) {
        const PoolMember* n = (const PoolMember*)node;
        free(n->name);
        free(node);
    }
}

static
int PoolMember_StrCmp(const char* s1, const char* s2)
{
    if( s1 == NULL && s2 == NULL ) {
        return 0;
    } else if( s1 == NULL ) {
        return -32767;
    } else if( s2 == NULL ) {
        return 32768;
    }
    return strcmp(s1, s2);
}

static
int PoolMember_Cmp(const BSTNode* item, const BSTNode* node)
{
    const PoolMember* i = (const PoolMember*)item;
    const PoolMember* n = (const PoolMember*)node;

    return PoolMember_StrCmp(i->name, n->name);
}

static
int PoolMember_FindByName(const void* item, const BSTNode* node)
{
    const char* name = (const char*)item;
    const PoolMember* n = (const PoolMember*)node;

    return PoolMember_StrCmp(name, n->name);
}

typedef struct PoolMember_FindByTag_Data_struct {
    uint8_t readid;
    const char* tag;
    const char* member_name;
} PoolMember_FindByTag_Data;

static
bool PoolMember_FindByTag(BSTNode *node, void *data)
{
    PoolMember_FindByTag_Data* d = (PoolMember_FindByTag_Data*)data;
    const PoolMember* n = (const PoolMember*)node;
    int i = 0;

    const ReadSpecXML_read_BASECALL* bc = n->spec[d->readid].item[i];
    while( bc != NULL ) {
        if( strcmp(bc->read_group_tag, d->tag) == 0 ) {
            d->member_name = n->name;
            return true;
        }
        bc = n->spec[d->readid].item[++i];
    }
    return false;
}

typedef struct PoolMember_FindReverse_Data_struct {
    uint8_t nreads;
    char* tags[MAX_SPOT_DESCRIPTOR_READS];
    const PoolMember* member;
} PoolMember_FindReverse_Data;

static
bool PoolMember_FindReverse(BSTNode *n, void *d)
{
    const PoolMember* node = (const PoolMember*)n;
    PoolMember_FindReverse_Data* data = (PoolMember_FindReverse_Data*)d;
    int r, j, found;

    for(r = 0; r < data->nreads; r++) {
        /* comparing pointers here because tags contains same pointers */
        if( node->spec[r].item[0] == NULL && data->tags[r] == NULL ) {
            continue;
        }
        found = 0;
        for(j = 0; node->spec[r].item[j] != NULL; j++ ) {
            if( node->spec[r].item[j]->read_group_tag == data->tags[r] ) {
                found = 1;
                break;
            }
        }
        if( found == 0 ) {
            return false;
        }
    }
    
#ifdef __GNUC__
#warning "TODO Check duplicate matches here"
#endif

    data->member = node;
    return true;
}

#if _DEBUGGING
static
void PoolMember_Dump( BSTNode *node, void *data )
{
    const PoolMember* n = (const PoolMember*)node;
    const ExperimentXML* self = (const ExperimentXML*)data;
    int r;

    DEBUG_MSG (9, ("Member '%s'\n", n->name));

    for(r = 0; r < self->reads->nreads; r++) {
        char b[4096];
        b[0] = '\0';
        if( self->reads->reads[r].coord_type == rdsp_ExpectedBaseCall_ct ) {
            strcpy(b, " ebc:'");
            strcat(b, n->spec[r].item[0]->basecall);
            strcat(b, "'");
        } else if( self->reads->reads[r].coord_type == rdsp_ExpectedBaseCallTable_ct ) {
            int i = -1;
            while( n->spec[r].item[++i] != NULL ) {
                strcat(b, " bc:'");
                strcat(b, n->spec[r].item[i]->basecall);
                strcat(b, "', tag:'");
                strcat(b, n->spec[r].item[i]->read_group_tag);
                strcat(b, "';");
            }
        }
        DEBUG_MSG (9, ("Read %hhu [%hd:%hd]%s\n",
                       r + 1, n->spec[r].start, n->spec[r].length, b));
    }
}
#endif

typedef struct PoolMember_Stat_struct {
    SLNode node;
    char* member;
    uint32_t added;
    uint32_t deleted;
} PoolMember_Stat;

typedef struct PoolMember_StatFindData_struct {
    const char* search;
    PoolMember_Stat* node;
} PoolMember_StatFindData;

static
bool PoolMemberStats_Find(SLNode *n, void *d)
{
    PoolMember_Stat* node = (PoolMember_Stat*)n;
    PoolMember_StatFindData* data = (PoolMember_StatFindData*)d;

    if( ((node->member == NULL || node->member[0] == '\0') && (data->search == NULL || data->search[0] == '\0')) ||
        (node->member != NULL && data->search != NULL && strcmp(node->member, data->search) == 0) ) {
        data->node = node;
        return true;
    }
    return false;
}

static
void PoolMemberStats_Whack( SLNode *n, void *data )
{
    PoolMember_Stat* node = (PoolMember_Stat*)n;
    bool print = *(bool*)data;

    if( print ) {
        const char* mm = node->member ? node->member : "default";
        PLOGMSG(klogInfo,(klogInfo, "Member stat $(m): +$(a), -$(d)",
                 PLOG_3(PLOG_S(m),PLOG_U32(a),PLOG_U32(d)), mm, node->added, node->deleted));
    }
    free(node->member);
    free(node);
}

static
rc_t PoolMemberStats(const ExperimentXML* self, const char* member_name, PoolMember_Stat** node)
{
    rc_t rc = 0;
    PoolMember_StatFindData data;

    data.node = NULL;
    data.search = member_name;

    if( !SLListDoUntil(&self->member_stats, PoolMemberStats_Find, &data) ) {
        char* nm = (member_name && member_name[0] != '\0') ? strdup(member_name) : NULL;
        data.node = calloc(1, sizeof(*data.node));
        if( data.node == NULL || (nm == NULL && member_name && member_name[0] != '\0') ) {
            free(data.node);
            free(nm);
            rc = RC(rcSRA, rcFormatter, rcResolving, rcMemory, rcExhausted);
        } else {
            data.node->member = nm;
            SLListPushTail((SLList*)&self->member_stats, &data.node->node);
        }
    }
    *node = data.node;
    return rc;
}

static
rc_t PoolMemberStats_Set(const ExperimentXML* self, const char* member_name)
{
    PoolMember_Stat* node = NULL;

    return PoolMemberStats(self, member_name, &node);
}

static
rc_t PoolMemberStats_Add(const ExperimentXML* self, const char* member_name)
{
    rc_t rc = 0;
    PoolMember_Stat* node = NULL;

    if( (rc = PoolMemberStats(self, member_name, &node)) == 0 ) {
        node->added++;
    }
    return rc;
}

static
rc_t PoolMemberStats_Delete(const ExperimentXML* self, const char* member_name)
{
    rc_t rc = 0;
    PoolMember_Stat* node = NULL;

    if( (rc = PoolMemberStats(self, member_name, &node)) == 0 ) {
        node->deleted++;
    }
    return rc;
}

static
rc_t PoolMember_Add(ExperimentXML* self, const char* name)
{
    rc_t rc = 0;
    PoolMember* member = NULL;

    if( self == NULL) {
        return RC(rcSRA, rcFormatter, rcAllocating, rcSelf, rcNull);
    }
    member = calloc(1, sizeof(*member));
    if( member == NULL ) {
        return RC(rcSRA, rcFormatter, rcAllocating, rcMemory, rcExhausted);
    }
    if( name == NULL ) {
        if( self->member_null == NULL ) {
            self->member_null = member;
        } else {
            rc = RC(rcSRA, rcFormatter, rcAllocating, rcId, rcDuplicate);
        }
    } else {
        member->name = strdup(name);
        if( member->name == NULL ) {
            rc = RC(rcSRA, rcFormatter, rcAllocating, rcMemory, rcExhausted);
        } else {
            if( self->member_pool == NULL ) {
                self->member_pool = calloc(1, sizeof(BSTree));
                if( self->member_pool == NULL ) {
                    rc = RC(rcSRA, rcFormatter, rcAllocating, rcMemory, rcExhausted);
                } else {
                    BSTreeInit(self->member_pool);
                }
            }
            if( rc == 0 ) {
                rc = BSTreeInsertUnique(self->member_pool, &member->node, NULL, PoolMember_Cmp);
            }
        }
    }
    if( rc == 0 ) {
        uint32_t i;
        for(i = 0; i < self->reads->nreads; i++) {
            switch(self->reads->reads[i].coord_type) {
                case rdsp_RelativeOrder_ct:
                    member->spec[i].start = -1;
                    member->spec[i].length = -1;
                    break;

                case rdsp_BaseCoord_ct:
                case rdsp_CycleCoord_ct:
                    member->spec[i].start = self->reads->reads[i].coord.start_coord - 1;
                    member->spec[i].length = -1;
                    break;

                case rdsp_ExpectedBaseCall_ct:
                    member->spec[i].start = -1;
                    member->spec[i].item[0] = self->reads->reads[i].coord.expected_basecalls.table;
                    member->spec[i].length = strlen(member->spec[i].item[0]->basecall);
                    break;

                case rdsp_ExpectedBaseCallTable_ct:
                    member->spec[i].start = -1;
                    member->spec[i].length = (!self->reads->reads[i].coord.expected_basecalls.pooled || name == NULL) ? -1 : 0;
                    break;
            }
        }
        rc = PoolMemberStats_Set(self, member->name);
    } else {
        PoolMember_Whack(&member->node, NULL);
    }
    return rc;
}

static
rc_t PoolMember_AddBasecalls(const ExperimentXML* self, const char* member, int readid, const ReadSpecXML_read_BASECALL* basecalls[])
{
    rc_t rc = 0;

    if( self == NULL || basecalls == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcParam, rcNull);
    } else if( self->reads->reads[readid].coord_type != rdsp_ExpectedBaseCallTable_ct ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcId, rcIncorrect);
    } else {
        PoolMember* pm = (PoolMember*)BSTreeFind(self->member_pool, member, PoolMember_FindByName);
        if( pm == NULL ) {
            rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcNotFound);
            PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)'", PLOG_S(m), member));
        } else {
            int k = -1, i = -1;
            while( pm->spec[readid].item[++k] != NULL );
            while( rc == 0 && basecalls[++i] != NULL ) {
                if( k >= MAX_POOL_MEMBER_BASECALLS ) {
                    rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcTooLong);
                    PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' basecalls", PLOG_S(m), member));
                }
                pm->spec[readid].item[k++] = basecalls[i];
            }
        }
    }
    return rc;
}

static
rc_t parse_PROCESSING(const KXMLNode* EXPERIMENT, ProcessingXML *processing)
{
    rc_t rc = 0;
    const KXMLNodeset* ns = NULL;

    if( (rc = KXMLNodeOpenNodesetRead(EXPERIMENT, &ns, "PROCESSING/*")) == 0 ) {
        uint32_t i, count = 0;
        if( (rc = KXMLNodesetCount(ns, &count)) == 0 && count < 2 ) {
            rc = RC(rcSRA, rcFormatter, rcConstructing, rcNode, rcIncomplete);
            LOGERR(klogErr, rc, "PROCESSING subnodes");
        } else {
            const char* name = NULL;
            for(i = 0; rc == 0 && i < count; i++) {
                const KXMLNode* n = NULL;
                if( (rc = KXMLNodesetGetNodeRead(ns, &n, i)) == 0 ) {
                    if( (rc = KXMLNodeElementName(n, &name)) == 0 ) {
                        if( strcmp(name, "BASE_CALLS") == 0 ) {
                            if( processing->sequence_space == eUndefined_Space ) {
                                char* sq_sp = 0;
                                if( (rc = XMLNode_get_strnode(n, "BASE_CALLER", false, &processing->base_caller)) == 0 &&
                                    (rc = XMLNode_get_strnode(n, "SEQUENCE_SPACE", false, &sq_sp)) == 0 ) {
                                    if( strcasecmp(sq_sp, "Base Space") == 0 ) {
                                        processing->sequence_space = eBase_Space;
                                    } else if( strcasecmp(sq_sp, "Color Space") == 0 ) {
                                        processing->sequence_space = eColor_Space;
                                    } else {
                                        name = "SEQUENCE_SPACE";
                                        rc = RC(rcSRA, rcFormatter, rcConstructing, rcNode, rcInvalid);
                                    }
                                    free(sq_sp);
                                }
                            } else {
                                rc = RC(rcSRA, rcFormatter, rcConstructing, rcNode, rcDuplicate);
                            }
                        } else if( strcmp(name, "QUALITY_SCORES") == 0 ) {
                            if( processing->quality_type == eExperimentQualityType_Undefined ) {
                                char* qtype;
                                if( (rc = KXMLNodeReadAttrCStr(n, "qtype", &qtype, NULL)) == 0 ) {
                                    if( strcasecmp(qtype, "phred") == 0 ) {
                                        processing->quality_type = eExperimentQualityType_Phred;
                                    } else if( strcasecmp(qtype, "other") == 0 ) {
                                        processing->quality_type = eExperimentQualityType_Other;
                                    } else {
                                        rc = RC(rcSRA, rcFormatter, rcParsing, rcAttr, rcUnrecognized);
                                    }
                                    free(qtype);
                                } else if( GetRCState(rc) == rcNotFound ) {
                                    processing->quality_type = eExperimentQualityType_Phred;
                                    rc = 0;
                                }
                                if( rc == 0 ) {
                                    char* qbc;
                                    rc = XMLNode_get_strnode(n, "QUALITY_SCORER", false, &processing->quality_scorer);
                                    if( rc == 0 && (rc = XMLNode_get_strnode(n, "QUALITY_BOOK_CHAR", true, &qbc)) == 0 && qbc != NULL ) {
                                        processing->quality_offset = qbc[0];
                                        free(qbc);
                                    }
                                    
                                }
                            } else {
                                LOGMSG(klogWarn, "Only 1st QUALITY_SCORES used");
                            }
                        } else {
                            rc = RC(rcSRA, rcFormatter, rcConstructing, rcNode, rcUnknown);
                        }
                    }
                    KXMLNodeRelease(n);
                }
            }
            if( rc != 0 ) {
                PLOGERR(klogErr,(klogErr, rc, "PROCESSING subnode '$(n)'", PLOG_S(n), name));
            }
        }
        KXMLNodesetRelease(ns);
    }
    return rc;
}

rc_t parse_POOL(const KXMLNode* EXPERIMENT, ExperimentXML* obj)
{
    rc_t rc = 0;

    /* create unspecified member -> NULL */
    if( (rc = PoolMember_Add(obj, NULL)) == 0 ) {
        const KXMLNodeset* MEMBERS;
        if( (rc = KXMLNodeOpenNodesetRead(EXPERIMENT, &MEMBERS, "DESIGN/SAMPLE_DESCRIPTOR/POOL/MEMBER")) == 0 ) {
            uint32_t i, num_members = 0;
            if( (rc = KXMLNodesetCount(MEMBERS, &num_members)) == 0 && num_members > 0 ) {
                /* create default member -> '' */
                rc = PoolMember_Add(obj, "");
                for(i = 0; rc == 0 && i < num_members; i++) {
                    const KXMLNode* MEMBER;
                    if( (rc = KXMLNodesetGetNodeRead(MEMBERS, &MEMBER, i)) == 0 ) {
                        char* member_name = NULL;
                        if( (rc = KXMLNodeReadAttrCStr(MEMBER, "member_name", &member_name, NULL)) == 0 ) {
                            const KXMLNodeset* LABELS;
                            if( strcasecmp(member_name, "default") == 0 ) {
                                member_name[0] = '\0';
                            }
                            if( member_name[0] != '\0' ) {
                                if( (rc = PoolMember_Add(obj, member_name)) != 0 ) {
                                    PLOGERR(klogErr, (klogErr, rc, "MEMBER '$(m)'", PLOG_S(m), member_name));
                                }
                            }
                            if( rc == 0 && (rc = KXMLNodeOpenNodesetRead(MEMBER, &LABELS, "READ_LABEL")) == 0 ) {
                                uint32_t l, num_labels = 0;
                                KXMLNodesetCount(LABELS, &num_labels);
                                if( num_labels > 0 ) {
                                    const ReadSpecXML_read_BASECALL* basecalls[MAX_POOL_MEMBER_BASECALLS];
                                    for(l = 0; rc == 0 && l < num_labels; l++) {
                                        const KXMLNode* LABEL;
                                        if( (rc = KXMLNodesetGetNodeRead(LABELS, &LABEL, l)) == 0 ) {
                                            char* label = NULL, *tag = NULL;
                                            if( (rc = KXMLNodeReadCStr(LABEL, &label, NULL)) != 0 || label == NULL || label[0] == '\0' ) {
                                                rc = rc ? rc : RC(rcSRA, rcFormatter, rcConstructing, rcData, rcEmpty);
                                                PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL value", PLOG_S(m), member_name));
                                            } else {
                                                if( (rc = KXMLNodeReadAttrCStr(LABEL, "read_group_tag", &tag, NULL)) != 0 ) {
                                                    if( GetRCState(rc) != rcNotFound ) {
                                                        PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' 'read_group_tag' attribute", PLOG_2(PLOG_S(m),PLOG_S(l)), member_name, label));
                                                    } else {
                                                        rc = 0;
                                                    }
                                                } else if( tag != NULL && tag[0] == '\0' )  {
                                                    rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcEmpty);
                                                    PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' 'read_group_tag' attribute", PLOG_2(PLOG_S(m),PLOG_S(l)), member_name, label));
                                                }
                                                if( rc == 0 ) {
                                                    /* find read_spec based on label and update basecall table */
                                                    uint32_t r, read_found = 0;
                                                    for(r = 0; rc == 0 && r < obj->reads->nreads; r++) {
                                                        if( obj->reads->reads[r].read_label != NULL &&
                                                            strcmp(label, obj->reads->reads[r].read_label) == 0 ) {
                                                            read_found++;
                                                            if( tag != NULL && tag[0] != '\0' ) {
                                                                if( obj->reads->reads[r].coord_type != rdsp_ExpectedBaseCallTable_ct     ) {
                                                                    rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInconsistent);
                                                                    PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' READ_SPEC must have EXPECTED_BASECALL_TABLE type",
                                                                            PLOG_2(PLOG_S(m),PLOG_S(l)), member_name, label));
                                                                } else {
                                                                    uint32_t b, bcall_found = 0;
                                                                    memset(basecalls, 0, sizeof(basecalls));
                                                                    /* find tag in table */
                                                                    for(b = 0; rc == 0 && b < obj->reads->reads[r].coord.expected_basecalls.count; b++) {
                                                                        if( obj->reads->reads[r].coord.expected_basecalls.table[b].read_group_tag != NULL &&
                                                                            strcmp(tag, obj->reads->reads[r].coord.expected_basecalls.table[b].read_group_tag) == 0 ) {
                                                                            if( bcall_found >= MAX_POOL_MEMBER_BASECALLS ) {
                                                                                rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcExcessive);
                                                                                PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' read_group_tag='$(t)' too many basecalls",
                                                                                         PLOG_3(PLOG_S(m),PLOG_S(l),PLOG_S(t)), member_name, label, tag));
                                                                            }
                                                                            basecalls[bcall_found++] = &obj->reads->reads[r].coord.expected_basecalls.table[b];
                                                                        }
                                                                    }
                                                                    if( rc == 0 && bcall_found == 0 ) {
                                                                        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcNotFound);
                                                                        PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' read_group_tag='$(t)' in READ_SPEC",
                                                                                 PLOG_3(PLOG_S(m),PLOG_S(l),PLOG_S(t)), member_name, label, tag));
                                                                    } else {
                                                                        rc = PoolMember_AddBasecalls(obj, member_name, r, basecalls);
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                    if( read_found == 0 ) {
                                                        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcNotFound);
                                                        PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' in READ_SPEC", PLOG_2(PLOG_S(l),PLOG_S(m)), label, member_name));
                                                    } else if( read_found > 1 ) {
                                                        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInconsistent);
                                                        PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL '$(l)' in READ_SPEC", PLOG_2(PLOG_S(l),PLOG_S(m)), label, member_name));
                                                    }
                                                }
                                            }
                                            free(tag);
                                            free(label);
                                            KXMLNodeRelease(LABEL);
                                        } else {
                                            PLOGERR(klogErr,(klogErr, rc, "MEMBER '$(m)' READ_LABEL #$(i)", PLOG_2(PLOG_S(m),PLOG_U32(i)), member_name, l));
                                        }
                                    }
                                } else {
                                    rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInvalid);
                                    PLOGERR(klogErr,(klogErr, rc, "missing READ_LABEL element(s) in MEMBER '$(m)'", PLOG_S(m), member_name));
                                }
                                KXMLNodesetRelease(LABELS);
                            }
                            free(member_name);
                        } else {
                            PLOGERR(klogErr,(klogErr, rc, "'member_name' in MEMBER node #$(i)", PLOG_U32(i), i));
                        }
                        KXMLNodeRelease(MEMBER);
                    } else {
                        PLOGERR(klogErr,(klogErr, rc, "MEMBER node #$(i)", PLOG_U32(i), i));
                    }
                }
            } else if( rc != 0 ) {
                LOGERR(klogErr, rc, "POOL/MEMBER(s)");
            }
            KXMLNodesetRelease(MEMBERS);
        } else {
            LOGERR(klogErr, rc, "POOL/MEMBER(s)");
        }
    }
#if _DEBUGGING
    PoolMember_Dump(&obj->member_null->node, obj);
    BSTreeForEach(obj->member_pool, false, PoolMember_Dump, obj);
#endif
    return rc;
}

typedef struct parse_BASECALL_mutate_data_struct {
    Trie* trie;
    uint16_t* max_mutation;
    uint16_t* min_mutation;
    ReadSpecXML_read_BASECALL* bc;
    uint32_t mismatches;
    size_t bc_sz;
    char tmp[1024];
    uint32_t qty;
    uint32_t dupes;
} parse_BASECALL_mutate_data;

typedef struct MuteNode_struct {
    TNode node;
    ReadSpecXML_read_BASECALL* basecall;
    uint32_t mismatches;
} MuteNode;

typedef struct Experiment_MatchData_struct {
    const char* tag;
    MuteNode* found;
} Experiment_MatchData;

static
bool Experiment_Match_Explore(TNode *node, void *data)
{
    MuteNode* n = (MuteNode*)node;
    Experiment_MatchData* d = (Experiment_MatchData*)data;
#if _TRACING

    DEBUG_MSG (10, ("explored %s for %s:%s:%u\n",
                    node->key.addr, n->basecall->basecall,
                    n->basecall->read_group_tag, n->mismatches));
#endif
    if( d->found == NULL || d->found->mismatches < n->mismatches ) {
        if( d->tag == NULL || (n->basecall->read_group_tag != NULL && strcmp(d->tag, n->basecall->read_group_tag) == 0) ) {
            d->found = n;
            if( n->mismatches == 0 ) {
                DEBUG_MSG (10, ("match best %s for %s:%s:%u\n",
                                node->key.addr, n->basecall->basecall, 
                                n->basecall->read_group_tag, n->mismatches));
                return true;
            }
            DEBUG_MSG (10, ("match %s for %s:%s:%u\n",
                            node->key.addr, n->basecall->basecall, n->basecall->read_group_tag, n->mismatches));
        }
    }
    return false;
}

static
void parse_BASECALL_mutations_whack( TNode *n, void *data )
{
    free((char*)n->key.addr);
    free(n);
}

static
rc_t parse_BASECALL_mutate_add(parse_BASECALL_mutate_data* data, char* t)
{
    rc_t rc = 0;

    MuteNode* n = malloc(sizeof(*n));
    char* key = strdup(data->tmp);
    if( n == NULL || key == NULL ) {
        free(n);
        free(key);
        rc = RC(rcSRA, rcFormatter, rcTokenizing, rcMemory, rcExhausted);
    } else {
        TNode* existing = NULL;
        StringInitCString(&n->node.key, key);
        n->basecall = data->bc;
        n->mismatches = data->mismatches;
        if( (rc = TrieInsertUnique(data->trie, &n->node, &existing)) != 0 ) {
            if( GetRCState(rc) == rcExists ) {
                MuteNode* dupe = (MuteNode*)existing;
                if( data->bc == dupe->basecall ) {
#if _TRACING
                    DEBUG_MSG(5, ("%s dupe ignored %s\n", t, key));
#endif
                    data->dupes++;
                    parse_BASECALL_mutations_whack(&n->node, NULL);
                    n = NULL;
                    rc = 0;
                } else {
#if _TRACING
                    if( n->mismatches == dupe->mismatches ) {
                        DEBUG_MSG(5, ("%s %s:%u same level for %s:%u\n",
                                      t, key, n->mismatches, dupe->basecall->basecall, dupe->mismatches));
                    }
#endif
                    if( (data->bc->read_group_tag == NULL && dupe->basecall->read_group_tag != NULL) ||
                        (data->bc->read_group_tag != NULL && dupe->basecall->read_group_tag == NULL) ||
                        (data->bc->read_group_tag != NULL && dupe->basecall->read_group_tag != NULL &&
                         strcmp(data->bc->read_group_tag, dupe->basecall->read_group_tag) != 0) ) {
                        if( (rc = TrieInsert(data->trie, &n->node)) != 0 ) {
                            parse_BASECALL_mutations_whack(&n->node, NULL);
                            n = NULL;
                        }
                    } else {
                        bool better = n->mismatches < dupe->mismatches;
#if _TRACING
                        DEBUG_MSG (5, ("%s %s:%u %s sibling for %s:%u\n",
                                       t, key, n->mismatches, 
                                       better ? "better " : "",
                                       dupe->basecall->basecall, dupe->mismatches));
#endif
                        if( better ) {
                            /* one of de-IUPACed calls gives mismatch when the other gives match on same letter:
                               RAT split into 2 basecall table entries: AAT, GAT.
                               After replace AAT will have GAT with 1 miss, and GAT will have self with 0 misses.
                               We keep GAT but with updated misses to 0 */
                            dupe->mismatches = n->mismatches;
                        }
                        parse_BASECALL_mutations_whack(&n->node, NULL);
                        n = NULL;
                        rc = 0;
                    }
                }
            }
        }
        if( n != NULL ) {
#if _TRACING
            DEBUG_MSG(5, ("%s added mutation %s:%u\n", t, key, data->mismatches));
#endif
            if( *data->max_mutation < n->node.key.len ) {
                *data->max_mutation = n->node.key.len;
            }
            if( *data->min_mutation > n->node.key.len ) {
                *data->min_mutation = n->node.key.len;
            }
            data->qty++;
        }
    }
    return rc;
}

/* Rule:
    Change sequentially each base to N (ACGT), always having changed bases == max_mismatch per case
*/
static
rc_t parse_BASECALL_mutate_replace(parse_BASECALL_mutate_data* data, size_t pos, uint32_t mis_left)
{
    rc_t rc = 0;

#if _TRACING
    DEBUG_MSG(3,("replace on basecall: %s, miss_left %u, pos %hu '%.*s'\n",
                 data->bc->basecall, mis_left, pos, pos, data->bc->basecall));
#endif
    if( pos >= data->bc_sz ) {
        rc = parse_BASECALL_mutate_add(data, NULL);
    } else if( mis_left <= 0 ) {
        strcat(data->tmp, &data->bc->basecall[pos]);
        rc = parse_BASECALL_mutate_add(data, NULL);
    } else {
        static const char* const REPL = "ACGT";
        size_t j, x;
        for(j = 0; rc == 0 && j < 4; j++) {
            /* detect is this is a real mutation or originally allowed value in basecall */
            x = (data->bc->basecall[pos] != REPL[j]) ? 1 : 0;
            data->tmp[pos] = REPL[j];
            data->tmp[pos + 1] = '\0';
            data->mismatches += x;
            rc = parse_BASECALL_mutate_replace(data, pos + 1, mis_left - x);
            data->mismatches -= x;
        }
    }
    return rc;
}

/* Rule:
    group of one or more identical bases -> -1, self, +1
    A -> '',A,AA; AA -> A,AA,AAA; AAA -> AA,AAA,AAAA; AAAA -> AAA,AAAA,AAAAA
*/
static
rc_t parse_BASECALL_mutate_insdel(parse_BASECALL_mutate_data* data, const size_t pos, const uint32_t mis_left, int l)
{
    rc_t rc = 0;
    char t[128000];

#if _TRACING
    l += 4;
    memset(t, '-', l);
#endif
    t[l] = '\0';
#if _TRACING
    DEBUG_MSG (3, ("%s insdel on basecall: %s, max_mismatch %u, pos %hu\n",
                   t, &(data->bc->basecall[pos]), mis_left, pos));
#endif
    if( pos <= data->bc_sz ) {
        if( mis_left == 0 ) {
            strcat(data->tmp, &data->bc->basecall[pos]);
            rc = parse_BASECALL_mutate_add(data, t);
        } else if( pos != data->bc_sz ) {
            uint32_t m;
            size_t i = pos, res;

            if( pos == 0 ) {
                /* add self w/0 misses */
                strcpy(data->tmp, data->bc->basecall);
                rc = parse_BASECALL_mutate_add(data, t);
                data->tmp[0] = '\0';
            }
            /* find group of 1 or more */
            while( data->bc->basecall[i] == data->bc->basecall[i + 1] ) {
                i++;
            }
            /* append - all and repeated - 1 chars */
            strncat(data->tmp, &data->bc->basecall[pos], i - pos);
            res = strlen(data->tmp);

            if( rc == 0 ) {
                /* -1 */
                data->mismatches++;
                for(m = mis_left; rc == 0 && m > 0; m--) {
#if _TRACING
                    DEBUG_MSG (3, ("%s -1 on basecall: %s, tmp: '%s', mis %hu\n",
                                   t, &(data->bc->basecall[pos]), data->tmp, m));
#endif
                    rc = parse_BASECALL_mutate_insdel(data, i + 1, m - 1, l);
                    data->tmp[res] = '\0';
                }
                data->mismatches--;
            }
            if( rc == 0 ) {
                /* not changed */
                strncat(data->tmp, &data->bc->basecall[i], 1);
                res++;
                for(m = mis_left; rc == 0 && m > 0; m--) {
#if _TRACING
                    DEBUG_MSG (3, ("%s 0 on basecall: %s, tmp: '%s', mis %hu\n",
                                   t, &(data->bc->basecall[pos]), data->tmp, m));
#endif
                    rc = parse_BASECALL_mutate_insdel(data, i + 1, m, l);
                    data->tmp[res] = '\0';
                }
            }
            if( rc == 0 ) {
                /* +1 */
                strncat(data->tmp, &data->bc->basecall[i], 1);
                res++;
                data->mismatches++;
                for(m = mis_left; rc == 0 && m > 0; m--) {
#if _TRACING
                    DEBUG_MSG (3, ("%s +1 on basecall: %s, tmp: '%d', mis %hu\n",
                                   t, &(data->bc->basecall[pos]), data->tmp, m));
#endif
                    rc = parse_BASECALL_mutate_insdel(data, i + 1, m - 1, l);
                    data->tmp[res] = '\0';
                }
                data->mismatches--;
            }
        }
    } else {
        rc = RC(rcSRA, rcFormatter, rcTokenizing, rcParam, rcOutofrange);
    }
    return rc;
}

#if _DEBUGGING
static
void parse_BASECALL_mutations_dump( TNode *n, void *data )
{
    uint32_t* q = (uint32_t*)data;
#if _TRACING
    MuteNode* node = (MuteNode*)n;
    DEBUG_MSG (10, ("dump %s for %s:%s:%u\n",
                    n->key.addr, node->basecall->basecall,
                    node->basecall->read_group_tag, node->mismatches));
#endif
    *q = *q + 1;
}
#endif

static
rc_t parse_BASECALL_mutations(const ReadSpecXML_read_BASECALL_TABLE* ctable, SRAPlatforms platform)
{
    rc_t rc = 0;
    int i;
    parse_BASECALL_mutate_data data;
    ReadSpecXML_read_BASECALL_TABLE* table = (ReadSpecXML_read_BASECALL_TABLE*)ctable;

    if( table->mutations != NULL ) {
        return RC(rcSRA, rcCondition, rcConstructing, rcData, rcDuplicate);
    } else if( (table->mutations = calloc(1, sizeof(*table->mutations))) == NULL ) {
        return RC(rcSRA, rcCondition, rcConstructing, rcMemory, rcExhausted);
    }
    rc = TrieInit(table->mutations, "ACGT", 2, false);
    table->min_mutation = ~0;

    data.trie = table->mutations;
    data.max_mutation = &table->max_mutation;
    data.min_mutation = &table->min_mutation;

    /* find maximum value for mismatch per experiment */
    for(i = 0; rc == 0 && i < table->count; i++) {
#ifdef __GNUC__    
#warning "Max mismatch truncated to 3 to avoid huge tables for now"
#endif
        int tmp_miss = table->table[i].max_mismatch > 3 ? 3 : table->table[i].max_mismatch;

        data.bc = &table->table[i];
        data.bc_sz = strlen(data.bc->basecall);
        data.qty = 0;
        data.mismatches = 0;
        data.dupes = 0;

        DEBUG_MSG (3, ("Mutilating basecall: %s, max_mismatch %u\n", data.bc->basecall, tmp_miss));
        data.tmp[0] = '\0';
        if( (rc = parse_BASECALL_mutate_replace(&data, 0, tmp_miss)) == 0 ) {
            if( platform == SRA_PLATFORM_454 ) {
                data.tmp[0] = '\0';
                rc = parse_BASECALL_mutate_insdel(&data, 0, tmp_miss, 0);
            }
        }
        if(rc) {
            PLOGERR(klogErr,(klogErr, rc, "$(b) mutations $(v) dupes $(d)",
                    PLOG_3(PLOG_S(b),PLOG_U32(v),PLOG_U32(d)), data.bc->basecall, data.qty, data.dupes));
        } else {
            DEBUG_MSG(3, ("%s mutations %u duptes %u\n", data.bc->basecall, data.qty, data.dupes));
        }
    }
#if _DEBUGGING
    {{
        uint32_t nqty = 0;
        TrieForEach(table->mutations, parse_BASECALL_mutations_dump, &nqty);
        DEBUG_MSG (3, ("tree keys %u, longest %hd, shortest %hd\n",
                       nqty, table->max_mutation, table->min_mutation));
    }}
#endif
    return rc;
}

rc_t Experiment_Match(const ReadSpecXML_read_BASECALL_TABLE* table, SRAPlatforms platform,
                      const char* bases, size_t base_len, uint32_t* left_miss_allowed, char** tag, int16_t* pos, int16_t* len, uint32_t* mismatches)
{
    rc_t rc = 0;

    if( table->mutations == NULL ) {
        rc = parse_BASECALL_mutations(table, platform);
    }
    if( rc == 0 ) {
        Experiment_MatchData data;
        char f2na[4096];
        String key;
        size_t from = 0, i = 0;

        if( base_len > sizeof(f2na) ) {
            return RC(rcSRA, rcFormatter, rcResolving, rcMemory, rcInsufficient);
        }
        data.found = NULL;
        data.tag = tag ? *tag : NULL;

        do {
            /* 1st convert up to max mutation length and each loop add one more letter */
            while( (i < from + table->max_mutation) && i < base_len ) {
                f2na[i] = toupper(bases[i]);
                switch(f2na[i]) {
                    case 'A':
                    case 'C':
                    case 'G':
                    case 'T':
                        break;
                    default:
                        f2na[i] = 'A';
                        break;
                }
                i++;
                f2na[i] = '\0';
            }

            StringInit(&key, &f2na[from], i - from, i - from);
#if _TRACING
            DEBUG_MSG (6, ("search for '%s' from %u, left %u\n",
                           key.addr, from, *left_miss_allowed));
#endif
            if( TrieExplore(table->mutations, &key, Experiment_Match_Explore, &data) || data.found != NULL ) {
                *pos = from;
                *len = data.found->node.key.len;
                if( tag != NULL && *tag == NULL ) {
                    /* whole table lookup, assign tag */
                    *tag = data.found->basecall->read_group_tag;
                }
                if( mismatches != NULL ) {
                    *mismatches = data.found->mismatches;
                }
                *left_miss_allowed = data.found->basecall->max_mismatch - data.found->mismatches;
                DEBUG_MSG (6, ("found %s:%u from %hi %hi bytes, left %u\n",
                               data.found->basecall->basecall, data.found->mismatches, *pos, *len, *left_miss_allowed));
                break;
            }
            from++;
        } while(from < *left_miss_allowed && (base_len - from >= table->min_mutation) );

        if( data.found == NULL ) {
            rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcNotFound);
            DEBUG_MSG(6, ("nothing found in %.*s\n", base_len, bases));
            *pos = 0;
            *len = 0;
        }
    }
    return rc;
}

static
rc_t Experiment_FillSeg(const ExperimentXML* self, PoolReadSpec* pool, const PoolMember* src,
                        uint32_t nbases, const char* bases,
                        const char** const new_member_name)
{
    rc_t rc = 0;
    uint32_t i, j, till;
    bool tailed = false;
    PoolMember_FindReverse_Data data;
    uint32_t mismatch_leftover = 0;

    memcpy(pool, src->spec, sizeof(src->spec[0]) * self->reads->nreads);
    if( new_member_name != NULL ) {
        data.nreads = self->reads->nreads;
        memset(data.tags, 0, sizeof(data.tags[0]) * self->reads->nreads);
        data.member = NULL;
    }

    for(i = 0; rc == 0 && i < self->reads->nreads; i++) {
        switch(self->reads->reads[i].coord_type) {
            case rdsp_RelativeOrder_ct:
                if( pool[i].start < 0 ) {
                    if( i == 0 ) {
                        pool[i].start = 0;
                    } else if( pool[i - 1].start >= 0 && pool[i - 1].length >= 0 ) {
                        pool[i].start = pool[i - 1].start + pool[i - 1].length;
                    }
                }
                if( pool[i].start >= 0 ) {
                    mismatch_leftover = ~0;
                }
                break;

            case rdsp_BaseCoord_ct:
            case rdsp_CycleCoord_ct:
                if( i > 0 && pool[i - 1].length < 0 && pool[i - 1].start >= 0 ) {
                    pool[i - 1].length = pool[i].start - pool[i - 1].start;
                }
                mismatch_leftover = ~0;
                break;

            case rdsp_ExpectedBaseCall_ct:
                if( pool[i].start < 0 ) {
                    if( i == 0 ) {
                        pool[i].start = 0;
                    } else if( pool[i - 1].start >= 0 ) {
                        pool[i].start = pool[i - 1].start;
                        if( pool[i - 1].length >= 0 ) {
                            pool[i].start += pool[i - 1].length;
                        }
                    }
                }
                if( pool[i].start >= 0 ) {
                    /* find where it really starts */
                    int16_t pos = 0;
                    if( i > 0 && pool[i - 1].length == 0 ) {
                        /* previous was not found, try to find current wherever */
                        mismatch_leftover = ~0;
                    }
                    till = nbases;
                    for(j = i + 1; j < self->reads->nreads; j++) {
                        /* look no further than next fixed start read */
                        if( self->reads->reads[j].coord_type == rdsp_BaseCoord_ct ||
                            self->reads->reads[j].coord_type == rdsp_CycleCoord_ct ) {
                            till = pool[j].start;
                            break;
                        }
                    }
                    rc = Experiment_Match(&self->reads->reads[i].coord.expected_basecalls, self->platform->id,
                                          &bases[pool[i].start], till - pool[i].start,
                                          &mismatch_leftover, NULL, &pos, &pool[i].length, NULL);
                    if( rc == 0 ) {
                        pool[i].start += pos;
                        if( i > 0 ) {
                            if( pos > 0 && pool[i - 1].length > 0) {
                                /* drop prefix to previous read tail, pos can be > 0 only if mismatch_leftover was > 0 before match */
                                pool[i - 1].length += pos;
                            } else if( pool[i - 1].length == 0 ) {
                                /* previous was not found, drop all misses into it */
                                pool[i - 1].length = pool[i].start - pool[i - 1].start;
                            }
                        }
                    } else if( GetRCState(rc) == rcNotFound ) {
                        mismatch_leftover = 0;
                        pool[i].length = 0;
                        rc = 0;
                    }
                    if( i > 0 && pool[i - 1].length < 0 && pool[i - 1].start >= 0 && pool[i].length > 0 ) {
                        pool[i - 1].length = pool[i].start - pool[i - 1].start;
                    }
                }
                break;

            case rdsp_ExpectedBaseCallTable_ct:
                if( pool[i].start < 0 ) {
                    if( i == 0 ) {
                        pool[i].start = 0;
                    } else if( pool[i - 1].start >= 0 ) {
                        pool[i].start = pool[i - 1].start;
                        if( pool[i - 1].length >= 0 ) {
                            pool[i].start += pool[i - 1].length;
                        }
                    }
                }
                if( pool[i].start >= 0 ) {
                    if( pool[i].item[0] != NULL && new_member_name != NULL ) {
                        /* an assert :) */
                        rc = RC(rcSRA, rcFormatter, rcResolving, rcParam, rcInconsistent);
                        PLOGERR(klogErr,(klogErr, rc, "member known while requested '$(m)'", PLOG_S(m), src->name));
                        break;
                    }
                    till = nbases;
                    for(j = i + 1; j < self->reads->nreads; j++) {
                        /* look no further than next fixed start read */
                        if( self->reads->reads[j].coord_type == rdsp_BaseCoord_ct ||
                            self->reads->reads[j].coord_type == rdsp_CycleCoord_ct ) {
                            till = pool[j].start;
                            break;
                        }
                    }
                    if( pool[i].item[0] == NULL ) {
                        if( (pool[i].length < 0 || new_member_name != NULL) ) {
                            /* find member and where it starts */
                            int16_t pos = 0;
                            if( i > 0 && pool[i - 1].length == 0 ) {
                                /* previous was not found, try to find current wherever */
                                mismatch_leftover = ~0;
                            }
                            rc = Experiment_Match(&self->reads->reads[i].coord.expected_basecalls, self->platform->id,
                                                  &bases[pool[i].start], till - pool[i].start,
                                                  &mismatch_leftover, &data.tags[i], &pos, &pool[i].length, NULL);
                            if( rc == 0 ) {
                                pool[i].start += pos;
                                if( i > 0 ) {
                                    if( pos > 0 && pool[i - 1].length > 0) {
                                        /* drop prefix to previous read tail, pos can be > 0 only if mismatch_leftover was > 0 before match */
                                        pool[i - 1].length += pos;
                                    } else if( pool[i - 1].length == 0 ) {
                                        /* previous was not found, drop all misses into it */
                                        pool[i - 1].length = pool[i].start - pool[i - 1].start;
                                    }
                                }
                            } else if( GetRCState(rc) == rcNotFound ) {
                                mismatch_leftover = 0;
                                /* set length to table length */
                                pool[i].length = self->reads->reads[i].coord.expected_basecalls.default_length;
                                rc = 0;
                            }
                        } else {
                            /* not present */
                            pool[i].length = 0;
                            mismatch_leftover = 0;
                        }
                    } else {
                        /* find where known basecall starts, find ONLY basecalls knwon through member (using tag), not the whole thing! */
                        int16_t pos_set = 0, len_set = 0;
                        uint32_t misses_set = ~0;
                        for(j = 0; pool[i].item[j] != NULL; j++) {
                            int16_t pos = 0, len = 0;
                            uint32_t misses = ~0, mismatch_left = mismatch_leftover;
                            rc = Experiment_Match(&self->reads->reads[i].coord.expected_basecalls, self->platform->id,
                                                  &bases[pool[i].start], till - pool[i].start,
                                                  &mismatch_left, (char**)&pool[i].item[j]->read_group_tag, &pos, &len, &misses);
                            if( rc != 0 && GetRCState(rc) != rcNotFound ) {
                                /* general error */
                                break;
                            } else if( rc == 0 && misses < misses_set ) {
                                /* remember better match based on mismatches used */
                                pos_set = pos;
                                len_set = len;
                                misses_set = misses;
                                mismatch_leftover = mismatch_left;
                            }
                        }
                        if( rc == 0 ) {
                            pool[i].start += pos_set;
                            pool[i].length += len_set;
                            if( i > 0 && pos_set > 0 && pool[i - 1].length > 0) {
                                /* drop prefix to previous read tail, pos_set can be > 0 only if mismatch_leftover was > 0 before match */
                                pool[i - 1].length += pos_set;
                            }
                        } else if( GetRCState(rc) == rcNotFound ) {
                            mismatch_leftover = 0;
                            /* set length to table length */
                            pool[i].length = self->reads->reads[i].coord.expected_basecalls.default_length;
                            rc = 0;
                        }
                    }
                    if( i > 0 && pool[i - 1].length < 0 && pool[i - 1].start >= 0 && pool[i].length > 0 ) {
                        pool[i - 1].length = pool[i].start - pool[i - 1].start;
                    }
                }
                break;
        }
    }
    if( rc == 0 && new_member_name != NULL ) {
        if( BSTreeDoUntil(self->member_pool, false, PoolMember_FindReverse, &data) ) {
            *new_member_name = data.member->name;
            DEBUG_MSG (2, ("Assigned member_name '%s'\n", *new_member_name));
        } else {
#if _DEBUGGING
            DEBUG_MSG(2, ("member_name reverse lookup failed:\n"));
            for(i = 0; i < self->reads->nreads; i++ ) {
                DEBUG_MSG (2, ("read %hhu read_group_tag '%s'\n", i + 1, data.tags[i]));
            }
#endif
        }
    }

    /* allocate tail of seq to 1st 0 or negative length read */
    for(i = 0; rc == 0 && i < self->reads->nreads; i++) {
        if( !tailed && pool[i].length < 0 && pool[i].start >=0 ) {
            pool[i].length = nbases - pool[i].start;
            tailed = true;
        } else if( tailed ) {
            pool[i].start = nbases;
            pool[i].length = 0;
        }
    }
    return rc;
}

rc_t Experiment_ReadSegDefault(const ExperimentXML* self, SRASegment* seg)
{
    rc_t rc = 0;

    if( self == NULL || seg == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcResolving, rcParam, rcNull);
    } else if( self->platform->id == SRA_PLATFORM_454 ||
               self->platform->id == SRA_PLATFORM_HELICOS ) {
        uint32_t i;
        /* HACK: fake READ_SEG starting with key_seq and 1 base long each next */
        for(i = 0; i < self->reads->nreads; i++) {
            seg[i].start = i;
            seg[i].len = 1;
        }
    } else {
        uint32_t i;
        int32_t spot_len = 0;

        if(self->platform->id == SRA_PLATFORM_ILLUMINA ) {
            spot_len = self->platform->param.illumina.sequence_length;
        } else if(self->platform->id == SRA_PLATFORM_ABSOLID ) {
            spot_len = self->platform->param.absolid.sequence_length;
        } else {
            rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcUnsupported);
            LOGERR(klogErr, rc, "platform");
            return rc;
        }
        i = self->reads->nreads;
        do {
            int16_t len = 0;
            --i;
            switch(self->reads->reads[i].coord_type) {
                case rdsp_RelativeOrder_ct:
                    if( i == 0 ) {
                        len = spot_len;
                    } else {
                        len = 0;
                    }
                    break;
                case rdsp_BaseCoord_ct:
                case rdsp_CycleCoord_ct:
                    len = spot_len - self->reads->reads[i].coord.start_coord + 1;
                    break;
                case rdsp_ExpectedBaseCall_ct:
                case rdsp_ExpectedBaseCallTable_ct:
                    len = self->reads->reads[i].coord.expected_basecalls.default_length;
                    break;
            }
            spot_len -= len;
            if( spot_len < 0 || len < 0 ) {
                rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInconsistent);
                LOGERR(klogErr, rc, "cumulative read lengths and SEQUENCE_LENGTH");
                return rc;
            } else {
                seg[i].start = spot_len;
                seg[i].len = len;
            }
        } while( i > 0 );
    }
    return rc;
}

rc_t Experiment_MemberSeg(const ExperimentXML* self,
                          const char* const file_member_name, const char* const data_block_member_name,
                          uint32_t nbases, const char* bases,
                          SRASegment* seg, const char** const new_member_name)
{
    rc_t rc = 0;
    uint32_t i;
    PoolReadSpec pool[MAX_SPOT_DESCRIPTOR_READS];

    if( self == NULL || bases == NULL || seg == NULL || new_member_name == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcResolving, rcParam, rcNull);
    } else if( nbases < 1 ) {
        rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcEmpty);
    }

    if( rc == 0 ) {
        const char* mm = NULL;
        switch(self->processing.barcode_rule) {
            case eBarcodeRule_not_set:
                if( data_block_member_name == NULL && file_member_name != NULL && file_member_name[0] != '\0'
                    && self->member_pool == NULL ) {
                    rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcExcessive);
                    LOGERR(klogErr, rc, "file contains barcode but sample pool is not defined");
                }
                mm = data_block_member_name ? data_block_member_name : file_member_name;
                break;

            case eBarcodeRule_use_file_spot_name:
                mm = file_member_name;
                break;

            case eBarcodeRule_use_table_in_experiment:
            case eBarcodeRule_ignore_barcode:
                mm = NULL;
                break;
        }
        if( rc == 0 && (self->member_pool == NULL || self->processing.barcode_rule == eBarcodeRule_ignore_barcode) ) {
            if( (rc = Experiment_FillSeg(self, pool, self->member_null, nbases, bases, NULL)) == 0 ) {
                *new_member_name = mm;
                rc = PoolMemberStats_Add(self, *new_member_name);
            }
        } else if( rc == 0 ) {
            if( mm == NULL ) {
                *new_member_name = NULL;
                if( (rc = Experiment_FillSeg(self, pool, self->member_null, nbases, bases, new_member_name)) == 0 ) {
                    if( (rc = PoolMemberStats_Add(self, *new_member_name)) == 0 ) {
                        if( *new_member_name != NULL ) {
                            rc = PoolMemberStats_Delete(self, NULL);
                        }
                    }
                }
            } else {
                PoolMember* pm = (PoolMember*)BSTreeFind(self->member_pool, mm, PoolMember_FindByName);
                if( pm == NULL ) {
                    rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcNotFound);
                    PLOGERR(klogErr,(klogErr, rc, "specified member_name '$(m)'", PLOG_S(m), mm));
                } else {
                    if( (rc = Experiment_FillSeg(self, pool, pm, nbases, bases, NULL)) == 0 ) {
                        *new_member_name = mm;
                        rc = PoolMemberStats_Add(self, *new_member_name);
                    }
                }
            }
        }
    }

    if( rc == 0 ) {
        uint32_t new_bases = 0;

        for(i = 0; i < self->reads->nreads; i++) {
            if( pool[i].start < 0 || pool[i].length < 0 ) {
                rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcIncomplete);
                PLOGERR(klogErr,(klogErr, rc, "Read #$(i) boundaries", PLOG_U32(i), i + 1));
            } else if( i == 0 ) {
                if( pool[i].start != 0 ) {
                    rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcInvalid);
                    PLOGERR(klogErr,(klogErr, rc, "Read #$(i) do not start at 0", PLOG_U32(i), i + 1));
                }
            } else if( i != 0 ) {
                if( pool[i].start - pool[i - 1].start != pool[i - 1].length ) {
                    rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcInconsistent);
                    PLOGERR(klogErr,(klogErr, rc, "Gap between segments #$(p) and #$(i) ", PLOG_2(PLOG_U32(p),PLOG_U32(i)), i, i + 1));
                }
            } else if( pool[i].start + pool[i].length > nbases ) {
                rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcInconsistent);
                PLOGERR(klogErr,(klogErr, rc, "Read #$(i) is longer than sequence", PLOG_U32(i), i + 1));
            }
            seg[i].start = pool[i].start;
            seg[i].len = pool[i].length;
            new_bases += seg[i].len;
        }
        if( nbases != new_bases ) {
            rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcInconsistent);
            PLOGERR(klogErr,(klogErr, rc, "total read length $(c)", PLOG_U32(i), new_bases));
        }
        DEBUG_MSG (3, ("%R, %s: bases %u: '%.*s'\n",
                       rc, *new_member_name, nbases, nbases, bases));
        for(i = 0; i < self->reads->nreads; i++) {
            const char* b = NULL;
            if( pool[i].start >= 0 ) {
                b = &bases[pool[i].start];
            }
            DEBUG_MSG (7, ("Read %u. [%hd:%hd] - '%.*s'\n",
                           i + 1, pool[i].start, pool[i].length, pool[i].length < 0 ? 0 : pool[i].length, b));
        }
    }
    return rc;
}

rc_t Experiment_MemberSegSimple(const ExperimentXML* self,
                                const char* const file_member_name, const char* const data_block_member_name,
                                const SRASegment* seg, const char** const new_member_name)
{
    rc_t rc = 0;
    if( self == NULL || seg == NULL || new_member_name == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcResolving, rcParam, rcNull);
    } else {
        switch(self->processing.barcode_rule) {
            case eBarcodeRule_not_set:
                if( data_block_member_name == NULL && file_member_name != NULL && file_member_name[0] != '\0'
                    && self->member_pool == NULL )
                {
                    rc = RC(rcSRA, rcFormatter, rcResolving, rcData, rcExcessive);
                    LOGERR(klogErr, rc, "file contains barcode but sample pool is not defined");
                }
                *new_member_name = data_block_member_name ? data_block_member_name : file_member_name;
                break;

            case eBarcodeRule_use_file_spot_name:
                *new_member_name = file_member_name;
                break;

            case eBarcodeRule_use_table_in_experiment:
            case eBarcodeRule_ignore_barcode:
                *new_member_name = NULL;
                break;
        }
        if( rc == 0 ) {
            rc = PoolMemberStats_Add(self, *new_member_name);
        }
    }
    return rc;
}

static
rc_t Experiment_Apply_RUN_ATTRIBUTES(const ExperimentXML* cself, RunAttributes* attr)
{
    rc_t rc = 0;

    if( cself == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcEvaluating, rcSelf, rcNull);
    } else if( attr != NULL ) {
        ExperimentXML* obj = (ExperimentXML*)cself;

        obj->processing.barcode_rule = attr->barcode_rule;
        if( attr->quality_offset != '\0' ) {
            obj->processing.quality_offset = attr->quality_offset;
        }
        if( attr->quality_type != eExperimentQualityType_Undefined ) {
            obj->processing.quality_type = attr->quality_type;
        }
    }
    return rc;
}

rc_t Experiment_Make(const ExperimentXML** self, const KXMLDoc* doc, RunAttributes* attr)
{
    rc_t rc = 0;
    ExperimentXML* obj = NULL;
    const KXMLNodeset* ns = NULL;

    if( self == NULL || doc == NULL ) {
        return RC(rcSRA, rcFormatter, rcConstructing, rcParam, rcNull);
    }
    obj = calloc(1, sizeof(*obj));
    if( obj == NULL ) {
        return RC(rcSRA, rcFormatter, rcConstructing, rcMemory, rcExhausted);
    }
    if( (rc = KXMLDocOpenNodesetRead(doc, &ns, "/EXPERIMENT_SET/EXPERIMENT | /EXPERIMENT")) == 0 ) {
        uint32_t count = 0;
        if( (rc = KXMLNodesetCount(ns, &count)) == 0 && count != 1 ) {
            rc = RC(rcSRA, rcFormatter, rcConstructing, rcTag, count ? rcExcessive : rcNotFound);
            LOGERR(klogErr, rc, "EXPERIMENT");
        } else {
            const KXMLNode* EXPERIMENT = NULL;
            if( (rc = KXMLNodesetGetNodeRead(ns, &EXPERIMENT, 0)) == 0 ) {
                if( (rc = PlatformXML_Make(&obj->platform, EXPERIMENT)) == 0 ) {
                    if( attr->platform != NULL ) {
                        if( obj->platform->id != attr->platform->id ) {
                            rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInconsistent);
                            LOGERR(klogErr, rc, "EXPERIMENT and RUN platforms differ");
                        } else {
                            PlatformXML_Whack(obj->platform);
                            obj->platform = attr->platform;
                            /* take ownership */
                            attr->platform = NULL;
                        }
                    }
                    if( rc == 0 && (rc = parse_PROCESSING(EXPERIMENT, &obj->processing)) == 0 ) {
                        if( (rc = ReadSpecXML_Make(&obj->reads, EXPERIMENT, "DESIGN/SPOT_DESCRIPTOR")) == 0 ) {
                            if( attr->reads != NULL ) {
                                ReadSpecXML_Whack(obj->reads);
                                obj->reads = attr->reads;
                                /* take ownership */
                                attr->reads = NULL;
                            }
                            if( rc == 0 && (rc = parse_POOL(EXPERIMENT, obj)) == 0 ) {
                                rc = Experiment_Apply_RUN_ATTRIBUTES(obj, attr);
                            }
                        } else {
                            LOGERR(klogErr, rc, "EXPERIMENT/.../READ_SPEC");
                        }
                    }
                } else {
                    LOGERR(klogErr, rc, "EXPERIMENT/PLATFORM");
                }
                KXMLNodeRelease(EXPERIMENT);
            }
        }
        KXMLNodesetRelease(ns);
    }
    if( rc != 0 ) {
        *self = NULL;
        Experiment_Whack(obj, false);
    } else {
        *self = obj;
    }
    return rc;
}

rc_t Experiment_GetPlatform(const ExperimentXML* cself, const PlatformXML** platform)
{
    rc_t rc = 0;
    if( cself == NULL || platform == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else {
        *platform = cself->platform;
    }
    return rc;
}

rc_t Experiment_GetProcessing(const ExperimentXML* cself, const ProcessingXML** processing)
{
    rc_t rc = 0;
    if( cself == NULL || processing == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else {
        *processing = &cself->processing;
    }
    return rc;
}

rc_t Experiment_GetReadNumber(const ExperimentXML* cself, uint8_t* nreads)
{
    rc_t rc = 0;
    if( cself == NULL || nreads == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else {
        *nreads = cself->reads->nreads;
    }
    return rc;
}

rc_t Experiment_GetRead(const ExperimentXML* cself, uint8_t read_id, const ReadSpecXML_read** read_spec)
{
    rc_t rc = 0;
    if( cself == NULL || read_spec == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else if( read_id >= cself->reads->nreads ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcId, rcOutofrange);
    } else {
        *read_spec = &(cself->reads->reads[read_id]);
    }
    return rc;
}

rc_t Experiment_FindReadInTable(const ExperimentXML* cself, uint8_t read_id, const char* key, const char** basecall, const char** member_name)
{
    rc_t rc = 0;
    if( cself == NULL || key == NULL || (basecall == NULL && member_name == NULL) ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else if( read_id >= cself->reads->nreads ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcId, rcOutofrange);
    } else {
        const ReadSpecXML_read* read_spec = &(cself->reads->reads[read_id]);
        if( read_spec->coord_type != rdsp_ExpectedBaseCallTable_ct ) {
            rc = RC(rcSRA, rcFormatter, rcAccessing, rcName, rcNotFound);
        } else {
            uint32_t i;
            /* try to find key as basecall */
            for(i = 0; i < read_spec->coord.expected_basecalls.count; i++) {
#ifdef __GNUC__
#warning "TBD: must do match here!"
#endif
                if( strcasecmp(key, read_spec->coord.expected_basecalls.table[i].basecall) == 0 ) {
                    if( basecall != NULL ) {
                        *basecall = read_spec->coord.expected_basecalls.table[i].basecall;
                    }
                    if( member_name != NULL ) {
                        if( cself->member_pool != NULL ) {
                            PoolMember_FindByTag_Data data;
                            data.readid = read_id;
                            data.tag = read_spec->coord.expected_basecalls.table[i].read_group_tag;
                            if( BSTreeDoUntil(cself->member_pool, false, PoolMember_FindByTag, &data) ) {
                                *member_name = data.member_name;
                            } else {
                                rc = RC(rcSRA, rcFormatter, rcAccessing, rcSelf, rcCorrupt);
                            }
                        } else {
                            *member_name = NULL;
                        }
                    }
                    /* key found as basecall value */
                    return rc;
                }
            }
            /* try to find key as member_name */
            if( cself->member_pool != NULL ) {
                PoolMember* pm = (PoolMember*)BSTreeFind(cself->member_pool, key, PoolMember_FindByName);
                if( pm != NULL ) {
                    if( member_name != NULL ) {
                        *member_name = pm->name;
                    }
                    if( basecall != NULL ) {
                        if( pm->spec[read_id].item[0] != NULL ) {
                            *basecall = pm->spec[read_id].item[0]->basecall;
                            if( pm->spec[read_id].item[1] != NULL ) {
                                rc = RC(rcSRA, rcFormatter, rcAccessing, rcData, rcAmbiguous);
                            }
                        } else {
                            rc = RC(rcSRA, rcFormatter, rcAccessing, rcData, rcViolated);
                        }
                    }
                    /* key found as member_name value */
                    return rc;
                }
            }
        }
    }
    return rc ? rc : RC(rcSRA, rcFormatter, rcAccessing, rcName, rcNotFound);
}

rc_t Experiment_HasPool(const ExperimentXML* cself, bool* has_pool)
{
    rc_t rc = 0;
    if( cself == NULL || has_pool == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcAccessing, rcParam, rcNull);
    } else {
        *has_pool = cself->member_pool != NULL;
    }
    return rc;
}

void Experiment_Whack(const ExperimentXML* cself, bool log_member_stats)
{
    if( cself != NULL ) {
        ExperimentXML* self = (ExperimentXML*)cself;

        free(self->processing.base_caller);
        free(self->processing.quality_scorer);
        ReadSpecXML_Whack(self->reads);
        BSTreeWhack(self->member_pool, PoolMember_Whack, NULL);
        free(self->member_pool);
        PoolMember_Whack(&self->member_null->node, NULL);
        SLListWhack(&self->member_stats, PoolMemberStats_Whack, &log_member_stats);
        PlatformXML_Whack(self->platform);
        free(self);
    }
}
