/*==============================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <kapp/main.h>
#include <klib/log.h>
#include <klib/out.h>
#include <klib/status.h>
#include <kdb/meta.h>
#include <klib/xml.h>
#include <kfs/md5.h>
#include <kfs/arc.h>
#include <kfs/bzip.h>
#include <kfs/gzip.h>
#include <kfs/file.h>
#include <kfs/directory.h>
#include <klib/rc.h>

#include <sra/srapath.h>
#include <sra/wsradb.h>

#include <os-native.h>

#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include "debug.h"
#include "loader.vers.h"
#include "loader-fmt.h"
#include "run-xml.h"
#include "experiment-xml.h"
#include "loader-file.h"

static KDirectory *s_Directory = NULL;
static const KXMLMgr *s_XmlMgr = NULL;

typedef enum EOutputMode {
    eUndefinedMode = 0,
    eCreateMode,
    eOutputMode,
    eAppendMode
} EOutputMode;

typedef struct ResultXML {
    char *file;
    char *name;
    char *region;
    char *sector;
    char *member_name;
} ResultXML;

typedef struct ExpectedXML {
    uint8_t number_of_results;
    ResultXML **result;
} ExpectedXML;

typedef struct SCmdLineArgs {
    /* help */
    bool _help;

    /* if spots_to_run >= 0 then must stop after writing given number of spots */
    int64_t _spots_to_run;
    uint32_t _spots_bad_allowed;
    int _spots_bad_allowed_percentage;

    /* output */
    EOutputMode _outputMode;
    const char *_repPath;
    const char *_volPath;
    const char *_runName; /* run path or accession */

    /* spot-descriptor */
    const char *_experimentXmlPath;

    /* input */
    const char *_runXmlPath;
    const char *_expectedPath;

    SLList _sraLoaderFileArgs;
    RunXML* _run;
    const ExperimentXML* _experiment;
    const ExpectedXML* _expected;
} SCmdLineArgs;

static const char* defaultLogLevel = "info";

rc_t Usage( const Args * args )
{
    /* dummy for newer args system will be filled when system is complete!!! */
    return 0;
}

static
void LocalUsage(const char *argv0, bool extended, rc_t rc, const char *msg, ...)
{
    printf("Usage:\n\n");
    printf("%s <help>\n", argv0);
    printf("%s [ <options> ] <input-spec> <spot-descriptor> <output-spec> [ <verbosity> ] \n\n", argv0);
    printf(
"input-spec:\n"
"  ( -r | --run-xml ) path           # path to run.xml describing input files\n"
"\n");
    printf(
"spot-descriptor:\n"
"  ( -e | --experiment ) path            # path to experiment.xml \n"
"                                        # with spot descriptor\n"
"\n");
   printf(
"output-spec:\n"
"  <output-mode> <run-name>           # produce output to run\n"
#if USE_NCBI_SRADB
"  ( -S | --server ) path             # path to replication server\n"
"  ( -V | --volume ) path             # server-relative path to output volume\n"
#endif
"\n");
    printf(
"output-mode:\n"
"  -c | --create                      # create only new run (fail if exists)\n"
"  -o | --output                      # create new or replace existing run\n"
"  -a | --append                      # create new or update existing run\n"
"\n"
"run-name:\n"
"  path                                  # full or working-directory relative\n"
"                                        # file-system path to run\n"
#if USE_NCBI_SRADB
"  accession                             # accession string\n"
#endif
"\n");
    printf(
"options:\n"
"  ( -n | --spotnumber) spots_to_run    # tells the loader to stop processing after that many spots\n"
"                                       # helpful in testing\n"
"\n"
);
    printf(
"  ( -b | --bad-spot-number ) bad_spot_number  # number of spot creation errors to be tolerated\n"
"                                              # default = 50\n"
"\n"
"  ( -p | --bad-spot-percentage ) bad_spot_percentage> # PERCENTAGE of the number of bad spots against\n"
"                                                      # the number of good spots, default = 5\n"
"\n"
);
    printf(
"help:\n"
"  ( -h | --help )                       # print usage and option details\n"
"\n"
);
    printf(
"verbosity:\n"
"  ( -l | --log-level ) <level>  #logging level: fatal|sys|int|err|warn|info default=%s\n", defaultLogLevel);
#if _DEBUGGING
    printf(
"debug:\n"
"  ( -+ | --debug ) <module>  #module: APP|KFS|XML|VDB|SRA|REF|LEGREF default=APP\n"
);
#endif

    printf("\nVersion: %u.%u.%u\n\n", KAppVersion() >> 24, (KAppVersion() >> 16) & 0x00FF, KAppVersion() & 0x0000FFFF );

    if(msg) {
        va_list args;
        va_start(args, msg);
        printf("\n");
        if( rc != 0 ) {
            char rcbuf[4096];
            size_t n = 0;
            RCExplain(rc, rcbuf, sizeof(rcbuf) - 1, &n);
            rcbuf[n < sizeof(rcbuf) ? n : (sizeof(rcbuf) - 1)] = '\0';
            printf("%s - ", rcbuf);
        }
        vprintf(msg, args);
        printf("\n");
        va_end(args);
    }
}

static bool GetArg(int *argI, int argc, char *argv[],
    const char *shortArg, const char *longArg,
    const char **out, const char *error, rc_t *errorRc)
{
    bool found = false;

    assert(argI && (*argI < argc) && longArg && argv && out && errorRc);

    if (strcmp(argv[*argI], longArg) == 0) {
        found = true;
    }
    else if (shortArg && (strcmp(argv[*argI], shortArg) == 0)) {
        found = true;
    }
    if (!found) {
        return false;
    }

    if ((*argI + 1) >= argc) {
        *errorRc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
        LocalUsage(argv[0], false, *errorRc, error);
    }
    else {
        *out = argv[++(*argI)];
    }

    return true;
}

static rc_t SCmdLineArgsParse(SCmdLineArgs *args,
    int argc, char *argv[], bool *quit)
{
    rc_t rc = 0;
    int i = 0;
    const char *outputMode = NULL;

    assert(args && argv && quit);

    *quit = false;

    for(i = 1; rc == 0 && i < argc; i++) {
        if( strcmp(argv[i], "-l") == 0 || strcmp(argv[i], "--log-level") == 0 ) {
            if( (rc = LogLevelSet(argv[++i])) != 0 ) {
                LocalUsage(argv[0], false, rc, "log level '%s'", argv[i - 1]);
            }
#if _DEBUGGING
        } else if( strcmp(argv[i], "-+") == 0 || strcmp(argv[i], "--debug") == 0 ) {
            if( (rc = KDbgSetString(argv[++i])) != 0 ) {
                LocalUsage(argv[0], false, rc, "debug level '%s'", argv[i - 1]);
            }
#endif
        } else if( strcmp(argv[i], "-n") == 0 || strcmp(argv[i], "--spotnumber") == 0 ) {
            char* end = NULL;
            errno = 0;
            args->_spots_to_run = strtol(argv[++i], &end, 10);
            if( errno != 0 || *end != '\0' || args->_spots_to_run <= 0 ) {
                rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                LocalUsage(argv[0], false, rc, "spot number %s", argv[i]);
                break;
            }
        } else if( strcmp(argv[i], "-b") == 0 || strcmp(argv[i], "--bad-spot-number") == 0 ) {
            char* end = NULL;
            long x;
            errno = 0;
            x = strtol(argv[++i], &end, 10);
            if( errno != 0 || *end != '\0' || x < 0 ) {
                rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                LocalUsage(argv[0], false, rc, "bad spot number %s", argv[i]);
                break;
            }
            args->_spots_bad_allowed = x;
        } else if( strcmp(argv[i], "-p") == 0 || strcmp(argv[i], "--bad-spot-percentage") == 0 ) {
            char* end = NULL;
            errno = 0;
            args->_spots_bad_allowed_percentage = strtol(argv[++i], &end, 10);
            if( errno != 0 || *end != '\0' || args->_spots_bad_allowed_percentage < 0 || args->_spots_bad_allowed_percentage > 100 ) {
                rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                LocalUsage(argv[0], false, rc, "bad spot precent %s", argv[i]);
                break;
            }
        } else if (strcmp(argv[i], "-h") == 0) {
            LocalUsage(argv[0], false, 0, NULL);
            *quit = true;
            break;
        } else if (strcmp(argv[i], "--help") == 0) {
            LocalUsage(argv[0], true, 0, NULL);
            *quit = true;
            break;
        } else if (GetArg(&i, argc, argv, "-c", "--create", &args->_runName, "expected run name", &rc)) {
            args->_outputMode = eCreateMode;
            if (rc == 0) {
                const char *mode = argv[i - 1];
                if (outputMode) {
                    rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                    LocalUsage(argv[0], false, rc, "Just one of %s and %s could be specified", outputMode, mode);
                    break;
                } else {
                    outputMode = mode;
                }
            }
        } else if (GetArg(&i, argc, argv, "-o", "--output", &args->_runName, "expected run name", &rc)) {
            args->_outputMode = eOutputMode;
            if (rc == 0) {
                const char *mode = argv[i - 1];
                if (outputMode) {
                    rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                    LocalUsage(argv[0], false, rc, "Just one of %s and %s could be specified", outputMode, mode);
                    break;
                } else {
                    outputMode = mode;
                }
            }
        } else if (GetArg(&i, argc, argv, "-a", "--append", &args->_runName, "expected run name", &rc)) {
            args->_outputMode = eAppendMode;
            if (rc == 0) {
                const char *mode = argv[i - 1];
                if (outputMode) {
                    rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
                    LocalUsage(argv[0], false, rc, "Just one of %s and %s could be specified", outputMode, mode);
                    break;
                } else {
                    outputMode = mode;
                }
            }
#if USE_NCBI_SRADB
        } else if (GetArg(&i, argc, argv, "-S", "--server", &args->_repPath, "expected server path", &rc)) {
        } else if (GetArg(&i, argc, argv, "-V", "--volume", &args->_volPath, "expected volume path", &rc)) {
#endif
        } else if (GetArg(&i, argc, argv, "-e", "--experiment", &args->_experimentXmlPath, "expected experiment path", &rc)) {
        } else if (GetArg(&i, argc, argv, "-r", "--run-xml", &args->_runXmlPath, "expected run.xml path", &rc)) {
        } else if (GetArg(&i, argc, argv, "-x", "--expected", &args->_expectedPath, "expected expected.xml", &rc)) {
        } else {
            rc = RC(rcExe, rcArgv, rcReading, rcParam, rcInvalid);
            LocalUsage(argv[0], false, rc, "unrecognized switch - '%s'", argv[i]);
        }
    }
    return rc;
}

static rc_t SCmdLineArgsMake(SCmdLineArgs **args,
    int argc, char *argv[])
{
    rc_t rc = 0;
    bool quit = false;

    SCmdLineArgs *r = (SCmdLineArgs*) calloc(1, sizeof *r);

    assert(args);
    *args = NULL;

    if (r == NULL) {
        return RC(rcExe, rcArgv, rcAllocating, rcMemory, rcExhausted);
    }

    SLListInit(&r->_sraLoaderFileArgs);
    r->_spots_to_run = -1;
    r->_spots_bad_allowed = 50;
    r->_spots_bad_allowed_percentage = 5;

    if (rc == 0) {
        rc = SCmdLineArgsParse(r, argc, argv, &quit);
    }

    if( !quit && rc == 0 ) {
        if( r->_runName == NULL ) {
            rc = RC(rcExe, rcArgv, rcValidating, rcParam, rcEmpty);
            LocalUsage(argv[0], false, rc, "Run name is not specified");
        } else if( r->_experimentXmlPath == NULL ) {
            rc = RC(rcExe, rcArgv, rcValidating, rcParam, rcEmpty);
            LocalUsage(argv[0], false, rc, "Experiment XML is not specified");
        } else if( r->_runXmlPath == NULL ) {
            rc = RC(rcExe, rcArgv, rcValidating, rcParam, rcEmpty);
            LocalUsage(argv[0], false, rc, "Run XML is not specified");
        }
    }

    if (!quit && rc == 0) {
        *args = r;
    }
    else {
        free(r);
        exit(1);
    }

    return rc;
}

static rc_t read_required_node_data_cstr(const KXMLNode *node,
    const char *attr, char **string)
{
    rc_t rc = attr == NULL ?
            KXMLNodeReadCStr(node, string, NULL) :
            KXMLNodeReadAttrCStr(node, attr, string, NULL);

    if (rc && GetRCState(rc) == rcNotFound) {
        const char *name = NULL;
        if( (rc = KXMLNodeElementName(node, &name)) == 0 ) {
            if (attr == NULL) {
                PLOGERR(klogErr, (klogErr, rc, "XML node $(name)", "name=%s", name));
            }
            else {
                PLOGERR(klogErr, (klogErr, rc, "XML attribute $(name)/@$(attr)",
                                  "name=%s,attr=%s", name, attr));
            }
        }
    }
    return rc;
}

static rc_t read_optional_node_data_cstr(const KXMLNode *node,
    const char *attr, char **string)
{
    rc_t rc = attr == NULL ?
            KXMLNodeReadCStr(node, string, NULL) :
            KXMLNodeReadAttrCStr(node, attr, string, NULL);

    if (rc) {
        if (GetRCState(rc) == rcNotFound) {
            rc = 0;
        }
        else {
            const char *name = NULL;
            if( (rc = KXMLNodeElementName(node, &name)) == 0 ) {
                PLOGERR(klogErr, (klogErr, rc, "XML attribute $(name)/@$(attr)",
                                  "name=%s,attr=%s", name, attr));
            }
        }
    }
    return rc;
}

static rc_t ResultXML_make(ResultXML **rslt, const KXMLNode *Result)
{
    rc_t rc = 0;
    ResultXML *r = NULL;

    assert(rslt && Result);

    r = calloc(1, sizeof *r);
    if (r == NULL) {
        rc = RC(rcExe, rcXmlDoc, rcAllocating, rcMemory, rcExhausted);
    }

    if (rc == 0) {
        rc = read_required_node_data_cstr(Result, "file", &r->file);
    }

    if (rc == 0) {
        rc = read_required_node_data_cstr(Result, "name", &r->name);
    }

    if (rc == 0) {
        rc = read_required_node_data_cstr(Result, "sector", &r->sector);
    }

    if (rc == 0) {
        rc = read_required_node_data_cstr(Result, "region", &r->region);
    }

    if (rc == 0) {
        rc = read_optional_node_data_cstr(Result, "member_name", &r->member_name);
    }

    if (rc) {
        free(r);
    }
    else {
        *rslt = r;
    }

    return rc;
}
static rc_t ExpectedXMLWhack(const ExpectedXML *r)
{
    return 0;
}
static rc_t ExpectedXMLMake(const ExpectedXML **rslt,
    const KXMLNode *Expected)
{
    rc_t rc = 0;
    const KXMLNodeset *Results = NULL;
    if (rslt == NULL) {
        return RC(rcExe, rcXmlDoc, rcConstructing, rcParam, rcNull);
    }
    *rslt = NULL;
    if (Expected == NULL) {
        return RC(rcExe, rcXmlDoc, rcConstructing, rcParam, rcNull);
    }
    rc = KXMLNodeOpenNodesetRead(Expected, &Results, "Result");
    if (rc == 0) {
        ExpectedXML *r = calloc(1, sizeof *r);
        if (r == NULL) {
            rc = RC(rcExe, rcXmlDoc, rcAllocating, rcMemory, rcExhausted);
        }
        else {
            uint32_t num = 0;
            rc = KXMLNodesetCount(Results, &num);
            if (rc == 0 && num > 0) {
                r->number_of_results = num;
                r->result = calloc(1, num * sizeof r->result[0]);
                if (r == NULL) {
                    rc = RC(rcExe, rcXmlDoc, rcAllocating,
                        rcMemory, rcExhausted);
                    free(r);
                }
                else {
                    int i = 0;
                    for (i = 0; rc == 0 && i < num; ++i) {
                        const KXMLNode *Result = NULL;
                        rc = KXMLNodesetGetNodeRead(Results, &Result, i);
                        if (rc == 0) {
                            rc = ResultXML_make(&r->result[i], Result);
                        }
                    }
                    if (rc == 0) {
                        *rslt = r;
                    }
                    else {
                        ExpectedXMLWhack(r);
                    }
                }
            }
        }
    }
    else if (GetRCState(rc) != rcNotFound) {
        LOGERR(klogErr, rc, "Cannot read Result node");
    }
    return rc;
}
static rc_t s_KXMLDocMake(const KXMLDoc **doc, const char *path)
{
    rc_t rc = 0;
    const KFile *file = NULL;
    if (doc == NULL || path == NULL) {
        return RC(rcExe, rcXmlDoc, rcConstructing, rcParam, rcNull);
    }
    if (rc == 0) {
        rc = KDirectoryOpenFileRead(s_Directory, &file, path);
        if (rc != 0) {
            PLOGERR(klogErr, (klogErr, rc, "Cannot open file $(file)", "file=%s", path));
        }
    }
    if (rc == 0) {
        rc = KXMLMgrMakeDocRead(s_XmlMgr, doc, file);
        if (rc != 0) {
            PLOGERR(klogErr, (klogErr, rc, "Cannot make XML document from $(file)",
                              "file=%s", path));
        }
    }
    KFileRelease(file);
    return rc;
}

static rc_t KXMLNodeMakeFromDoc(const KXMLNode **node, const KXMLDoc *doc,
    const char *root)
{
    rc_t rc = 0;
    const KXMLNodeset *ns = NULL;
    if (node == NULL || doc == NULL || root == NULL) {
        return RC(rcExe, rcXmlDoc, rcConstructing, rcParam, rcNull);
    }
    if (rc == 0) {
        rc = KXMLDocOpenNodesetRead(doc, &ns, root);
    }
    if (rc == 0) {
        uint32_t count = 0;
        rc = KXMLNodesetCount(ns, &count);
        if (rc == 0) {
            if (count == 0) {
                rc = RC(rcExe, rcXmlDoc, rcReading, rcNode, rcNotFound);
                PLOGERR(klogErr, (klogErr, rc, "$(path)", "path=%s", root));
            }
            else if (count > 1) {
                rc = RC(rcExe, rcXmlDoc, rcParsing, rcData, rcExcessive);
                PLOGERR(klogErr, (klogErr, rc,
                                  "too many data $(node) nodes", "node=%s", root));
            }
        }
    }
    if (rc == 0) {
        rc = KXMLNodesetGetNodeRead(ns, node, 0);
    }
    KXMLNodesetRelease(ns);
    return rc;
}

static rc_t KXMLNodeMakeFromFile(const KXMLNode **node, const char *path,
    const char *root)
{
    rc_t rc = 0;
    const KXMLDoc *doc = NULL;
    const char *rootPath = "/";
    if (node == NULL || path == NULL) {
        return RC(rcExe, rcXmlDoc, rcConstructing, rcParam, rcNull);
    }
    rc = s_KXMLDocMake(&doc, path);
    if (rc == 0) {
        if (root != NULL) {
            rootPath = root;
        }
        rc = KXMLNodeMakeFromDoc(node, doc, rootPath);
    }
    KXMLDocRelease(doc);
    return rc;
}

static
rc_t SCmdLineArgsParseRunXml(SCmdLineArgs *args)
{
    rc_t rc = 0;
    const KXMLNode *node = NULL;
    const KXMLDoc *doc = NULL;
    assert(args);
    if (args->_runXmlPath == NULL) {
        return rc;
    }
    if (rc == 0) {
        rc = s_KXMLDocMake(&doc, args->_runXmlPath);
    }
    if (rc == 0) {
        rc= KXMLNodeMakeFromDoc(&node, doc, "/RUN_SET/RUN | /RUN");
    }
    if (rc == 0) {
        rc = RunXML_Make(&args->_run, node);
    }
    KXMLNodeRelease(node);
    KXMLDocRelease(doc);
    return rc;
}

static rc_t SCmdLineArgsParseExpectedXml(SCmdLineArgs *self)
{
    rc_t rc = 0;
    const KXMLNode *node = NULL;
    assert(self);
    if (self->_expectedPath == NULL) {
        return rc;
    }
    if (rc == 0) {
        rc = KXMLNodeMakeFromFile(&node, self->_expectedPath, "//Expected");
    }
    if (rc == 0) {
        rc = ExpectedXMLMake(&self->_expected, node);
    }
    KXMLNodeRelease(node);
    return rc;
}

static rc_t SCmdLineArgsParseXmls(SCmdLineArgs *self)
{
    rc_t rc = 0;
    const KXMLDoc *doc = NULL;

    if( (rc = SCmdLineArgsParseRunXml(self)) == 0 &&
        (rc = s_KXMLDocMake(&doc, self->_experimentXmlPath)) == 0 &&
        (rc = Experiment_Make(&self->_experiment, doc, &self->_run->attributes)) == 0 ) {
        rc = SCmdLineArgsParseExpectedXml(self);
    }
    KXMLDocRelease(doc);
    return rc;
}

static rc_t SCmdLineArgsRelease(const SCmdLineArgs *cself)
{
    if (cself) {
        SCmdLineArgs *self = (SCmdLineArgs*) cself;
        RunXML_Whack(self->_run);
        Experiment_Whack(self->_experiment, true);
        free(self);
    }
    return 0;
}

static rc_t SCmdLineArgsLogResultTotal(const SCmdLineArgs *self,
    uint64_t spotCount, uint64_t baseCount)
{
    rc_t rc = 0;
    int i = 0;
    if (self == NULL) {
        return rc;
    }
    for (i = 0;
            self->_expected != NULL && i < self->_expected->number_of_results
        && rc == 0; ++i)
    {
        ResultXML *f = self->_expected->result[i];
        const char *member_name = f->member_name;
        if (member_name == NULL) {
            member_name = "";
        }
    }
    if (rc == 0) {
        PLOGMSG ( klogInfo, ( klogInfo,
            "loaded accession=$(accession): spot_count=$(spot_count), base_count=$(base_count), status=$(status)",
            "accession=%s,spot_count=%lu,base_count=%lu,status=success",
            self->_runName, spotCount, baseCount));
    }
    return rc;
}

typedef struct SInput_struct {
    struct {
        const SRALoaderFile** files;
        uint32_t count;
    } *blocks;
    uint32_t count;
} SInput;

static
void SInput_Release(const SInput* cself)
{
    if(cself) {
        SInput* self = (SInput*)cself;
        uint32_t i;
        for(i = 0; i < self->count; i++) {
            /* actual files must be released elsewhere */
            free(self->blocks[i].files);
        }
        free(self->blocks);
        free(self);
    }
}

typedef struct SInputOpen_TarVisitData_struct {
    uint32_t alloc;
    char** files;
    uint32_t count;
    
} SInputOpen_TarVisitData;

static
rc_t SInputOpen_TarVisit(const KDirectory *dir, uint32_t type, const char *name, void *data )
{
    rc_t rc = 0;

    if( type == kptFile ) {
        SInputOpen_TarVisitData* d = (SInputOpen_TarVisitData*)data;

        if( d->alloc <= d->count ) {
            void* p;
            d->alloc = d->files == NULL ? 1024 : (d->alloc * 2);
            if( (p = realloc(d->files, d->alloc * sizeof(*d->files))) == NULL ) {
                rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
            } else {
                d->files = p;
            }
        }
        if( rc == 0 ) {
            char buf[1024];
            if( (rc = KDirectoryResolvePath(dir, true, buf, sizeof(buf), name)) == 0 ) { 
                d->files[d->count++] = strdup(buf);
                if( d->files[d->count - 1] == NULL ) {
                    rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
                }
            }
        }
    }
    return rc;
}

static
rc_t SInputOpen(const SInput **cself, SCmdLineArgs *args)
{
    rc_t rc = 0;
    SInput *self = calloc(1, sizeof(*self));

    if(self == NULL) {
        rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
    } else {
        self->blocks = calloc(args->_run->datablock_count, sizeof(*(self->blocks)));
        if(self->blocks == NULL) {
            rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
        } else {
            uint32_t i;
            self->count = args->_run->datablock_count;
            for(i = 0; rc == 0 && i < self->count; i++) {
                const DataBlock* b = &args->_run->datablocks[i];
                uint32_t f, j;
                char fname[10240];

                self->blocks[i].files = calloc(b->files_count, sizeof(*(self->blocks->files)));
                if(self->blocks == NULL) {
                    rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
                    continue;
                }
                self->blocks[i].count = b->files_count;
                for(f = 0, j = 0; rc == 0 && j < b->files_count; j++ ) {
                    int fnlen = strlen(b->files[j].filename);
                    if( fnlen >= sizeof(fname) ) {
                        rc = RC(rcExe, rcStorage, rcAllocating, rcFile, rcTooLong);
                        PLOGERR(klogErr, (klogErr, rc, "$(file)", PLOG_S(file), b->files[j].filename));
                        continue;
                    }
                    strcpy(fname, b->files[j].filename);
                    /* compressed archives supposed to be decompressed beforehand */
                    if( fnlen > 7 && strcmp(&fname[fnlen - 7], ".tar.gz") == 0 ) {
                        fnlen -= 3;
                        fname[fnlen] = '\0';
                    } else if( fnlen > 4 && strcmp(&fname[fnlen - 4], ".tgz") == 0 ) {
                        fname[fnlen - 2] = 'a';
                        fname[fnlen - 1] = 'r';
                    } else if( fnlen > 7 && strcmp(&fname[fnlen - 7], ".tar.bz2") == 0 ) {
                        fnlen -= 4;
                        fname[fnlen] = '\0';
                    }
                    if( strcmp(&fname[fnlen - 4], ".tar") == 0 ) {
                        const KDirectory* tar;
                        if( (rc = KDirectoryOpenArcDirRead(s_Directory, &tar, true, fname,
                                                           tocKFile, KTocParseTAR, NULL, NULL)) == 0 ) {
                            uint32_t k;
                            void* p;
                            SInputOpen_TarVisitData data = {0, NULL, 0};

                            if( (rc = KDirectoryVisit(tar, true, SInputOpen_TarVisit, &data, ".")) != 0 ) {
                                continue;
                            }
                            if( data.count == 0 ) {
                                rc = RC(rcExe, rcStorage, rcAllocating, rcFile, rcEmpty);
                                PLOGERR(klogErr,(klogErr, rc, "Empty tar $(file)", PLOG_S(file), fname));
                                continue;
                            }
                            self->blocks[i].count += data.count - 1; /* minus tar file */
                            p = realloc(self->blocks[i].files, self->blocks[i].count * sizeof(*(self->blocks->files)));
                            if( p == NULL) {
                                rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
                                continue;
                            }
                            self->blocks[i].files = p;
                            for(k = 0; k < data.count; k++) {
                                rc = SRALoaderFile_Make(&self->blocks[i].files[f++], tar, data.files[k], b, NULL);
                            }
                            free(data.files);
                            KDirectoryRelease(tar);
                        }
                    } else {
                        char* fn = strdup(fname);
                        if( fn == NULL ) {
                            rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
                        } else {
                            rc = SRALoaderFile_Make(&self->blocks[i].files[f++], s_Directory, fn, b, b->files[j].md5_digest);
                        }
                    }
                }
                /* we really do not need this check here since tars can produce unwanted errors
                if( rc == 0 && i > 0 && self->blocks[i].count != self->blocks[i - 1].count ) {
                    rc = RC(rcExe, rcStorage, rcAllocating, rcData, rcInconsistent);
                    LOGERR(klogErr, rc, "number of files in DATA_BLOCKs");
                }*/
            }
        }
    }
    if( rc == 0 ) {
        *cself = self;
    } else {
        SInput_Release(self);
    }
    return rc;
}

typedef struct SSraDb {
    char *_tablePath;
    char *_tablePathWork;

#if USE_NCBI_SRADB
    SRAPath *_sraPath;
#endif
    SRAMgr *_mgr;
} SSraDb;
static rc_t SSraDbWhack(const SSraDb *self)
{
    if (self) {
#if USE_NCBI_SRADB
        SRAPathRelease(self->_sraPath);
#endif
        SRAMgrRelease(self->_mgr);
        free(self->_tablePath);
        free(self->_tablePathWork);
        free((void*) self);
    }
    return 0;
}

static
rc_t SSraDbMake(SSraDb **db, const SCmdLineArgs *args, const ExperimentXML* experiment)
{
    rc_t rc = 0;

    SSraDb *r = NULL;

    assert(db && args);

    r = (SSraDb*) calloc(1, sizeof *r);
    if (r == NULL) {
        rc = RC(rcExe, rcStorage, rcAllocating, rcMemory, rcExhausted);
    }

    if (rc == 0) {
#if USE_NCBI_SRADB
        if (args->_repPath) {
            char path[512];
            rc = SRAPathMake(&r->_sraPath, NULL);
            if (rc == 0) {
                rc = SRAPathClear(r->_sraPath);
            }
            if (rc == 0) {
                rc = SRAPathAddRepPath(r->_sraPath, args->_repPath);
            }
            if (rc == 0) {
                rc = SRAPathAddVolPath(r->_sraPath, args->_volPath);
            }
            if (rc == 0) {
                rc = SRAPathFull(r->_sraPath,
                    args->_repPath, args->_volPath, args->_runName,
                    path, sizeof path);
            }
            if (rc == 0) {
                r->_tablePath = (char*) malloc(strlen(path) + 1);
                strcpy(r->_tablePath, path);
            }
        }
        else
#endif
        {
            r->_tablePath = (char*) malloc(strlen(args->_runName) + 1);
            strcpy(r->_tablePath, args->_runName);
        }
    }

    if (rc == 0 && args->_outputMode == eCreateMode) {
        /* thack that r->_tablePath don't exist */
        uint32_t pathType = KDirectoryPathType(s_Directory, r->_tablePath);
        if (pathType != kptNotFound) {
            rc = RC(rcExe, rcTable, rcCreating, rcTable, rcExists);
            PLOGERR(klogErr, (klogErr, rc, "$(path)", "path=%s", r->_tablePath));
        }
    }

    if (rc == 0) {
        char nameTail[] = ".tmp";
        size_t sz = strlen(r->_tablePath) + sizeof nameTail;
        r->_tablePathWork = (char*) malloc(sz);
        if (r->_tablePathWork == NULL) {
            rc = RC(rcExe, rcString, rcAllocating, rcMemory, rcExhausted);
        }
        else {
            snprintf(r->_tablePathWork, sz, "%s%s", r->_tablePath, nameTail);
        }
    }

    if (rc == 0) {
        rc = SRAMgrMakeUpdate(&r->_mgr, 0);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot Make SRAMgr");
        }
        else {
            rc = SRAMgrSetMD5Mode(r->_mgr, true);
            if (rc != 0) {
                LOGERR(klogErr, rc, "Cannot SetMD5Mode");
            }
        }
    }

    assert(args);
    if (rc == 0) {
        uint32_t pathType
            = KDirectoryPathType(s_Directory, r->_tablePathWork);
        switch (pathType) {
            case kptNotFound:
                break;
            case kptDir:
                rc = SRAMgrDropTable(r->_mgr, true, r->_tablePathWork);
                if (rc) {
                    PLOGERR(klogErr, (klogErr, rc, "Cannot drop table $(path)",
                                      "path=%s", r->_tablePathWork));
                }
                break;
            default:
                rc = RC(rcExe, rcDirectory, rcResolving, rcType, rcUnexpected);
                PLOGERR(klogErr, (klogErr, rc, "$(path) is not a directory",
                                  "path=%s", r->_tablePathWork));
                break;
        }

        if (rc == 0) {
            char parent[512];
            char resolved[512];
            snprintf(parent, sizeof parent, "%s/..", r->_tablePathWork);
            rc = KDirectoryResolvePath(s_Directory, true,
                resolved, sizeof resolved, parent);
            if (rc == 0) {
                pathType = KDirectoryPathType(s_Directory, resolved);
                switch (pathType) {
                    case kptNotFound:
                        rc = KDirectoryCreateDir(s_Directory,
                            0775, kcmCreate | kcmParents, resolved);
                        if (rc != 0) {
                            PLOGERR(klogErr, (klogErr, rc, "Creating directory $(name)",
                                              "name=%s", resolved));
                        }
                        break;
                    case kptDir:
                        break;
                    default:
                        rc = RC(rcExe, rcDirectory, rcCreating, rcType, rcUnexpected);
                        break;
                }
            }
        }
    }
    if (rc) {
        SSraDbWhack(r);
    } else {
        *db = r;
    }
    return rc;
}

static
void MakeVersion(uint32_t vers, char *version)
{
    assert(version);
    sprintf(version, "%u.%u.%u", vers >> 24, (vers >> 16) & 0x00FF, vers & 0x0000FFFF);
}

static
rc_t SSraDbWriteSoftwareVersions(SSraDb *db, SRATable* table, const char *argv0, uint32_t frontEndVersion, const char* fe_name)
{
    rc_t rc = 0;
    KMDataNode *node = NULL;
    KMDataNode *subNode = NULL;

    assert(db);
    assert(table);

    rc = SRATableOpenMDataNodeUpdate(table, &node, "SOFTWARE");
    if (rc != 0) {
        LOGERR(klogErr, rc, "Cannot open SOFTWARE node");
    }

    if (rc == 0) {
        rc = KMDataNodeOpenNodeUpdate(node, &subNode, "loader");
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot open loader node");
        }
    }

    if (rc == 0) {
        char vers[64];
        MakeVersion(LOADER_VERS, vers);
        rc = KMDataNodeWriteAttr(subNode, "vers", vers);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot write loader/vers attr");
        }
    }

    if (rc == 0) {
        rc = KMDataNodeWriteAttr(subNode, "date", __DATE__);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot write loader/date attr");
        }
    }

    if (rc == 0) {
        const char *tool_name = strrchr(argv0, '/');
        if (tool_name++ == NULL) {
            tool_name = argv0;
        }
        rc = KMDataNodeWriteAttr(subNode, "name", tool_name);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot write loader/name attr");
        }
    }
    KMDataNodeRelease(subNode);
    if (rc == 0) {
        rc = KMDataNodeOpenNodeUpdate(node, &subNode, "formatter");
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot open formatter node");
        }
    }
    if (rc == 0) {
        char vers[64];
        MakeVersion(frontEndVersion, vers);
        rc = KMDataNodeWriteAttr(subNode, "vers", vers);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot write formatter/vers attr");
        }
    }
    if (rc == 0) {
        fe_name = fe_name ? fe_name : "internal";
        rc = KMDataNodeWriteAttr(subNode, "name", fe_name);
        if (rc != 0) {
            LOGERR(klogErr, rc, "Cannot write formatter/name attr");
        }
    }

    KMDataNodeRelease(subNode);
    KMDataNodeRelease(node);

    return rc;
}

static rc_t SSraDbFixRunNameOnAllVolumes(const SSraDb *db)
{
    rc_t rc = 0;

    uint32_t pathType = 0;

    assert(db);
    
    pathType = KDirectoryPathType(s_Directory, db->_tablePath);
    switch (pathType) {
        case kptNotFound:
            break;
        case kptDir: {
            int i = 1;
            char renamed[512];
            for (i = 1; ; ++i) {
                snprintf(renamed, sizeof renamed, "%s.%d", db->_tablePath, i);
                pathType = KDirectoryPathType(s_Directory, renamed);
                if (pathType == kptNotFound) {
                    break;
                }
            }
            rc = KDirectoryRename(s_Directory, false, db->_tablePath, renamed);
            if (rc != 0) {
                return rc;
            }
            break;
        }
        default:
            rc = RC(rcExe, rcTable, rcRenaming, rcType, rcUnexpected);
            return rc;
    }

    return rc;
}

uint32_t KAppVersion(void)
{
    return LOADER_VERS;
}

rc_t KMain(int argc, char *argv[])
{
    rc_t rc = 0, lastRc = 0;

    SCmdLineArgs *args = NULL;
    const SInput *input = NULL;
    SSraDb *db = NULL;
    SRATable* table = NULL;
    SRALoaderFmt* fe = NULL;

    SRALoaderConfig feInput;
    memset(&feInput, 0, sizeof feInput);
    feInput.columnFilter = efltrNOISE;
    feInput.spots_to_run = -1;

    /* Re-route logging to stdout and others to stderr */
    KLogHandlerSetStdOut();
    KLogLibHandlerSetStdOut();
    KOutHandlerSetStdErr();
    KStsHandlerSetStdErr();
    KStsLibHandlerSetStdErr();

    /* Set default logging level */
    LogLevelSet(defaultLogLevel);

    /* Parse command line arguments */
    if (rc == 0) {
        rc = SCmdLineArgsMake(&args, argc, argv);
    }

    /* Initialize statics */
    if (rc == 0) {
        rc = KDirectoryNativeDir(&s_Directory);
    }

    if (rc == 0) {
        rc = KXMLMgrMakeRead(&s_XmlMgr);
    }

    if( rc == 0 && (rc = SCmdLineArgsParseXmls(args)) == 0) {
        feInput.spots_to_run = args->_spots_to_run;
        feInput.experiment = args->_experiment;
        rc = SRALoaderFmtMake(&fe, &feInput);
    }

    /* Prepare front-end input */
    if (rc == 0) {
        rc = SInputOpen(&input, args);
    }

    /* Create SRAMgr, SRATable */
    if (rc == 0) {
        if( (rc = SSraDbMake(&db, args, feInput.experiment)) == 0 ) {
            feInput.sra_mgr = db->_mgr;
            feInput.table_path = db->_tablePathWork;
        }
    }

    if(rc == 0) {
        rc_t rc2 = 0;
        uint64_t spotCount = 0;
        uint64_t numBases = 0;
        uint32_t fe_vers = 0;
        const char* fe_name;

        rc = SRALoaderFmtVersion(fe, &fe_vers, &fe_name);
        if(rc == 0) {
            uint32_t i, j;
            for(i = 0; rc == 0 && i < input->count; i++) {
                rc = SRALoaderFmtWriteData(fe, input->blocks[i].count, input->blocks[i].files);
                for(j = 0; j < input->blocks[i].count; j++) {
                    bool eof = true;
                    if( rc == 0 && SRALoaderFile_IsEof(input->blocks[i].files[j], &eof) == 0 ) {
                        if( args->_spots_to_run < 0 && eof != true ) {
                            const char* fn;
                            SRALoaderFileName(input->blocks[i].files[j], &fn);
                            rc2 = RC(rcExe, rcFormatter, rcReading, rcFile, rcExcessive);
                            PLOGERR(klogErr, (klogErr, rc2, "File $(file) not completely parsed", PLOG_S(file), fn));
                        }
                    }
                    SRALoaderFile_Release(input->blocks[i].files[j]);
                }
                if( GetRCObject(rc) == rcTransfer && GetRCState(rc) == rcDone ) {
                    /* interrupted by requested spot count */
                    rc = 0;
                    break;
                }
                rc = rc ? rc : rc2;
            }
        }
        /* release object to make sure it free's it's data and commits */
        rc2 = SRALoaderFmtRelease(fe, &table);
        rc = rc ? rc : rc2;
        fe = NULL;

        if( rc == 0 && table == NULL ) {
            rc = RC(rcExe, rcTable, rcClosing, rcTable, rcNull);
        }
        /* Write versions */
        if (rc == 0) {
            rc = SSraDbWriteSoftwareVersions(db, table, argv[0], fe_vers, fe_name);
        }
        if (rc == 0) {
            rc = SRATableCommit(table);
        }
        if (rc == 0) {
            uint32_t badSpotCount = 0;

            rc = SRATableSpotCount(table, &spotCount);
            if (rc == 0) {
                /*rc = SRATableBadSpotCount(table, &badSpotCount);*/
            }
            if (rc == 0) {
                rc = SRATableBaseCount(table, &numBases);
            }
            if (rc == 0) {
                if (spotCount > 0) {
                    double percentage = ((double) badSpotCount) / spotCount;
                    double allowed = args->_spots_bad_allowed_percentage / 100.0;
                    if(percentage > allowed ) {
                        rc = RC(rcExe, rcTable, rcClosing, rcData, rcInvalid);
                        PLOGERR(klogErr, 
                                (klogErr, rc,
                                 "Too many bad spots: "
                                 "spots: $(spots), bad spots: $(bad_spots), "
                                 "bad spots percentage: $(percentage), "
                                 "allowed percentage: $(allowed)",
                                 "spots=%d, bad_spots=%d, percentage=%f, allowed=%f",
                                 spotCount, badSpotCount,
                                 percentage, allowed));
                    }
                }
            }
        }

        if (rc == 0) {
            rc = SRATableRelease(table);
            table = NULL;
            if (rc == 0) {
                rc = SRAMgrLock(db->_mgr, db->_tablePathWork);
            }
        }
        if (rc == 0) {
            rc = SSraDbFixRunNameOnAllVolumes(db);
        }
        if (rc == 0) {
            rc = KDirectoryRename(s_Directory, true, db->_tablePathWork, db->_tablePath);
            if (rc == 0) {
                strcpy(db->_tablePathWork, db->_tablePath);
            }
        }
        /* TODO Log result & total here in the old loader.
           But we should have list of files in order to do it.*/

        if (rc == 0) {
            if (spotCount == 0) {
                rc = RC(rcExe, rcFormatter, rcClosing, rcTable, rcEmpty);
                LOGERR(klogErr, rc, "0 spots were loaded after successful front-end run");
            }
            else {
                rc = SCmdLineArgsLogResultTotal(args, spotCount, numBases);
            }
        }
    }

    if( rc != 0 ) {
        if( db && table ) {
            rc_t status = SRATableRelease(table);
            table = NULL;
            if(status == 0) {
                status = SRAMgrDropTable(db->_mgr, true, db->_tablePathWork);
            }
            if(status != 0) {
                lastRc = KLogLastErrorCode();
                LOGERR(klogInt, status, "cannot DropRun after load failure");
            }
        }
    }

    /* make sure the error was reported */
    if( rc != 0 &&
        rc != RC(rcExe, rcProcess, rcExecuting, rcProcess, rcCanceled) &&
        rc != RC(rcExe, rcProcess, rcExecuting, rcTransfer, rcDone)) {
        if (lastRc == 0) {
            lastRc = KLogLastErrorCode();
        }
        if (rc != lastRc) {
            LOGERR(klogInt, rc, "no message text available");
        }
    }

    SSraDbWhack(db);
    SInput_Release(input);
    SRALoaderFmtRelease(fe, NULL);
    SCmdLineArgsRelease(args);
    KDirectoryRelease(s_Directory);
    KXMLMgrRelease(s_XmlMgr);

    return rc;
}
