/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <klib/log.h>
#include <klib/rc.h>
#include <kapp/main.h>

#include <endian.h>
#include <byteswap.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

typedef struct SRFAbsolidLoaderFmt SRFAbsolidLoaderFmt;

#define SRALOADERFMT_IMPL SRFAbsolidLoaderFmt
#include "loader-fmt.h"

#include "srf.h"
#include "ztr-absolid.h"
#include "srf-fmt.h"
#include "pstring.h"
#include "writer-absolid.h"
#include "srf-absolid.vers.h"
#include "debug.h"

typedef struct fe_context_t_struct {
    SRF_context ctx;
    bool skip_signal;
    const SRAWriteAbsolid* writer;

    pstring name_prefix;
    
    struct {
        uint8_t nreads;
        uint16_t start[ABSOLID_FMT_MAX_NUM_READS];
        pstring label[ABSOLID_FMT_MAX_NUM_READS];
        char cs_key[ABSOLID_FMT_MAX_NUM_READS];
        EAbisolidReadType type[ABSOLID_FMT_MAX_NUM_READS];
    } region;
} fe_context_t;

static
rc_t set_label_type(const char* label, pstring* name, EAbisolidReadType* type)
{
    rc_t rc = 0;

    assert(name && type);
    if( label && strcmp(label, "F3") == 0 ) {
        *type = eForward;
        rc = pstring_assign(name, label, strlen(label));
    } else if( label && strcmp(label, "R3") == 0 ) {
        *type = eReverse;
        rc = pstring_assign(name, label, strlen(label));
    } else {
        *type = eUnknown;
        rc = pstring_clear(name);
        if( label ) {
            DEBUG_MSG(3, ("LABEL ignored '%s'\n", label));
        }
    }
    return rc;
}

static
rc_t fe_new_region(fe_context_t *self, size_t region_count, const region_t region[])
{
    rc_t rc = 0;
    int i;

    self->region.nreads = region_count / 2;
    DEBUG_MSG(3, ("REGION: %u -> %u reads\n", region_count, self->region.nreads));
    if( self->region.nreads <= 0 || self->region.nreads > ABSOLID_FMT_MAX_NUM_READS ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnsupported);
        PLOGERR(klogErr, (klogErr, rc, "read count $(c)", PLOG_U8(c), self->region.nreads));
    }
    for(i = 0; rc == 0 && i < self->region.nreads; i++ ) {
        int j = i * 2 + 1;
        self->region.start[i] = region[j].start;
        rc = set_label_type(region[j].name, &self->region.label[i], &self->region.type[i]);
        self->region.cs_key[i] = region[j - 1].name[0];
        DEBUG_MSG(3, ("REGION[%u]: '%s', %u, '%c', start: %u\n",
                      i, self->region.label[i].data, self->region.type[i], self->region.cs_key[i], self->region.start[i]));
        switch(region[j].type) {
            case Biological:
            case Normal:
            case Paired:
            case Technical:
                break;
            default:
                rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnexpected);
                PLOGERR(klogErr, (klogErr, rc, "read #$(read_id) type mismatch; expected $(expected), got $(got)",
                        "read_id=%u,expected=%s,got=%u", i, "(B|N|P|T)", region[j].type));
                return rc;
        }
    }
    if( self->region.nreads > 1 && self->region.type[eForward] != eUnknown && self->region.type[eForward] == self->region.type[eReverse] ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcDuplicate);
        LOGERR(klogErr, rc, "both reads have same type");
    }
    return rc;
}

static
rc_t report_file_error(SRF_context* ctx, rc_t rc, const char* msg)
{
    uint64_t pos = 0;
    SRALoaderFile_Pos(ctx->file, &pos);
    PLOGERR(klogErr, (klogErr, rc, "file '$(file)', offset $(offset): $(msg)",
        PLOG_3(PLOG_S(file),PLOG_U64(offset),PLOG_S(msg)), ctx->file_name, pos, msg));
    return rc;
}

static
rc_t parse_v1_header(SRF_context *ctx, ZTR_Context *ztr_ctx, const uint8_t *data, size_t size)
{
    rc_t rc = 0;
    size_t parsed;
    char prefixType;
    uint32_t counter;
    ztr_raw_t ztr_raw;
    ztr_t ztr;
    enum ztr_chunk_type type;
    fe_context_t* fe = (fe_context_t*)ctx;
    
    if( (rc = SRF_ParseDataChunk(data, size, &parsed, &fe->name_prefix, &prefixType, &counter)) != 0 ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInvalid);
        return report_file_error(ctx, rc, "parse_v1_header - failed to parse SRF chunk");
    }
    DEBUG_MSG(3, ("HEADER PREFIX: '%s'\n", fe->name_prefix.data));
    if((rc = ABI_ZTR_AddToBuffer(ztr_ctx, data + parsed, size - parsed)) != 0) {
        return rc;
    }
    if((rc = ABI_ZTR_ParseHeader(ztr_ctx)) != 0) {
        return report_file_error(ctx, rc, "parse_v1_header - failed to parse ZTR header");
    }
    while(rc == 0 && !ABI_ZTR_BufferIsEmpty(ztr_ctx)) {
        if((rc = ABI_ZTR_ParseBlock(ztr_ctx, &ztr_raw)) != 0) {
            return report_file_error(ctx, rc, "parse_v1_header - failed to parse ZTR chunk");
        }
        if((rc = ABI_ZTR_ProcessBlock(ztr_ctx, &ztr_raw, &ztr, &type)) != 0) {
            report_file_error(ctx, rc, "parse_v1_header - failed to process ZTR chunk");
        }
        if(type == REGN) {
            rc = fe_new_region(fe, ztr.region->count, ztr.region->region);
        }
        if(*(void **)&ztr != NULL) {
            free(*(void **)&ztr);
        }
    }
    return rc;
}

static
void set_read_filter(uint8_t* filter, int flags)
{
    if( ( flags & SRF_READ_WITHDRAWN ) != 0 ) {
        *filter = SRA_READ_FILTER_REJECT;
    } else if( ( flags & SRF_READ_BAD ) != 0 ) {
        *filter = SRA_READ_FILTER_CRITERIA;
    } else {
        *filter = SRA_READ_FILTER_PASS;
    }
}

static
rc_t fe_new_read(fe_context_t *self, pstring *readId, EAbisolidReadType* type, pstring* label)
{
    rc_t rc = 0;
    pstring name_suffix;

    assert(self && readId && type && label);
    DEBUG_MSG(3, ("READ_LABEL: '%s'\n", readId->data));
    /* spot name suffix may end with '(F|R)3' */
    if( readId->len >= 2 ) {
        rc = set_label_type(&readId->data[readId->len - 2], label, type);
        if( rc == 0 && *type != eUnknown ) {
            /* cut label */
            readId->len -= label->len;
            /* cut trailing stuff */
            while( !isdigit(readId->data[readId->len - 1]) ) {
                readId->len--;
            }
        }
    } else {
        pstring_clear(label);
        *type = eUnknown;
    }
    if( rc == 0 && (rc = pstring_copy(&name_suffix, readId)) == 0 ) {
        rc = SRAWriteAbsolid_MakeName(&self->name_prefix, &name_suffix, readId);
    }
    return rc;
}

static
rc_t parse_v1_read(SRF_context *ctx, ZTR_Context *ztr_ctx, const uint8_t *data, size_t size)
{
    rc_t rc = 0;
    size_t i, parsed;
    ztr_raw_t ztr_raw;
    ztr_t ztr;
    enum ztr_chunk_type type;
    fe_context_t* fe = (fe_context_t*)ctx;

    uint8_t flags;
    pstring readId;
    EAbisolidReadType read_type;
    pstring label;

    AbsolidRead read[ABSOLID_FMT_MAX_NUM_READS];
        
    if( fe->region.nreads == 0 ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcNotFound);
        return report_file_error(ctx, rc, "missing region chunk before 1st read chunk");
    }
    if( (rc = SRF_ParseReadChunk(data, size, &parsed, &flags, &readId)) != 0 ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rc);
        return report_file_error(ctx, rc, "SRF parsing failure");
    }
    ABI_ZTR_AddToBuffer(ztr_ctx, data + parsed, size - parsed);

    /* readId will have spotname */
    if( (rc = fe_new_read(fe, &readId, &read_type, &label)) != 0 ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rc);
        return report_file_error(ctx, rc, "parsing suffix");
    }
    AbsolidRead_Init(&read[eForward]);
    AbsolidRead_Init(&read[eReverse]);

    while(!ABI_ZTR_BufferIsEmpty(ztr_ctx)) {
        if( (rc = ABI_ZTR_ParseBlock(ztr_ctx, &ztr_raw)) != 0 ||
            (rc = ABI_ZTR_ProcessBlock(ztr_ctx, &ztr_raw, &ztr, &type)) != 0 ) {
            report_file_error(ctx, rc, "ZTR parsing failure");
            break;
        }
        switch (type) {
        case BASE:
            if(ztr.sequence->datatype != i8) {
                rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnexpected);
                report_file_error(ctx, rc, "read: expected 8-bit datatype");
            } else if( read_type != eUnknown ) {
                if( (rc = pstring_assign(&read[read_type].seq, ztr.sequence->data, ztr.sequence->datasize)) == 0 ) {
                    /* grab 1st, may be the only cs_key */
                    read[read_type].cs_key = fe->region.cs_key[0];
                    for(i = 1; i < fe->region.nreads; i++) {
                        if( read_type == fe->region.type[i] ) {
                            read[read_type].cs_key = fe->region.cs_key[i];
                            break;
                        }
                    }
                    set_read_filter(&read[read_type].filter, flags);
                    rc = pstring_copy(&read[read_type].label, &label);
                    DEBUG_MSG(3, ("SRF READ: '%s'\n", read[read_type].seq.data));
                }
                if( rc != 0 ) {
                    report_file_error(ctx, rc, "copying read");
                }
            } else {
                for(i = 0; rc == 0 && i < fe->region.nreads; i++) {
                    EAbisolidReadType t = fe->region.type[i];
                    size_t len = (i + 1 >= fe->region.nreads ? ztr.sequence->datasize : fe->region.start[i + 1]) - fe->region.start[i];
                    rc = pstring_assign(&read[t].seq, &ztr.sequence->data[fe->region.start[i]], len);
                    read[t].cs_key = fe->region.cs_key[i];
                    set_read_filter(&read[t].filter, flags);
                    if( fe->region.label[i].len != 0 ) {
                        rc = pstring_copy(&read[t].label, &fe->region.label[i]);
                    }
                    DEBUG_MSG(3, ("SRF READ[%u]: '%s'\n", i, read[fe->region.type[i]].seq.data));
                }
                if( rc != 0 ) {
                    report_file_error(ctx, rc, "copying reads");
                }
            }
            break;
        case CNF1:
            if(ztr.quality1->datatype != i8) {
                rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnexpected);
                report_file_error(ctx, rc, "quality: expected 8-bit datatype");
            } else if( read_type != eUnknown ) {
                if( (rc = pstring_assign(&read[read_type].qual, ztr.quality1->data, ztr.quality1->datasize)) == 0 ) {
                    DEBUG_MSG(3, ("SRF QUAL: %u bytes\n", read[read_type].qual.len));
                }
                if( rc != 0 ) {
                    report_file_error(ctx, rc, "copying quality");
                }
            } else {
                for(i = 0; rc == 0 && i < fe->region.nreads; i++) {
                    EAbisolidReadType t = fe->region.type[i];
                    size_t len = (i + 1 >= fe->region.nreads ? ztr.quality1->datasize : fe->region.start[i + 1]) - fe->region.start[i];
                    rc = pstring_assign(&read[t].qual, &ztr.quality1->data[fe->region.start[i]], len);
                    DEBUG_MSG(3, ("SRF QUAL[%u]: %u bytes\n", i, read[fe->region.type[i]].qual.len));
                }
                if( rc != 0 ) {
                    report_file_error(ctx, rc, "copying qualities");
                }
            }
            break;
        case SAMP:
            if( !fe->skip_signal ) {
                size_t i;
                int stype = ABSOLID_FMT_COLMASK_NOTSET;
                if(ztr.signal->datatype != f32) {
                    rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnexpected);
                    report_file_error(ctx, rc, "signal: expected 32-bit float datatype");
                } else if( (ztr.signal->datasize % sizeof(float)) != 0 ) {
                    rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInvalid);
                    report_file_error(ctx, rc, "signal: size not 32-bit float aligned");
                } else if (ztr.signal->channel == NULL) {
                    rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcIncomplete);
                    report_file_error(ctx, rc, "SIGNAL column: missing channel type");
                } else if(strcmp(ztr.signal->channel, "0FAM") == 0) {
                    stype = ABSOLID_FMT_COLMASK_FAM;
                } else if(strcmp(ztr.signal->channel, "1CY3") == 0) {
                    stype = ABSOLID_FMT_COLMASK_CY3;
                } else if(strcmp(ztr.signal->channel, "2TXR") == 0) {
                    stype = ABSOLID_FMT_COLMASK_TXR;
                } else if(strcmp(ztr.signal->channel, "3CY5") == 0) {
                    stype = ABSOLID_FMT_COLMASK_CY5;
                } else {
                    rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcUnexpected);
                    report_file_error(ctx, rc, "SIGNAL column: unexpected channel type");
                }
#if __BYTE_ORDER == __LITTLE_ENDIAN
                for(i = 0; rc == 0 && i < ztr.signal->datasize; i += 4) {
                    uint32_t* r = (uint32_t*)&ztr.signal->data[i];
                    *r = bswap_32(*r);
                }
#endif
                if( rc == 0 ) {
                    if( read_type != eUnknown ) {
                        pstring* d;
                        switch(stype) {
                            case ABSOLID_FMT_COLMASK_FAM:
                                read[read_type].fs_type = eSignalTypeFAM;
                                d = &read[read_type].fxx;
                                break;
                            case ABSOLID_FMT_COLMASK_CY3:
                                d = &read[read_type].cy3;
                               break;
                            case ABSOLID_FMT_COLMASK_TXR:
                                d = &read[read_type].txr;
                                break;
                            case ABSOLID_FMT_COLMASK_CY5:
                                d = &read[read_type].cy5;
                                break;
                        }
                        if( (rc = pstring_assign(d, ztr.signal->data, ztr.signal->datasize)) == 0 ) {
                            DEBUG_MSG(3, ("SRF SIGNAL[%s]: %u bytes\n", ztr.signal->channel, d->len));
                        }
                        if( rc != 0 ) {
                            report_file_error(ctx, rc, "copying signal");
                        }
                    } else {
                        for(i = 0; rc == 0 && i < fe->region.nreads; i++) {
                            pstring* d;
                            EAbisolidReadType t = fe->region.type[i];
                            size_t len = (i + 1 >= fe->region.nreads) ? ztr.signal->datasize : (fe->region.start[i + 1] * sizeof(float));
                            len -= fe->region.start[i] * sizeof(float);
                            switch(stype) {
                                case ABSOLID_FMT_COLMASK_FAM:
                                    read[t].fs_type = eSignalTypeFAM;
                                    d = &read[t].fxx;
                                    break;
                                case ABSOLID_FMT_COLMASK_CY3:
                                    d = &read[t].cy3;
                                   break;
                                case ABSOLID_FMT_COLMASK_TXR:
                                    d = &read[t].txr;
                                    break;
                                case ABSOLID_FMT_COLMASK_CY5:
                                    d = &read[t].cy5;
                                    break;
                            }
                            rc = pstring_assign(d, &ztr.signal->data[fe->region.start[i] * sizeof(float)], len);
                            DEBUG_MSG(3, ("SRF SIGNAL[%s]: %u bytes\n", ztr.signal->channel, d->len));
                        }
                        if( rc != 0 ) {
                            report_file_error(ctx, rc, "copying signals");
                        }
                    }
                }
            }
            break;
        default:
            break;
        }
        if(type != none && type != ignore) {
            free(*(void **)&ztr);
        }
    }
    if(rc == 0) {
        if(flags != 0) {
            DEBUG_MSG(5, ("SRF_FLAGS='%02X'\n", flags));
        }
        switch(read_type) {
            case eForward:
                rc = SRAWriteAbsolid_Write(fe->writer, ctx->file, &readId, NULL, &read[eForward], NULL);
                break;
            case eReverse:
                rc = SRAWriteAbsolid_Write(fe->writer, ctx->file, &readId, NULL, NULL, &read[eReverse]);
                break;
            case eUnknown:
                rc = SRAWriteAbsolid_Write(fe->writer, ctx->file, &readId, NULL, &read[eForward], &read[eReverse]);
                break;
        }
    }
    return rc;
}

struct SRFAbsolidLoaderFmt {
    SRALoaderFmt dad;
    fe_context_t fe;
};

static
rc_t SRFAbsolidLoaderFmt_WriteData(SRFAbsolidLoaderFmt *self, uint32_t argc, const SRALoaderFile *const argv [])
{
    rc_t rc = 0;
    uint32_t i;
    
    for(i = 0; rc == 0 && i < argc; i++) {
        self->fe.ctx.file = argv[i];
        if( (rc = SRALoaderFileName(argv[i], &self->fe.ctx.file_name)) == 0 ) {
            rc = SRF_parse(&self->fe.ctx, parse_v1_header, parse_v1_read, ABI_ZTR_CreateContext, ABI_ZTR_ContextRelease);
        }
    }
    return rc;
}

static
rc_t SRFAbsolidLoaderFmt_Whack(SRFAbsolidLoaderFmt* self, SRATable** table)
{
    SRAWriteAbsolid_Whack(self->fe.writer, table);
    free(self);
    return 0;
}

static
rc_t SRFAbsolidLoaderFmt_Version( const SRFAbsolidLoaderFmt* self, uint32_t* vers, const char** name )
{
    *vers = SRF_ABSOLID_VERS;
    *name = "AB SOLiD SRF";
    return 0;
}

static SRALoaderFmt_vt_v1 SRFAbsolidLoaderFmt_vt_v1 =
{
    1, 0,
    SRFAbsolidLoaderFmt_Whack,
    SRFAbsolidLoaderFmt_Version,
    SRFAbsolidLoaderFmt_WriteData
};

rc_t SRFABSolidLoaderFmt_Make(SRALoaderFmt** self, const SRALoaderConfig* config)
{
    rc_t rc = 0;
    SRFAbsolidLoaderFmt* fmt;

    if( self == NULL || config == NULL ) {
        return RC(rcSRA, rcFormatter, rcConstructing, rcParam, rcNull);
    }

    fmt = calloc(1, sizeof(*fmt));
    if(fmt == NULL) {
        rc = RC(rcSRA, rcFormatter, rcConstructing, rcMemory, rcExhausted);
        LOGERR(klogInt, rc, "failed to initialize; out of memory");
        return rc;
    }
    if( (rc = SRALoaderFmtInit(&fmt->dad, (const SRALoaderFmt_vt*)&SRFAbsolidLoaderFmt_vt_v1)) != 0 ) {
        LOGERR(klogInt, rc, "failed to initialize parent object");
    } else if( (rc = SRAWriteAbsolid_Make(&fmt->fe.writer, config)) != 0 ) {
        LOGERR(klogInt, rc, "failed to initialize writer");
    }
    if( rc == 0 ) {
        fmt->fe.skip_signal = (config->columnFilter & efltrSIGNAL);
        *self = &fmt->dad;
    } else {
        free(fmt);
    }
    return rc;
}
