/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <klib/log.h>
#include <klib/rc.h>
#include <sra/wsradb.h>
#include <sra/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "debug.h"
#include "writer-absolid.h"
#include "sra-writer.h"

void AbsolidRead_Init(AbsolidRead* read)
{
    if( read ) {
        pstring_clear(&read->label);
        read->cs_key = 0;
        pstring_clear(&read->seq);
        pstring_clear(&read->qual);
        read->fs_type = eSignalNotSet;
        pstring_clear(&read->fxx);
        pstring_clear(&read->cy3);
        pstring_clear(&read->txr);
        pstring_clear(&read->cy5);
        read->filter = 0;
    }
}

struct SRAWriteAbsolid {
    SRAWriter* base;
    const PlatformXML* platform;
    uint8_t nreads;
    SRASegment read_seg[ABSOLID_FMT_MAX_NUM_READS];
    uint8_t read_id[ABSOLID_FMT_MAX_NUM_READS];
    AbsolidRead dummy[ABSOLID_FMT_MAX_NUM_READS];

    uint32_t ci_name;
    uint32_t ci_cs_key;
    uint32_t ci_label;
    uint32_t ci_label_start;
    uint32_t ci_label_len;
    uint32_t ci_csread;
    uint32_t ci_qual;
    uint32_t ci_fxx;
    uint32_t ci_cy3;
    uint32_t ci_txr;
    uint32_t ci_cy5;
    uint32_t ci_read_filter;
};

static
rc_t fill_dummy(AbsolidRead* read, EAbisolidReadType type, uint16_t len)
{
    rc_t rc = 0;

    if( (rc = pstring_clear(&read->seq)) == 0 &&
        (rc = pstring_append_chr(&read->seq, '.', len)) == 0 &&
        (rc = pstring_clear(&read->qual)) == 0 &&
        (rc = pstring_append_chr(&read->qual, 1, len)) == 0 &&
        (rc = pstring_clear(&read->label)) == 0 &&
        (rc = pstring_clear(&read->fxx)) == 0 &&
        (rc = pstring_append_chr(&read->fxx, 0, len * sizeof(float))) == 0 &&
        (rc = pstring_clear(&read->cy3)) == 0 &&
        (rc = pstring_append_chr(&read->cy3, 0, len * sizeof(float))) == 0 &&
        (rc = pstring_clear(&read->txr)) == 0 &&
        (rc = pstring_append_chr(&read->txr, 0, len * sizeof(float))) == 0 &&
        (rc = pstring_clear(&read->cy5)) == 0 &&
        (rc = pstring_append_chr(&read->cy5, 0, len * sizeof(float))) == 0 ) {
        read->cs_key = type == eForward ? 'T' : 'G';
        /* needs to be reset per spot */
        read->fs_type = eSignalNotSet;
        read->filter = SRA_READ_FILTER_REJECT;
    }
    return rc;
}

rc_t SRAWriteAbsolid_Make(const SRAWriteAbsolid** cself, const SRALoaderConfig* config)
{
    rc_t rc = 0;
    SRAWriteAbsolid* self;
    const PlatformXML* platform;
    const ReadSpecXML_read* read;
    EAbisolidReadType read_type[ABSOLID_FMT_MAX_NUM_READS];
    uint8_t nreads;

    if( cself == NULL || config == NULL ) {
        return RC(rcSRA, rcFormatter, rcConstructing, rcParam, rcNull);
    }
    if( (rc = Experiment_GetPlatform(config->experiment, &platform)) != 0 ||
        (rc = Experiment_GetReadNumber(config->experiment, &nreads)) != 0 ) {
        return rc;
    }
    if( platform->id != SRA_PLATFORM_ABSOLID ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInvalid);
        LOGERR(klogErr, rc, "platform type");
        return rc;
    }
    if( nreads > ABSOLID_FMT_MAX_NUM_READS ) {
        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcUnsupported);
        PLOGERR(klogErr, (klogErr, rc, "expected up to $(max) reads", "max=%u", ABSOLID_FMT_MAX_NUM_READS));
        return rc;
    }
    self = calloc(1, sizeof(*self));
    if( self == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcConstructing, rcMemory, rcExhausted);
        return rc;
    }
    if( (rc = SRAWriter_Make(&self->base, config)) != 0 ) {
        LOGERR(klogInt, rc, "failed to initialize base writer");
    } else if( (rc = Experiment_GetRead(config->experiment, 0, &read)) != 0 ) {
        LOGERR(klogInt, rc, "failed to get read 1 descriptor");
    } else if( read->coord_type != rdsp_RelativeOrder_ct && read->coord_type != rdsp_BaseCoord_ct ) {
        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInvalid);
        LOGERR(klogErr, rc, "1st read can be BASE_COORD or RELATIVE_ORDER only");
    } else if( read->read_type == rdsp_Forward_rt ) {
        read_type[0] = eForward;
        self->read_id[eForward] = 0;
    } else if( read->read_type == rdsp_Reverse_rt ) {
        read_type[0] = eReverse;
        self->read_id[eReverse] = 0;
        LOGMSG(klogWarn, "expected 1st read to be of type 'Forward'");
    } else {
        rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcUnrecognized);
        LOGMSG(klogErr, "type of 1st read");
    }
    self->read_seg[0].start = 0;
    self->read_seg[0].len = platform->param.absolid.sequence_length;
    if( rc == 0 && nreads >= ABSOLID_FMT_MAX_NUM_READS ) {
        if( (rc = Experiment_GetRead(config->experiment, 1, &read)) != 0 ) {
            LOGERR(klogInt, rc, "failed to get read 2 descriptor");
        } else if( read->coord_type != rdsp_BaseCoord_ct ) {
            rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcInvalid);
            LOGERR(klogErr, rc, "2nd read can be BASE_COORD only");
        } else if( read->read_type == rdsp_Forward_rt ) {
            if( read_type[0] == eForward ) {
                rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcDuplicate);
                LOGMSG(klogErr, "2 reads of type 'Forward'");
            } else {
                LOGMSG(klogWarn, "expected 2nd read to be of type 'Reverse'");
                read_type[1] = eForward;
                self->read_id[eForward] = 1;
            }
        } else if( read->read_type == rdsp_Reverse_rt ) {
            if( read_type[0] == eReverse ) {
                rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcDuplicate);
                LOGMSG(klogErr, "2 reads of type 'Reverse'");
            } else {
                read_type[1] = eReverse;
                self->read_id[eReverse] = 1;
            }
        } else {
            rc = RC(rcSRA, rcFormatter, rcConstructing, rcData, rcUnrecognized);
            LOGMSG(klogErr, "type of 2nd read");
        }
        self->read_seg[1].start = read->coord.start_coord - 1;
        self->read_seg[1].len = platform->param.absolid.sequence_length - self->read_seg[1].start;
        self->read_seg[0].len = self->read_seg[1].start;
        rc = fill_dummy(&self->dummy[read_type[1]], read_type[1], self->read_seg[1].len);
    }
    if( rc == 0 ) {
        rc = fill_dummy(&self->dummy[read_type[0]], read_type[0], self->read_seg[0].len);
    }
#if _DEBUGGING
    DEBUG_MSG(3, ("1. READ_SEG[%hu,%hu]\n", self->read_seg[0].start, self->read_seg[0].len));
    if( nreads > 1 ) {
        DEBUG_MSG(3, ("2. READ_SEG[%hu,%hu]\n", self->read_seg[1].start, self->read_seg[1].len));
    }
#endif
    if( rc == 0 ) {
        self->platform = platform;
        self->nreads = nreads;
        *cself = self;
    } else {
        SRAWriteAbsolid_Whack(self, NULL);
    }
    return rc;
}

rc_t SRAWriteAbsolid_MakeName(const pstring* prefix, const pstring* suffix, pstring* name)
{
    rc_t rc = 0;
    if( prefix == NULL || name == NULL ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcParam, rcNull);
    } else if( (rc = pstring_copy(name, prefix)) == 0 ) {
        if( suffix && suffix->len > 0 ) {
            if( name->data[name->len - 1] != '_' && suffix->data[0] != '_' ) {
                rc = pstring_append(name, "_", 1);
            }
            if( rc == 0 ) {
                pstring_concat(name, suffix);
            }
        }
    }
    if( rc != 0 ) {
        LOGERR(klogErr, rc, "preparing spot name");
    }
    return rc;
}

void SRAWriteAbsolid_Whack(const SRAWriteAbsolid* cself, SRATable** table)
{
    if( cself != NULL ) {
        SRAWriteAbsolid* self = (SRAWriteAbsolid*)cself;
        SRAWriter_Whack(self->base, table);
        free(self);
    }
}

rc_t SRAWriteAbsolid_Write(const SRAWriteAbsolid* cself, const SRALoaderFile* file, 
                           const pstring* spot_name, const pstring* spot_group,
                           const AbsolidRead* fwd, const AbsolidRead* rev)
{
    rc_t rc = 0;
    uint32_t i, nreads = ((fwd ? 1 : 0) + (rev ? 1 : 0));
    uint8_t fid = cself->read_id[eForward];
    uint8_t rid = cself->read_id[eReverse];

    struct {
        pstring label;
        INSDC_coord_zero label_start[ABSOLID_FMT_MAX_NUM_READS];
        uint32_t label_len[ABSOLID_FMT_MAX_NUM_READS];
        char cs_key[ABSOLID_FMT_MAX_NUM_READS];
        pstring seq;
        pstring qual;
        EAbisolidFSignalType fs_type;
        pstring fxx;
        pstring cy3;
        pstring txr;
        pstring cy5;
        uint8_t filter[ABSOLID_FMT_MAX_NUM_READS];
    } spot;

    assert(cself != NULL);
    assert(file != NULL);
    assert(spot_name);

    DEBUG_MSG(3, ("spot name '%s', group '%s'\n", spot_name->data, spot_group ? spot_group->data : NULL));

    if( nreads == 0 || nreads > cself->nreads ) {
        rc = RC(rcSRA, rcFormatter, rcReading, rcData, rcExcessive);
        PLOGERR(klogErr, (klogErr, rc, "read count do not match experiment in spot $(n): $(c)",
                "n=%s,c=%u", spot_name->data, nreads));
        return rc;
    }
    /* each read length must match experiment segmentation, unless it's the a whole spot */
    if( fwd && fwd->seq.len != cself->read_seg[fid].len ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInvalid);
        PLOGERR(klogErr, (klogErr, rc, "Forward length do not match experiment in spot '$(n)': $(l)",
                 "n=%s,l=%u", spot_name->data, fwd->seq.len));
        return rc;
    }
    if( rev && rev->seq.len != cself->read_seg[rid].len ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInvalid);
        PLOGERR(klogErr, (klogErr, rc, "Reverse length do not match experiment in spot '$(n)': $(l)",
                 "n=%s,l=%u", spot_name->data, rev->seq.len));
        return rc;
    }
    if( cself->nreads == 2 ) {
        /* check for missing pair */
        if( rev == NULL ) {
            LOGMSG(klogInfo, "adding missing Reverse read");
            rev = &cself->dummy[eReverse];
            ((AbsolidRead*)rev)->filter = fwd->filter;
            ((AbsolidRead*)rev)->fs_type = fwd->fs_type;
        } else if( fwd == NULL ) {
            LOGMSG(klogInfo, "adding missing Forward read");
            fwd = &cself->dummy[eForward];
            ((AbsolidRead*)fwd)->filter = rev->filter;
            ((AbsolidRead*)fwd)->fs_type = rev->fs_type;
        }
    }
    if( fid > rid ) {
        /* reorder as per experiment xml */
        const AbsolidRead* x = fwd;
        fwd = rev;
        rev = x;
    }
    /* concat fwd + rev data */
    if( fwd != NULL ) {
        if( (rc = pstring_copy(&spot.seq, &fwd->seq)) == 0 &&
            (rc = pstring_copy(&spot.qual, &fwd->qual)) == 0 &&
            (rc = pstring_copy(&spot.label, &fwd->label)) == 0 ) {
            spot.label_start[eForward] = 0;
            spot.label_len[eForward] = fwd->label.len;
            spot.cs_key[eForward] = fwd->cs_key;
            spot.filter[eForward] = fwd->filter;
            if( fwd->fs_type != eSignalNotSet ) { 
                if( (rc = pstring_copy(&spot.fxx, &fwd->fxx)) == 0 &&
                    (rc = pstring_copy(&spot.cy3, &fwd->cy3)) == 0 &&
                    (rc = pstring_copy(&spot.txr, &fwd->txr)) == 0 &&
                    (rc = pstring_copy(&spot.cy5, &fwd->cy5)) == 0 ) {
                    spot.fs_type = fwd->fs_type;
                }
            } else {
                pstring_clear(&spot.fxx);
                pstring_clear(&spot.cy3);
                pstring_clear(&spot.txr);
                pstring_clear(&spot.cy5);
                spot.fs_type = eSignalNotSet;
            }
        }
    } else {
        /* this is needed to actualy init spot */
        pstring_clear(&spot.seq);
        pstring_clear(&spot.qual);
        pstring_clear(&spot.label);
        pstring_clear(&spot.fxx);
        pstring_clear(&spot.cy3);
        pstring_clear(&spot.txr);
        pstring_clear(&spot.cy5);
        spot.fs_type = eSignalNotSet;
    }
    if(rev != NULL ) {
        int x = fwd != NULL ? eReverse : eForward;
        if( (rc = pstring_concat(&spot.seq, &rev->seq)) == 0 &&
            (rc = pstring_concat(&spot.qual, &rev->qual)) == 0 &&
            (rc = pstring_concat(&spot.label, &rev->label)) == 0 ) {
            spot.label_start[x] =  x == eReverse ? spot.label_len[eForward] : 0;
            spot.label_len[x] = rev->label.len;
            spot.cs_key[x] = rev->cs_key;
            spot.filter[x] = rev->filter;
            if( rev->fs_type != eSignalNotSet ) {
                if( (rc = pstring_concat(&spot.fxx, &rev->fxx)) == 0 &&
                    (rc = pstring_concat(&spot.cy3, &rev->cy3)) == 0 &&
                    (rc = pstring_concat(&spot.txr, &rev->txr)) == 0 &&
                    (rc = pstring_concat(&spot.cy5, &rev->cy5)) == 0 ) {
                    if( fwd == NULL ) {
                        spot.fs_type = rev->fs_type;
                    }
                }
            }
        }
    }
    
#if _DEBUGGING
    {{
        char qbuf[40960], q1[512];

        DEBUG_MSG(13, ("READ[%u]: %s\n", spot.seq.len, spot.seq.data));
        qbuf[0] = '\0';
        for(i = 0; i < spot.qual.len; i++ ) {
            sprintf(q1, "%i ", spot.qual.data[i]);
            strcat(qbuf, q1);
        }
        DEBUG_MSG(13, ("QUAL[%u]: %s\n", spot.qual.len, qbuf));
        for(i = 0; i < cself->nreads; i++) {
            DEBUG_MSG(13, ("READ_FLT[%u]: %s\n", i, spot.filter[i] == SRA_READ_FILTER_PASS ? "PASS" : "REJECT"));
        }
        if( spot.fs_type != eSignalNotSet ) {
            float* f;
            qbuf[0] = '\0';
            f = (float*)spot.fxx.data;
            for(i = 0; i < spot.fxx.len / sizeof(*f); i++) {
                snprintf(q1, sizeof(q1) - 1, "%f ", f[i]);
                strcat(qbuf, q1);
            }
            DEBUG_MSG(13, ("FXX[%u]: %s\n", spot.fxx.len, qbuf));
            qbuf[0] = '\0';
            f = (float*)spot.cy3.data;
            for(i = 0; i < spot.cy3.len / sizeof(*f); i++) {
                snprintf(q1, sizeof(q1) - 1, "%f ", f[i]);
                strcat(qbuf, q1);
            }
            DEBUG_MSG(13, ("CY3[%u]: %s\n", spot.cy3.len, qbuf));
            qbuf[0] = '\0';
            f = (float*)spot.txr.data;
            for(i = 0; i < spot.txr.len / sizeof(*f); i++) {
                snprintf(q1, sizeof(q1) - 1, "%f ", f[i]);
                strcat(qbuf, q1);
            }
            DEBUG_MSG(13, ("TXR[%u]: %s\n", spot.txr.len, qbuf));
            qbuf[0] = '\0';
            f = (float*)spot.cy5.data;
            for(i = 0; i < spot.cy5.len / sizeof(*f); i++) {
                snprintf(q1, sizeof(q1) - 1, "%f ", f[i]);
                strcat(qbuf, q1);
            }
            DEBUG_MSG(13, ("CY5[%u]: %s\n", spot.cy5.len, qbuf));
        }
    }}
#endif
    if( spot.qual.len != spot.seq.len ) {
        rc = RC(rcSRA, rcFormatter, rcWriting, rcData, rcInconsistent);
        LOGERR(klogErr, rc, "quality");
    }
    if( fwd != NULL && rev != NULL && fwd->fs_type != rev->fs_type ) {
        rc = RC(rcSRA, rcFormatter, rcParsing, rcData, rcInconsistent);
        LOGERR(klogErr, rc, "signal type mismatch FAM/FTC");
    } else if( spot.fs_type != eSignalNotSet ) {
        if( spot.fxx.len / spot.seq.len != sizeof(float) ) {
            rc = RC(rcSRA, rcFormatter, rcWriting, rcData, rcInconsistent);
            LOGERR(klogErr, rc, (spot.fs_type == eSignalTypeFTC) ? "signal FTC" :  "signal FAM");
        }
        if( spot.cy3.len / spot.seq.len != sizeof(float) ) {
            rc = RC(rcSRA, rcFormatter, rcWriting, rcData, rcInconsistent);
            LOGERR(klogErr, rc, "signal CY3");
        }
        if( spot.txr.len / spot.seq.len != sizeof(float) ) {
            rc = RC(rcSRA, rcFormatter, rcWriting, rcData, rcInconsistent);
            LOGERR(klogErr, rc, "signal TXR");
        }
        if( spot.cy5.len / spot.seq.len != sizeof(float) ) {
            rc = RC(rcSRA, rcFormatter, rcWriting, rcData, rcInconsistent);
            LOGERR(klogErr, rc, "signal CY5");
        }
    }
    if( rc != 0 ) {
        return rc;
    }
    if( cself->ci_name == 0 ) {
        SRAWriteAbsolid* self =(SRAWriteAbsolid*)cself;
        if( (rc = SRAWriter_CreateTable(cself->base, "NCBI:SRA:ABI:tbl:v2")) != 0 ) {
            return rc;
        }
        if( (rc = SRAWriter_WriteDefaults(cself->base)) != 0 ) {
            LOGERR(klogInt, rc, "failed to write table defaults");
            return rc;
        }
        COL_OPEN(self->ci_name, "NAME", vdb_ascii_t);
        COL_OPEN(self->ci_cs_key, "CS_KEY", insdc_fasta_t);
        COL_OPEN(self->ci_csread, "CSREAD", insdc_csfasta_t);
        COL_OPEN(self->ci_qual, "QUALITY", insdc_phred_t);
        COL_OPEN(self->ci_read_filter, "READ_FILTER", sra_read_filter_t);

        if( spot.label.len != 0 ) {
            COL_OPEN(self->ci_label, "LABEL", vdb_ascii_t);
            COL_OPEN(self->ci_label_start, "LABEL_START", "INSDC:coord:zero");
            COL_OPEN(self->ci_label_len, "LABEL_LEN", "U32");
        }
        if( spot.fs_type != eSignalNotSet ) {
            if( spot.fs_type == eSignalTypeFTC ) {
                COL_OPEN(self->ci_fxx, "FTC", ncbi_fsamp1_t);
            } else {
                COL_OPEN(self->ci_fxx, "FAM", ncbi_fsamp1_t);
            }
            COL_OPEN(self->ci_cy3, "CY3", ncbi_fsamp1_t);
            COL_OPEN(self->ci_txr, "TXR", ncbi_fsamp1_t);
            COL_OPEN(self->ci_cy5, "CY5", ncbi_fsamp1_t);
        }
    }

    if( rc != 0 || (rc = SRAWriter_NewSpot(cself->base, NULL)) != 0 ) {
        return rc;
    }
    
    /* this column is the one that is most likely to fail as it is required to be unique
     * so we put it first so that we can still close the spot if it fails
     */
    COL_WRITE(cself->ci_name, "NAME", (*spot_name));
    COL_WRITE(cself->ci_csread, "CSREAD", spot.seq);
    COL_WRITE(cself->ci_qual, "QUALITY", spot.qual);
    if( cself->ci_fxx != 0 ) {
        if( spot.fs_type == eSignalTypeFTC ) {
            COL_WRITE(cself->ci_fxx, "FTC", spot.fxx);
        } else {
            COL_WRITE(cself->ci_fxx, "FAM", spot.fxx);
        }
        COL_WRITE(cself->ci_cy3, "CY3", spot.cy3);
        COL_WRITE(cself->ci_txr, "TXR", spot.txr);
        COL_WRITE(cself->ci_cy5, "CY5", spot.cy5);
    }
    if( (rc = SRAWriter_WriteSpotSegmentSimple(cself->base, file, spot_group ? spot_group->data : NULL, cself->read_seg)) != 0 ) {
        PLOGERR(klogErr, (klogErr, rc, "failed to write $(column) column", "column=SPOT_GROUP/READ_SEG"));
        return rc;
    }
    if( cself->ci_label != 0 ) {
        COL_WRITE(cself->ci_label, "LABEL", spot.label);
        rc = SRAWriter_WriteIdxColumn(cself->base, cself->ci_label_start, spot.label_start, nreads * sizeof(spot.label_start[0]), false);
        if(rc) {
            PLOGERR(klogErr, (klogErr, rc, "failed to write '$(column)' column", "column=LABEL_START"));
            return rc;
        }
        rc = SRAWriter_WriteIdxColumn(cself->base, cself->ci_label_len, spot.label_len, nreads * sizeof(spot.label_len[0]), false);
        if(rc) {
            PLOGERR(klogErr, (klogErr, rc, "failed to write '$(column)' column", "column=LABEL_LEN"));
            return rc;
        }
    }

    rc = SRAWriter_WriteIdxColumn(cself->base, cself->ci_cs_key, spot.cs_key, nreads * sizeof(spot.cs_key[0]), false);
    if(rc) {
        PLOGERR(klogErr, (klogErr, rc, "failed to write '$(column)' column", "column=CS_KEY"));
        return rc;
    }
    rc = SRAWriter_WriteIdxColumn(cself->base, cself->ci_read_filter, spot.filter, nreads * sizeof(spot.filter[0]), false);
    if(rc) {
        PLOGERR(klogInt, (klogInt, rc, "failed to write $(column) column", "column=READ_FILTER"));
        return rc;
    }
    return SRAWriter_CloseSpot(cself->base);
}
