/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include "sra-stat.vers.h"

#include <sra/sradb.h>
#include <sra/srapath.h>
#include <vdb/manager.h>
#include <sra/types.h>
#include <kapp/main.h>
#include <klib/log.h>
#include <klib/out.h>
#include <klib/container.h>
#include <klib/rc.h>
#include <fmtdef.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

typedef struct srastat_parms srastat_parms;
struct srastat_parms
{
    const char *src_spec;
    const char *src_path;

    bool xml; /* output format (txt or xml) */

    const char **columns;
    uint32_t column_cnt;

    spotid_t start, stop;

    bool hasSPOT_GROUP;
};

typedef struct SraStats {
    BSTNode n;
    char     spot_group[1024]; /* SPOT_GROUP Column */
    uint64_t spot_count; /*** spot count ***/
    uint64_t spot_count_m; /*** spots with mates ***/
    uint64_t blen; /** biological len **/
    uint64_t blen_m; /** biological len when mates are present**/
    uint64_t tlen; /** total len **/
    uint64_t bad_spot_count; /** number of spots flagged with rd_filter=2 **/
    uint64_t bad_blen;        /** biological length of bad spots ***/
    uint64_t filtered_spot_count; /** number of spots flagged with rd_filter=2 **/ 
    uint64_t filtered_blen;	  /** biological length of filtered spots **/
}SraStats;

static
void CC srastat_print ( BSTNode *n, void *data )
{
   const srastat_parms *pb = (const srastat_parms*) data;
   const SraStats *ss = ( const SraStats* ) n;
   assert(pb && ss);
   if (pb->xml) {
    if (pb->hasSPOT_GROUP) {
        OUTMSG(("  <Member member_name=\"%s\"", ss->spot_group));
    }
    OUTMSG((" spot_count=\"%ld\" base_count=\"%ld\"", ss->spot_count, ss->tlen));
    OUTMSG((" base_count_bio=\"%ld\"", ss->blen));
    OUTMSG((" spot_count_mates=\"%ld\" base_count_bio_mates=\"%ld\"", ss->spot_count_m, ss->blen_m));
    OUTMSG((" spot_count_bad=\"%ld\" base_count_bio_bad=\"%ld\"", ss->bad_spot_count, ss->bad_blen));
    OUTMSG((" spot_count_filtered=\"%ld\" base_count_bio_filtered=\"%ld\"", ss->filtered_spot_count, ss->filtered_blen));
    if (pb->hasSPOT_GROUP) {
        OUTMSG(("/>\n"));
    }
   }
   else {
       OUTMSG(("%s|%s|%ld:%ld:%ld|%ld:%ld|%ld:%ld|%ld:%ld\n",
           pb->src_spec,ss->spot_group,ss->spot_count,ss->tlen,ss->blen,ss->spot_count_m,ss->blen_m,
               ss->bad_spot_count,ss->bad_blen,ss->filtered_spot_count,ss->filtered_blen));
   }
}


static
void CC bst_whack_free ( BSTNode *n, void *ignore )
{
    free ( n );
}

static
int CC srastats_cmp ( const void *item, const BSTNode *n )
{
    const char *sg = item;
    const SraStats *ss = ( const SraStats* ) n;

    return strcmp(sg,ss->spot_group);
}

static 
int CC srastats_sort ( const BSTNode *item, const BSTNode *n )
{
    const SraStats *ss = ( const SraStats* ) item;
    return srastats_cmp(ss->spot_group,n);
}





static
rc_t sra_stat ( srastat_parms *pb, const SRATable *tbl )
{
    rc_t	rc=0;
    const SRAColumn *len_col;
    const SRAColumn *rtype_col;
    const SRAColumn *sgrp_col=NULL;
    const SRAColumn *rdf_col=NULL;
    if(   (rc = SRATableOpenColumnRead ( tbl, &len_col, "READ_LEN",vdb_uint32_t))==0
       && (rc = SRATableOpenColumnRead ( tbl, &rtype_col, "READ_TYPE",sra_read_type_t))==0)
    {

        rc = SRATableOpenColumnRead ( tbl, &sgrp_col, "SPOT_GROUP",vdb_ascii_t);
        if(rc==0 || GetRCState(rc)==rcNotFound)
        {
		rc = SRATableOpenColumnRead ( tbl, &rdf_col, "RD_FILTER",sra_read_filter_t);
		if(rc==0 || GetRCState(rc)==rcNotFound)
		{

		    spotid_t spotid;
		    pb -> hasSPOT_GROUP =  0;
		    rc = SRATableMaxSpotId ( tbl, &spotid );
		    if ( rc != 0 )
			LOGERR ( klogInt, rc, "failed to read max spot id" );
		    else
		    {
			BSTree tr;

			BSTreeInit(&tr);

			if ( pb -> start == 0 ) pb -> start = 1;
			if ( pb -> stop == 0 || pb -> stop > spotid ) pb -> stop = spotid;

			for ( spotid = pb -> start; spotid <= pb -> stop && (rc==0); ++ spotid ){
			    SraStats	*ss;
			    uint32_t	len[255];
			    uint8_t	rtype[255];
			    uint8_t	rdf[255];
			    char	sgrp[1024]="NULL";

			    const void *base;
			    bitsz_t boff, row_bits;
			    rc = SRAColumnRead ( len_col, spotid, &base, &boff, & row_bits );
			    if(rc==0){
				int nreads;
				assert((boff&7)==0);
				assert((row_bits>>3)  <= sizeof(len));
				memcpy(len,((const char*)base) + (boff>>3),row_bits>>3);
				nreads=(row_bits>>3)/sizeof(*len);

				rc = SRAColumnRead ( rtype_col, spotid, &base, &boff, & row_bits );
				if(rc==0){
				    int i,blen,bcount,bad_cnt,filt_cnt;
				    assert((boff&7)==0);
				    assert((row_bits>>3)  <= sizeof(rtype));
				    assert((row_bits>>3)==nreads);
				    memcpy(rtype,((const char*)base) + (boff>>3),row_bits>>3);
				    if(sgrp_col){
					rc = SRAColumnRead ( sgrp_col, spotid, &base, & boff, & row_bits );
					if(rc==0){
					    if(row_bits > 0) {
						pb -> hasSPOT_GROUP = 1;
						assert((boff&7)==0);
						assert(row_bits>>3  <= sizeof(sgrp)-1);	
						memcpy(sgrp,((const char*)base) + (boff>>3),row_bits>>3);
						sgrp[row_bits>>3]='\0';
					   } else {
						sgrp[0]='\0';
					   }
					} else {
					    break;
					}
				    }
				    if(rdf_col){
					rc = SRAColumnRead ( rdf_col, spotid, &base, & boff, & row_bits );
					if(rc==0){
					    assert((boff&7)==0);
					    assert(row_bits>>3  <= sizeof(rdf));	
					    memcpy(rdf,((const char*)base) + (boff>>3),row_bits>>3);
					} else {
					    break;
					}
				    }
				   
				
				    ss=(SraStats*)BSTreeFind( &tr,sgrp, srastats_cmp );
				    if(ss==NULL){
					ss=malloc(sizeof(*ss));
					memset(ss,0,sizeof(*ss));
					strcpy(ss->spot_group,sgrp);
					BSTreeInsert( &tr, (BSTNode*)ss, srastats_sort );
				    }
				    ss->spot_count++;
				    for(blen=bcount=i=bad_cnt=filt_cnt=0;i<nreads;i++){
					if(len[i] > 0){
					ss->tlen+=len[i];
					    if(rtype[i]==SRA_READ_TYPE_BIOLOGICAL){
						blen+=len[i];
						bcount++;
						if(rdf_col){
							switch(rdf[i]){
							 case 2:
								ss->bad_blen+=len[i];	
								bad_cnt++;
								break;
							 case 3:
								ss->filtered_blen+=len[i];
								filt_cnt++;
								break;
							 default:
								break;
							}
						}
					    }
					}
				    }
				    ss->blen+=blen;
				    if(bcount > 1){
					ss->spot_count_m++;
					ss->blen_m+=blen;
				    }
				    if(bad_cnt){
					ss->bad_spot_count++;
				    }
				    if(filt_cnt){
					ss->filtered_spot_count++;
				    }
				}
			    }
			}
	/******* Output Results here ************/
			if (pb->xml) {
			    OUTMSG(("<Run accession=\"%s\"", pb->src_spec));
			    if (pb->hasSPOT_GROUP) {
				OUTMSG((">\n"));
			    }
			}
			BSTreeForEach( &tr, false, srastat_print, pb );
			if (pb->xml) {
			    if (!pb->hasSPOT_GROUP) {
				OUTMSG((">"));
			    }
			    OUTMSG(("</Run>\n"));
			}
			BSTreeWhack( &tr, bst_whack_free, NULL );
		    }
		    if(sgrp_col) SRAColumnRelease ( sgrp_col );
		}
		if(rdf_col) SRAColumnRelease ( rdf_col );
	}
        SRAColumnRelease ( len_col );
        SRAColumnRelease ( rtype_col );
    }
    return rc;
}

static
rc_t run ( srastat_parms *pb )
{
    const SRAMgr *mgr;
    rc_t rc = SRAMgrMakeRead ( & mgr );
    if ( rc != 0 )
        LOGERR ( klogInt, rc, "failed to open SRAMgr" );
    else
    {
        SRAPath *pmgr;
        char buffer [ 4096 ];

        /* make path manager */
        rc = SRAPathMake ( & pmgr, NULL );
        if ( rc != 0 )
        {
            if ( GetRCState ( rc ) != rcNotFound || GetRCTarget ( rc ) != rcDylib )
                LOGERR ( klogInt, rc, "failed to open SRAPath" );
            else
                rc = 0;
        }

        /* test path for being an accession */
        else if ( ! SRAPathTest ( pmgr, pb -> src_path ) && strchr ( pb -> src_path, '/' ) == NULL )
        {
            rc = SRAPathFind ( pmgr, pb -> src_path, buffer, sizeof buffer );
            if ( rc == 0 )
                pb -> src_path = buffer;
        }

        if ( rc == 0 )
        {
            const SRATable *tbl;
            rc = SRAMgrOpenTableRead ( mgr, & tbl, pb -> src_path );
            if ( rc != 0 )
                PLOGERR ( klogInt, (klogInt, rc, "failed to open SRATable '$(spec)'", "spec=%s", pb -> src_spec ));
            else
            {
                rc = sra_stat ( pb, tbl );
                SRATableRelease ( tbl );
            }
        }

        SRAPathRelease ( pmgr );
        SRAMgrRelease ( mgr );
    }

    return rc;
}

/* Version  EXTERN
 *  return 4-part version code: 0xMMmmrrrr, where
 *      MM = major release
 *      mm = minor release
 *    rrrr = bug-fix release
 */
ver_t CC KAppVersion ( void )
{
    return SRA_STAT_VERS;
}


/* Usage
 */
static
KLogLevel default_log_level;

rc_t CC Usage (const Args * args)
{
    const char * progname = "sra-dump";
    char * pc;
    size_t rem;
    size_t used;
    char logbuffer [1024];
    KLogLevel level;
    rc_t rc;

    ArgsArgvValue (args, 0, &progname);

    KOutMsg ("\n"
             "Usage: %s [ options ] table [ column-spec ... ]\n"
             "\n%s"
             "    options:\n"
             "      -h                      give tool help\n"
             "     --help\n"
             "\n"
             "     --xml                    output in XML (default is ASCII)\n"
             "     --start ID               starting spot id ( default 1 )\n"
             "     --stop ID                ending spot id ( default max )\n"
             "\n"
             "      -l, --level             Log level: %u-%u or one of \n",
             progname,
             "    display column statistics.\n",
             klogLevelMin, klogLevelMax);
    

    pc = logbuffer;
    rem = sizeof logbuffer;

    for (level = klogLevelMin; level <= klogLevelMax; ++level)
    {
        static const char div[] = " | ";
        rc = KLogLevelExplain (level, pc, rem, &used);
        if (rc || (used < 0))
        {
            used = -1;
            break;
        }
        pc += used;
        rem -= used;
        memcpy (pc, div, 3);
        pc += 3;
        rem -= 3;
    }
    pc -= 3;
    rem += 3;
    if (used >= 0)
    {
        *pc = '\0';

        KOutMsg("                               %s",logbuffer);
    }

    rc = KLogLevelExplain (default_log_level, logbuffer, sizeof logbuffer, &used);
    if (rc == 0)
    {
        logbuffer[used] = '\0';
        KOutMsg ("                               ( default %s )",logbuffer);
    }
    KOutMsg ("\n"
             "\n"
             "    column-spec:\n"
             "      NAME                     simple column name\n"
             "      (typedecl)NAME           specifically typed column name\n"
             "\n");
    return rc;
}


rc_t MiniUsage (const Args * args)
{
    const char * progname = "sra-dump";

    ArgsArgvValue (args, 0, &progname);

    KOutMsg ( "\n"
              "Usage: %s [ options ] table [ column-spec... ]\n"
              "    run with option '--help' for help\n",
              progname);
    return 0;
}

rc_t CC Version ( const Args * args )
{
    const char * progname = "sra-stat";
    rc_t rc;
    ver_t version;

    rc = ArgsArgvValue (args, 0, &progname);
    version = KAppVersion();

    KOutMsg ( "%s: %V\n", progname, version);

    return 0;
}


#define OPTION_XML   "xml"
#define OPTION_START "start"
#define OPTION_STOP  "stop"
static
OptDef Options[] = 
{
    { OPTION_XML,   "", 0, false, false, NULL, "output as XML (default is text)" },
    { OPTION_START, "", 1, true,  false, NULL, "starting spot id ( default 1 )" },
    { OPTION_STOP,  "", 1, true,  false, NULL, "ending spot id ( default max )" }
};

/* KMain - EXTERN
 *  executable entrypoint "main" is implemented by
 *  an OS-specific wrapper that takes care of establishing
 *  signal handlers, logging, etc.
 *
 *  in turn, OS-specific "main" will invoke "KMain" as
 *  platform independent main entrypoint.
 *
 *  "argc" [ IN ] - the number of textual parameters in "argv"
 *  should never be < 0, but has been left as a signed int
 *  for reasons of tradition.
 *
 *  "argv" [ IN ] - array of NUL terminated strings expected
 *  to be in the shell-native character set: ASCII or UTF-8
 *  element 0 is expected to be executable identity or path.
 */
rc_t CC KMain ( int argc, char *argv [] )
{
    Args * args;
    rc_t rc;

    default_log_level = KLogLevelGet();

    rc = ArgsMakeAndHandle (&args, argc, argv, 1, Options, sizeof Options / sizeof (OptDef));
    if (rc == 0)
    {
        do
        {
            uint32_t pcount;
            const char * pc;
            srastat_parms pb;

            memset (&pb, 0, sizeof pb);

            rc = ArgsOptionCount (args, OPTION_START, &pcount);
            if (rc)
                break;

            if (pcount == 1)
            {
                rc = ArgsOptionValue (args, OPTION_START, 0, &pc);
                if (rc)
                    break;

                pb.start = AsciiToU32 (pc, NULL, NULL);
            }

            rc = ArgsOptionCount (args, OPTION_STOP, &pcount);
            if (rc)
                break;

            if (pcount == 1)
            {
                rc = ArgsOptionValue (args, OPTION_STOP, 0, &pc);
                if (rc)
                    break;

                pb.stop = AsciiToU32 (pc, NULL, NULL);
            }

            rc = ArgsOptionCount (args, OPTION_XML, &pcount);
            if (rc)
                break;

            if (pcount)
                pb.xml = true;

            rc = ArgsParamCount (args, &pcount);
            if (rc)
                break;

            if (pcount == 0)
            {
                KOutMsg ("missing source table\n");
                rc = MiniUsage (args);
                break;
            }

            rc = ArgsParamValue (args, 0, &pb.src_path);
            if (rc)
                break;

            pb.src_spec = pb.src_path;

            pb.column_cnt = pcount - 1; /* pcount is used below */

            if (pb.column_cnt) {
                pb.columns = malloc (pb.column_cnt * sizeof (const char *));
                if (pb.columns == NULL)
                    rc = RC (rcExe, rcArgv, rcParsing, rcMemory, rcExhausted);
            }
            else
                pb.columns = NULL;
            {
                uint32_t ix;
                const char ** ppc;

                ppc = pb.columns;
                for (ix = 1; ix < pcount; ix++)
                {
                    rc = ArgsParamValue (args, ix, ppc);
                    if (rc)
                        break;
                    ppc++;
                }

                if (rc == 0)
                {
                    rc = run (&pb);
                }

                if ( pcount != 1 )
                    free ( pb.columns );
            }
        } while (0);
    }
    return rc;
}
