/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <kapp/main.h>
#include <klib/log.h>
#include <klib/rc.h>
#include <klib/namelist.h>
#include <klib/container.h>
#include <kdb/database.h>
#include <kdb/table.h>
#include <kdb/namelist.h>
#include <kdb/meta.h>
#include <kfs/directory.h>
#include <kfs/file.h>
#include <kfs/md5.h>
#include <kfs/toc.h>

#include <vdb/table.h>
#include <vdb/cursor.h>
#include <vdb/schema.h>
#include <vdb/types.h>
#include <vdb/vdb-priv.h>

#include <sra/types.h>
#include <sra/sradb.h>

#include "sra-dbcc.vers.h"

typedef struct SColumn_struct {
    enum SRAPlatforms platform_id;
    const char* name;
    uint32_t v_idx;
    bitsz_t elem_sz;
    const KColumn* kcol;
    int64_t c_from, c_to;
} SColumn;

static
rc_t ValidateSKeyIndex(const SRATable* tbl)
{
    rc_t rc_ret = 0, rc = 0;
    const SRAColumn* col = NULL;
    spotid_t spots = 0;

    LOGMSG(klogInfo, "verifying skey index");
    if( (rc = SRATableMaxSpotId(tbl, &spots)) != 0 ) {
        LOGERR(klogErr, rc, "spot count");
    } else if( (rc = SRATableOpenColumnRead(tbl, &col, "NAME",  NULL)) != 0 ) {
        LOGERR(klogErr, rc, "NAME");
    } else {
        const char* nm = NULL;
        bitsz_t off = 0, sz = 0;
        spotid_t spot = 0;
        for(spot = 1; spot <= spots; spot++) {
            if( (spot % 1000) == 0 ) {
                PLOGMSG((klogDebug5, "spot $(s) is $(n)", PLOG_2(PLOG_U32(s),PLOG_S(n)), spot, nm));
            }
            if( (rc = SRAColumnRead(col, spot, (const void**)(&nm), &off, &sz)) != 0 ) {
                if( GetRCState(rc) != rcNotFound ) {
                    PLOGERR((klogErr, rc, "spot $(s)", PLOG_U32(s), spot));
                    rc_ret = rc;
                } else {
                    rc = 0;
                }
            } else {
                spotid_t id = 0;
                if( (rc = SRATableGetSpotId(tbl, &id, nm)) != 0 ) {
                    PLOGERR((klogErr, rc, "GetSpotId $(s): $(n)", PLOG_2(PLOG_U32(s),PLOG_S(n)), spot, nm));
                    rc_ret = rc;
                } else if(spot != id) {
                    rc = RC(rcExe, rcIndex, rcValidating, rcId, rcInconsistent);
                    PLOGERR((klogErr, rc, "spot $(n): $(s) <-> $(i)", PLOG_3(PLOG_S(n),PLOG_U32(s),PLOG_U32(i)), nm, spot, id));
                    rc_ret = rc;
                }
            }
            if( rc == 0 && (rc = Quitting()) != 0 ) {
                rc_ret = rc;
                break;
            }
        }
    }
    if( col != NULL ) {
        rc_ret = SRAColumnRelease(col);
    }
    LOGERR(rc_ret ? klogErr : klogInfo, rc, "verifying skey index");
    return rc_ret;
}

static
rc_t ValidateColumns(uint32_t col_qty, SColumn* cols)
{
    rc_t rc_ret = 0, rc = 0;

    uint32_t col = 0;
    for(col = 0; col < col_qty; col++) {
        int64_t b_from = 0;
        int64_t b_to = 0;
        uint32_t b_count = 0;
        const KColumnBlob* blob = NULL;
        
        while( b_to + 1 <= cols[col].c_to ) {
            if( (rc = KColumnOpenBlobRead(cols[col].kcol, &blob, b_to + 1)) != 0 ) {
                rc_ret = rc;
                PLOGERR((klogErr, rc, "$(c) spot $(s)", PLOG_2(PLOG_S(c),PLOG_I64(s)), cols[col].name, b_to + 1));
                break;
            }
            else if( (rc = KColumnBlobIdRange(blob, &b_from, &b_count)) != 0 ) {
                rc_ret = rc;
                PLOGERR((klogErr, rc, "$(c) spot $(s)", PLOG_2(PLOG_S(c),PLOG_I64(s)), cols[col].name, b_to + 1));
                break;
            }
            b_to = b_from + b_count;
            if( (rc = KColumnBlobValidate(blob)) != 0 ) {
                rc_ret = rc;
                PLOGERR((klogErr, rc, "$(c) $(a):$(b)", PLOG_3(PLOG_S(c),PLOG_I64(a),PLOG_I64(b)),
                         cols[col].name, b_from, b_to));
            }
            if( (rc = KColumnBlobRelease(blob)) != 0 ) {
                rc_ret = rc;
                PLOGERR((klogErr, rc, "$(c) $(a):$(b)", PLOG_3(PLOG_S(c),PLOG_I64(a),PLOG_I64(b)),
                         cols[col].name, b_from, b_to));
            }
            if( rc == 0 && (rc = Quitting()) != 0 ) {
                return rc;
            }
        }
        PLOGMSG((klogDebug5, "$(c) max id $(b)", PLOG_2(PLOG_S(c),PLOG_I64(b)), cols[col].name, b_to));
    }
    return rc_ret;
}

static
rc_t CheckTable(const SRATable* tbl, enum SRAPlatforms platform)
{
    rc_t rc = 0;
    const KTable* ktbl = NULL;
    const VTable* vtbl = NULL;
    char* mand_cols_454[] = {"+READ", "+QUALITY", "?SIGNAL", "?POSITION", NULL};
    char* mand_cols_ILL[] = {"+READ", "+QUALITY", "?SIGNAL", "?NOISE", "?INTENSITY", NULL};
    char* mand_cols_ABS[] = {"+CSREAD", "+QUALITY", "?SIGNAL", NULL};
    char** mand_cols = NULL;
    int mand_cols_found[5] = {0, 0, 0, 0, 0};

    switch(platform) {
        case SRA_PLATFORM_454:
            mand_cols = mand_cols_454;
            break;
        case SRA_PLATFORM_ILLUMINA:
            mand_cols = mand_cols_ILL;
            break;
        case SRA_PLATFORM_ABSOLID:
            mand_cols = mand_cols_ABS;
            break;
        default:
            rc = RC(rcExe, rcAttr, rcReading, rcName, rcUnrecognized);
            PLOGERR((klogErr, rc, "platform $(platform_id)", PLOG_U16(platform_id), platform));
            break;
    }
    rc = SRATableGetVTableRead(tbl, &vtbl);
    if ( rc )
        LOGERR(klogErr, rc, "SRATableGetVTableRead");
    else
        rc = VTableOpenKTableRead(vtbl, &ktbl);
    if (rc)
        LOGERR(klogErr, rc, "VTableGetKTableRead");
    else {
        const VCursor *curs;
        
        rc = VTableCreateCursorRead(vtbl, &curs);
        if (rc == 0) {
            struct KNamelist* col_names = NULL;

            rc = KTableListCol(ktbl, &col_names);
            if( rc != 0 ) {
                LOGERR(klogErr, rc, "KTableListCol");
            } else {
                uint32_t col_qty = 0;

                rc = KNamelistCount(col_names, &col_qty);
                if(rc != 0) {
                    LOGERR(klogErr, rc, "KNamelistCount");
                } else {
                    SColumn* cols = calloc(col_qty, sizeof(SColumn));
                    if( cols == NULL ) {
                        rc = RC(rcExe, rcBuffer, rcAllocating, rcMemory, rcExhausted);
                        LOGERR(klogErr, rc, "column list");
                    } else {
                        uint32_t i = 0;
                        VTypedecl tdec;
                        VTypedesc tdesc;
                        uint64_t c_count;

                        /* open */
                        for(i = 0; i < col_qty; i++) {
                            int k, mcol = 0;
                            if( (rc = KNamelistGet(col_names, i, &cols[i].name)) != 0) {
                                PLOGERR((klogErr, rc, "Column #$(c)", PLOG_U32(c), i));
                                continue;
                            }
                            cols[i].platform_id = platform;
                            for(k = 0; rc == 0 && mand_cols[k] != NULL; k++) {
                                if( strcmp(cols[i].name, &mand_cols[k][1]) == 0 ) {
                                    mcol = k + 1;
                                    switch(mand_cols[k][0]) {
                                        case '?':
                                        case '+':
                                            if( mand_cols_found[k] != 0 ) {
                                                rc = RC(rcExe, rcColumn, rcOpening, rcName, rcDuplicate);
                                                PLOGERR((klogErr, rc, "$(c)", PLOG_S(c), cols[i].name));
                                            } else {
                                                mand_cols_found[k] = 1;
                                            }
                                            break;
                                        default:
                                            rc = RC(rcExe, rcColumn, rcValidating, rcId, rcCorrupt);
                                            PLOGERR((klogErr, rc, "$(c)", PLOG_S(c), mand_cols[k]));
                                            break;
                                    }
                                    break;
                                }
                            }
                            if( rc != 0 ) {
                                continue;
                            }
                            if( (rc = KTableOpenColumnRead(ktbl, &cols[i].kcol, cols[i].name)) != 0 ) {
                                PLOGERR((klogErr, rc, "KColumn $(c)", PLOG_S(c), cols[i].name));
                                continue;
                            }
                            if( (rc = KColumnIdRange(cols[i].kcol, &cols[i].c_from, &c_count)) != 0 ) {
                                PLOGERR((klogErr, rc, "KColumn $(c)", PLOG_S(c), cols[i].name));
                                continue;
                            }
                            cols[i].c_to = cols[i].c_from + c_count;
                            rc = VCursorAddColumn(curs, &cols[i].v_idx, cols[i].name);
                            if( rc != 0 ) {
                                if( mcol > 0 && mand_cols[mcol - 1][0] == '+' ) {
                                    PLOGERR((klogErr, rc, "VColumn $(c)", PLOG_S(c), cols[i].name));
                                }
                                continue;
                            }
                            rc = VCursorDatatype(curs, cols[i].v_idx, &tdec, &tdesc);
                            if( rc != 0 ) {
                                PLOGERR((klogErr, rc, "VColumn $(c)", PLOG_S(c), cols[i].name));
                                continue;
                            }
                            cols[i].elem_sz = VTypedescSizeof(&tdesc);
                        }
                        for(i = 0; i < col_qty; i++) {
                            PLOGMSG((klogInfo, "$(n). $(c) $(a):$(b), V element size $(d)",
                                PLOG_5(PLOG_I32(n),PLOG_S(c),PLOG_I64(a),PLOG_I64(b),PLOG_U32(d)),
                                i + 1, cols[i].name, cols[i].c_from, cols[i].c_to, cols[i].elem_sz));
                            if( cols[i].elem_sz <= 0 ) {
                                rc = RC(rcExe, rcColumn, rcValidating, rcNoObj, rcInvalid);
                                PLOGERR((cols[i].v_idx ? klogErr : klogWarn, rc,
                                    "$(c) element size $(d)", PLOG_2(PLOG_S(c),PLOG_U32(d)), cols[i].name, cols[i].elem_sz));
                            }
                        }
                        LOGMSG(klogInfo, "verifying blobs");
                        rc = ValidateColumns(col_qty, cols);
                        LOGERR(rc ? klogErr : klogInfo, rc, "verifying blobs");

                        /* release */
                        for(i = 0; i < col_qty; i++) {
                            if( cols[i].kcol != NULL ) {
                                KColumnRelease(cols[i].kcol);
                            }
                        }
                        free(cols);
                    }
                }
                KNamelistRelease(col_names);
            }
            VCursorRelease(curs);
        }
        else {
            LOGERR(klogErr, rc, "VTableCreateCursorRead");
        }
    }
    if( vtbl != NULL ) {
        VTableRelease(vtbl);
    }
    if( ktbl != NULL ) {
        KTableRelease(ktbl);
    }
    return rc;
}

typedef struct CheckMD5Node_struct {
    BSTNode dad;
    char file[4096];
    bool md5;
    uint8_t digest[16];
} CheckMD5Node;

typedef struct CheckMD5_Data_struct {
    BSTree files;
    bool md5;
    uint8_t digest[16];
} CheckMD5_Data;

static
int CheckMD5Node_Cmp( const BSTNode *item, const BSTNode *n )
{
    return strcmp(((CheckMD5Node*)item)->file, ((CheckMD5Node*)n)->file);
}

static
rc_t CheckMD5Visitor(const KDirectory *dir, uint32_t type, const char *name, void *data)
{
    rc_t rc = 0;
    CheckMD5_Data* d = (CheckMD5_Data*)data;

    if(type == kptFile ) {
        if( strcmp(name, "md5") == 0 || strcmp(&name[strlen(name) - 4], ".md5") == 0 ) {
            const KFile* kfmd5;
            if( (rc = KDirectoryOpenFileRead(dir, &kfmd5, name)) == 0 ) {
                const KMD5SumFmt *md5;
                if( (rc = KMD5SumFmtMakeRead(&md5, kfmd5)) == 0 ) {
                    uint32_t count = 0;
                    if( (rc = KMD5SumFmtCount(md5, &count)) == 0 ) {
                        char buf[4096];
                        bool bin;
                        do {
                            if( (rc = KMD5SumFmtGet(md5, --count, buf, sizeof(buf) - 1, d->digest, &bin)) == 0 ) {
                                PLOGMSG((klogDebug4, "md5 for $(nm) ", PLOG_S(nm), buf));
                                d->md5 = true;
                                rc = CheckMD5Visitor(dir, kptFile, buf, data);
                                d->md5 = false;
                            }
                        } while( rc == 0 && count > 0 );
                    }
                    rc = KMD5SumFmtRelease(md5);
                } else {
                    KFileRelease(kfmd5);
                }
            }
        } else {
            CheckMD5Node* node = calloc(1, sizeof(*node));
            if( (rc = KDirectoryResolvePath(dir, true, node->file, sizeof(node->file) - 1, name)) == 0 ) {
                BSTNode* existing = NULL;
                if( (rc = BSTreeInsertUnique(&d->files, &node->dad, &existing, CheckMD5Node_Cmp)) == 0 ) {
                    PLOGMSG((klogDebug4, "adding $(nm) ", PLOG_S(nm), node->file));
                    if( d->md5 ) {
                        node->md5 = true;
                        memcpy(node->digest, d->digest, sizeof(d->digest));
                    }
                } else if( GetRCState(rc) == rcExists ) {
                    PLOGMSG((klogDebug4, "updating $(nm) ", PLOG_S(nm), node->file));
                    if( d->md5 ) {
                        ((CheckMD5Node*)existing)->md5 = true;
                        memcpy(((CheckMD5Node*)existing)->digest, d->digest, sizeof(d->digest));
                    }
                    free(node);
                    rc = 0;
                }
                if( rc != 0 ) {
                    free(node);
                }
            }
        }
    } else if( (type != kptDir) && !(type & kptAlias) ) {
        rc = RC(rcExe, rcTable, rcVisiting, rcDirEntry, rcInvalid);
        PLOGERR((klogInfo, rc, "$(nm)", PLOG_S(nm), name));
    }
    return rc;
}

typedef struct CheckMD5_WalkData_struct {
    const KDirectory* dir;
    rc_t rc;
} CheckMD5_WalkData;

static
bool CheckMD5_Walk( BSTNode *n, void *data )
{
    rc_t rc = 0;
    CheckMD5Node* m = (CheckMD5Node*)n;
    CheckMD5_WalkData* d = (CheckMD5_WalkData*)data;

    if( m->md5 ) {
        const KFile* kf;
        PLOGMSG((klogDebug4, "verifying $(nm) ", PLOG_S(nm), m->file));
        if( (rc = KDirectoryOpenFileRead(d->dir, &kf, m->file)) == 0 ) {
            const KFile* kfmd5;
            if( (rc = KFileMakeMD5Read(&kfmd5, kf, m->digest)) == 0 ) {
                char buf[10240];
                size_t nr;
                if( (rc = KFileRead(kfmd5, ~0, buf, sizeof(buf), &nr)) != 0 ) {
                    PLOGERR((klogErr, rc, "MD5 check failed for file $(name)", PLOG_S(name), m->file));
                }
                KFileRelease(kfmd5);
            } else {
                KFileRelease(kf);
            }
        }
    } else if( strcmp(&m->file[strlen(m->file) - 5 ], "/lock") != 0 &&
               strcmp(&m->file[strlen(m->file) - 7 ], "/sealed") != 0) {
        rc = RC(rcExe, rcTable, rcValidating, rcConstraint, rcInconsistent);
        PLOGERR((klogErr, rc, "missing MD5 for file $(name)", PLOG_S(name), m->file));
    }
    d->rc = rc ? rc : d->rc;
    if( d->rc == 0 && (d->rc = Quitting()) != 0 ) {
        return true;
    }
    return false;
}

static
rc_t CheckMD5(const char* path)
{
    rc_t rc = 0;
    const KDirectory* dir = NULL;


    PLOGMSG((klogInfo, "verifying MD5 for '$(path)'", PLOG_S(path), path));
    if( (rc = KDirectoryNativeDir((KDirectory**)&dir)) == 0 ) {
        uint32_t type = KDirectoryPathType(dir, path) & ~kptAlias;

        if( type == kptFile ) {
            const KDirectory* adir = NULL;
            if( (rc = KDirectoryOpenArcDirRead(dir, &adir, false, path, tocKFile, KTocParseSRA, NULL, NULL)) == 0 ) {
                KDirectoryRelease(dir);
                dir = adir;
                path = ".";
            }
        } else if( type != kptDir ) {
            rc = RC(rcExe, rcTable, rcValidating, rcArc, rcUnknown);
        }
        if( rc == 0 ) {
            CheckMD5_Data data;
            BSTreeInit(&data.files);
            data.md5 = false;
            if( (rc = KDirectoryVisit(dir, true, CheckMD5Visitor, &data, path)) == 0 ) {
                CheckMD5_WalkData wd;
                wd.dir = dir;
                wd.rc = 0;
                BSTreeDoUntil(&data.files, false, CheckMD5_Walk, &wd);
                rc = wd.rc;
                LOGERR(rc ? klogErr : klogInfo, rc, "verifying MD5");
            }
        }
        KDirectoryRelease(dir);
    }
    return rc;
}


static char const* const defaultLogLevel = 
#if _DEBUGGING
        "debug5";
#else
        "info";
#endif

/*******************************************************************************
* Usage
*******************************************************************************/
static void Usage(const char *progName)
{
    const char* p = strrchr(progName, '/');
    p = p ? p + 1 : progName;
    printf("\nUsage: %s [options] table\n\n", p);
    printf(
        "    -5, --md5        Check components md5s if present, fail otherwise, unless other checks are requested (default: yes)\n"
        "    -b, --blob-crc   Check blobs CRC32 (default: no)\n"
        "    -i, --index      Check 'skey' index (default: no)\n\n"
        "    -l, --log-level  Log level: 0-13 or fatal|sys|int|err|warn|info|debug[1-10]. Default: %s\n"
        "    -h, --help       This help\n"
        "    -v, --version    Program version\n"
        "\n", defaultLogLevel);
}

uint32_t KAppVersion(void)
{
    return DBCC_VERS;
}

const char* GetArg(char const* const abbr, char const* const lng, int* i, int argc, char *argv[])
{
    size_t la = strlen(abbr);
    char* arg = argv[*i];
    if( strncmp(arg, abbr, la) == 0 || strcmp(arg, lng) == 0 ) {
        const char* a = arg + (arg[1] == '-' ? strlen(arg) : la) - 1;
        PLOGMSG((klogDebug9, "GetArg key: '$(a)'",  "a=%s", arg));
        if( strcmp(abbr, "-l") == 0 ) {
            NextLogLevel(&a, i, argc, argv, NULL, NULL);
            a = abbr;
        } else {
            a = NextArg(&a, i, argc, argv, NULL, NULL);
        }
        PLOGMSG((klogDebug9, "GetArg val: '$(a)'",  "a=%s", a));
        return a;
    }
    return NULL;
}

rc_t KMain ( int argc, char *argv [] )
{
    rc_t rc = 0;
    int i;
    const char* arg;
    const char* src_path = NULL;
    bool md5_chk = true, md5_chk_explicit = false;
    bool blob_crc = false;
    bool index_chk = false;



#if _DEBUGGING
    KLogLevelSet(klogDebug5);
#else
    KLogLevelSet(klogInfo);
#endif
    
    for(i = 1; rc == 0 && i < argc; i++) {
        if( argv[i][0] != '-' ) {
            src_path = argv[i];
            break;
        } else if((arg = GetArg("-l", "--log-level", &i, argc, argv)) != NULL) {
        } else if(strcmp(argv[i], "-5") == 0 || strcmp(argv[i], "--md5") == 0 ) {
            md5_chk_explicit = true;
        } else if(strcmp(argv[i], "-b") == 0 || strcmp(argv[i], "--blob-crc") == 0 ) {
            blob_crc = true;
        } else if(strcmp(argv[i], "-i") == 0 || strcmp(argv[i], "--index") == 0 ) {
            index_chk = true;
        } else if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0 ) {
            Usage(argv[0]);
            return 0;
        } else if(strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--version") == 0 ) {
            printf("Version: %u.%u.%u\n", KAppVersion() >> 24, (KAppVersion() >> 16) & 0xFF, KAppVersion() & 0xFFFF);
            return 0;
        } else {
            Usage(argv[0]);
            rc = RC(rcExe, rcArgv, rcReading, rcParam, rcUnknown);
        }
    }
    if(i != argc - 1) {
        Usage(argv[0]);
        rc = RC(rcExe, rcArgv, rcReading, rcParam, rcUnexpected);
    }

    if( rc != 0 ) {
    }
    else if( src_path == NULL ) {
        rc = RC(rcExe, rcArgv, rcReading, rcParam, rcNotFound);
    }
    else {
        if( blob_crc || index_chk ) {
            /* if blob or index is requested, md5 is off unless explicitly requested */
            md5_chk = md5_chk_explicit;
        }
        if( md5_chk ) {
            rc = CheckMD5(src_path);
            if( GetRCObject(rc) == rcConstraint && GetRCState(rc) == rcInconsistent ) {
                blob_crc = true;
                rc = 0;
            }
        }
        if (blob_crc || index_chk) {
            const SRAMgr* mgr;
            if( (rc = SRAMgrMakeRead(&mgr)) == 0 ) {
                const SRATable* tbl;
                rc = SRAMgrOpenTableRead(mgr, &tbl, src_path);
                if (rc == 0) {
                    enum SRAPlatforms platform = SRA_PLATFORM_UNDEFINED;
                    const char* platform_str = "UNDEFINED";
                    const SRAColumn* c = NULL;

                    if( (rc = SRATableOpenColumnRead(tbl, &c, "PLATFORM", sra_platform_id_t)) == 0 ) {
                        const uint8_t *pltf;
                        bitsz_t o, z;
                        if( (rc = SRAColumnRead(c, 1, (const void **)&pltf, &o, &z)) == 0 ) {
                            platform = *pltf;
                            switch(platform) {
                            case SRA_PLATFORM_454:
                                platform_str = "454";
                                break;
                            case SRA_PLATFORM_ILLUMINA:
                                platform_str = "Illumina";
                                break;
                            case SRA_PLATFORM_ABSOLID:
                                platform_str = "AB SOLiD";
                                break;
                            default:
                                rc = RC(rcExe, rcAttr, rcReading, rcName, rcUnrecognized);
                                PLOGERR((klogErr, rc, "platform $(platform_id)", PLOG_U8(platform_id), pltf));
                                break;
                            }
                            if( rc == 0 ) {
                                PLOGMSG((klogInfo, "Platform $(platform)", PLOG_S(platform), platform_str));
                            }
                        }
                        SRAColumnRelease(c);
                    }
                    if( rc == 0 && blob_crc ) {
                        rc = CheckTable(tbl, platform);
                    }
                    if( rc == 0 && index_chk ) {
                        rc = ValidateSKeyIndex(tbl);
                    }
                    SRATableRelease(tbl);
                }
                SRAMgrRelease(mgr);
            }
        }
    }
    if( rc == 0 ) {
        LOGMSG(klogInfo, "check ok");
    } else {
        PLOGERR((klogErr, rc, "check failed: '$(table)'", PLOG_S(table), src_path));
    }
    return rc;
}
