#include <kapp/main.h>
#include <klib/log.h>
#include <klib/status.h>
#include <klib/rc.h>

#include <klib/vector.h>
#include <kfs/directory.h>
#include <kfs/dyload.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


/* update libpath
 */
static
void testld_update_libpath ( KDyld *dl, const char *path )
{
    rc_t rc = 0;

    const char *end;
    for ( end = strchr ( path, ':' ); end != NULL; end = strchr ( path = end + 1, ':' ) )
    {
        if ( path < end )
        {
            rc = KDyldAddSearchPath ( dl, "%.*s",
                ( int ) ( end - path ), path );
            if ( rc != 0 )
            {
                PLOGERR ( klogWarn, ( klogWarn, rc, "failed to add library path '$(path)'",
                                      "path=%.*s", ( int ) ( end - path ), path ));
            }
        }
    }

    if ( path [ 0 ] != 0 )
    {
        rc = KDyldAddSearchPath ( dl, path );
        if ( rc != 0 )
        {
            PLOGERR ( klogWarn, ( klogWarn, rc, "failed to add library path '$(path)'", "path=s", path ));
        }
    }
}


/* load library
 */
static
void testld_load_library ( KDyld *dl, KDlset *libs, const char *libname )
{
    KDylib *lib;
    rc_t rc = KDyldLoadLib ( dl, & lib, libname );
    if ( rc != 0 )
    {
        PLOGERR ( klogErr, ( klogErr, rc, "failed to load library '$(name)'",
                             "name=%s", libname ));
    }
    else
    {
        printf ( "loaded library '%s'\n", libname);

        rc = KDlsetAddLib ( libs, lib );
        if ( rc != 0 )
        {
            PLOGERR ( klogInt, ( klogInt, rc, "failed to retain library '$(name)'",
                                 "name=%s", libname ));
        }

        KDylibRelease ( lib );
    }
}

/* Version  EXTERN
 *  return 4-part version code: 0xMMmmrrrr, where
 *      MM = major release
 *      mm = minor release
 *    rrrr = bug-fix release
 */
ver_t CC KAppVersion ( void )
{
    return 0;
}
    

/* KMain - EXTERN
 *  executable entrypoint "main" is implemented by
 *  an OS-specific wrapper that takes care of establishing
 *  signal handlers, logging, etc.
 *
 *  in turn, OS-specific "main" will invoke "KMain" as
 *  platform independent main entrypoint.
 *
 *  "argc" [ IN ] - the number of textual parameters in "argv"
 *  should never be < 0, but has been left as a signed int
 *  for reasons of tradition.
 *
 *  "argv" [ IN ] - array of NUL terminated strings expected
 *  to be in the shell-native character set: ASCII or UTF-8
 *  element 0 is expected to be executable identity or path.
 */
rc_t CC Usage (const Args * args)
{
    const char * progname = "testld";
    rc_t rc;

    rc = ArgsArgvValue (args, 0, &progname);

    STSMSG (0, ("\n"
                "Usage: %s [ options ]\n"
                "\n"
                "    options:\n"
                "      -h               give tool help\n"
                "      -l <path>        path(s) for loading dynamic libraries\n"
                "\n",
                progname));
    return 0;
}

rc_t MiniUsage (const Args * args)
{
    const char * progname = "testld";

    ArgsArgvValue (args, 0, &progname);

    STSMSG (0, ( "\n"
                 "Usage: %s [ options ]\n"
                 "    run with option '--help' for help\n",
                 progname));
    return 0;
}

rc_t CC Version (const Args * args)
{
    const char * progname = "testld";
    rc_t rc;
    ver_t version;

    rc = ArgsArgvValue (args, 0, &progname);
    version = KAppVersion();

    STSMSG(0,( "%s: %V\n", progname, version));

    return 0;
}

#define OPTION_LOAD "lib-path"

OptDef MyOptions[] =
{
    { OPTION_LOAD, "l", 0, true, false, NULL, "load path for dynamic libraries" }
};

rc_t CC KMain ( int argc, char *argv [] )
{
    Args * args;
    rc_t rc;

    rc = ArgsMakeAndHandle (&args, argc, argv, 1, MyOptions, sizeof MyOptions / sizeof (OptDef));
    if (rc == 0)
    {
        KDyld *dl;

        /* create loader */
        rc = KDyldMake (&dl);
        if (rc)
            LOGERR ( klogInt, rc, "failed to create dynamic loader" );
        else
        {
            uint32_t num_libs;

            rc = ArgsOptionCount (args, OPTION_LOAD, &num_libs);
            if (rc == 0)
            {
                uint32_t ix;

                for (ix = 0; ix < num_libs; ix++)
                {
                    const char * path;

                    rc = ArgsOptionValue (args, OPTION_LOAD, ix, &path);
                    if (rc)
                        break;

                    testld_update_libpath (dl, path);
                }
                if (rc == 0)
                {
                    const char * path;
                    KDlset *libs;

                    /* append contents of LD_LIBRARY_PATH */
                    path = getenv ( "LD_LIBRARY_PATH" );

                    if (path)
                        testld_update_libpath (dl, path);

                    /* create libset */
                    rc = KDyldMakeSet (dl, & libs);
                    if (rc)
                        LOGERR (klogInt, rc, "failed to create dl library set");
                    else
                    {
                        rc = ArgsParamCount (args, &num_libs);
                        if (rc == 0)
                        {
                            /* test it */
                            for (ix = 0; ix < num_libs; ++ ix )
                            {
                                rc = ArgsParamValue (args, ix, &path);
                                if (rc)
                                    break;

                                testld_load_library (dl, libs, path);
                            }
                        }
                        KDlsetRelease ( libs );
                    }
                }
            }
            KDyldRelease ( dl );
        }
        ArgsWhack (args);
    }
    return rc;
}
